/* Generated by Cython 3.2.4 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "cyjs/bridge.h",
            "quickjs/quickjs.h"
        ],
        "extra_compile_args": [
            "-Wno-implicit-fallthrough",
            "-Wno-sign-compare",
            "-Wno-missing-field-initializers",
            "-Wno-unused-parameter",
            "-Wno-unused-but-set-variable",
            "-Wno-unused-result",
            "-Wno-stringop-truncation",
            "-Wno-array-bounds"
        ],
        "include_dirs": [
            "cyjs",
            "quickjs"
        ],
        "name": "cyjs._cyjs",
        "sources": [
            "cyjs/_cyjs.pyx",
            "quickjs/cutils.c",
            "quickjs/dtoa.c",
            "quickjs/libregexp.c",
            "quickjs/libunicode.c",
            "quickjs/quickjs.c"
        ]
    },
    "module_name": "cyjs._cyjs"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_4"
#define CYTHON_HEX_VERSION 0x030204F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CInitCode */
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__cyjs___cyjs
#define __PYX_HAVE_API__cyjs___cyjs
/* Early includes */
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include "quickjs.h"
#include "bridge.h"

    #if __PYX_LIMITED_VERSION_HEX < 0x030d0000
    static CYTHON_INLINE PyObject *
    __Pyx_CAPI_PyList_GetItemRef(PyObject *list, Py_ssize_t index)
    {
        PyObject *item = PyList_GetItem(list, index);
        Py_XINCREF(item);
        return item;
    }
    #else
    #define __Pyx_CAPI_PyList_GetItemRef PyList_GetItemRef
    #endif

    #if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030d0000
    static CYTHON_INLINE int
    __Pyx_CAPI_PyList_Extend(PyObject *list, PyObject *iterable)
    {
        return PyList_SetSlice(list, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, iterable);
    }

    static CYTHON_INLINE int
    __Pyx_CAPI_PyList_Clear(PyObject *list)
    {
        return PyList_SetSlice(list, 0, PY_SSIZE_T_MAX, NULL);
    }
    #else
    #define __Pyx_CAPI_PyList_Extend PyList_Extend
    #define __Pyx_CAPI_PyList_Clear PyList_Clear
    #endif
    
#include <stddef.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "cyjs/_cyjs.pyx",
  "<stringsource>",
  "cyjs/_cyjs.pxd",
  "cpython/bool.pxd",
  "cpython/type.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto (used by UnpackUnboundCMethod) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* CriticalSectionsDefinition.proto (used by CriticalSections) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_End(cs)
#define __Pyx_PyCriticalSection2_End(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_End PyCriticalSection_End
#define __Pyx_PyCriticalSection2_End PyCriticalSection2_End
#endif

/* CriticalSections.proto (used by ParseKeywordsImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection_Begin(cs, arg) (void)(cs)
#define __Pyx_PyCriticalSection2_Begin(cs, arg1, arg2) (void)(cs)
#else
#define __Pyx_PyCriticalSection_Begin PyCriticalSection_Begin
#define __Pyx_PyCriticalSection2_Begin PyCriticalSection2_Begin
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* IncludeStructmemberH.proto (used by FixUpExtensionType) */
#include <structmember.h>

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_4cyjs_5_cyjs_JSError;
struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage;
struct __pyx_obj_4cyjs_5_cyjs_PromiseHook;
struct __pyx_obj_4cyjs_5_cyjs_Runtime;
struct __pyx_obj_4cyjs_5_cyjs_Object;
struct __pyx_obj_4cyjs_5_cyjs_JSFunction;
struct __pyx_obj_4cyjs_5_cyjs_Context;
struct __pyx_obj_4cyjs_5_cyjs__OView;
struct __pyx_obj_4cyjs_5_cyjs__ObjectItemsView;
struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView;
struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView;
struct __pyx_obj_4cyjs_5_cyjs_CancelledError;
struct __pyx_obj_4cyjs_5_cyjs_InvalidStateError;
struct __pyx_obj_4cyjs_5_cyjs_Promise;
struct __pyx_obj_4cyjs_5_cyjs_NODEFAULT;
struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__;
struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__;
struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__;
struct __pyx_opt_args_4cyjs_5_cyjs_6Object_eval;
struct __pyx_opt_args_4cyjs_5_cyjs_6Object_eval_module;
struct __pyx_opt_args_4cyjs_5_cyjs_7Context_ceval;
struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval;
struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_module;
struct __pyx_opt_args_4cyjs_5_cyjs_7Context_add_function;
struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_this;
struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_this_with_module;
struct __pyx_fuse_0__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this;
struct __pyx_fuse_1__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this;
struct __pyx_fuse_2__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this;
struct __pyx_fuse_3__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this;
struct __pyx_fuse_4__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this;
struct __pyx_fuse_5__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this;
struct __pyx_fuse_6__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this;
struct __pyx_fuse_7__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this;
struct __pyx_fuse_8__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this;
struct __pyx_fuse_9__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this;
struct __pyx_t_4cyjs_5_cyjs_arg_parser_t;

/* "cyjs/_cyjs.pxd":158
 *     cpdef bytes to_json(self)
 * 
 *     cpdef object eval(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
struct __pyx_opt_args_4cyjs_5_cyjs_6Object_eval {
  int __pyx_n;
  PyObject *filename;
  int strict;
  int backtrace_barrier;
  int promise;
};

/* "cyjs/_cyjs.pxd":167
 *     )
 * 
 *     cpdef object eval_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
struct __pyx_opt_args_4cyjs_5_cyjs_6Object_eval_module {
  int __pyx_n;
  PyObject *filename;
  int strict;
  int backtrace_barrier;
  int promise;
};

/* "cyjs/_cyjs.pxd":232
 * 
 *     cdef object raise_exception(self)
 *     cdef object ceval(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
struct __pyx_opt_args_4cyjs_5_cyjs_7Context_ceval {
  int __pyx_n;
  PyObject *filename;
  int module;
  int strict;
  int backtrace_barrier;
  int promise;
};

/* "cyjs/_cyjs.pxd":242
 *     )
 * 
 *     cpdef object eval(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval {
  int __pyx_n;
  PyObject *filename;
  int strict;
  int backtrace_barrier;
  int promise;
};

/* "cyjs/_cyjs.pxd":251
 *     )
 * 
 *     cpdef object eval_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_module {
  int __pyx_n;
  PyObject *filename;
  int strict;
  int backtrace_barrier;
  int promise;
};

/* "cyjs/_cyjs.pxd":260
 *     )
 * 
 *     cpdef JSFunction add_function(             # <<<<<<<<<<<<<<
 *         self,
 *         object func,
*/
struct __pyx_opt_args_4cyjs_5_cyjs_7Context_add_function {
  int __pyx_n;
  PyObject *name;
  int count;
  int magic;
};

/* "cyjs/_cyjs.pxd":279
 *     )
 * 
 *     cpdef object eval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_this {
  int __pyx_n;
  PyObject *filename;
  int strict;
  int backtrace_barrier;
  int promise;
};

/* "cyjs/_cyjs.pxd":289
 *     )
 * 
 *     cpdef object eval_this_with_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_this_with_module {
  int __pyx_n;
  PyObject *filename;
  int strict;
  int backtrace_barrier;
  int promise;
};

/* "cyjs/_cyjs.pyx":104
 *     )
 * 
 * cpdef enum PromiseHookType:             # <<<<<<<<<<<<<<
 *     INIT = 0
 *     BEFORE = 1
*/
enum __pyx_t_4cyjs_5_cyjs_PromiseHookType {
  __pyx_e_4cyjs_5_cyjs_INIT = 0,
  __pyx_e_4cyjs_5_cyjs_BEFORE = 1,
  __pyx_e_4cyjs_5_cyjs_AFTER = 2,
  __pyx_e_4cyjs_5_cyjs_RESOLVE = 3
};

/* "cyjs/_cyjs.pxd":268
 *     )
 * 
 *     cdef object ceval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
struct __pyx_fuse_0__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this {
  int __pyx_n;
  PyObject *filename;
  int module;
  int strict;
  int backtrace_barrier;
  int promise;
};
struct __pyx_fuse_1__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this {
  int __pyx_n;
  PyObject *filename;
  int module;
  int strict;
  int backtrace_barrier;
  int promise;
};
struct __pyx_fuse_2__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this {
  int __pyx_n;
  PyObject *filename;
  int module;
  int strict;
  int backtrace_barrier;
  int promise;
};
struct __pyx_fuse_3__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this {
  int __pyx_n;
  PyObject *filename;
  int module;
  int strict;
  int backtrace_barrier;
  int promise;
};
struct __pyx_fuse_4__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this {
  int __pyx_n;
  PyObject *filename;
  int module;
  int strict;
  int backtrace_barrier;
  int promise;
};
struct __pyx_fuse_5__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this {
  int __pyx_n;
  PyObject *filename;
  int module;
  int strict;
  int backtrace_barrier;
  int promise;
};
struct __pyx_fuse_6__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this {
  int __pyx_n;
  PyObject *filename;
  int module;
  int strict;
  int backtrace_barrier;
  int promise;
};
struct __pyx_fuse_7__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this {
  int __pyx_n;
  PyObject *filename;
  int module;
  int strict;
  int backtrace_barrier;
  int promise;
};
struct __pyx_fuse_8__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this {
  int __pyx_n;
  PyObject *filename;
  int module;
  int strict;
  int backtrace_barrier;
  int promise;
};
struct __pyx_fuse_9__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this {
  int __pyx_n;
  PyObject *filename;
  int module;
  int strict;
  int backtrace_barrier;
  int promise;
};

/* "cyjs/_cyjs.pyx":381
 * # Argument parser (Not public)
 * 
 * cdef struct arg_parser_t:             # <<<<<<<<<<<<<<
 *     Py_ssize_t size # heap size
 *     Py_ssize_t len # objects held
*/
struct __pyx_t_4cyjs_5_cyjs_arg_parser_t {
  Py_ssize_t size;
  Py_ssize_t len;
  JSValue *values;
  JSContext *ctx;
  int heap;
};

/* "cyjs/_cyjs.pxd":46
 *     )
 * 
 * cdef class JSError(Exception):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef JSError new(JSContext* ctx, JSValue value)
*/
struct __pyx_obj_4cyjs_5_cyjs_JSError {
  PyBaseExceptionObject __pyx_base;
  struct __pyx_vtabstruct_4cyjs_5_cyjs_JSError *__pyx_vtab;
};


/* "cyjs/_cyjs.pxd":52
 * 
 * 
 * @cython.final             # <<<<<<<<<<<<<<
 * cdef class MemoryUsage:
 *     cdef:
*/
struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage {
  PyObject_HEAD
  struct __pyx_vtabstruct_4cyjs_5_cyjs_MemoryUsage *__pyx_vtab;
  int64_t malloc_size;
  int64_t malloc_limit;
  int64_t memory_used_size;
  int64_t malloc_count;
  int64_t memory_used_count;
  int64_t atom_count;
  int64_t atom_size;
  int64_t str_count;
  int64_t str_size;
  int64_t obj_count;
  int64_t obj_size;
  int64_t prop_count;
  int64_t prop_size;
  int64_t shape_count;
  int64_t shape_size;
  int64_t js_func_count;
  int64_t js_func_size;
  int64_t js_func_code_size;
  int64_t js_func_pc2line_count;
  int64_t array_count;
  int64_t fast_array_count;
  int64_t fast_array_elements;
  int64_t binary_object_count;
  int64_t binary_object_size;
};


/* "cyjs/_cyjs.pxd":90
 * # Used for hooking a promise callback without triggering a deref
 * # This class is not exactly public nor should it be imported.
 * @cython.internal             # <<<<<<<<<<<<<<
 * cdef class PromiseHook:
 *     cdef:
*/
struct __pyx_obj_4cyjs_5_cyjs_PromiseHook {
  PyObject_HEAD
  struct __pyx_vtabstruct_4cyjs_5_cyjs_PromiseHook *__pyx_vtab;
  PyObject *func;
  struct __pyx_obj_4cyjs_5_cyjs_Runtime *rt;
};


/* "cyjs/_cyjs.pxd":107
 * 
 * 
 * cdef class Runtime:             # <<<<<<<<<<<<<<
 *     """
 *     Represents a JavaScript runtime corresponding to an
*/
struct __pyx_obj_4cyjs_5_cyjs_Runtime {
  PyObject_HEAD
  struct __pyx_vtabstruct_4cyjs_5_cyjs_Runtime *__pyx_vtab;
  JSRuntime *rt;
  struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *promise_hook;
  int has_promise_hook;
};


/* "cyjs/_cyjs.pxd":147
 * 
 * 
 * cdef class Object:             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly Context context
*/
struct __pyx_obj_4cyjs_5_cyjs_Object {
  PyObject_HEAD
  struct __pyx_vtabstruct_4cyjs_5_cyjs_Object *__pyx_vtab;
  struct __pyx_obj_4cyjs_5_cyjs_Context *context;
  int32_t tag;
  JSValue value;
  JSContext *ctx;
};


/* "cyjs/_cyjs.pxd":178
 * 
 * 
 * cdef class JSFunction:             # <<<<<<<<<<<<<<
 *     """Used for acting as a bridge between Javascript and Python
 *     it is meant to be used as an access-point for Javascript (ECMA) to be called upon
*/
struct __pyx_obj_4cyjs_5_cyjs_JSFunction {
  PyObject_HEAD
  struct __pyx_vtabstruct_4cyjs_5_cyjs_JSFunction *__pyx_vtab;
  struct __pyx_obj_4cyjs_5_cyjs_Context *context;
  JSContext *ctx;
  JSValue value;
  PyObject *func;
};


/* "cyjs/_cyjs.pxd":210
 *     bint
 * 
 * cdef class Context:             # <<<<<<<<<<<<<<
 *     cdef:
 *         # public runtime for python access to value (can't be deleted)
*/
struct __pyx_obj_4cyjs_5_cyjs_Context {
  PyObject_HEAD
  struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *__pyx_vtab;
  struct __pyx_obj_4cyjs_5_cyjs_Runtime *runtime;
  JSRuntime *rt;
  JSContext *ctx;
  PyObject *_cb_exception;
};


/* "cyjs/_cyjs.pyx":450
 * 
 * # base for gathering iterations
 * cdef class _OView:             # <<<<<<<<<<<<<<
 *     cdef:
 *         # parent object to prevent GC
*/
struct __pyx_obj_4cyjs_5_cyjs__OView {
  PyObject_HEAD
  struct __pyx_vtabstruct_4cyjs_5_cyjs__OView *__pyx_vtab;
  struct __pyx_obj_4cyjs_5_cyjs_Object *obj;
  JSValue value;
  JSPropertyEnum *tab;
  JSContext *ctx;
  uint32_t len;
};


/* "cyjs/_cyjs.pyx":495
 * 
 * 
 * cdef class _ObjectItemsView(_OView):             # <<<<<<<<<<<<<<
 *     def __iter__(self):
 *         cdef JSAtom key
*/
struct __pyx_obj_4cyjs_5_cyjs__ObjectItemsView {
  struct __pyx_obj_4cyjs_5_cyjs__OView __pyx_base;
};


/* "cyjs/_cyjs.pyx":505
 * 
 * 
 * cdef class _ObjectKeysView(_OView):             # <<<<<<<<<<<<<<
 *     def __iter__(self):
 *         cdef size_t i
*/
struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView {
  struct __pyx_obj_4cyjs_5_cyjs__OView __pyx_base;
};


/* "cyjs/_cyjs.pyx":522
 *         return False
 * 
 * cdef class _ObjectValuesView(_OView):             # <<<<<<<<<<<<<<
 *     def __iter__(self):
 *         cdef JSAtom key
*/
struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView {
  struct __pyx_obj_4cyjs_5_cyjs__OView __pyx_base;
};


/* "cyjs/_cyjs.pyx":1137
 * 
 * 
 * cdef class CancelledError(Exception):             # <<<<<<<<<<<<<<
 *     """Promise was rejected"""
 * 
*/
struct __pyx_obj_4cyjs_5_cyjs_CancelledError {
  PyBaseExceptionObject __pyx_base;
};


/* "cyjs/_cyjs.pyx":1140
 *     """Promise was rejected"""
 * 
 * cdef class InvalidStateError(Exception):             # <<<<<<<<<<<<<<
 *     """Promise on inavlid state"""
 * 
*/
struct __pyx_obj_4cyjs_5_cyjs_InvalidStateError {
  PyBaseExceptionObject __pyx_base;
};


/* "cyjs/_cyjs.pyx":1144
 * 
 * 
 * cdef class Promise(Object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list _callbacks
*/
struct __pyx_obj_4cyjs_5_cyjs_Promise {
  struct __pyx_obj_4cyjs_5_cyjs_Object __pyx_base;
  PyObject *_callbacks;
  PyObject *_exception;
  PyObject *_result;
  int completed;
};


/* "cyjs/_cyjs.pyx":1243
 * # SENTINAL value (NOT PUBLIC!!!!) It's never used (Only ever once)
 * # And should not be publically called upon
 * @cython.internal             # <<<<<<<<<<<<<<
 * cdef class NODEFAULT:
 *     pass
*/
struct __pyx_obj_4cyjs_5_cyjs_NODEFAULT {
  PyObject_HEAD
};


/* "cyjs/_cyjs.pyx":496
 * 
 * cdef class _ObjectItemsView(_OView):
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef JSAtom key
 *         cdef JSValue val
*/
struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__ {
  PyObject_HEAD
  JSAtom __pyx_v_key;
  size_t __pyx_v_pos;
  struct __pyx_obj_4cyjs_5_cyjs__ObjectItemsView *__pyx_v_self;
  JSValue __pyx_v_val;
};


/* "cyjs/_cyjs.pyx":506
 * 
 * cdef class _ObjectKeysView(_OView):
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self.len):
*/
struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ {
  PyObject_HEAD
  size_t __pyx_v_i;
  struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView *__pyx_v_self;
  uint32_t __pyx_t_0;
  uint32_t __pyx_t_1;
  size_t __pyx_t_2;
};


/* "cyjs/_cyjs.pyx":523
 * 
 * cdef class _ObjectValuesView(_OView):
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef JSAtom key
 *         cdef JSValue val
*/
struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ {
  PyObject_HEAD
  JSAtom __pyx_v_key;
  size_t __pyx_v_pos;
  struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView *__pyx_v_self;
  JSValue __pyx_v_val;
};



/* "cyjs/_cyjs.pyx":33
 * 
 * 
 * cdef class JSError(Exception):             # <<<<<<<<<<<<<<
 *     """Represents Numerous Exceptions raised from CYJS"""
 * 
*/

struct __pyx_vtabstruct_4cyjs_5_cyjs_JSError {
  struct __pyx_obj_4cyjs_5_cyjs_JSError *(*__pyx_new)(JSContext *, JSValue);
};
static struct __pyx_vtabstruct_4cyjs_5_cyjs_JSError *__pyx_vtabptr_4cyjs_5_cyjs_JSError;


/* "cyjs/_cyjs.pyx":63
 * 
 * 
 * @cython.final             # <<<<<<<<<<<<<<
 * cdef class MemoryUsage:
 * 
*/

struct __pyx_vtabstruct_4cyjs_5_cyjs_MemoryUsage {
  struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *(*__pyx_new)(JSMemoryUsage const *);
};
static struct __pyx_vtabstruct_4cyjs_5_cyjs_MemoryUsage *__pyx_vtabptr_4cyjs_5_cyjs_MemoryUsage;
static struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_f_4cyjs_5_cyjs_11MemoryUsage_new(JSMemoryUsage const *);


/* "cyjs/_cyjs.pyx":110
 *     RESOLVE = 3
 * 
 * @cython.internal             # <<<<<<<<<<<<<<
 * cdef class PromiseHook:
 *     @staticmethod
*/

struct __pyx_vtabstruct_4cyjs_5_cyjs_PromiseHook {
  struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *(*__pyx_new)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, PyObject *);
  void (*hook)(struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *, JSContext *, JSPromiseHookType, JSValue, JSValue);
};
static struct __pyx_vtabstruct_4cyjs_5_cyjs_PromiseHook *__pyx_vtabptr_4cyjs_5_cyjs_PromiseHook;


/* "cyjs/_cyjs.pyx":137
 * 
 * 
 * cdef class Runtime:             # <<<<<<<<<<<<<<
 *     def __init__(self) -> None:
 *         self.rt = CYJS_NewRuntime(<void*>self)
*/

struct __pyx_vtabstruct_4cyjs_5_cyjs_Runtime {
  struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *(*compute_memory_usage)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, int __pyx_skip_dispatch);
  PyObject *(*dump_memory_usage)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*execute_pending_job)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, int __pyx_skip_dispatch);
  int (*is_job_pending)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, int __pyx_skip_dispatch);
  JSContext *(*new_context)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *);
  void (*run_gc)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, int __pyx_skip_dispatch);
  void (*set_memory_limit)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, size_t, int __pyx_skip_dispatch);
  void (*set_max_stack_size)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, size_t, int __pyx_skip_dispatch);
  void (*update_statck_top)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, int __pyx_skip_dispatch);
  PyObject *(*set_promise_hook)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4cyjs_5_cyjs_Runtime *__pyx_vtabptr_4cyjs_5_cyjs_Runtime;


/* "cyjs/_cyjs.pyx":543
 * 
 * # Inspired by Old the & retired Quickjs ObjectData class obejct
 * cdef class Object:             # <<<<<<<<<<<<<<
 * 
 *     cdef void init(self, Context ctx, JSValue value):
*/

struct __pyx_vtabstruct_4cyjs_5_cyjs_Object {
  void (*init)(struct __pyx_obj_4cyjs_5_cyjs_Object *, struct __pyx_obj_4cyjs_5_cyjs_Context *, JSValue);
  PyObject *(*to_json)(struct __pyx_obj_4cyjs_5_cyjs_Object *, int __pyx_skip_dispatch);
  PyObject *(*eval)(struct __pyx_obj_4cyjs_5_cyjs_Object *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_6Object_eval *__pyx_optional_args);
  PyObject *(*eval_module)(struct __pyx_obj_4cyjs_5_cyjs_Object *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_6Object_eval_module *__pyx_optional_args);
};
static struct __pyx_vtabstruct_4cyjs_5_cyjs_Object *__pyx_vtabptr_4cyjs_5_cyjs_Object;


/* "cyjs/_cyjs.pyx":706
 * # An async binding for python asynchronous functions might be in consideration for the future
 * # as part of the aiojs project planned for the future of 2026 - Vizonex
 * cdef class JSFunction:             # <<<<<<<<<<<<<<
 *     """Used for acting as a bridge between Javascript and Python
 *     it is meant to be used as an access-point for Javascript (ECMA) to be called upon
*/

struct __pyx_vtabstruct_4cyjs_5_cyjs_JSFunction {
  struct __pyx_obj_4cyjs_5_cyjs_JSFunction *(*__pyx_new)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *);
  JSValue (*call_js)(struct __pyx_obj_4cyjs_5_cyjs_JSFunction *, JSContext *, JSValue, int, JSValue *, int);
};
static struct __pyx_vtabstruct_4cyjs_5_cyjs_JSFunction *__pyx_vtabptr_4cyjs_5_cyjs_JSFunction;


/* "cyjs/_cyjs.pyx":776
 * #   - get_exception
 * 
 * cdef class Context: # type: ignore             # <<<<<<<<<<<<<<
 *     def __init__(self,
 *         Runtime runtime = Runtime(),
*/

struct __pyx_vtabstruct_4cyjs_5_cyjs_Context {
  int (*has_exception)(struct __pyx_obj_4cyjs_5_cyjs_Context *);
  JSValue (*get_exception)(struct __pyx_obj_4cyjs_5_cyjs_Context *);
  PyObject *(*raise_exception)(struct __pyx_obj_4cyjs_5_cyjs_Context *);
  PyObject *(*ceval)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_ceval *__pyx_optional_args);
  PyObject *(*eval)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval *__pyx_optional_args);
  PyObject *(*eval_module)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_module *__pyx_optional_args);
  struct __pyx_obj_4cyjs_5_cyjs_JSFunction *(*add_function)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_add_function *__pyx_optional_args);
  PyObject *(*eval_this)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_this *__pyx_optional_args);
  PyObject *(*eval_this_with_module)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_this_with_module *__pyx_optional_args);
  PyObject *(*__pyx_fuse_0ceval_this)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, struct __pyx_obj_4cyjs_5_cyjs_JSFunction *, struct __pyx_fuse_0__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args);
  PyObject *(*__pyx_fuse_1ceval_this)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, struct __pyx_obj_4cyjs_5_cyjs_Object *, struct __pyx_fuse_1__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args);
  PyObject *(*__pyx_fuse_2ceval_this)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, PyObject *, struct __pyx_fuse_2__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args);
  PyObject *(*__pyx_fuse_3ceval_this)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, PyObject *, struct __pyx_fuse_3__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args);
  PyObject *(*__pyx_fuse_4ceval_this)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, PyObject *, struct __pyx_fuse_4__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args);
  PyObject *(*__pyx_fuse_5ceval_this)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, PyObject *, struct __pyx_fuse_5__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args);
  PyObject *(*__pyx_fuse_6ceval_this)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, int, struct __pyx_fuse_6__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args);
  PyObject *(*__pyx_fuse_7ceval_this)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, float, struct __pyx_fuse_7__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args);
  PyObject *(*__pyx_fuse_8ceval_this)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, PyObject *, struct __pyx_fuse_8__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args);
  PyObject *(*__pyx_fuse_9ceval_this)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, int, struct __pyx_fuse_9__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args);
};
static struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *__pyx_vtabptr_4cyjs_5_cyjs_Context;
static CYTHON_INLINE int __pyx_f_4cyjs_5_cyjs_7Context_has_exception(struct __pyx_obj_4cyjs_5_cyjs_Context *);
static CYTHON_INLINE JSValue __pyx_f_4cyjs_5_cyjs_7Context_get_exception(struct __pyx_obj_4cyjs_5_cyjs_Context *);


/* "cyjs/_cyjs.pyx":450
 * 
 * # base for gathering iterations
 * cdef class _OView:             # <<<<<<<<<<<<<<
 *     cdef:
 *         # parent object to prevent GC
*/

struct __pyx_vtabstruct_4cyjs_5_cyjs__OView {
  int (*cnext)(struct __pyx_obj_4cyjs_5_cyjs__OView *, size_t *, JSAtom *, JSValue *);
};
static struct __pyx_vtabstruct_4cyjs_5_cyjs__OView *__pyx_vtabptr_4cyjs_5_cyjs__OView;


/* "cyjs/_cyjs.pyx":495
 * 
 * 
 * cdef class _ObjectItemsView(_OView):             # <<<<<<<<<<<<<<
 *     def __iter__(self):
 *         cdef JSAtom key
*/

struct __pyx_vtabstruct_4cyjs_5_cyjs__ObjectItemsView {
  struct __pyx_vtabstruct_4cyjs_5_cyjs__OView __pyx_base;
};
static struct __pyx_vtabstruct_4cyjs_5_cyjs__ObjectItemsView *__pyx_vtabptr_4cyjs_5_cyjs__ObjectItemsView;


/* "cyjs/_cyjs.pyx":505
 * 
 * 
 * cdef class _ObjectKeysView(_OView):             # <<<<<<<<<<<<<<
 *     def __iter__(self):
 *         cdef size_t i
*/

struct __pyx_vtabstruct_4cyjs_5_cyjs__ObjectKeysView {
  struct __pyx_vtabstruct_4cyjs_5_cyjs__OView __pyx_base;
};
static struct __pyx_vtabstruct_4cyjs_5_cyjs__ObjectKeysView *__pyx_vtabptr_4cyjs_5_cyjs__ObjectKeysView;


/* "cyjs/_cyjs.pyx":522
 *         return False
 * 
 * cdef class _ObjectValuesView(_OView):             # <<<<<<<<<<<<<<
 *     def __iter__(self):
 *         cdef JSAtom key
*/

struct __pyx_vtabstruct_4cyjs_5_cyjs__ObjectValuesView {
  struct __pyx_vtabstruct_4cyjs_5_cyjs__OView __pyx_base;
};
static struct __pyx_vtabstruct_4cyjs_5_cyjs__ObjectValuesView *__pyx_vtabptr_4cyjs_5_cyjs__ObjectValuesView;


/* "cyjs/_cyjs.pyx":1144
 * 
 * 
 * cdef class Promise(Object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list _callbacks
*/

struct __pyx_vtabstruct_4cyjs_5_cyjs_Promise {
  struct __pyx_vtabstruct_4cyjs_5_cyjs_Object __pyx_base;
  PyObject *(*add_done_callback)(struct __pyx_obj_4cyjs_5_cyjs_Promise *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*exception)(struct __pyx_obj_4cyjs_5_cyjs_Promise *, int __pyx_skip_dispatch);
  int (*done)(struct __pyx_obj_4cyjs_5_cyjs_Promise *, int __pyx_skip_dispatch);
  Py_ssize_t (*remove_done_callback)(struct __pyx_obj_4cyjs_5_cyjs_Promise *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*result)(struct __pyx_obj_4cyjs_5_cyjs_Promise *, int __pyx_skip_dispatch);
  PyObject *(*complete)(struct __pyx_obj_4cyjs_5_cyjs_Promise *);
  PyObject *(*poll)(struct __pyx_obj_4cyjs_5_cyjs_Promise *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4cyjs_5_cyjs_Promise *__pyx_vtabptr_4cyjs_5_cyjs_Promise;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto (used by PyErrFetchRestore) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto (used by GetBuiltinName) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto (used by GetModuleGlobalName) */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyDictVersioning.proto (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* TupleAndListFromArray.proto (used by fastcall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto (used by BytesEquals) */
#include <string.h>

/* BytesEquals.proto (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* py_dict_items.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallCFunction.proto (used by CallUnboundCMethod0) */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* PyObjectCallOneArg.proto (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* UnpackUnboundCMethod.proto (used by CallUnboundCMethod0) */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod0.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* py_dict_values.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* OwnedDictNext.proto (used by RejectKeywords) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue);
#else
CYTHON_INLINE
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue);
#endif

/* RejectKeywords.export */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* RaiseDoubleKeywords.proto (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywordsImpl.export */
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name
);
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* GetTopmostException.proto (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* ReRaiseException.proto */
static CYTHON_INLINE void __Pyx_ReraiseException(void);

/* PyObjectFastCallMethod.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyObject_FastCallMethod(name, args, nargsf) PyObject_VectorcallMethod(name, args, nargsf, NULL)
#else
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf);
#endif

/* RaiseException.export */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* CallableCheck.proto */
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyCallable_Check(obj)   (Py_TYPE(obj)->tp_call != NULL)
#else
#define __Pyx_PyCallable_Check(obj)   PyCallable_Check(obj)
#endif

/* PyTypeError_Check.proto */
#define __Pyx_PyExc_TypeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_TypeError)

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* ArgTypeTestFunc.export */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))

/* PyRuntimeError_Check.proto */
#define __Pyx_PyExc_RuntimeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_RuntimeError)

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* PyAttributeError_Check.proto */
#define __Pyx_PyExc_AttributeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_AttributeError)

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* JoinPyUnicode.export */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* KeywordStringCheck.proto */
static CYTHON_INLINE int __Pyx_CheckKeywordStrings(const char* function_name, PyObject *kw);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* SliceObject.proto */
#define __Pyx_PyObject_DelSlice(obj, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)\
    __Pyx_PyObject_SetSlice(obj, (PyObject*)NULL, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(
        PyObject* obj, PyObject* value, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* CallTypeTraverse.proto */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* AllocateExtensionType.proto */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final);

/* CheckTypeForFreelists.proto */
#if CYTHON_USE_FREELISTS
#if CYTHON_USE_TYPE_SPECS
#define __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, expected_tp, expected_size) ((int) ((t) == (expected_tp)))
#define __PYX_CHECK_TYPE_FOR_FREELIST_FLAGS  Py_TPFLAGS_IS_ABSTRACT
#else
#define __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, expected_tp, expected_size) ((int) ((t)->tp_basicsize == (expected_size)))
#define __PYX_CHECK_TYPE_FOR_FREELIST_FLAGS  (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)
#endif
#define __PYX_CHECK_TYPE_FOR_FREELISTS(t, expected_tp, expected_size)\
    (__PYX_CHECK_FINAL_TYPE_FOR_FREELISTS((t), (expected_tp), (expected_size)) &\
     (int) (!__Pyx_PyType_HasFeature((t), __PYX_CHECK_TYPE_FOR_FREELIST_FLAGS)))
#endif

/* LimitedApiGetTypeDict.proto (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* FormatTypeName.proto (used by ValidateExternBase) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* ValidateExternBase.proto */
static int __Pyx_validate_extern_base(PyTypeObject *base);

/* PyObjectCallNoArg.proto (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);
#endif

/* PyObjectCallMethod0.proto (used by PyType_Ready) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* ValidateBasesTuple.proto (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto (used by MergeVTables) */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
static int __Pyx_MergeVtables(PyTypeObject *type);

/* DelItemOnTypeDict.proto (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k);
#define __Pyx_DelItemOnTypeDict(tp, k) __Pyx__DelItemOnTypeDict((PyTypeObject*)tp, k)

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_2_4
#define __PYX_HAVE_RT_ImportType_proto_3_2_4
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_2_4 {
   __Pyx_ImportType_CheckSize_Error_3_2_4 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_2_4 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_2_4 = 2
};
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size);
#endif

/* HasAttr.proto (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* ImportImpl.export */
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level);

/* Import.proto */
static CYTHON_INLINE PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* ListPack.proto */
static PyObject *__Pyx_PyList_Pack(Py_ssize_t n, ...);

/* GetAttr.proto (used by Globals) */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* Globals.proto */
static PyObject* __Pyx_Globals(void);

/* PyObjectVectorCallKwBuilder.proto */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* PyObjectDelAttr.proto (used by PyObjectSetAttrStr) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
#define __Pyx_PyObject_DelAttr(o, n) PyObject_SetAttr(o, n, NULL)
#else
#define __Pyx_PyObject_DelAttr(o, n) PyObject_DelAttr(o, n)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   __Pyx_PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* dict_setdefault.proto (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* AddModuleRef.proto (used by FetchSharedCythonModule) */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* FetchSharedCythonModule.proto (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto (used by CommonTypesMetaclass) */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto (used by CythonFunctionShared) */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* PyMethodNew.proto (used by CythonFunctionShared) */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto (used by CythonFunction) */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_dict;
#endif
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CheckUnpickleChecksum.proto */
static CYTHON_INLINE int __Pyx_CheckUnpickleChecksum(long checksum, long checksum1, long checksum2, long checksum3, const char *members);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int64_t(int64_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE int64_t __Pyx_PyLong_As_int64_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyLong_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint32_t __Pyx_PyLong_As_uint32_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint32_t(uint32_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int32_t(int32_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(enum __pyx_t_4cyjs_5_cyjs_PromiseHookType value);

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* UpdateUnpickledDict.proto */
static int __Pyx_UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* IterNextPlain.proto (used by CoroutineBase) */
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next_Plain(PyObject *iterator);
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *__Pyx_GetBuiltinNext_LimitedAPI(void);
#endif

/* ReturnWithStopIteration.proto (used by CoroutineBase) */
static CYTHON_INLINE void __Pyx_ReturnWithStopIteration(PyObject* value, int async, int iternext);

/* CoroutineBase.proto (used by Generator) */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *gi_weakreflist;
#endif
    PyObject *classobj;
    PyObject *yieldfrom;
    __Pyx_pyiter_sendfunc yieldfrom_am_send;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
#if CYTHON_USE_SYS_MONITORING && (CYTHON_PROFILE || CYTHON_TRACE)
    PyMonitoringState __pyx_pymonitoring_state[__Pyx_MonitoringEventTypes_CyGen_count];
    uint64_t __pyx_pymonitoring_version;
#endif
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static __Pyx_PySendResult __Pyx_Coroutine_AmSend(PyObject *self, PyObject *value, PyObject **retval);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static __Pyx_PySendResult __Pyx_Coroutine_Close(PyObject *self, PyObject **retval);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);
static char __Pyx_Coroutine_test_and_set_is_running(__pyx_CoroutineObject *gen);
static void __Pyx_Coroutine_unset_is_running(__pyx_CoroutineObject *gen);
static char __Pyx_Coroutine_get_is_running(__pyx_CoroutineObject *gen);
static PyObject *__Pyx_Coroutine_get_is_running_getter(PyObject *gen, void *closure);
#if __PYX_HAS_PY_AM_SEND == 2
static void __Pyx_SetBackportTypeAmSend(PyTypeObject *type, __Pyx_PyAsyncMethodsStruct *static_amsend_methods, __Pyx_pyiter_sendfunc am_send);
#endif
static PyObject *__Pyx_Coroutine_fail_reduce_ex(PyObject *self, PyObject *arg);

/* Generator.proto */
#define __Pyx_Generator_USED
#define __Pyx_Generator_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_mstate_global->__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(PyObject *module);
static CYTHON_INLINE PyObject *__Pyx_Generator_GetInlinedResult(PyObject *self);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."

static CYTHON_INLINE int __pyx_f_4cyjs_5_cyjs_7Context_has_exception(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self); /* proto*/
static CYTHON_INLINE JSValue __pyx_f_4cyjs_5_cyjs_7Context_get_exception(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self); /* proto*/
static struct __pyx_obj_4cyjs_5_cyjs_JSError *__pyx_f_4cyjs_5_cyjs_7JSError_new(JSContext *__pyx_v_ctx, JSValue __pyx_v_value); /* proto*/
static struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_f_4cyjs_5_cyjs_11MemoryUsage_new(JSMemoryUsage const *__pyx_v_ptr); /* proto*/
static struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *__pyx_f_4cyjs_5_cyjs_11PromiseHook_new(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_rt, PyObject *__pyx_v_func); /* proto*/
static void __pyx_f_4cyjs_5_cyjs_11PromiseHook_hook(struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *__pyx_v_self, JSContext *__pyx_v_context, JSPromiseHookType __pyx_v_type, JSValue __pyx_v_promise, JSValue __pyx_v_parent_promise); /* proto*/
static struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_f_4cyjs_5_cyjs_7Runtime_compute_memory_usage(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Runtime_dump_memory_usage(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, PyObject *__pyx_v_file, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Runtime_execute_pending_job(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_4cyjs_5_cyjs_7Runtime_is_job_pending(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static JSContext *__pyx_f_4cyjs_5_cyjs_7Runtime_new_context(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self); /* proto*/
static void __pyx_f_4cyjs_5_cyjs_7Runtime_run_gc(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_4cyjs_5_cyjs_7Runtime_set_memory_limit(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, size_t __pyx_v_limit, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_4cyjs_5_cyjs_7Runtime_set_max_stack_size(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, size_t __pyx_v_max_stack_size, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_4cyjs_5_cyjs_7Runtime_update_statck_top(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Runtime_set_promise_hook(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, PyObject *__pyx_v_func, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_4cyjs_5_cyjs_6_OView_cnext(struct __pyx_obj_4cyjs_5_cyjs__OView *__pyx_v_self, size_t *__pyx_v_pos, JSAtom *__pyx_v_key, JSValue *__pyx_v_val); /* proto*/
static void __pyx_f_4cyjs_5_cyjs_6Object_init(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_ctx, JSValue __pyx_v_value); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_6Object_to_json(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_6Object_eval(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, PyObject *__pyx_v_code, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_6Object_eval *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_6Object_eval_module(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, PyObject *__pyx_v_code, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_6Object_eval_module *__pyx_optional_args); /* proto*/
static struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_f_4cyjs_5_cyjs_10JSFunction_new(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_context, PyObject *__pyx_v_func); /* proto*/
static JSValue __pyx_f_4cyjs_5_cyjs_10JSFunction_call_js(struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_self, JSContext *__pyx_v_ctx, CYTHON_UNUSED JSValue __pyx_v_this_val, int __pyx_v_argc, JSValue *__pyx_v_argv, CYTHON_UNUSED int __pyx_v_magic); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Context_raise_exception(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Context_ceval(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_ceval *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Context_eval(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Context_eval_module(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_module *__pyx_optional_args); /* proto*/
static struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_f_4cyjs_5_cyjs_7Context_add_function(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_func, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_add_function *__pyx_optional_args); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_this, struct __pyx_fuse_0__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_this, struct __pyx_fuse_1__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, struct __pyx_fuse_2__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, struct __pyx_fuse_3__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, struct __pyx_fuse_4__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args); /* proto*/
static PyObject *__pyx_fuse_5__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, struct __pyx_fuse_5__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args); /* proto*/
static PyObject *__pyx_fuse_6__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, int __pyx_v_this, struct __pyx_fuse_6__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args); /* proto*/
static PyObject *__pyx_fuse_7__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, float __pyx_v_this, struct __pyx_fuse_7__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args); /* proto*/
static PyObject *__pyx_fuse_8__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, struct __pyx_fuse_8__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args); /* proto*/
static PyObject *__pyx_fuse_9__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, int __pyx_v_this, struct __pyx_fuse_9__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Context_eval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_this *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Context_eval_this_with_module(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_this_with_module *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Promise_add_done_callback(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, PyObject *__pyx_v_fn, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Promise_exception(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_4cyjs_5_cyjs_7Promise_done(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static Py_ssize_t __pyx_f_4cyjs_5_cyjs_7Promise_remove_done_callback(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, PyObject *__pyx_v_fn, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Promise_result(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Promise_complete(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Promise_poll(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from "cython" */

/* Module declarations from "libc.stdint" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "cyjs.quickjs" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.bool" */

/* Module declarations from "cpython.buffer" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.exc" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cpython.long" */

/* Module declarations from "cpython.mem" */

/* Module declarations from "cpython.tuple" */

/* Module declarations from "libc.stddef" */

/* Module declarations from "cpython.unicode" */

/* Module declarations from "cyjs._cyjs" */
static PyObject *__Pyx_EnumBase = 0;
static PyObject *__Pyx_FlagBase = 0;
static void __pyx_f_4cyjs_5_cyjs_on_promise_hook(JSContext *, JSPromiseHookType, JSValue, JSValue, void *); /*proto*/
static int __pyx_f_4cyjs_5_cyjs_py_to_atom(JSContext *, PyObject *, JSAtom *); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_atom_to_py(JSContext *, JSAtom); /*proto*/
static CYTHON_INLINE int __pyx_f_4cyjs_5_cyjs_fast_parse_json_from_obj(JSContext *, JSValue *, PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_4cyjs_5_cyjs_fast_serlize_dict_or_list(JSContext *, JSValue *, PyObject *); /*proto*/
static int __pyx_f_4cyjs_5_cyjs_fast_serlize_string(JSContext *, JSValue *, PyObject *); /*proto*/
static CYTHON_INLINE JSValue __pyx_f_4cyjs_5_cyjs_py_to_js_exception(JSContext *, PyObject *); /*proto*/
static void __pyx_f_4cyjs_5_cyjs_arg_parser_init(struct __pyx_t_4cyjs_5_cyjs_arg_parser_t *, JSContext *, JSValue *, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_f_4cyjs_5_cyjs_arg_parser_append(struct __pyx_t_4cyjs_5_cyjs_arg_parser_t *, JSValue); /*proto*/
static int __pyx_f_4cyjs_5_cyjs_arg_parser_pack(struct __pyx_t_4cyjs_5_cyjs_arg_parser_t *, PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4cyjs_5_cyjs_arg_parser_unpack(struct __pyx_t_4cyjs_5_cyjs_arg_parser_t *); /*proto*/
static void __pyx_f_4cyjs_5_cyjs_arg_parser_finish(struct __pyx_t_4cyjs_5_cyjs_arg_parser_t *); /*proto*/
static JSValue __pyx_f_4cyjs_5_cyjs_on_cclosure_callback(JSContext *, JSValue, int, JSValue *, int, void *); /*proto*/
static void __pyx_f_4cyjs_5_cyjs_on_cclosure_opaque_finalize(void *); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_jsv_to_str(JSContext *, JSValue); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_jsv_to_big_int(JSContext *, JSValue); /*proto*/
static struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_f_4cyjs_5_cyjs_jsv_to_object(JSContext *, JSValue); /*proto*/
static struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_f_4cyjs_5_cyjs_jsv_to_promise(JSContext *, JSValue); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_to_python(JSContext *, JSValue); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs___pyx_unpickle_JSError__set_state(struct __pyx_obj_4cyjs_5_cyjs_JSError *, PyObject *); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs___pyx_unpickle_MemoryUsage__set_state(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *, PyObject *); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs___pyx_unpickle_PromiseHook__set_state(struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *, PyObject *); /*proto*/
static int __pyx_fuse_0__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *, JSValue *, struct __pyx_obj_4cyjs_5_cyjs_JSFunction *); /*proto*/
static int __pyx_fuse_1__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *, JSValue *, struct __pyx_obj_4cyjs_5_cyjs_Object *); /*proto*/
static int __pyx_fuse_2__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *, JSValue *, PyObject *); /*proto*/
static int __pyx_fuse_3__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *, JSValue *, PyObject *); /*proto*/
static int __pyx_fuse_4__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *, JSValue *, PyObject *); /*proto*/
static int __pyx_fuse_5__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *, JSValue *, PyObject *); /*proto*/
static int __pyx_fuse_6__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *, JSValue *, int); /*proto*/
static int __pyx_fuse_7__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *, JSValue *, float); /*proto*/
static int __pyx_fuse_8__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *, JSValue *, PyObject *); /*proto*/
static int __pyx_fuse_9__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *, JSValue *, int); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs___pyx_unpickle_CancelledError__set_state(struct __pyx_obj_4cyjs_5_cyjs_CancelledError *, PyObject *); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs___pyx_unpickle_InvalidStateError__set_state(struct __pyx_obj_4cyjs_5_cyjs_InvalidStateError *, PyObject *); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs___pyx_unpickle_NODEFAULT__set_state(struct __pyx_obj_4cyjs_5_cyjs_NODEFAULT *, PyObject *); /*proto*/
static PyObject *__Pyx_Enum_4cyjs_5_cyjs_enum__dunderpyx_t_4cyjs_5_cyjs_PromiseHookType_to_py(enum __pyx_t_4cyjs_5_cyjs_PromiseHookType); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "cyjs._cyjs"
extern int __pyx_module_is_main_cyjs___cyjs;
int __pyx_module_is_main_cyjs___cyjs = 0;

/* Implementation of "cyjs._cyjs" */
/* #### Code section: global_var ### */
/* #### Code section: string_decls ### */
static const char __pyx_k__2[] = "";
static const char __pyx_k_input[] = "<input>";
static const char __pyx_k_stack[] = "stack";
static const char __pyx_k_func_rt[] = "func, rt";
static const char __pyx_k_Unable_to_create_exception_from[] = "Unable to create exception from a string object";
static const char __pyx_k_array_count_atom_count_atom_size[] = "array_count, atom_count, atom_size, binary_object_count, binary_object_size, fast_array_count, fast_array_elements, js_func_code_size, js_func_count, js_func_pc2line_count, js_func_size, malloc_count, malloc_limit, malloc_size, memory_used_count, memory_used_size, obj_count, obj_size, prop_count, prop_size, shape_count, shape_size, str_count, str_size";
/* #### Code section: decls ### */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7JSError___reduce_cython__(struct __pyx_obj_4cyjs_5_cyjs_JSError *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7JSError_2__setstate_cython__(struct __pyx_obj_4cyjs_5_cyjs_JSError *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_11malloc_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_12malloc_limit___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_16memory_used_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_12malloc_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_17memory_used_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_10atom_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_9atom_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_9str_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_8str_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_9obj_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_8obj_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_10prop_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_9prop_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_11shape_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_10shape_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_13js_func_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_12js_func_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_17js_func_code_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_21js_func_pc2line_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_11array_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_16fast_array_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_19fast_array_elements___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_19binary_object_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_18binary_object_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage___reduce_cython__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_2__setstate_cython__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11PromiseHook___reduce_cython__(struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_11PromiseHook_2__setstate_cython__(struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_4cyjs_5_cyjs_7Runtime___init__(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_2compute_memory_usage(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_4dump_memory_usage(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, PyObject *__pyx_v_file); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_6execute_pending_job(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_8is_job_pending(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_10run_gc(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_12set_memory_limit(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, size_t __pyx_v_limit); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_14set_max_stack_size(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, size_t __pyx_v_max_stack_size); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_16update_statck_top(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_18set_promise_hook(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static void __pyx_pf_4cyjs_5_cyjs_7Runtime_20__dealloc__(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_22__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_24__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_4cyjs_5_cyjs_6_OView___init__(struct __pyx_obj_4cyjs_5_cyjs__OView *__pyx_v_self, struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_obj); /* proto */
static void __pyx_pf_4cyjs_5_cyjs_6_OView_2__dealloc__(struct __pyx_obj_4cyjs_5_cyjs__OView *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_4cyjs_5_cyjs_6_OView_4__len__(struct __pyx_obj_4cyjs_5_cyjs__OView *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6_OView_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs__OView *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6_OView_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs__OView *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_16_ObjectItemsView___iter__(struct __pyx_obj_4cyjs_5_cyjs__ObjectItemsView *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_16_ObjectItemsView_3__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs__ObjectItemsView *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_16_ObjectItemsView_5__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs__ObjectItemsView *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_15_ObjectKeysView___iter__(struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView *__pyx_v_self); /* proto */
static int __pyx_pf_4cyjs_5_cyjs_15_ObjectKeysView_3__contains__(struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_15_ObjectKeysView_5__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_15_ObjectKeysView_7__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_17_ObjectValuesView___iter__(struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView *__pyx_v_self); /* proto */
static int __pyx_pf_4cyjs_5_cyjs_17_ObjectValuesView_3__contains__(struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_17_ObjectValuesView_5__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_17_ObjectValuesView_7__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_to_json(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_2get(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_4set(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_6__call__(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_8invoke(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_args); /* proto */
static void __pyx_pf_4cyjs_5_cyjs_6Object_10__dealloc__(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_12items(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_14values(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_16keys(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_18eval(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_filename, int __pyx_v_strict, int __pyx_v_backtrace_barrier, int __pyx_v_promise); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_20eval_module(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_filename, int __pyx_v_strict, int __pyx_v_backtrace_barrier, int __pyx_v_promise); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_7context___get__(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_3tag___get__(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_22__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_24__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_pf_4cyjs_5_cyjs_10JSFunction___dealloc__(struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_10JSFunction_2__call__(struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwds); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_10JSFunction_6object___get__(struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_10JSFunction_7context___get__(struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_10JSFunction_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_10JSFunction_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_4cyjs_5_cyjs_7Context___init__(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_runtime, int __pyx_v_base_objects, int __pyx_v_date, int __pyx_v_intrinsic_eval, int __pyx_v_regexp_compiler, int __pyx_v_regexp, int __pyx_v_json, int __pyx_v_proxy, int __pyx_v_map_set, int __pyx_v_typed_arrays, int __pyx_v_bigint, int __pyx_v_weak_ref, int __pyx_v_performance, int __pyx_v_dom_exception, int __pyx_v_promise); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_2eval(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_filename, int __pyx_v_strict, int __pyx_v_backtrace_barrier, int __pyx_v_promise); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_4eval_module(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_filename, int __pyx_v_strict, int __pyx_v_backtrace_barrier, int __pyx_v_promise); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_6get_global(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_8json_parse(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_json); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_10get(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_12set(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_item); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_14add_function(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_name, int __pyx_v_count, int __pyx_v_magic); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_16eval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, PyObject *__pyx_v_filename, int __pyx_v_strict, int __pyx_v_backtrace_barrier, int __pyx_v_promise); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_18eval_this_with_module(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, PyObject *__pyx_v_filename, int __pyx_v_strict, int __pyx_v_backtrace_barrier, int __pyx_v_promise); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_7runtime___get__(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_20__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_22__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_14CancelledError___reduce_cython__(struct __pyx_obj_4cyjs_5_cyjs_CancelledError *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_14CancelledError_2__setstate_cython__(struct __pyx_obj_4cyjs_5_cyjs_CancelledError *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_17InvalidStateError___reduce_cython__(struct __pyx_obj_4cyjs_5_cyjs_InvalidStateError *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_17InvalidStateError_2__setstate_cython__(struct __pyx_obj_4cyjs_5_cyjs_InvalidStateError *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Promise_add_done_callback(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, PyObject *__pyx_v_fn); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Promise_2exception(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Promise_4done(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Promise_6remove_done_callback(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, PyObject *__pyx_v_fn); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Promise_8result(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Promise_10poll(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Promise_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_7Promise_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_9NODEFAULT___reduce_cython__(struct __pyx_obj_4cyjs_5_cyjs_NODEFAULT *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_9NODEFAULT_2__setstate_cython__(struct __pyx_obj_4cyjs_5_cyjs_NODEFAULT *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs___pyx_unpickle_JSError(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_2__pyx_unpickle_MemoryUsage(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_4__pyx_unpickle_PromiseHook(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_6__pyx_unpickle_CancelledError(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_8__pyx_unpickle_InvalidStateError(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4cyjs_5_cyjs_10__pyx_unpickle_NODEFAULT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_4cyjs_5_cyjs_JSError(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs_MemoryUsage(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs_PromiseHook(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs_Runtime(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs_Object(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs_JSFunction(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs_Context(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs__OView(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs__ObjectItemsView(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs__ObjectKeysView(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs__ObjectValuesView(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs_CancelledError(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs_InvalidStateError(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs_Promise(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs_NODEFAULT(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs___pyx_scope_struct____iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs___pyx_scope_struct_1___iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4cyjs_5_cyjs___pyx_scope_struct_2___iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  PyTypeObject *__pyx_ptype_7cpython_4bool_bool;
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  PyObject *__pyx_type_4cyjs_5_cyjs_JSError;
  PyObject *__pyx_type_4cyjs_5_cyjs_MemoryUsage;
  PyObject *__pyx_type_4cyjs_5_cyjs_PromiseHook;
  PyObject *__pyx_type_4cyjs_5_cyjs_Runtime;
  PyObject *__pyx_type_4cyjs_5_cyjs_Object;
  PyObject *__pyx_type_4cyjs_5_cyjs_JSFunction;
  PyObject *__pyx_type_4cyjs_5_cyjs_Context;
  PyObject *__pyx_type_4cyjs_5_cyjs__OView;
  PyObject *__pyx_type_4cyjs_5_cyjs__ObjectItemsView;
  PyObject *__pyx_type_4cyjs_5_cyjs__ObjectKeysView;
  PyObject *__pyx_type_4cyjs_5_cyjs__ObjectValuesView;
  PyObject *__pyx_type_4cyjs_5_cyjs_CancelledError;
  PyObject *__pyx_type_4cyjs_5_cyjs_InvalidStateError;
  PyObject *__pyx_type_4cyjs_5_cyjs_Promise;
  PyObject *__pyx_type_4cyjs_5_cyjs_NODEFAULT;
  PyObject *__pyx_type_4cyjs_5_cyjs___pyx_scope_struct____iter__;
  PyObject *__pyx_type_4cyjs_5_cyjs___pyx_scope_struct_1___iter__;
  PyObject *__pyx_type_4cyjs_5_cyjs___pyx_scope_struct_2___iter__;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs_JSError;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs_MemoryUsage;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs_PromiseHook;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs_Runtime;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs_Object;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs_JSFunction;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs_Context;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs__OView;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs_CancelledError;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs_InvalidStateError;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs_Promise;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs_NODEFAULT;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct____iter__;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_1___iter__;
  PyTypeObject *__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_2___iter__;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values;
  struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_k_;
  PyObject *__pyx_tuple[3];
  PyObject *__pyx_codeobj_tab[72];
  PyObject *__pyx_string_tab[289];
  PyObject *__pyx_number_tab[5];
/* #### Code section: module_state_contents ### */

#if CYTHON_USE_FREELISTS
struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__ *__pyx_freelist_4cyjs_5_cyjs___pyx_scope_struct____iter__[8];
int __pyx_freecount_4cyjs_5_cyjs___pyx_scope_struct____iter__;
#endif

#if CYTHON_USE_FREELISTS
struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ *__pyx_freelist_4cyjs_5_cyjs___pyx_scope_struct_1___iter__[8];
int __pyx_freecount_4cyjs_5_cyjs___pyx_scope_struct_1___iter__;
#endif

#if CYTHON_USE_FREELISTS
struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ *__pyx_freelist_4cyjs_5_cyjs___pyx_scope_struct_2___iter__[8];
int __pyx_freecount_4cyjs_5_cyjs___pyx_scope_struct_2___iter__;
#endif
/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CythonFunctionShared.module_state_decls */
PyTypeObject *__pyx_CyFunctionType;

/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* IterNextPlain.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
PyObject *__Pyx_GetBuiltinNext_LimitedAPI_cache;
#endif

/* Generator.module_state_decls */
PyTypeObject *__pyx_GeneratorType;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_Could_not_set __pyx_string_tab[0]
#define __pyx_kp_u_Failed_setting_the_variable __pyx_string_tab[1]
#define __pyx_kp_u_Failed_to_convert_python_value_t __pyx_string_tab[2]
#define __pyx_kp_u_Note_that_Cython_is_deliberately __pyx_string_tab[3]
#define __pyx_kp_u_Promise_hook_must_be_callable __pyx_string_tab[4]
#define __pyx_kp_u_Result_is_not_ready __pyx_string_tab[5]
#define __pyx_kp_u__3 __pyx_string_tab[6]
#define __pyx_kp_u__4 __pyx_string_tab[7]
#define __pyx_kp_u_add_note __pyx_string_tab[8]
#define __pyx_kp_u_cyjs__cyjs __pyx_string_tab[9]
#define __pyx_kp_u_cyjs__cyjs_pyx __pyx_string_tab[10]
#define __pyx_kp_u_disable __pyx_string_tab[11]
#define __pyx_kp_u_done_callbacks_must_be_callable __pyx_string_tab[12]
#define __pyx_kp_u_enable __pyx_string_tab[13]
#define __pyx_kp_u_failed_to_get_object_s_Property __pyx_string_tab[14]
#define __pyx_kp_u_function_must_be_callable __pyx_string_tab[15]
#define __pyx_kp_u_gc __pyx_string_tab[16]
#define __pyx_kp_u_isenabled __pyx_string_tab[17]
#define __pyx_kp_u_raised_exception_without_it_bein __pyx_string_tab[18]
#define __pyx_kp_u_self_ctx_self_rt_cannot_be_conve __pyx_string_tab[19]
#define __pyx_kp_u_self_ctx_self_tab_self_value_can __pyx_string_tab[20]
#define __pyx_kp_u_self_ctx_self_value_cannot_be_co __pyx_string_tab[21]
#define __pyx_kp_u_self_rt_cannot_be_converted_to_a __pyx_string_tab[22]
#define __pyx_kp_u_stringsource __pyx_string_tab[23]
#define __pyx_kp_u_to_Object __pyx_string_tab[24]
#define __pyx_kp_u_unknown_object __pyx_string_tab[25]
#define __pyx_n_u_AFTER __pyx_string_tab[26]
#define __pyx_n_u_BEFORE __pyx_string_tab[27]
#define __pyx_n_u_CancelledError __pyx_string_tab[28]
#define __pyx_n_u_CancelledError___reduce_cython __pyx_string_tab[29]
#define __pyx_n_u_CancelledError___setstate_cython __pyx_string_tab[30]
#define __pyx_n_u_Context __pyx_string_tab[31]
#define __pyx_n_u_Context___reduce_cython __pyx_string_tab[32]
#define __pyx_n_u_Context___setstate_cython __pyx_string_tab[33]
#define __pyx_n_u_Context_add_function __pyx_string_tab[34]
#define __pyx_n_u_Context_eval __pyx_string_tab[35]
#define __pyx_n_u_Context_eval_module __pyx_string_tab[36]
#define __pyx_n_u_Context_eval_this __pyx_string_tab[37]
#define __pyx_n_u_Context_eval_this_with_module __pyx_string_tab[38]
#define __pyx_n_u_Context_get __pyx_string_tab[39]
#define __pyx_n_u_Context_get_global __pyx_string_tab[40]
#define __pyx_n_u_Context_json_parse __pyx_string_tab[41]
#define __pyx_n_u_Context_set __pyx_string_tab[42]
#define __pyx_n_u_INIT __pyx_string_tab[43]
#define __pyx_n_u_IntEnum __pyx_string_tab[44]
#define __pyx_n_u_IntFlag __pyx_string_tab[45]
#define __pyx_n_u_InvalidStateError __pyx_string_tab[46]
#define __pyx_n_u_InvalidStateError___reduce_cytho __pyx_string_tab[47]
#define __pyx_n_u_InvalidStateError___setstate_cyt __pyx_string_tab[48]
#define __pyx_n_u_JSError __pyx_string_tab[49]
#define __pyx_n_u_JSError___reduce_cython __pyx_string_tab[50]
#define __pyx_n_u_JSError___setstate_cython __pyx_string_tab[51]
#define __pyx_n_u_JSFunction __pyx_string_tab[52]
#define __pyx_n_u_JSFunction___reduce_cython __pyx_string_tab[53]
#define __pyx_n_u_JSFunction___setstate_cython __pyx_string_tab[54]
#define __pyx_n_u_MemoryUsage __pyx_string_tab[55]
#define __pyx_n_u_MemoryUsage___reduce_cython __pyx_string_tab[56]
#define __pyx_n_u_MemoryUsage___setstate_cython __pyx_string_tab[57]
#define __pyx_n_u_NODEFAULT___reduce_cython __pyx_string_tab[58]
#define __pyx_n_u_NODEFAULT___setstate_cython __pyx_string_tab[59]
#define __pyx_n_u_OView __pyx_string_tab[60]
#define __pyx_n_u_OView___reduce_cython __pyx_string_tab[61]
#define __pyx_n_u_OView___setstate_cython __pyx_string_tab[62]
#define __pyx_n_u_Object __pyx_string_tab[63]
#define __pyx_n_u_ObjectItemsView __pyx_string_tab[64]
#define __pyx_n_u_ObjectItemsView___iter __pyx_string_tab[65]
#define __pyx_n_u_ObjectItemsView___reduce_cython __pyx_string_tab[66]
#define __pyx_n_u_ObjectItemsView___setstate_cyth __pyx_string_tab[67]
#define __pyx_n_u_ObjectKeysView __pyx_string_tab[68]
#define __pyx_n_u_ObjectKeysView___iter __pyx_string_tab[69]
#define __pyx_n_u_ObjectKeysView___reduce_cython __pyx_string_tab[70]
#define __pyx_n_u_ObjectKeysView___setstate_cytho __pyx_string_tab[71]
#define __pyx_n_u_ObjectValuesView __pyx_string_tab[72]
#define __pyx_n_u_ObjectValuesView___iter __pyx_string_tab[73]
#define __pyx_n_u_ObjectValuesView___reduce_cytho __pyx_string_tab[74]
#define __pyx_n_u_ObjectValuesView___setstate_cyt __pyx_string_tab[75]
#define __pyx_n_u_Object___reduce_cython __pyx_string_tab[76]
#define __pyx_n_u_Object___setstate_cython __pyx_string_tab[77]
#define __pyx_n_u_Object_eval __pyx_string_tab[78]
#define __pyx_n_u_Object_eval_module __pyx_string_tab[79]
#define __pyx_n_u_Object_get __pyx_string_tab[80]
#define __pyx_n_u_Object_invoke __pyx_string_tab[81]
#define __pyx_n_u_Object_items __pyx_string_tab[82]
#define __pyx_n_u_Object_keys __pyx_string_tab[83]
#define __pyx_n_u_Object_set __pyx_string_tab[84]
#define __pyx_n_u_Object_to_json __pyx_string_tab[85]
#define __pyx_n_u_Object_values __pyx_string_tab[86]
#define __pyx_n_u_Promise __pyx_string_tab[87]
#define __pyx_n_u_PromiseHookType __pyx_string_tab[88]
#define __pyx_n_u_PromiseHook___reduce_cython __pyx_string_tab[89]
#define __pyx_n_u_PromiseHook___setstate_cython __pyx_string_tab[90]
#define __pyx_n_u_Promise___reduce_cython __pyx_string_tab[91]
#define __pyx_n_u_Promise___setstate_cython __pyx_string_tab[92]
#define __pyx_n_u_Promise_add_done_callback __pyx_string_tab[93]
#define __pyx_n_u_Promise_done __pyx_string_tab[94]
#define __pyx_n_u_Promise_exception __pyx_string_tab[95]
#define __pyx_n_u_Promise_poll __pyx_string_tab[96]
#define __pyx_n_u_Promise_remove_done_callback __pyx_string_tab[97]
#define __pyx_n_u_Promise_result __pyx_string_tab[98]
#define __pyx_n_u_Pyx_PyDict_NextRef __pyx_string_tab[99]
#define __pyx_n_u_RESOLVE __pyx_string_tab[100]
#define __pyx_n_u_Runtime __pyx_string_tab[101]
#define __pyx_n_u_Runtime___reduce_cython __pyx_string_tab[102]
#define __pyx_n_u_Runtime___setstate_cython __pyx_string_tab[103]
#define __pyx_n_u_Runtime_compute_memory_usage __pyx_string_tab[104]
#define __pyx_n_u_Runtime_dump_memory_usage __pyx_string_tab[105]
#define __pyx_n_u_Runtime_execute_pending_job __pyx_string_tab[106]
#define __pyx_n_u_Runtime_is_job_pending __pyx_string_tab[107]
#define __pyx_n_u_Runtime_run_gc __pyx_string_tab[108]
#define __pyx_n_u_Runtime_set_max_stack_size __pyx_string_tab[109]
#define __pyx_n_u_Runtime_set_memory_limit __pyx_string_tab[110]
#define __pyx_n_u_Runtime_set_promise_hook __pyx_string_tab[111]
#define __pyx_n_u_Runtime_update_statck_top __pyx_string_tab[112]
#define __pyx_n_u_add_done_callback __pyx_string_tab[113]
#define __pyx_n_u_add_function __pyx_string_tab[114]
#define __pyx_n_u_annotations __pyx_string_tab[115]
#define __pyx_n_u_args __pyx_string_tab[116]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[117]
#define __pyx_n_u_atom __pyx_string_tab[118]
#define __pyx_n_u_backtrace_barrier __pyx_string_tab[119]
#define __pyx_n_u_base_objects __pyx_string_tab[120]
#define __pyx_n_u_bigint __pyx_string_tab[121]
#define __pyx_n_u_call __pyx_string_tab[122]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[123]
#define __pyx_n_u_close __pyx_string_tab[124]
#define __pyx_n_u_code __pyx_string_tab[125]
#define __pyx_n_u_compute_memory_usage __pyx_string_tab[126]
#define __pyx_n_u_count __pyx_string_tab[127]
#define __pyx_n_u_ctx __pyx_string_tab[128]
#define __pyx_n_u_cyjs__cyjs __pyx_string_tab[129]
#define __pyx_n_u_date __pyx_string_tab[130]
#define __pyx_n_u_dict __pyx_string_tab[131]
#define __pyx_n_u_dict_2 __pyx_string_tab[132]
#define __pyx_n_u_dom_exception __pyx_string_tab[133]
#define __pyx_n_u_done __pyx_string_tab[134]
#define __pyx_n_u_dump_memory_usage __pyx_string_tab[135]
#define __pyx_n_u_dumps __pyx_string_tab[136]
#define __pyx_n_u_enum __pyx_string_tab[137]
#define __pyx_n_u_eval __pyx_string_tab[138]
#define __pyx_n_u_eval_module __pyx_string_tab[139]
#define __pyx_n_u_eval_this __pyx_string_tab[140]
#define __pyx_n_u_eval_this_with_module __pyx_string_tab[141]
#define __pyx_n_u_exception __pyx_string_tab[142]
#define __pyx_n_u_execute_pending_job __pyx_string_tab[143]
#define __pyx_n_u_file __pyx_string_tab[144]
#define __pyx_n_u_filename __pyx_string_tab[145]
#define __pyx_n_u_fn __pyx_string_tab[146]
#define __pyx_n_u_func __pyx_string_tab[147]
#define __pyx_n_u_func_2 __pyx_string_tab[148]
#define __pyx_n_u_get __pyx_string_tab[149]
#define __pyx_n_u_get_global __pyx_string_tab[150]
#define __pyx_n_u_getstate __pyx_string_tab[151]
#define __pyx_n_u_glob __pyx_string_tab[152]
#define __pyx_n_u_i __pyx_string_tab[153]
#define __pyx_n_u_intrinsic_eval __pyx_string_tab[154]
#define __pyx_n_u_invoke __pyx_string_tab[155]
#define __pyx_n_u_is_coroutine __pyx_string_tab[156]
#define __pyx_n_u_is_job_pending __pyx_string_tab[157]
#define __pyx_n_u_item __pyx_string_tab[158]
#define __pyx_n_u_items __pyx_string_tab[159]
#define __pyx_n_u_iter __pyx_string_tab[160]
#define __pyx_n_u_js_args __pyx_string_tab[161]
#define __pyx_n_u_js_value __pyx_string_tab[162]
#define __pyx_n_u_json __pyx_string_tab[163]
#define __pyx_n_u_json_parse __pyx_string_tab[164]
#define __pyx_n_u_key __pyx_string_tab[165]
#define __pyx_n_u_keys __pyx_string_tab[166]
#define __pyx_n_u_limit __pyx_string_tab[167]
#define __pyx_n_u_magic __pyx_string_tab[168]
#define __pyx_n_u_main __pyx_string_tab[169]
#define __pyx_n_u_map_set __pyx_string_tab[170]
#define __pyx_n_u_max_stack_size __pyx_string_tab[171]
#define __pyx_n_u_member_names __pyx_string_tab[172]
#define __pyx_n_u_members __pyx_string_tab[173]
#define __pyx_n_u_module __pyx_string_tab[174]
#define __pyx_n_u_module_2 __pyx_string_tab[175]
#define __pyx_n_u_name __pyx_string_tab[176]
#define __pyx_n_u_name_2 __pyx_string_tab[177]
#define __pyx_n_u_new __pyx_string_tab[178]
#define __pyx_n_u_next __pyx_string_tab[179]
#define __pyx_n_u_obj __pyx_string_tab[180]
#define __pyx_n_u_p __pyx_string_tab[181]
#define __pyx_n_u_performance __pyx_string_tab[182]
#define __pyx_n_u_poll __pyx_string_tab[183]
#define __pyx_n_u_pop __pyx_string_tab[184]
#define __pyx_n_u_pos __pyx_string_tab[185]
#define __pyx_n_u_promise __pyx_string_tab[186]
#define __pyx_n_u_proxy __pyx_string_tab[187]
#define __pyx_n_u_pyx_checksum __pyx_string_tab[188]
#define __pyx_n_u_pyx_result __pyx_string_tab[189]
#define __pyx_n_u_pyx_state __pyx_string_tab[190]
#define __pyx_n_u_pyx_type __pyx_string_tab[191]
#define __pyx_n_u_pyx_unpickle_CancelledError __pyx_string_tab[192]
#define __pyx_n_u_pyx_unpickle_InvalidStateError __pyx_string_tab[193]
#define __pyx_n_u_pyx_unpickle_JSError __pyx_string_tab[194]
#define __pyx_n_u_pyx_unpickle_MemoryUsage __pyx_string_tab[195]
#define __pyx_n_u_pyx_unpickle_NODEFAULT __pyx_string_tab[196]
#define __pyx_n_u_pyx_unpickle_PromiseHook __pyx_string_tab[197]
#define __pyx_n_u_pyx_vtable __pyx_string_tab[198]
#define __pyx_n_u_qualname __pyx_string_tab[199]
#define __pyx_n_u_raise_exception __pyx_string_tab[200]
#define __pyx_n_u_reduce __pyx_string_tab[201]
#define __pyx_n_u_reduce_cython __pyx_string_tab[202]
#define __pyx_n_u_reduce_ex __pyx_string_tab[203]
#define __pyx_n_u_regexp __pyx_string_tab[204]
#define __pyx_n_u_regexp_compiler __pyx_string_tab[205]
#define __pyx_n_u_remove_done_callback __pyx_string_tab[206]
#define __pyx_n_u_result __pyx_string_tab[207]
#define __pyx_n_u_ret __pyx_string_tab[208]
#define __pyx_n_u_run_gc __pyx_string_tab[209]
#define __pyx_n_u_runtime __pyx_string_tab[210]
#define __pyx_n_u_self __pyx_string_tab[211]
#define __pyx_n_u_send __pyx_string_tab[212]
#define __pyx_n_u_set __pyx_string_tab[213]
#define __pyx_n_u_set_max_stack_size __pyx_string_tab[214]
#define __pyx_n_u_set_memory_limit __pyx_string_tab[215]
#define __pyx_n_u_set_name __pyx_string_tab[216]
#define __pyx_n_u_set_promise_hook __pyx_string_tab[217]
#define __pyx_n_u_setdefault __pyx_string_tab[218]
#define __pyx_n_u_setstate __pyx_string_tab[219]
#define __pyx_n_u_setstate_cython __pyx_string_tab[220]
#define __pyx_n_u_state __pyx_string_tab[221]
#define __pyx_n_u_strict __pyx_string_tab[222]
#define __pyx_n_u_test __pyx_string_tab[223]
#define __pyx_n_u_this __pyx_string_tab[224]
#define __pyx_n_u_throw __pyx_string_tab[225]
#define __pyx_n_u_to_json __pyx_string_tab[226]
#define __pyx_n_u_typed_arrays __pyx_string_tab[227]
#define __pyx_n_u_update __pyx_string_tab[228]
#define __pyx_n_u_update_statck_top __pyx_string_tab[229]
#define __pyx_n_u_use_setstate __pyx_string_tab[230]
#define __pyx_n_u_v __pyx_string_tab[231]
#define __pyx_n_u_val __pyx_string_tab[232]
#define __pyx_n_u_value __pyx_string_tab[233]
#define __pyx_n_u_values __pyx_string_tab[234]
#define __pyx_n_u_view __pyx_string_tab[235]
#define __pyx_n_u_weak_ref __pyx_string_tab[236]
#define __pyx_kp_b_input __pyx_string_tab[237]
#define __pyx_kp_b_iso88591_A __pyx_string_tab[238]
#define __pyx_kp_b_iso88591_AV1 __pyx_string_tab[239]
#define __pyx_kp_b_iso88591_A_0_4xq_1A_S_A_uCq_2_7_6_1_AV1_n __pyx_string_tab[240]
#define __pyx_kp_b_iso88591_A_2 __pyx_string_tab[241]
#define __pyx_kp_b_iso88591_A_4uA __pyx_string_tab[242]
#define __pyx_kp_b_iso88591_A_4xq_1A_4vT_A_7_1_aq __pyx_string_tab[243]
#define __pyx_kp_b_iso88591_A_4xq_1A_t1F_e_4G4q __pyx_string_tab[244]
#define __pyx_kp_b_iso88591_A_6_l_4v_fD_a_1AQ_y_V1 __pyx_string_tab[245]
#define __pyx_kp_b_iso88591_A_G1F_a_vWE_Q_q_q_q_2_awk_PQ_2 __pyx_string_tab[246]
#define __pyx_kp_b_iso88591_A_G1F_a_vWE_Q_q_q_q_AWKwa_AWKq __pyx_string_tab[247]
#define __pyx_kp_b_iso88591_A_G1F_a_vWE_Q_q_q_q_awk_awk __pyx_string_tab[248]
#define __pyx_kp_b_iso88591_A_G1F_a_vWE_Q_q_q_q_t1G_gQ_t1G __pyx_string_tab[249]
#define __pyx_kp_b_iso88591_A_Q __pyx_string_tab[250]
#define __pyx_kp_b_iso88591_A_Qd_avRq_Qd_b_a_Qd_fG2Q_iq_V1 __pyx_string_tab[251]
#define __pyx_kp_b_iso88591_A_Qd_q_aq __pyx_string_tab[252]
#define __pyx_kp_b_iso88591_A_a __pyx_string_tab[253]
#define __pyx_kp_b_iso88591_A_a_4_a_1D_Qe2Q_A_q_5Qa_z_j_q __pyx_string_tab[254]
#define __pyx_kp_b_iso88591_A_a_Qd_avRq_nAT_vQ_AT_q_y_V1 __pyx_string_tab[255]
#define __pyx_kp_b_iso88591_A_aq __pyx_string_tab[256]
#define __pyx_kp_b_iso88591_A_at1 __pyx_string_tab[257]
#define __pyx_kp_b_iso88591_A_d_Qe5_r_N_5_HA_y_k_q __pyx_string_tab[258]
#define __pyx_kp_b_iso88591_A_d_Qe5_r_Qe1JgRq_t86_2Q_1 __pyx_string_tab[259]
#define __pyx_kp_b_iso88591_A_d_Qe6_q_D_it1_aq_6_1_1E_XV1F_1 __pyx_string_tab[260]
#define __pyx_kp_b_iso88591_A_d_T_4uA_1_1_1D_b_H_A_4uA_1_1 __pyx_string_tab[261]
#define __pyx_kp_b_iso88591_A_e1 __pyx_string_tab[262]
#define __pyx_kp_b_iso88591_A_fD_a_q_F_6_Q_2_6_at6_q __pyx_string_tab[263]
#define __pyx_kp_b_iso88591_A_q __pyx_string_tab[264]
#define __pyx_kp_b_iso88591_A_q_E_r __pyx_string_tab[265]
#define __pyx_kp_b_iso88591_A_q_F_a_6_A_3a_1A_t1 __pyx_string_tab[266]
#define __pyx_kp_b_iso88591_A_q_F_a_6_A_6_A_t_wa_d_4q_1A __pyx_string_tab[267]
#define __pyx_kp_b_iso88591_A_q_F_a_vS_S_c __pyx_string_tab[268]
#define __pyx_kp_b_iso88591_A_r_A_3hd_3a_S_Cq_1_6_q __pyx_string_tab[269]
#define __pyx_kp_b_iso88591_A_t6_1_XXsst __pyx_string_tab[270]
#define __pyx_kp_b_iso88591_A_t6_1_YYttu __pyx_string_tab[271]
#define __pyx_kp_b_iso88591_A_t8_avV_V8K __pyx_string_tab[272]
#define __pyx_kp_b_iso88591_A_t8_avV_WHL __pyx_string_tab[273]
#define __pyx_kp_b_iso88591_A_t_avV_V8CVVW __pyx_string_tab[274]
#define __pyx_kp_b_iso88591_A_t_avV_WHDWWX __pyx_string_tab[275]
#define __pyx_kp_b_iso88591_A_y_V_5Qd __pyx_string_tab[276]
#define __pyx_kp_b_iso88591_Q __pyx_string_tab[277]
#define __pyx_kp_b_iso88591_QfA __pyx_string_tab[278]
#define __pyx_kp_b_iso88591_T_A_G1F_a_vWE_Q_q_t6_S_D_q_q_D __pyx_string_tab[279]
#define __pyx_kp_b_iso88591_T_t_L_DZZ_ssw_x_K_K_O_O_e_e_i_i __pyx_string_tab[280]
#define __pyx_kp_b_iso88591__5 __pyx_string_tab[281]
#define __pyx_kp_b_iso88591_q_0_kQR_7_0_1B_PQ_1 __pyx_string_tab[282]
#define __pyx_kp_b_iso88591_q_0_kQR_7_1_7_N_1 __pyx_string_tab[283]
#define __pyx_kp_b_iso88591_q_0_kQR_9HAQ_7_1L_a_1 __pyx_string_tab[284]
#define __pyx_kp_b_iso88591_q_0_kQR_HAQ_7_314H_VW_1 __pyx_string_tab[285]
#define __pyx_kp_b_iso88591_q_0_kQR_haq_7_QnN_1 __pyx_string_tab[286]
#define __pyx_kp_b_iso88591_q_a __pyx_string_tab[287]
#define __pyx_kp_b_iso88591_q_a_2 __pyx_string_tab[288]
#define __pyx_int_0 __pyx_number_tab[0]
#define __pyx_int_neg_1 __pyx_number_tab[1]
#define __pyx_int_139263547 __pyx_number_tab[2]
#define __pyx_int_238750788 __pyx_number_tab[3]
#define __pyx_int_239704551 __pyx_number_tab[4]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs_JSError);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs_JSError);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs_MemoryUsage);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs_PromiseHook);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs_PromiseHook);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs_Runtime);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs_Runtime);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs_Object);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs_Object);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs_JSFunction);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs_JSFunction);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs_Context);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs_Context);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs__OView);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs__OView);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs__ObjectItemsView);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs__ObjectKeysView);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs__ObjectValuesView);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs_CancelledError);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs_CancelledError);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs_InvalidStateError);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs_Promise);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs_Promise);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs_NODEFAULT);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs_NODEFAULT);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct____iter__);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs___pyx_scope_struct____iter__);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_1___iter__);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs___pyx_scope_struct_1___iter__);
  Py_CLEAR(clear_module_state->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_2___iter__);
  Py_CLEAR(clear_module_state->__pyx_type_4cyjs_5_cyjs___pyx_scope_struct_2___iter__);
  Py_CLEAR(clear_module_state->__pyx_k_);
  for (int i=0; i<3; ++i) { Py_CLEAR(clear_module_state->__pyx_tuple[i]); }
  for (int i=0; i<72; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<289; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<5; ++i) { Py_CLEAR(clear_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* CommonTypesMetaclass.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CyFunctionType);

/* Generator.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_GeneratorType);

/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs_JSError);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs_JSError);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs_MemoryUsage);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs_PromiseHook);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs_PromiseHook);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs_Runtime);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs_Runtime);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs_Object);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs_Object);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs_JSFunction);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs_JSFunction);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs_Context);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs_Context);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs__OView);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs__OView);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs__ObjectItemsView);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs__ObjectKeysView);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs__ObjectValuesView);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs_CancelledError);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs_CancelledError);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs_InvalidStateError);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs_Promise);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs_Promise);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs_NODEFAULT);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs_NODEFAULT);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct____iter__);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs___pyx_scope_struct____iter__);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_1___iter__);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs___pyx_scope_struct_1___iter__);
  Py_VISIT(traverse_module_state->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_2___iter__);
  Py_VISIT(traverse_module_state->__pyx_type_4cyjs_5_cyjs___pyx_scope_struct_2___iter__);
  Py_VISIT(traverse_module_state->__pyx_k_);
  for (int i=0; i<3; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_tuple[i]); }
  for (int i=0; i<72; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<289; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<5; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* CommonTypesMetaclass.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CyFunctionType);

/* Generator.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_GeneratorType);

/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */

/* "EnumTypeToPy":2
 * 
 * @cname("__Pyx_Enum_4cyjs_5_cyjs_enum__dunderpyx_t_4cyjs_5_cyjs_PromiseHookType_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_4cyjs_5_cyjs_enum__dunderpyx_t_4cyjs_5_cyjs_PromiseHookType_to_py(PromiseHookType c_val):
 *     cdef object __pyx_enum
*/

static PyObject *__Pyx_Enum_4cyjs_5_cyjs_enum__dunderpyx_t_4cyjs_5_cyjs_PromiseHookType_to_py(enum __pyx_t_4cyjs_5_cyjs_PromiseHookType __pyx_v_c_val) {
  PyObject *__pyx_v___pyx_enum = 0;
  int __pyx_v_underlying_c_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_Enum_4cyjs_5_cyjs_enum__dunderpyx_t_4cyjs_5_cyjs_PromiseHookType_to_py", 0);

  /* "EnumTypeToPy":5
 * cdef __Pyx_Enum_4cyjs_5_cyjs_enum__dunderpyx_t_4cyjs_5_cyjs_PromiseHookType_to_py(PromiseHookType c_val):
 *     cdef object __pyx_enum
 *     __pyx_enum = PromiseHookType             # <<<<<<<<<<<<<<
 * 
 *     if 0:
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_PromiseHookType); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v___pyx_enum = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "EnumTypeToPy":7
 *     __pyx_enum = PromiseHookType
 * 
 *     if 0:             # <<<<<<<<<<<<<<
 *         pass
 *     elif c_val == PromiseHookType.INIT:
*/
  switch (__pyx_v_c_val) {
    case __pyx_e_4cyjs_5_cyjs_INIT:

    /* "EnumTypeToPy":10
 *         pass
 *     elif c_val == PromiseHookType.INIT:
 *         return __pyx_enum.INIT             # <<<<<<<<<<<<<<
 *     elif c_val == PromiseHookType.BEFORE:
 *         return __pyx_enum.BEFORE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_INIT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":9
 *     if 0:
 *         pass
 *     elif c_val == PromiseHookType.INIT:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.INIT
 *     elif c_val == PromiseHookType.BEFORE:
*/
    break;
    case __pyx_e_4cyjs_5_cyjs_BEFORE:

    /* "EnumTypeToPy":12
 *         return __pyx_enum.INIT
 *     elif c_val == PromiseHookType.BEFORE:
 *         return __pyx_enum.BEFORE             # <<<<<<<<<<<<<<
 *     elif c_val == PromiseHookType.AFTER:
 *         return __pyx_enum.AFTER
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_BEFORE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":11
 *     elif c_val == PromiseHookType.INIT:
 *         return __pyx_enum.INIT
 *     elif c_val == PromiseHookType.BEFORE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.BEFORE
 *     elif c_val == PromiseHookType.AFTER:
*/
    break;
    case __pyx_e_4cyjs_5_cyjs_AFTER:

    /* "EnumTypeToPy":14
 *         return __pyx_enum.BEFORE
 *     elif c_val == PromiseHookType.AFTER:
 *         return __pyx_enum.AFTER             # <<<<<<<<<<<<<<
 *     elif c_val == PromiseHookType.RESOLVE:
 *         return __pyx_enum.RESOLVE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_AFTER); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":13
 *     elif c_val == PromiseHookType.BEFORE:
 *         return __pyx_enum.BEFORE
 *     elif c_val == PromiseHookType.AFTER:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.AFTER
 *     elif c_val == PromiseHookType.RESOLVE:
*/
    break;
    case __pyx_e_4cyjs_5_cyjs_RESOLVE:

    /* "EnumTypeToPy":16
 *         return __pyx_enum.AFTER
 *     elif c_val == PromiseHookType.RESOLVE:
 *         return __pyx_enum.RESOLVE             # <<<<<<<<<<<<<<
 *     else:
 *         underlying_c_val = <int>c_val
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RESOLVE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 16, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":15
 *     elif c_val == PromiseHookType.AFTER:
 *         return __pyx_enum.AFTER
 *     elif c_val == PromiseHookType.RESOLVE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RESOLVE
 *     else:
*/
    break;
    default:

    /* "EnumTypeToPy":18
 *         return __pyx_enum.RESOLVE
 *     else:
 *         underlying_c_val = <int>c_val             # <<<<<<<<<<<<<<
 *         return __pyx_enum(underlying_c_val)
*/
    __pyx_v_underlying_c_val = ((int)__pyx_v_c_val);

    /* "EnumTypeToPy":19
 *     else:
 *         underlying_c_val = <int>c_val
 *         return __pyx_enum(underlying_c_val)             # <<<<<<<<<<<<<<
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = NULL;
    __Pyx_INCREF(__pyx_v___pyx_enum);
    __pyx_t_3 = __pyx_v___pyx_enum; 
    __pyx_t_4 = __Pyx_PyLong_From_int(__pyx_v_underlying_c_val); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 19, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 19, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
    break;
  }

  /* "EnumTypeToPy":2
 * 
 * @cname("__Pyx_Enum_4cyjs_5_cyjs_enum__dunderpyx_t_4cyjs_5_cyjs_PromiseHookType_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_4cyjs_5_cyjs_enum__dunderpyx_t_4cyjs_5_cyjs_PromiseHookType_to_py(PromiseHookType c_val):
 *     cdef object __pyx_enum
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_4cyjs_5_cyjs_enum__dunderpyx_t_4cyjs_5_cyjs_PromiseHookType_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_enum);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":225
 *     # inlines are in fact correct even though cyright says otherwise.
 * 
 *     cdef inline bint has_exception(self): # type: ignore             # <<<<<<<<<<<<<<
 *         return JS_HasException(self.ctx)
 * 
*/

static CYTHON_INLINE int __pyx_f_4cyjs_5_cyjs_7Context_has_exception(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self) {
  int __pyx_r;

  /* "cyjs/_cyjs.pxd":226
 * 
 *     cdef inline bint has_exception(self): # type: ignore
 *         return JS_HasException(self.ctx)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline JSValue get_exception(self): # type: ignore
*/
  __pyx_r = JS_HasException(__pyx_v_self->ctx);
  goto __pyx_L0;

  /* "cyjs/_cyjs.pxd":225
 *     # inlines are in fact correct even though cyright says otherwise.
 * 
 *     cdef inline bint has_exception(self): # type: ignore             # <<<<<<<<<<<<<<
 *         return JS_HasException(self.ctx)
 * 
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":228
 *         return JS_HasException(self.ctx)
 * 
 *     cdef inline JSValue get_exception(self): # type: ignore             # <<<<<<<<<<<<<<
 *         return JS_GetException(self.ctx)
 * 
*/

static CYTHON_INLINE JSValue __pyx_f_4cyjs_5_cyjs_7Context_get_exception(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self) {
  JSValue __pyx_r;

  /* "cyjs/_cyjs.pxd":229
 * 
 *     cdef inline JSValue get_exception(self): # type: ignore
 *         return JS_GetException(self.ctx)             # <<<<<<<<<<<<<<
 * 
 *     cdef object raise_exception(self)
*/
  __pyx_r = JS_GetException(__pyx_v_self->ctx);
  goto __pyx_L0;

  /* "cyjs/_cyjs.pxd":228
 *         return JS_HasException(self.ctx)
 * 
 *     cdef inline JSValue get_exception(self): # type: ignore             # <<<<<<<<<<<<<<
 *         return JS_GetException(self.ctx)
 * 
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":36
 *     """Represents Numerous Exceptions raised from CYJS"""
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef JSError new(JSContext* ctx, JSValue value):
 *         cdef JSError err
*/

static struct __pyx_obj_4cyjs_5_cyjs_JSError *__pyx_f_4cyjs_5_cyjs_7JSError_new(JSContext *__pyx_v_ctx, JSValue __pyx_v_value) {
  struct __pyx_obj_4cyjs_5_cyjs_JSError *__pyx_v_err = 0;
  char const *__pyx_v_cstring;
  char const *__pyx_v_stack_cstring;
  JSValue __pyx_v_stack;
  struct __pyx_obj_4cyjs_5_cyjs_JSError *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 0);

  /* "cyjs/_cyjs.pyx":39
 *     cdef JSError new(JSContext* ctx, JSValue value):
 *         cdef JSError err
 *         cdef const char* cstring = JS_ToCString(ctx, value)             # <<<<<<<<<<<<<<
 *         cdef const char* stack_cstring = NULL
 *         cdef JSValue stack = JS_NULL
*/
  __pyx_v_cstring = JS_ToCString(__pyx_v_ctx, __pyx_v_value);

  /* "cyjs/_cyjs.pyx":40
 *         cdef JSError err
 *         cdef const char* cstring = JS_ToCString(ctx, value)
 *         cdef const char* stack_cstring = NULL             # <<<<<<<<<<<<<<
 *         cdef JSValue stack = JS_NULL
 * 
*/
  __pyx_v_stack_cstring = NULL;

  /* "cyjs/_cyjs.pyx":41
 *         cdef const char* cstring = JS_ToCString(ctx, value)
 *         cdef const char* stack_cstring = NULL
 *         cdef JSValue stack = JS_NULL             # <<<<<<<<<<<<<<
 * 
 *         if not (JS_IsNull(value) or JS_IsUndefined(value)):
*/
  __pyx_v_stack = JS_NULL;

  /* "cyjs/_cyjs.pyx":43
 *         cdef JSValue stack = JS_NULL
 * 
 *         if not (JS_IsNull(value) or JS_IsUndefined(value)):             # <<<<<<<<<<<<<<
 *             stack = JS_GetPropertyStr(ctx, value, "stack")
 *             if not JS_IsException(stack):
*/
  __pyx_t_2 = JS_IsNull(__pyx_v_value);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = JS_IsUndefined(__pyx_v_value);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":44
 * 
 *         if not (JS_IsNull(value) or JS_IsUndefined(value)):
 *             stack = JS_GetPropertyStr(ctx, value, "stack")             # <<<<<<<<<<<<<<
 *             if not JS_IsException(stack):
 *                 stack_cstring = JS_ToCString(ctx, stack)
*/
    __pyx_v_stack = JS_GetPropertyStr(__pyx_v_ctx, __pyx_v_value, __pyx_k_stack);

    /* "cyjs/_cyjs.pyx":45
 *         if not (JS_IsNull(value) or JS_IsUndefined(value)):
 *             stack = JS_GetPropertyStr(ctx, value, "stack")
 *             if not JS_IsException(stack):             # <<<<<<<<<<<<<<
 *                 stack_cstring = JS_ToCString(ctx, stack)
 * 
*/
    __pyx_t_2 = (!JS_IsException(__pyx_v_stack));
    if (__pyx_t_2) {

      /* "cyjs/_cyjs.pyx":46
 *             stack = JS_GetPropertyStr(ctx, value, "stack")
 *             if not JS_IsException(stack):
 *                 stack_cstring = JS_ToCString(ctx, stack)             # <<<<<<<<<<<<<<
 * 
 *         err = JSError(
*/
      __pyx_v_stack_cstring = JS_ToCString(__pyx_v_ctx, __pyx_v_stack);

      /* "cyjs/_cyjs.pyx":45
 *         if not (JS_IsNull(value) or JS_IsUndefined(value)):
 *             stack = JS_GetPropertyStr(ctx, value, "stack")
 *             if not JS_IsException(stack):             # <<<<<<<<<<<<<<
 *                 stack_cstring = JS_ToCString(ctx, stack)
 * 
*/
    }

    /* "cyjs/_cyjs.pyx":43
 *         cdef JSValue stack = JS_NULL
 * 
 *         if not (JS_IsNull(value) or JS_IsUndefined(value)):             # <<<<<<<<<<<<<<
 *             stack = JS_GetPropertyStr(ctx, value, "stack")
 *             if not JS_IsException(stack):
*/
  }

  /* "cyjs/_cyjs.pyx":48
 *                 stack_cstring = JS_ToCString(ctx, stack)
 * 
 *         err = JSError(             # <<<<<<<<<<<<<<
 *             PyUnicode_FromString(cstring) if cstring != NULL else None,
 *             PyUnicode_FromString(stack_cstring) if stack_cstring != NULL else None
*/
  __pyx_t_4 = NULL;

  /* "cyjs/_cyjs.pyx":49
 * 
 *         err = JSError(
 *             PyUnicode_FromString(cstring) if cstring != NULL else None,             # <<<<<<<<<<<<<<
 *             PyUnicode_FromString(stack_cstring) if stack_cstring != NULL else None
 *         )
*/
  __pyx_t_2 = (__pyx_v_cstring != NULL);
  if (__pyx_t_2) {
    __pyx_t_6 = PyUnicode_FromString(__pyx_v_cstring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __pyx_t_6;
    __pyx_t_6 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_5 = Py_None;
  }

  /* "cyjs/_cyjs.pyx":50
 *         err = JSError(
 *             PyUnicode_FromString(cstring) if cstring != NULL else None,
 *             PyUnicode_FromString(stack_cstring) if stack_cstring != NULL else None             # <<<<<<<<<<<<<<
 *         )
 *         if cstring != NULL:
*/
  __pyx_t_2 = (__pyx_v_stack_cstring != NULL);
  if (__pyx_t_2) {
    __pyx_t_7 = PyUnicode_FromString(__pyx_v_stack_cstring); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __pyx_t_7;
    __pyx_t_7 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_6 = Py_None;
  }
  __pyx_t_8 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_JSError, __pyx_callargs+__pyx_t_8, (3-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_3);
  }
  __pyx_v_err = ((struct __pyx_obj_4cyjs_5_cyjs_JSError *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "cyjs/_cyjs.pyx":52
 *             PyUnicode_FromString(stack_cstring) if stack_cstring != NULL else None
 *         )
 *         if cstring != NULL:             # <<<<<<<<<<<<<<
 *             JS_FreeCString(ctx, cstring)
 * 
*/
  __pyx_t_2 = (__pyx_v_cstring != NULL);
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":53
 *         )
 *         if cstring != NULL:
 *             JS_FreeCString(ctx, cstring)             # <<<<<<<<<<<<<<
 * 
 *         if stack_cstring != NULL:
*/
    JS_FreeCString(__pyx_v_ctx, __pyx_v_cstring);

    /* "cyjs/_cyjs.pyx":52
 *             PyUnicode_FromString(stack_cstring) if stack_cstring != NULL else None
 *         )
 *         if cstring != NULL:             # <<<<<<<<<<<<<<
 *             JS_FreeCString(ctx, cstring)
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":55
 *             JS_FreeCString(ctx, cstring)
 * 
 *         if stack_cstring != NULL:             # <<<<<<<<<<<<<<
 *             JS_FreeCString(ctx, stack_cstring)
 *             JS_FreeValue(ctx, stack)
*/
  __pyx_t_2 = (__pyx_v_stack_cstring != NULL);
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":56
 * 
 *         if stack_cstring != NULL:
 *             JS_FreeCString(ctx, stack_cstring)             # <<<<<<<<<<<<<<
 *             JS_FreeValue(ctx, stack)
 * 
*/
    JS_FreeCString(__pyx_v_ctx, __pyx_v_stack_cstring);

    /* "cyjs/_cyjs.pyx":57
 *         if stack_cstring != NULL:
 *             JS_FreeCString(ctx, stack_cstring)
 *             JS_FreeValue(ctx, stack)             # <<<<<<<<<<<<<<
 * 
 *         JS_FreeValue(ctx, value)
*/
    JS_FreeValue(__pyx_v_ctx, __pyx_v_stack);

    /* "cyjs/_cyjs.pyx":55
 *             JS_FreeCString(ctx, cstring)
 * 
 *         if stack_cstring != NULL:             # <<<<<<<<<<<<<<
 *             JS_FreeCString(ctx, stack_cstring)
 *             JS_FreeValue(ctx, stack)
*/
  }

  /* "cyjs/_cyjs.pyx":59
 *             JS_FreeValue(ctx, stack)
 * 
 *         JS_FreeValue(ctx, value)             # <<<<<<<<<<<<<<
 *         return err
 * 
*/
  JS_FreeValue(__pyx_v_ctx, __pyx_v_value);

  /* "cyjs/_cyjs.pyx":60
 * 
 *         JS_FreeValue(ctx, value)
 *         return err             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_err);
  __pyx_r = __pyx_v_err;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":36
 *     """Represents Numerous Exceptions raised from CYJS"""
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef JSError new(JSContext* ctx, JSValue value):
 *         cdef JSError err
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cyjs._cyjs.JSError.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_err);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7JSError_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7JSError_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7JSError_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7JSError_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7JSError___reduce_cython__(((struct __pyx_obj_4cyjs_5_cyjs_JSError *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7JSError___reduce_cython__(struct __pyx_obj_4cyjs_5_cyjs_JSError *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_empty_tuple);
  __pyx_v_state = __pyx_mstate_global->__pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(1, 7, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_JSError, (type(self), 0xe3b0c44, None), state
*/
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_JSError, (type(self), 0xe3b0c44, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_JSError, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_JSError, (type(self), 0xe3b0c44, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_JSError); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_238750788) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_JSError, (type(self), 0xe3b0c44, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_JSError, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_JSError, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_JSError__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_JSError); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_238750788) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cyjs._cyjs.JSError.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_JSError, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_JSError__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7JSError_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7JSError_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7JSError_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7JSError_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.JSError.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7JSError_2__setstate_cython__(((struct __pyx_obj_4cyjs_5_cyjs_JSError *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7JSError_2__setstate_cython__(struct __pyx_obj_4cyjs_5_cyjs_JSError *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_JSError, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_JSError__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(1, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_4cyjs_5_cyjs___pyx_unpickle_JSError__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_JSError, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_JSError__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cyjs._cyjs.JSError.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":66
 * cdef class MemoryUsage:
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef MemoryUsage new(const JSMemoryUsage *ptr):
 *         cdef MemoryUsage self = MemoryUsage.__new__(MemoryUsage)
*/

static struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_f_4cyjs_5_cyjs_11MemoryUsage_new(JSMemoryUsage const *__pyx_v_ptr) {
  struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self = 0;
  struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int64_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 0);

  /* "cyjs/_cyjs.pyx":68
 *     @staticmethod
 *     cdef MemoryUsage new(const JSMemoryUsage *ptr):
 *         cdef MemoryUsage self = MemoryUsage.__new__(MemoryUsage)             # <<<<<<<<<<<<<<
 *         self.malloc_size = ptr.malloc_size
 *         self.malloc_limit = ptr.malloc_limit
*/
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_4cyjs_5_cyjs_MemoryUsage(((PyTypeObject *)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage), __pyx_mstate_global->__pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_self = ((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":69
 *     cdef MemoryUsage new(const JSMemoryUsage *ptr):
 *         cdef MemoryUsage self = MemoryUsage.__new__(MemoryUsage)
 *         self.malloc_size = ptr.malloc_size             # <<<<<<<<<<<<<<
 *         self.malloc_limit = ptr.malloc_limit
 *         self.memory_used_size = ptr.memory_used_size
*/
  __pyx_t_2 = __pyx_v_ptr->malloc_size;
  __pyx_v_self->malloc_size = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":70
 *         cdef MemoryUsage self = MemoryUsage.__new__(MemoryUsage)
 *         self.malloc_size = ptr.malloc_size
 *         self.malloc_limit = ptr.malloc_limit             # <<<<<<<<<<<<<<
 *         self.memory_used_size = ptr.memory_used_size
 *         self.malloc_count = ptr.malloc_count
*/
  __pyx_t_2 = __pyx_v_ptr->malloc_limit;
  __pyx_v_self->malloc_limit = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":71
 *         self.malloc_size = ptr.malloc_size
 *         self.malloc_limit = ptr.malloc_limit
 *         self.memory_used_size = ptr.memory_used_size             # <<<<<<<<<<<<<<
 *         self.malloc_count = ptr.malloc_count
 *         self.memory_used_count = ptr.memory_used_count
*/
  __pyx_t_2 = __pyx_v_ptr->memory_used_size;
  __pyx_v_self->memory_used_size = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":72
 *         self.malloc_limit = ptr.malloc_limit
 *         self.memory_used_size = ptr.memory_used_size
 *         self.malloc_count = ptr.malloc_count             # <<<<<<<<<<<<<<
 *         self.memory_used_count = ptr.memory_used_count
 *         self.atom_count = ptr.atom_count
*/
  __pyx_t_2 = __pyx_v_ptr->malloc_count;
  __pyx_v_self->malloc_count = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":73
 *         self.memory_used_size = ptr.memory_used_size
 *         self.malloc_count = ptr.malloc_count
 *         self.memory_used_count = ptr.memory_used_count             # <<<<<<<<<<<<<<
 *         self.atom_count = ptr.atom_count
 *         self.atom_size = ptr.atom_size
*/
  __pyx_t_2 = __pyx_v_ptr->memory_used_count;
  __pyx_v_self->memory_used_count = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":74
 *         self.malloc_count = ptr.malloc_count
 *         self.memory_used_count = ptr.memory_used_count
 *         self.atom_count = ptr.atom_count             # <<<<<<<<<<<<<<
 *         self.atom_size = ptr.atom_size
 *         self.str_count = ptr.str_count
*/
  __pyx_t_2 = __pyx_v_ptr->atom_count;
  __pyx_v_self->atom_count = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":75
 *         self.memory_used_count = ptr.memory_used_count
 *         self.atom_count = ptr.atom_count
 *         self.atom_size = ptr.atom_size             # <<<<<<<<<<<<<<
 *         self.str_count = ptr.str_count
 *         self.str_size = ptr.str_size
*/
  __pyx_t_2 = __pyx_v_ptr->atom_size;
  __pyx_v_self->atom_size = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":76
 *         self.atom_count = ptr.atom_count
 *         self.atom_size = ptr.atom_size
 *         self.str_count = ptr.str_count             # <<<<<<<<<<<<<<
 *         self.str_size = ptr.str_size
 *         self.obj_count = ptr.obj_count
*/
  __pyx_t_2 = __pyx_v_ptr->str_count;
  __pyx_v_self->str_count = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":77
 *         self.atom_size = ptr.atom_size
 *         self.str_count = ptr.str_count
 *         self.str_size = ptr.str_size             # <<<<<<<<<<<<<<
 *         self.obj_count = ptr.obj_count
 *         self.obj_size = ptr.obj_size
*/
  __pyx_t_2 = __pyx_v_ptr->str_size;
  __pyx_v_self->str_size = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":78
 *         self.str_count = ptr.str_count
 *         self.str_size = ptr.str_size
 *         self.obj_count = ptr.obj_count             # <<<<<<<<<<<<<<
 *         self.obj_size = ptr.obj_size
 *         self.prop_count = ptr.prop_count
*/
  __pyx_t_2 = __pyx_v_ptr->obj_count;
  __pyx_v_self->obj_count = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":79
 *         self.str_size = ptr.str_size
 *         self.obj_count = ptr.obj_count
 *         self.obj_size = ptr.obj_size             # <<<<<<<<<<<<<<
 *         self.prop_count = ptr.prop_count
 *         self.prop_size = ptr.prop_size
*/
  __pyx_t_2 = __pyx_v_ptr->obj_size;
  __pyx_v_self->obj_size = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":80
 *         self.obj_count = ptr.obj_count
 *         self.obj_size = ptr.obj_size
 *         self.prop_count = ptr.prop_count             # <<<<<<<<<<<<<<
 *         self.prop_size = ptr.prop_size
 *         self.shape_count = ptr.shape_count
*/
  __pyx_t_2 = __pyx_v_ptr->prop_count;
  __pyx_v_self->prop_count = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":81
 *         self.obj_size = ptr.obj_size
 *         self.prop_count = ptr.prop_count
 *         self.prop_size = ptr.prop_size             # <<<<<<<<<<<<<<
 *         self.shape_count = ptr.shape_count
 *         self.shape_size = ptr.shape_size
*/
  __pyx_t_2 = __pyx_v_ptr->prop_size;
  __pyx_v_self->prop_size = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":82
 *         self.prop_count = ptr.prop_count
 *         self.prop_size = ptr.prop_size
 *         self.shape_count = ptr.shape_count             # <<<<<<<<<<<<<<
 *         self.shape_size = ptr.shape_size
 *         self.js_func_count = ptr.js_func_count
*/
  __pyx_t_2 = __pyx_v_ptr->shape_count;
  __pyx_v_self->shape_count = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":83
 *         self.prop_size = ptr.prop_size
 *         self.shape_count = ptr.shape_count
 *         self.shape_size = ptr.shape_size             # <<<<<<<<<<<<<<
 *         self.js_func_count = ptr.js_func_count
 *         self.js_func_size = ptr.js_func_size
*/
  __pyx_t_2 = __pyx_v_ptr->shape_size;
  __pyx_v_self->shape_size = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":84
 *         self.shape_count = ptr.shape_count
 *         self.shape_size = ptr.shape_size
 *         self.js_func_count = ptr.js_func_count             # <<<<<<<<<<<<<<
 *         self.js_func_size = ptr.js_func_size
 *         self.js_func_code_size = ptr.js_func_code_size
*/
  __pyx_t_2 = __pyx_v_ptr->js_func_count;
  __pyx_v_self->js_func_count = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":85
 *         self.shape_size = ptr.shape_size
 *         self.js_func_count = ptr.js_func_count
 *         self.js_func_size = ptr.js_func_size             # <<<<<<<<<<<<<<
 *         self.js_func_code_size = ptr.js_func_code_size
 *         self.js_func_pc2line_count = ptr.js_func_pc2line_count
*/
  __pyx_t_2 = __pyx_v_ptr->js_func_size;
  __pyx_v_self->js_func_size = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":86
 *         self.js_func_count = ptr.js_func_count
 *         self.js_func_size = ptr.js_func_size
 *         self.js_func_code_size = ptr.js_func_code_size             # <<<<<<<<<<<<<<
 *         self.js_func_pc2line_count = ptr.js_func_pc2line_count
 *         self.array_count = ptr.array_count
*/
  __pyx_t_2 = __pyx_v_ptr->js_func_code_size;
  __pyx_v_self->js_func_code_size = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":87
 *         self.js_func_size = ptr.js_func_size
 *         self.js_func_code_size = ptr.js_func_code_size
 *         self.js_func_pc2line_count = ptr.js_func_pc2line_count             # <<<<<<<<<<<<<<
 *         self.array_count = ptr.array_count
 *         self.fast_array_count = ptr.fast_array_count
*/
  __pyx_t_2 = __pyx_v_ptr->js_func_pc2line_count;
  __pyx_v_self->js_func_pc2line_count = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":88
 *         self.js_func_code_size = ptr.js_func_code_size
 *         self.js_func_pc2line_count = ptr.js_func_pc2line_count
 *         self.array_count = ptr.array_count             # <<<<<<<<<<<<<<
 *         self.fast_array_count = ptr.fast_array_count
 *         self.fast_array_elements = ptr.fast_array_elements
*/
  __pyx_t_2 = __pyx_v_ptr->array_count;
  __pyx_v_self->array_count = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":89
 *         self.js_func_pc2line_count = ptr.js_func_pc2line_count
 *         self.array_count = ptr.array_count
 *         self.fast_array_count = ptr.fast_array_count             # <<<<<<<<<<<<<<
 *         self.fast_array_elements = ptr.fast_array_elements
 *         self.binary_object_count = ptr.binary_object_count
*/
  __pyx_t_2 = __pyx_v_ptr->fast_array_count;
  __pyx_v_self->fast_array_count = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":90
 *         self.array_count = ptr.array_count
 *         self.fast_array_count = ptr.fast_array_count
 *         self.fast_array_elements = ptr.fast_array_elements             # <<<<<<<<<<<<<<
 *         self.binary_object_count = ptr.binary_object_count
 *         self.binary_object_size = ptr.binary_object_size
*/
  __pyx_t_2 = __pyx_v_ptr->fast_array_elements;
  __pyx_v_self->fast_array_elements = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":91
 *         self.fast_array_count = ptr.fast_array_count
 *         self.fast_array_elements = ptr.fast_array_elements
 *         self.binary_object_count = ptr.binary_object_count             # <<<<<<<<<<<<<<
 *         self.binary_object_size = ptr.binary_object_size
 *         return self
*/
  __pyx_t_2 = __pyx_v_ptr->binary_object_count;
  __pyx_v_self->binary_object_count = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":92
 *         self.fast_array_elements = ptr.fast_array_elements
 *         self.binary_object_count = ptr.binary_object_count
 *         self.binary_object_size = ptr.binary_object_size             # <<<<<<<<<<<<<<
 *         return self
 * 
*/
  __pyx_t_2 = __pyx_v_ptr->binary_object_size;
  __pyx_v_self->binary_object_size = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":93
 *         self.binary_object_count = ptr.binary_object_count
 *         self.binary_object_size = ptr.binary_object_size
 *         return self             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":66
 * cdef class MemoryUsage:
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef MemoryUsage new(const JSMemoryUsage *ptr):
 *         cdef MemoryUsage self = MemoryUsage.__new__(MemoryUsage)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":55
 * cdef class MemoryUsage:
 *     cdef:
 *         readonly int64_t malloc_size             # <<<<<<<<<<<<<<
 *         readonly int64_t malloc_limit
 *         readonly int64_t memory_used_size
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_11malloc_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_11malloc_size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_11malloc_size___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_11malloc_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->malloc_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.malloc_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":56
 *     cdef:
 *         readonly int64_t malloc_size
 *         readonly int64_t malloc_limit             # <<<<<<<<<<<<<<
 *         readonly int64_t memory_used_size
 *         readonly int64_t malloc_count
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_12malloc_limit_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_12malloc_limit_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_12malloc_limit___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_12malloc_limit___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->malloc_limit); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.malloc_limit.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":57
 *         readonly int64_t malloc_size
 *         readonly int64_t malloc_limit
 *         readonly int64_t memory_used_size             # <<<<<<<<<<<<<<
 *         readonly int64_t malloc_count
 *         readonly int64_t memory_used_count
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_16memory_used_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_16memory_used_size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_16memory_used_size___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_16memory_used_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->memory_used_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.memory_used_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":58
 *         readonly int64_t malloc_limit
 *         readonly int64_t memory_used_size
 *         readonly int64_t malloc_count             # <<<<<<<<<<<<<<
 *         readonly int64_t memory_used_count
 *         readonly int64_t atom_count
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_12malloc_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_12malloc_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_12malloc_count___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_12malloc_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->malloc_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.malloc_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":59
 *         readonly int64_t memory_used_size
 *         readonly int64_t malloc_count
 *         readonly int64_t memory_used_count             # <<<<<<<<<<<<<<
 *         readonly int64_t atom_count
 *         readonly int64_t atom_size
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_17memory_used_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_17memory_used_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_17memory_used_count___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_17memory_used_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->memory_used_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.memory_used_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":60
 *         readonly int64_t malloc_count
 *         readonly int64_t memory_used_count
 *         readonly int64_t atom_count             # <<<<<<<<<<<<<<
 *         readonly int64_t atom_size
 *         readonly int64_t str_count
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_10atom_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_10atom_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_10atom_count___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_10atom_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->atom_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.atom_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":61
 *         readonly int64_t memory_used_count
 *         readonly int64_t atom_count
 *         readonly int64_t atom_size             # <<<<<<<<<<<<<<
 *         readonly int64_t str_count
 *         readonly int64_t str_size
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_9atom_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_9atom_size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_9atom_size___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_9atom_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->atom_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.atom_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":62
 *         readonly int64_t atom_count
 *         readonly int64_t atom_size
 *         readonly int64_t str_count             # <<<<<<<<<<<<<<
 *         readonly int64_t str_size
 *         readonly int64_t obj_count
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_9str_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_9str_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_9str_count___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_9str_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->str_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.str_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":63
 *         readonly int64_t atom_size
 *         readonly int64_t str_count
 *         readonly int64_t str_size             # <<<<<<<<<<<<<<
 *         readonly int64_t obj_count
 *         readonly int64_t obj_size
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_8str_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_8str_size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_8str_size___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_8str_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->str_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.str_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":64
 *         readonly int64_t str_count
 *         readonly int64_t str_size
 *         readonly int64_t obj_count             # <<<<<<<<<<<<<<
 *         readonly int64_t obj_size
 *         readonly int64_t prop_count
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_9obj_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_9obj_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_9obj_count___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_9obj_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->obj_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.obj_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":65
 *         readonly int64_t str_size
 *         readonly int64_t obj_count
 *         readonly int64_t obj_size             # <<<<<<<<<<<<<<
 *         readonly int64_t prop_count
 *         readonly int64_t prop_size
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_8obj_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_8obj_size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_8obj_size___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_8obj_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->obj_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.obj_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":66
 *         readonly int64_t obj_count
 *         readonly int64_t obj_size
 *         readonly int64_t prop_count             # <<<<<<<<<<<<<<
 *         readonly int64_t prop_size
 *         readonly int64_t shape_count
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_10prop_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_10prop_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_10prop_count___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_10prop_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->prop_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.prop_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":67
 *         readonly int64_t obj_size
 *         readonly int64_t prop_count
 *         readonly int64_t prop_size             # <<<<<<<<<<<<<<
 *         readonly int64_t shape_count
 *         readonly int64_t shape_size
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_9prop_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_9prop_size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_9prop_size___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_9prop_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->prop_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.prop_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":68
 *         readonly int64_t prop_count
 *         readonly int64_t prop_size
 *         readonly int64_t shape_count             # <<<<<<<<<<<<<<
 *         readonly int64_t shape_size
 *         readonly int64_t js_func_count
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_11shape_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_11shape_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_11shape_count___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_11shape_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->shape_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.shape_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":69
 *         readonly int64_t prop_size
 *         readonly int64_t shape_count
 *         readonly int64_t shape_size             # <<<<<<<<<<<<<<
 *         readonly int64_t js_func_count
 *         readonly int64_t js_func_size
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_10shape_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_10shape_size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_10shape_size___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_10shape_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->shape_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.shape_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":70
 *         readonly int64_t shape_count
 *         readonly int64_t shape_size
 *         readonly int64_t js_func_count             # <<<<<<<<<<<<<<
 *         readonly int64_t js_func_size
 *         readonly int64_t js_func_code_size
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_13js_func_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_13js_func_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_13js_func_count___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_13js_func_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->js_func_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.js_func_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":71
 *         readonly int64_t shape_size
 *         readonly int64_t js_func_count
 *         readonly int64_t js_func_size             # <<<<<<<<<<<<<<
 *         readonly int64_t js_func_code_size
 *         readonly int64_t js_func_pc2line_count
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_12js_func_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_12js_func_size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_12js_func_size___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_12js_func_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->js_func_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.js_func_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":72
 *         readonly int64_t js_func_count
 *         readonly int64_t js_func_size
 *         readonly int64_t js_func_code_size             # <<<<<<<<<<<<<<
 *         readonly int64_t js_func_pc2line_count
 *         readonly int64_t array_count
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_17js_func_code_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_17js_func_code_size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_17js_func_code_size___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_17js_func_code_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->js_func_code_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.js_func_code_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":73
 *         readonly int64_t js_func_size
 *         readonly int64_t js_func_code_size
 *         readonly int64_t js_func_pc2line_count             # <<<<<<<<<<<<<<
 *         readonly int64_t array_count
 *         readonly int64_t fast_array_count
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_21js_func_pc2line_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_21js_func_pc2line_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_21js_func_pc2line_count___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_21js_func_pc2line_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->js_func_pc2line_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.js_func_pc2line_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":74
 *         readonly int64_t js_func_code_size
 *         readonly int64_t js_func_pc2line_count
 *         readonly int64_t array_count             # <<<<<<<<<<<<<<
 *         readonly int64_t fast_array_count
 *         readonly int64_t fast_array_elements
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_11array_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_11array_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_11array_count___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_11array_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->array_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.array_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":75
 *         readonly int64_t js_func_pc2line_count
 *         readonly int64_t array_count
 *         readonly int64_t fast_array_count             # <<<<<<<<<<<<<<
 *         readonly int64_t fast_array_elements
 *         readonly int64_t binary_object_count
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_16fast_array_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_16fast_array_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_16fast_array_count___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_16fast_array_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->fast_array_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.fast_array_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":76
 *         readonly int64_t array_count
 *         readonly int64_t fast_array_count
 *         readonly int64_t fast_array_elements             # <<<<<<<<<<<<<<
 *         readonly int64_t binary_object_count
 *         readonly int64_t binary_object_size
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_19fast_array_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_19fast_array_elements_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_19fast_array_elements___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_19fast_array_elements___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->fast_array_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.fast_array_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":77
 *         readonly int64_t fast_array_count
 *         readonly int64_t fast_array_elements
 *         readonly int64_t binary_object_count             # <<<<<<<<<<<<<<
 *         readonly int64_t binary_object_size
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_19binary_object_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_19binary_object_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_19binary_object_count___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_19binary_object_count___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->binary_object_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.binary_object_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":78
 *         readonly int64_t fast_array_elements
 *         readonly int64_t binary_object_count
 *         readonly int64_t binary_object_size             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_18binary_object_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_18binary_object_size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_18binary_object_size___get__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_18binary_object_size___get__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->binary_object_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.binary_object_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_11MemoryUsage_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage___reduce_cython__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage___reduce_cython__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  int __pyx_t_26;
  int __pyx_t_27;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.array_count, self.atom_count, self.atom_size, self.binary_object_count, self.binary_object_size, self.fast_array_count, self.fast_array_elements, self.js_func_code_size, self.js_func_count, self.js_func_pc2line_count, self.js_func_size, self.malloc_count, self.malloc_limit, self.malloc_size, self.memory_used_count, self.memory_used_size, self.obj_count, self.obj_size, self.prop_count, self.prop_size, self.shape_count, self.shape_size, self.str_count, self.str_size)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->array_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_v_self->atom_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyLong_From_int64_t(__pyx_v_self->atom_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyLong_From_int64_t(__pyx_v_self->binary_object_count); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyLong_From_int64_t(__pyx_v_self->binary_object_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyLong_From_int64_t(__pyx_v_self->fast_array_count); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyLong_From_int64_t(__pyx_v_self->fast_array_elements); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyLong_From_int64_t(__pyx_v_self->js_func_code_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyLong_From_int64_t(__pyx_v_self->js_func_count); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyLong_From_int64_t(__pyx_v_self->js_func_pc2line_count); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyLong_From_int64_t(__pyx_v_self->js_func_size); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyLong_From_int64_t(__pyx_v_self->malloc_count); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyLong_From_int64_t(__pyx_v_self->malloc_limit); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_14 = __Pyx_PyLong_From_int64_t(__pyx_v_self->malloc_size); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = __Pyx_PyLong_From_int64_t(__pyx_v_self->memory_used_count); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_16 = __Pyx_PyLong_From_int64_t(__pyx_v_self->memory_used_size); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_17 = __Pyx_PyLong_From_int64_t(__pyx_v_self->obj_count); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_18 = __Pyx_PyLong_From_int64_t(__pyx_v_self->obj_size); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_19 = __Pyx_PyLong_From_int64_t(__pyx_v_self->prop_count); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_20 = __Pyx_PyLong_From_int64_t(__pyx_v_self->prop_size); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_21 = __Pyx_PyLong_From_int64_t(__pyx_v_self->shape_count); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_22 = __Pyx_PyLong_From_int64_t(__pyx_v_self->shape_size); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_23 = __Pyx_PyLong_From_int64_t(__pyx_v_self->str_count); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_24 = __Pyx_PyLong_From_int64_t(__pyx_v_self->str_size); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_25 = PyTuple_New(24); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_t_1) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_2) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 2, __pyx_t_3) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 3, __pyx_t_4) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 4, __pyx_t_5) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 5, __pyx_t_6) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 6, __pyx_t_7) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 7, __pyx_t_8) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_9);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 8, __pyx_t_9) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_10);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 9, __pyx_t_10) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_11);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 10, __pyx_t_11) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_12);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 11, __pyx_t_12) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_13);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 12, __pyx_t_13) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_14);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 13, __pyx_t_14) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_15);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 14, __pyx_t_15) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_16);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 15, __pyx_t_16) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_17);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 16, __pyx_t_17) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_18);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 17, __pyx_t_18) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_19);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 18, __pyx_t_19) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_20);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 19, __pyx_t_20) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_21);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 20, __pyx_t_21) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_22);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 21, __pyx_t_22) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_23);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 22, __pyx_t_23) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_24);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 23, __pyx_t_24) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_t_15 = 0;
  __pyx_t_16 = 0;
  __pyx_t_17 = 0;
  __pyx_t_18 = 0;
  __pyx_t_19 = 0;
  __pyx_t_20 = 0;
  __pyx_t_21 = 0;
  __pyx_t_22 = 0;
  __pyx_t_23 = 0;
  __pyx_t_24 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_25);
  __pyx_t_25 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.array_count, self.atom_count, self.atom_size, self.binary_object_count, self.binary_object_size, self.fast_array_count, self.fast_array_elements, self.js_func_code_size, self.js_func_count, self.js_func_pc2line_count, self.js_func_size, self.malloc_count, self.malloc_limit, self.malloc_size, self.memory_used_count, self.memory_used_size, self.obj_count, self.obj_size, self.prop_count, self.prop_size, self.shape_count, self.shape_size, self.str_count, self.str_size)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_25 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_v__dict = __pyx_t_25;
  __pyx_t_25 = 0;

  /* "(tree fragment)":7
 *     state = (self.array_count, self.atom_count, self.atom_size, self.binary_object_count, self.binary_object_size, self.fast_array_count, self.fast_array_elements, self.js_func_code_size, self.js_func_count, self.js_func_pc2line_count, self.js_func_size, self.malloc_count, self.malloc_limit, self.malloc_size, self.memory_used_count, self.memory_used_size, self.obj_count, self.obj_size, self.prop_count, self.prop_size, self.shape_count, self.shape_size, self.str_count, self.str_size)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_27 = (__pyx_v__dict != Py_None);
  if (__pyx_t_27) {
  } else {
    __pyx_t_26 = __pyx_t_27;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_27 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_27 < 0))) __PYX_ERR(1, 7, __pyx_L1_error)
  __pyx_t_26 = __pyx_t_27;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_26) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_25 = PyTuple_New(1); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_24 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_25); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_24));
    __pyx_t_24 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.array_count, self.atom_count, self.atom_size, self.binary_object_count, self.binary_object_size, self.fast_array_count, self.fast_array_elements, self.js_func_code_size, self.js_func_count, self.js_func_pc2line_count, self.js_func_size, self.malloc_count, self.malloc_limit, self.malloc_size, self.memory_used_count, self.memory_used_size, self.obj_count, self.obj_size, self.prop_count, self.prop_size, self.shape_count, self.shape_size, self.str_count, self.str_size)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_MemoryUsage, (type(self), 0xe4999e7, None), state
*/
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_MemoryUsage, (type(self), 0xe4999e7, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_MemoryUsage, (type(self), 0xe4999e7, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_MemoryUsage, (type(self), 0xe4999e7, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_MemoryUsage); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_25 = PyTuple_New(3); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_239704551);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_239704551);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_mstate_global->__pyx_int_239704551) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_23 = PyTuple_New(3); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_GIVEREF(__pyx_t_24);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_24) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_25);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_25) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_23, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_24 = 0;
    __pyx_t_25 = 0;
    __pyx_r = __pyx_t_23;
    __pyx_t_23 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_MemoryUsage, (type(self), 0xe4999e7, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_MemoryUsage, (type(self), 0xe4999e7, None), state
 *     else:
 *         return __pyx_unpickle_MemoryUsage, (type(self), 0xe4999e7, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_MemoryUsage__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_MemoryUsage); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_25 = PyTuple_New(3); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_239704551);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_239704551);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_mstate_global->__pyx_int_239704551) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_23);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_23) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_25);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_25) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_23 = 0;
    __pyx_t_25 = 0;
    __pyx_r = __pyx_t_24;
    __pyx_t_24 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MemoryUsage, (type(self), 0xe4999e7, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MemoryUsage__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_11MemoryUsage_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11MemoryUsage_2__setstate_cython__(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11MemoryUsage_2__setstate_cython__(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_MemoryUsage, (type(self), 0xe4999e7, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_MemoryUsage__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(1, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_4cyjs_5_cyjs___pyx_unpickle_MemoryUsage__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MemoryUsage, (type(self), 0xe4999e7, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MemoryUsage__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cyjs._cyjs.MemoryUsage.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":97
 * 
 * 
 * cdef void on_promise_hook(JSContext *ctx, JSPromiseHookType type,             # <<<<<<<<<<<<<<
 *     JSValue promise, JSValue parent_promise, void *opaque) noexcept with gil:
 *     cdef PromiseHook hook = <PromiseHook>opaque
*/

static void __pyx_f_4cyjs_5_cyjs_on_promise_hook(JSContext *__pyx_v_ctx, JSPromiseHookType __pyx_v_type, JSValue __pyx_v_promise, JSValue __pyx_v_parent_promise, void *__pyx_v_opaque) {
  struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *__pyx_v_hook = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_RefNannySetupContext("on_promise_hook", 0);

  /* "cyjs/_cyjs.pyx":99
 * cdef void on_promise_hook(JSContext *ctx, JSPromiseHookType type,
 *     JSValue promise, JSValue parent_promise, void *opaque) noexcept with gil:
 *     cdef PromiseHook hook = <PromiseHook>opaque             # <<<<<<<<<<<<<<
 *     hook.hook(
 *         ctx, type, promise, parent_promise
*/
  __pyx_t_1 = ((PyObject *)__pyx_v_opaque);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_hook = ((struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":100
 *     JSValue promise, JSValue parent_promise, void *opaque) noexcept with gil:
 *     cdef PromiseHook hook = <PromiseHook>opaque
 *     hook.hook(             # <<<<<<<<<<<<<<
 *         ctx, type, promise, parent_promise
 *     )
*/
  ((struct __pyx_vtabstruct_4cyjs_5_cyjs_PromiseHook *)__pyx_v_hook->__pyx_vtab)->hook(__pyx_v_hook, __pyx_v_ctx, __pyx_v_type, __pyx_v_promise, __pyx_v_parent_promise);

  /* "cyjs/_cyjs.pyx":97
 * 
 * 
 * cdef void on_promise_hook(JSContext *ctx, JSPromiseHookType type,             # <<<<<<<<<<<<<<
 *     JSValue promise, JSValue parent_promise, void *opaque) noexcept with gil:
 *     cdef PromiseHook hook = <PromiseHook>opaque
*/

  /* function exit code */
  __Pyx_XDECREF((PyObject *)__pyx_v_hook);
  __Pyx_RefNannyFinishContext();
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
}

/* "cyjs/_cyjs.pyx":112
 * @cython.internal
 * cdef class PromiseHook:
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef PromiseHook new(Runtime rt, object func):
 *         cdef PromiseHook self = PromiseHook.__new__(PromiseHook)
*/

static struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *__pyx_f_4cyjs_5_cyjs_11PromiseHook_new(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_rt, PyObject *__pyx_v_func) {
  struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *__pyx_v_self = 0;
  struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 0);

  /* "cyjs/_cyjs.pyx":114
 *     @staticmethod
 *     cdef PromiseHook new(Runtime rt, object func):
 *         cdef PromiseHook self = PromiseHook.__new__(PromiseHook)             # <<<<<<<<<<<<<<
 *         self.func = func
 *         self.rt = rt
*/
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_4cyjs_5_cyjs_PromiseHook(((PyTypeObject *)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_PromiseHook), __pyx_mstate_global->__pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_self = ((struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":115
 *     cdef PromiseHook new(Runtime rt, object func):
 *         cdef PromiseHook self = PromiseHook.__new__(PromiseHook)
 *         self.func = func             # <<<<<<<<<<<<<<
 *         self.rt = rt
 *         return self
*/
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->func);
  __Pyx_DECREF(__pyx_v_self->func);
  __pyx_v_self->func = __pyx_v_func;

  /* "cyjs/_cyjs.pyx":116
 *         cdef PromiseHook self = PromiseHook.__new__(PromiseHook)
 *         self.func = func
 *         self.rt = rt             # <<<<<<<<<<<<<<
 *         return self
 * 
*/
  __Pyx_INCREF((PyObject *)__pyx_v_rt);
  __Pyx_GIVEREF((PyObject *)__pyx_v_rt);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->rt);
  __Pyx_DECREF((PyObject *)__pyx_v_self->rt);
  __pyx_v_self->rt = __pyx_v_rt;

  /* "cyjs/_cyjs.pyx":117
 *         self.func = func
 *         self.rt = rt
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cdef void hook(self, JSContext *context, JSPromiseHookType type,
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":112
 * @cython.internal
 * cdef class PromiseHook:
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef PromiseHook new(Runtime rt, object func):
 *         cdef PromiseHook self = PromiseHook.__new__(PromiseHook)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.PromiseHook.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":119
 *         return self
 * 
 *     cdef void hook(self, JSContext *context, JSPromiseHookType type,             # <<<<<<<<<<<<<<
 *         JSValue promise, JSValue parent_promise) noexcept:
 *         cdef Context ctx = <Context>JS_GetContextOpaque(context)
*/

static void __pyx_f_4cyjs_5_cyjs_11PromiseHook_hook(struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *__pyx_v_self, JSContext *__pyx_v_context, JSPromiseHookType __pyx_v_type, JSValue __pyx_v_promise, JSValue __pyx_v_parent_promise) {
  struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_ctx = 0;
  PyObject *__pyx_v_e = NULL;
  __Pyx_RefNannyDeclarations
  void *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("hook", 0);

  /* "cyjs/_cyjs.pyx":121
 *     cdef void hook(self, JSContext *context, JSPromiseHookType type,
 *         JSValue promise, JSValue parent_promise) noexcept:
 *         cdef Context ctx = <Context>JS_GetContextOpaque(context)             # <<<<<<<<<<<<<<
 *         try:
 *             # Convert promiseHook to a public version and
*/
  __pyx_t_1 = JS_GetContextOpaque(__pyx_v_context);
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_ctx = ((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "cyjs/_cyjs.pyx":122
 *         JSValue promise, JSValue parent_promise) noexcept:
 *         cdef Context ctx = <Context>JS_GetContextOpaque(context)
 *         try:             # <<<<<<<<<<<<<<
 *             # Convert promiseHook to a public version and
 *             # convert all objects to python acceptable types before sending and ensure nothing can crash via duping the object
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "cyjs/_cyjs.pyx":125
 *             # Convert promiseHook to a public version and
 *             # convert all objects to python acceptable types before sending and ensure nothing can crash via duping the object
 *             PyObject_CallObject(self.func, (ctx, <PromiseHookType>type,  to_python(context, JS_DupValue(context, promise)), to_python(context, JS_DupValue(context, parent_promise))))             # <<<<<<<<<<<<<<
 *         except BaseException as e:
 *             # We can regain this exception later if grabbed immediately...
*/
      __pyx_t_2 = __pyx_v_self->func;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_Enum_4cyjs_5_cyjs_enum__dunderpyx_t_4cyjs_5_cyjs_PromiseHookType_to_py(((enum __pyx_t_4cyjs_5_cyjs_PromiseHookType)__pyx_v_type)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 125, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_context, JS_DupValue(__pyx_v_context, __pyx_v_promise)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 125, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_context, JS_DupValue(__pyx_v_context, __pyx_v_parent_promise)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 125, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 125, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_INCREF((PyObject *)__pyx_v_ctx);
      __Pyx_GIVEREF((PyObject *)__pyx_v_ctx);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_ctx)) != (0)) __PYX_ERR(0, 125, __pyx_L3_error);
      __Pyx_GIVEREF(__pyx_t_6);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_6) != (0)) __PYX_ERR(0, 125, __pyx_L3_error);
      __Pyx_GIVEREF(__pyx_t_7);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_7) != (0)) __PYX_ERR(0, 125, __pyx_L3_error);
      __Pyx_GIVEREF(__pyx_t_8);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_8) != (0)) __PYX_ERR(0, 125, __pyx_L3_error);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_8 = 0;
      __pyx_t_8 = PyObject_CallObject(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 125, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "cyjs/_cyjs.pyx":122
 *         JSValue promise, JSValue parent_promise) noexcept:
 *         cdef Context ctx = <Context>JS_GetContextOpaque(context)
 *         try:             # <<<<<<<<<<<<<<
 *             # Convert promiseHook to a public version and
 *             # convert all objects to python acceptable types before sending and ensure nothing can crash via duping the object
*/
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "cyjs/_cyjs.pyx":126
 *             # convert all objects to python acceptable types before sending and ensure nothing can crash via duping the object
 *             PyObject_CallObject(self.func, (ctx, <PromiseHookType>type,  to_python(context, JS_DupValue(context, promise)), to_python(context, JS_DupValue(context, parent_promise))))
 *         except BaseException as e:             # <<<<<<<<<<<<<<
 *             # We can regain this exception later if grabbed immediately...
 *             ctx._cb_exception = e
*/
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_BaseException))));
    if (__pyx_t_10) {
      __Pyx_AddTraceback("cyjs._cyjs.PromiseHook.hook", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_2) < 0) __PYX_ERR(0, 126, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_9);
      __pyx_v_e = __pyx_t_9;
      /*try:*/ {

        /* "cyjs/_cyjs.pyx":128
 *         except BaseException as e:
 *             # We can regain this exception later if grabbed immediately...
 *             ctx._cb_exception = e             # <<<<<<<<<<<<<<
 * 
 * 
*/
        __Pyx_INCREF(__pyx_v_e);
        __Pyx_GIVEREF(__pyx_v_e);
        __Pyx_GOTREF(__pyx_v_ctx->_cb_exception);
        __Pyx_DECREF(__pyx_v_ctx->_cb_exception);
        __pyx_v_ctx->_cb_exception = __pyx_v_e;
      }

      /* "cyjs/_cyjs.pyx":126
 *             # convert all objects to python acceptable types before sending and ensure nothing can crash via duping the object
 *             PyObject_CallObject(self.func, (ctx, <PromiseHookType>type,  to_python(context, JS_DupValue(context, promise)), to_python(context, JS_DupValue(context, parent_promise))))
 *         except BaseException as e:             # <<<<<<<<<<<<<<
 *             # We can regain this exception later if grabbed immediately...
 *             ctx._cb_exception = e
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
          goto __pyx_L15;
        }
        __pyx_L15:;
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "cyjs/_cyjs.pyx":122
 *         JSValue promise, JSValue parent_promise) noexcept:
 *         cdef Context ctx = <Context>JS_GetContextOpaque(context)
 *         try:             # <<<<<<<<<<<<<<
 *             # Convert promiseHook to a public version and
 *             # convert all objects to python acceptable types before sending and ensure nothing can crash via duping the object
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    __pyx_L8_try_end:;
  }

  /* "cyjs/_cyjs.pyx":119
 *         return self
 * 
 *     cdef void hook(self, JSContext *context, JSPromiseHookType type,             # <<<<<<<<<<<<<<
 *         JSValue promise, JSValue parent_promise) noexcept:
 *         cdef Context ctx = <Context>JS_GetContextOpaque(context)
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("cyjs._cyjs.PromiseHook.hook", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ctx);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11PromiseHook_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_11PromiseHook_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_11PromiseHook_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_11PromiseHook_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11PromiseHook___reduce_cython__(((struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11PromiseHook___reduce_cython__(struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.func, self.rt)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->func);
  __Pyx_GIVEREF(__pyx_v_self->func);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->func) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->rt);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->rt);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->rt)) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.func, self.rt)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.func, self.rt)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(1, 7, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.func is not None or self.rt is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.func, self.rt)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.func is not None or self.rt is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PromiseHook, (type(self), 0x84cfe3b, None), state
*/
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->func != Py_None);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_3 = (((PyObject *)__pyx_v_self->rt) != Py_None);
    __pyx_t_2 = __pyx_t_3;
    __pyx_L6_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.func is not None or self.rt is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PromiseHook, (type(self), 0x84cfe3b, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.func is not None or self.rt is not None
 *     if use_setstate:
 *         return __pyx_unpickle_PromiseHook, (type(self), 0x84cfe3b, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PromiseHook, (type(self), 0x84cfe3b, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_PromiseHook); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_139263547);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_139263547);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_139263547) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.func is not None or self.rt is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PromiseHook, (type(self), 0x84cfe3b, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PromiseHook, (type(self), 0x84cfe3b, None), state
 *     else:
 *         return __pyx_unpickle_PromiseHook, (type(self), 0x84cfe3b, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PromiseHook__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_PromiseHook); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_139263547);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_139263547);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_139263547) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cyjs._cyjs.PromiseHook.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromiseHook, (type(self), 0x84cfe3b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromiseHook__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11PromiseHook_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_11PromiseHook_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_11PromiseHook_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_11PromiseHook_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.PromiseHook.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_11PromiseHook_2__setstate_cython__(((struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_11PromiseHook_2__setstate_cython__(struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PromiseHook, (type(self), 0x84cfe3b, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PromiseHook__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(1, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_4cyjs_5_cyjs___pyx_unpickle_PromiseHook__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromiseHook, (type(self), 0x84cfe3b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromiseHook__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cyjs._cyjs.PromiseHook.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":138
 * 
 * cdef class Runtime:
 *     def __init__(self) -> None:             # <<<<<<<<<<<<<<
 *         self.rt = CYJS_NewRuntime(<void*>self)
 *         # We can set this attribute up later...
*/

/* Python wrapper */
static int __pyx_pw_4cyjs_5_cyjs_7Runtime_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4cyjs_5_cyjs_7Runtime_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__init__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Runtime___init__(((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4cyjs_5_cyjs_7Runtime___init__(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self) {
  int __pyx_r;

  /* "cyjs/_cyjs.pyx":139
 * cdef class Runtime:
 *     def __init__(self) -> None:
 *         self.rt = CYJS_NewRuntime(<void*>self)             # <<<<<<<<<<<<<<
 *         # We can set this attribute up later...
 *         self.has_promise_hook = False
*/
  __pyx_v_self->rt = CYJS_NewRuntime(((void *)__pyx_v_self));

  /* "cyjs/_cyjs.pyx":141
 *         self.rt = CYJS_NewRuntime(<void*>self)
 *         # We can set this attribute up later...
 *         self.has_promise_hook = False             # <<<<<<<<<<<<<<
 * 
 *     cpdef MemoryUsage compute_memory_usage(self):
*/
  __pyx_v_self->has_promise_hook = 0;

  /* "cyjs/_cyjs.pyx":138
 * 
 * cdef class Runtime:
 *     def __init__(self) -> None:             # <<<<<<<<<<<<<<
 *         self.rt = CYJS_NewRuntime(<void*>self)
 *         # We can set this attribute up later...
*/

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":143
 *         self.has_promise_hook = False
 * 
 *     cpdef MemoryUsage compute_memory_usage(self):             # <<<<<<<<<<<<<<
 *         cdef JSMemoryUsage mu
 *         JS_ComputeMemoryUsage(self.rt, &mu)
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_3compute_memory_usage(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_f_4cyjs_5_cyjs_7Runtime_compute_memory_usage(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, int __pyx_skip_dispatch) {
  JSMemoryUsage __pyx_v_mu;
  struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_memory_usage", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_compute_memory_usage); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Runtime_3compute_memory_usage)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 143, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage))))) __PYX_ERR(0, 143, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":145
 *     cpdef MemoryUsage compute_memory_usage(self):
 *         cdef JSMemoryUsage mu
 *         JS_ComputeMemoryUsage(self.rt, &mu)             # <<<<<<<<<<<<<<
 *         return MemoryUsage.new(&mu)
 * 
*/
  JS_ComputeMemoryUsage(__pyx_v_self->rt, (&__pyx_v_mu));

  /* "cyjs/_cyjs.pyx":146
 *         cdef JSMemoryUsage mu
 *         JS_ComputeMemoryUsage(self.rt, &mu)
 *         return MemoryUsage.new(&mu)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object dump_memory_usage(self, object file):
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_4cyjs_5_cyjs_11MemoryUsage_new((&__pyx_v_mu))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":143
 *         self.has_promise_hook = False
 * 
 *     cpdef MemoryUsage compute_memory_usage(self):             # <<<<<<<<<<<<<<
 *         cdef JSMemoryUsage mu
 *         JS_ComputeMemoryUsage(self.rt, &mu)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.compute_memory_usage", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_3compute_memory_usage(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Runtime_3compute_memory_usage = {"compute_memory_usage", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Runtime_3compute_memory_usage, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_3compute_memory_usage(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_memory_usage (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("compute_memory_usage", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("compute_memory_usage", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Runtime_2compute_memory_usage(((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_2compute_memory_usage(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_memory_usage", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_4cyjs_5_cyjs_7Runtime_compute_memory_usage(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.compute_memory_usage", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":148
 *         return MemoryUsage.new(&mu)
 * 
 *     cpdef object dump_memory_usage(self, object file):             # <<<<<<<<<<<<<<
 *         if CYJS_DumpMemoryUsage(self.rt, file) < 0:
 *             raise
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_5dump_memory_usage(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Runtime_dump_memory_usage(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, PyObject *__pyx_v_file, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("dump_memory_usage", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dump_memory_usage); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Runtime_5dump_memory_usage)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_file};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 148, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":149
 * 
 *     cpdef object dump_memory_usage(self, object file):
 *         if CYJS_DumpMemoryUsage(self.rt, file) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_6 = CYJS_DumpMemoryUsage(__pyx_v_self->rt, __pyx_v_file); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 149, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 < 0);
  if (unlikely(__pyx_t_7)) {

    /* "cyjs/_cyjs.pyx":150
 *     cpdef object dump_memory_usage(self, object file):
 *         if CYJS_DumpMemoryUsage(self.rt, file) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *     cpdef object execute_pending_job(self):
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 150, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":149
 * 
 *     cpdef object dump_memory_usage(self, object file):
 *         if CYJS_DumpMemoryUsage(self.rt, file) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":148
 *         return MemoryUsage.new(&mu)
 * 
 *     cpdef object dump_memory_usage(self, object file):             # <<<<<<<<<<<<<<
 *         if CYJS_DumpMemoryUsage(self.rt, file) < 0:
 *             raise
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.dump_memory_usage", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_5dump_memory_usage(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Runtime_5dump_memory_usage = {"dump_memory_usage", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Runtime_5dump_memory_usage, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_5dump_memory_usage(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_file = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dump_memory_usage (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_file,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 148, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 148, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "dump_memory_usage", 0) < (0)) __PYX_ERR(0, 148, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("dump_memory_usage", 1, 1, 1, i); __PYX_ERR(0, 148, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 148, __pyx_L3_error)
    }
    __pyx_v_file = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dump_memory_usage", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 148, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.dump_memory_usage", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Runtime_4dump_memory_usage(((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)__pyx_v_self), __pyx_v_file);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_4dump_memory_usage(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, PyObject *__pyx_v_file) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("dump_memory_usage", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_7Runtime_dump_memory_usage(__pyx_v_self, __pyx_v_file, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.dump_memory_usage", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":152
 *             raise
 * 
 *     cpdef object execute_pending_job(self):             # <<<<<<<<<<<<<<
 *         cdef JSContext* ctx = NULL
 *         cdef Context py_context
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_7execute_pending_job(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Runtime_execute_pending_job(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, int __pyx_skip_dispatch) {
  JSContext *__pyx_v_ctx;
  struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_py_context = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("execute_pending_job", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_execute_pending_job); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Runtime_7execute_pending_job)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 152, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":153
 * 
 *     cpdef object execute_pending_job(self):
 *         cdef JSContext* ctx = NULL             # <<<<<<<<<<<<<<
 *         cdef Context py_context
 *         if self.is_job_pending():
*/
  __pyx_v_ctx = NULL;

  /* "cyjs/_cyjs.pyx":155
 *         cdef JSContext* ctx = NULL
 *         cdef Context py_context
 *         if self.is_job_pending():             # <<<<<<<<<<<<<<
 *             if JS_ExecutePendingJob(self.rt, &ctx) < 0:
 *                 self.raise_exception()
*/
  __pyx_t_6 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Runtime *)__pyx_v_self->__pyx_vtab)->is_job_pending(__pyx_v_self, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 155, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "cyjs/_cyjs.pyx":156
 *         cdef Context py_context
 *         if self.is_job_pending():
 *             if JS_ExecutePendingJob(self.rt, &ctx) < 0:             # <<<<<<<<<<<<<<
 *                 self.raise_exception()
 *                 return None
*/
    __pyx_t_6 = (JS_ExecutePendingJob(__pyx_v_self->rt, (&__pyx_v_ctx)) < 0);
    if (__pyx_t_6) {

      /* "cyjs/_cyjs.pyx":157
 *         if self.is_job_pending():
 *             if JS_ExecutePendingJob(self.rt, &ctx) < 0:
 *                 self.raise_exception()             # <<<<<<<<<<<<<<
 *                 return None
 * 
*/
      __pyx_t_2 = ((PyObject *)__pyx_v_self);
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
        __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_raise_exception, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 157, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "cyjs/_cyjs.pyx":158
 *             if JS_ExecutePendingJob(self.rt, &ctx) < 0:
 *                 self.raise_exception()
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *             py_context = <Context>JS_GetContextOpaque(ctx)
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "cyjs/_cyjs.pyx":156
 *         cdef Context py_context
 *         if self.is_job_pending():
 *             if JS_ExecutePendingJob(self.rt, &ctx) < 0:             # <<<<<<<<<<<<<<
 *                 self.raise_exception()
 *                 return None
*/
    }

    /* "cyjs/_cyjs.pyx":160
 *                 return None
 * 
 *             py_context = <Context>JS_GetContextOpaque(ctx)             # <<<<<<<<<<<<<<
 *             if py_context._cb_exception:
 *                 raise py_context._cb_exception
*/
    __pyx_t_7 = JS_GetContextOpaque(__pyx_v_ctx);
    __pyx_t_1 = ((PyObject *)__pyx_t_7);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_py_context = ((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "cyjs/_cyjs.pyx":161
 * 
 *             py_context = <Context>JS_GetContextOpaque(ctx)
 *             if py_context._cb_exception:             # <<<<<<<<<<<<<<
 *                 raise py_context._cb_exception
 *         return None
*/
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_py_context->_cb_exception); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 161, __pyx_L1_error)
    if (unlikely(__pyx_t_6)) {

      /* "cyjs/_cyjs.pyx":162
 *             py_context = <Context>JS_GetContextOpaque(ctx)
 *             if py_context._cb_exception:
 *                 raise py_context._cb_exception             # <<<<<<<<<<<<<<
 *         return None
 * 
*/
      __Pyx_Raise(__pyx_v_py_context->_cb_exception, 0, 0, 0);
      __PYX_ERR(0, 162, __pyx_L1_error)

      /* "cyjs/_cyjs.pyx":161
 * 
 *             py_context = <Context>JS_GetContextOpaque(ctx)
 *             if py_context._cb_exception:             # <<<<<<<<<<<<<<
 *                 raise py_context._cb_exception
 *         return None
*/
    }

    /* "cyjs/_cyjs.pyx":155
 *         cdef JSContext* ctx = NULL
 *         cdef Context py_context
 *         if self.is_job_pending():             # <<<<<<<<<<<<<<
 *             if JS_ExecutePendingJob(self.rt, &ctx) < 0:
 *                 self.raise_exception()
*/
  }

  /* "cyjs/_cyjs.pyx":163
 *             if py_context._cb_exception:
 *                 raise py_context._cb_exception
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":152
 *             raise
 * 
 *     cpdef object execute_pending_job(self):             # <<<<<<<<<<<<<<
 *         cdef JSContext* ctx = NULL
 *         cdef Context py_context
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.execute_pending_job", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_context);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_7execute_pending_job(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Runtime_7execute_pending_job = {"execute_pending_job", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Runtime_7execute_pending_job, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_7execute_pending_job(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("execute_pending_job (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("execute_pending_job", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("execute_pending_job", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Runtime_6execute_pending_job(((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_6execute_pending_job(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("execute_pending_job", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_7Runtime_execute_pending_job(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.execute_pending_job", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":166
 * 
 * 
 *     cpdef bint is_job_pending(self):             # <<<<<<<<<<<<<<
 *         return JS_IsJobPending(self.rt)
 * 
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_9is_job_pending(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_4cyjs_5_cyjs_7Runtime_is_job_pending(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_job_pending", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_is_job_pending); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Runtime_9is_job_pending)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 166, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":167
 * 
 *     cpdef bint is_job_pending(self):
 *         return JS_IsJobPending(self.rt)             # <<<<<<<<<<<<<<
 * 
 *     cdef JSContext* new_context(self) except NULL:
*/
  __pyx_r = JS_IsJobPending(__pyx_v_self->rt);
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":166
 * 
 * 
 *     cpdef bint is_job_pending(self):             # <<<<<<<<<<<<<<
 *         return JS_IsJobPending(self.rt)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.is_job_pending", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_9is_job_pending(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Runtime_9is_job_pending = {"is_job_pending", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Runtime_9is_job_pending, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_9is_job_pending(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_job_pending (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("is_job_pending", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("is_job_pending", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Runtime_8is_job_pending(((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_8is_job_pending(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_job_pending", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_7Runtime_is_job_pending(__pyx_v_self, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 166, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.is_job_pending", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":169
 *         return JS_IsJobPending(self.rt)
 * 
 *     cdef JSContext* new_context(self) except NULL:             # <<<<<<<<<<<<<<
 *         # Use a raw context to enable higher performance when we can disable things...
 *         cdef JSContext* ctx = JS_NewContextRaw(self.rt)
*/

static JSContext *__pyx_f_4cyjs_5_cyjs_7Runtime_new_context(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self) {
  JSContext *__pyx_v_ctx;
  JSContext *__pyx_r;
  int __pyx_t_1;
  PyObject *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyjs/_cyjs.pyx":171
 *     cdef JSContext* new_context(self) except NULL:
 *         # Use a raw context to enable higher performance when we can disable things...
 *         cdef JSContext* ctx = JS_NewContextRaw(self.rt)             # <<<<<<<<<<<<<<
 *         if ctx == NULL:
 *             PyErr_NoMemory()
*/
  __pyx_v_ctx = JS_NewContextRaw(__pyx_v_self->rt);

  /* "cyjs/_cyjs.pyx":172
 *         # Use a raw context to enable higher performance when we can disable things...
 *         cdef JSContext* ctx = JS_NewContextRaw(self.rt)
 *         if ctx == NULL:             # <<<<<<<<<<<<<<
 *             PyErr_NoMemory()
 *         return ctx
*/
  __pyx_t_1 = (__pyx_v_ctx == NULL);
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":173
 *         cdef JSContext* ctx = JS_NewContextRaw(self.rt)
 *         if ctx == NULL:
 *             PyErr_NoMemory()             # <<<<<<<<<<<<<<
 *         return ctx
 * 
*/
    __pyx_t_2 = PyErr_NoMemory(); if (unlikely(__pyx_t_2 == ((void *)NULL))) __PYX_ERR(0, 173, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":172
 *         # Use a raw context to enable higher performance when we can disable things...
 *         cdef JSContext* ctx = JS_NewContextRaw(self.rt)
 *         if ctx == NULL:             # <<<<<<<<<<<<<<
 *             PyErr_NoMemory()
 *         return ctx
*/
  }

  /* "cyjs/_cyjs.pyx":174
 *         if ctx == NULL:
 *             PyErr_NoMemory()
 *         return ctx             # <<<<<<<<<<<<<<
 * 
 *     cpdef void run_gc(self):
*/
  __pyx_r = __pyx_v_ctx;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":169
 *         return JS_IsJobPending(self.rt)
 * 
 *     cdef JSContext* new_context(self) except NULL:             # <<<<<<<<<<<<<<
 *         # Use a raw context to enable higher performance when we can disable things...
 *         cdef JSContext* ctx = JS_NewContextRaw(self.rt)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.new_context", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":176
 *         return ctx
 * 
 *     cpdef void run_gc(self):             # <<<<<<<<<<<<<<
 *         """
 *         Runs QuickJS-NG's internal Garbage Collector
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_11run_gc(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_4cyjs_5_cyjs_7Runtime_run_gc(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run_gc", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_run_gc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Runtime_11run_gc)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 176, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":182
 *         **Warning!** Use at your own risk.
 *         """
 *         JS_RunGC(self.rt)             # <<<<<<<<<<<<<<
 * 
 *     # While I would go ahead start to inline these, python also needs a way to call them...
*/
  JS_RunGC(__pyx_v_self->rt);

  /* "cyjs/_cyjs.pyx":176
 *         return ctx
 * 
 *     cpdef void run_gc(self):             # <<<<<<<<<<<<<<
 *         """
 *         Runs QuickJS-NG's internal Garbage Collector
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.run_gc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_11run_gc(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4cyjs_5_cyjs_7Runtime_10run_gc, "\n        Runs QuickJS-NG's internal Garbage Collector\n\n        **Warning!** Use at your own risk.\n        ");
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Runtime_11run_gc = {"run_gc", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Runtime_11run_gc, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4cyjs_5_cyjs_7Runtime_10run_gc};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_11run_gc(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("run_gc (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("run_gc", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("run_gc", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Runtime_10run_gc(((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_10run_gc(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run_gc", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_4cyjs_5_cyjs_7Runtime_run_gc(__pyx_v_self, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 176, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.run_gc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":185
 * 
 *     # While I would go ahead start to inline these, python also needs a way to call them...
 *     cpdef void set_memory_limit(self, size_t limit):             # <<<<<<<<<<<<<<
 *         JS_SetMemoryLimit(self.rt, limit)
 * 
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_13set_memory_limit(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_4cyjs_5_cyjs_7Runtime_set_memory_limit(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, size_t __pyx_v_limit, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_memory_limit", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_set_memory_limit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Runtime_13set_memory_limit)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = __Pyx_PyLong_FromSize_t(__pyx_v_limit); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_6 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 185, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":186
 *     # While I would go ahead start to inline these, python also needs a way to call them...
 *     cpdef void set_memory_limit(self, size_t limit):
 *         JS_SetMemoryLimit(self.rt, limit)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void set_max_stack_size(self, size_t max_stack_size):
*/
  JS_SetMemoryLimit(__pyx_v_self->rt, __pyx_v_limit);

  /* "cyjs/_cyjs.pyx":185
 * 
 *     # While I would go ahead start to inline these, python also needs a way to call them...
 *     cpdef void set_memory_limit(self, size_t limit):             # <<<<<<<<<<<<<<
 *         JS_SetMemoryLimit(self.rt, limit)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.set_memory_limit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_13set_memory_limit(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Runtime_13set_memory_limit = {"set_memory_limit", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Runtime_13set_memory_limit, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_13set_memory_limit(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  size_t __pyx_v_limit;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_memory_limit (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_limit,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 185, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 185, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_memory_limit", 0) < (0)) __PYX_ERR(0, 185, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_memory_limit", 1, 1, 1, i); __PYX_ERR(0, 185, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 185, __pyx_L3_error)
    }
    __pyx_v_limit = __Pyx_PyLong_As_size_t(values[0]); if (unlikely((__pyx_v_limit == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 185, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_memory_limit", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 185, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.set_memory_limit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Runtime_12set_memory_limit(((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)__pyx_v_self), __pyx_v_limit);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_12set_memory_limit(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, size_t __pyx_v_limit) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_memory_limit", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_4cyjs_5_cyjs_7Runtime_set_memory_limit(__pyx_v_self, __pyx_v_limit, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 185, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.set_memory_limit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":188
 *         JS_SetMemoryLimit(self.rt, limit)
 * 
 *     cpdef void set_max_stack_size(self, size_t max_stack_size):             # <<<<<<<<<<<<<<
 *         JS_SetMaxStackSize(self.rt, max_stack_size)
 * 
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_15set_max_stack_size(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_4cyjs_5_cyjs_7Runtime_set_max_stack_size(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, size_t __pyx_v_max_stack_size, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_max_stack_size", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_set_max_stack_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Runtime_15set_max_stack_size)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = __Pyx_PyLong_FromSize_t(__pyx_v_max_stack_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 188, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_6 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 188, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":189
 * 
 *     cpdef void set_max_stack_size(self, size_t max_stack_size):
 *         JS_SetMaxStackSize(self.rt, max_stack_size)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void update_statck_top(self):
*/
  JS_SetMaxStackSize(__pyx_v_self->rt, __pyx_v_max_stack_size);

  /* "cyjs/_cyjs.pyx":188
 *         JS_SetMemoryLimit(self.rt, limit)
 * 
 *     cpdef void set_max_stack_size(self, size_t max_stack_size):             # <<<<<<<<<<<<<<
 *         JS_SetMaxStackSize(self.rt, max_stack_size)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.set_max_stack_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_15set_max_stack_size(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Runtime_15set_max_stack_size = {"set_max_stack_size", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Runtime_15set_max_stack_size, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_15set_max_stack_size(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  size_t __pyx_v_max_stack_size;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_max_stack_size (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_max_stack_size,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 188, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 188, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_max_stack_size", 0) < (0)) __PYX_ERR(0, 188, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_max_stack_size", 1, 1, 1, i); __PYX_ERR(0, 188, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 188, __pyx_L3_error)
    }
    __pyx_v_max_stack_size = __Pyx_PyLong_As_size_t(values[0]); if (unlikely((__pyx_v_max_stack_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 188, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_max_stack_size", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 188, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.set_max_stack_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Runtime_14set_max_stack_size(((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)__pyx_v_self), __pyx_v_max_stack_size);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_14set_max_stack_size(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, size_t __pyx_v_max_stack_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_max_stack_size", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_4cyjs_5_cyjs_7Runtime_set_max_stack_size(__pyx_v_self, __pyx_v_max_stack_size, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 188, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.set_max_stack_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":191
 *         JS_SetMaxStackSize(self.rt, max_stack_size)
 * 
 *     cpdef void update_statck_top(self):             # <<<<<<<<<<<<<<
 *         JS_UpdateStackTop(self.rt)
 * 
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_17update_statck_top(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_4cyjs_5_cyjs_7Runtime_update_statck_top(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update_statck_top", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_update_statck_top); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Runtime_17update_statck_top)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 191, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":192
 * 
 *     cpdef void update_statck_top(self):
 *         JS_UpdateStackTop(self.rt)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object set_promise_hook(self, object func):
*/
  JS_UpdateStackTop(__pyx_v_self->rt);

  /* "cyjs/_cyjs.pyx":191
 *         JS_SetMaxStackSize(self.rt, max_stack_size)
 * 
 *     cpdef void update_statck_top(self):             # <<<<<<<<<<<<<<
 *         JS_UpdateStackTop(self.rt)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.update_statck_top", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_17update_statck_top(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Runtime_17update_statck_top = {"update_statck_top", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Runtime_17update_statck_top, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_17update_statck_top(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update_statck_top (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("update_statck_top", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("update_statck_top", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Runtime_16update_statck_top(((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_16update_statck_top(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update_statck_top", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_4cyjs_5_cyjs_7Runtime_update_statck_top(__pyx_v_self, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 191, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.update_statck_top", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":194
 *         JS_UpdateStackTop(self.rt)
 * 
 *     cpdef object set_promise_hook(self, object func):             # <<<<<<<<<<<<<<
 *         if not callable(func):
 *             raise TypeError("Promise hook must be callable.")
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_19set_promise_hook(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Runtime_set_promise_hook(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, PyObject *__pyx_v_func, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_promise_hook", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_set_promise_hook); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Runtime_19set_promise_hook)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_func};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 194, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":195
 * 
 *     cpdef object set_promise_hook(self, object func):
 *         if not callable(func):             # <<<<<<<<<<<<<<
 *             raise TypeError("Promise hook must be callable.")
 *         self.promise_hook = PromiseHook.new(self, func)
*/
  __pyx_t_6 = __Pyx_PyCallable_Check(__pyx_v_func); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 195, __pyx_L1_error)
  __pyx_t_7 = (!__pyx_t_6);
  if (unlikely(__pyx_t_7)) {

    /* "cyjs/_cyjs.pyx":196
 *     cpdef object set_promise_hook(self, object func):
 *         if not callable(func):
 *             raise TypeError("Promise hook must be callable.")             # <<<<<<<<<<<<<<
 *         self.promise_hook = PromiseHook.new(self, func)
 *         self.has_promise_hook = True
*/
    __pyx_t_2 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Promise_hook_must_be_callable};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 196, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 196, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":195
 * 
 *     cpdef object set_promise_hook(self, object func):
 *         if not callable(func):             # <<<<<<<<<<<<<<
 *             raise TypeError("Promise hook must be callable.")
 *         self.promise_hook = PromiseHook.new(self, func)
*/
  }

  /* "cyjs/_cyjs.pyx":197
 *         if not callable(func):
 *             raise TypeError("Promise hook must be callable.")
 *         self.promise_hook = PromiseHook.new(self, func)             # <<<<<<<<<<<<<<
 *         self.has_promise_hook = True
 *         JS_SetPromiseHook(self.rt, on_promise_hook, <void*>self.promise_hook)
*/
  __pyx_t_1 = ((PyObject *)__pyx_f_4cyjs_5_cyjs_11PromiseHook_new(__pyx_v_self, __pyx_v_func)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->promise_hook);
  __Pyx_DECREF((PyObject *)__pyx_v_self->promise_hook);
  __pyx_v_self->promise_hook = ((struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":198
 *             raise TypeError("Promise hook must be callable.")
 *         self.promise_hook = PromiseHook.new(self, func)
 *         self.has_promise_hook = True             # <<<<<<<<<<<<<<
 *         JS_SetPromiseHook(self.rt, on_promise_hook, <void*>self.promise_hook)
 * 
*/
  __pyx_v_self->has_promise_hook = 1;

  /* "cyjs/_cyjs.pyx":199
 *         self.promise_hook = PromiseHook.new(self, func)
 *         self.has_promise_hook = True
 *         JS_SetPromiseHook(self.rt, on_promise_hook, <void*>self.promise_hook)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  JS_SetPromiseHook(__pyx_v_self->rt, __pyx_f_4cyjs_5_cyjs_on_promise_hook, ((void *)__pyx_v_self->promise_hook));

  /* "cyjs/_cyjs.pyx":194
 *         JS_UpdateStackTop(self.rt)
 * 
 *     cpdef object set_promise_hook(self, object func):             # <<<<<<<<<<<<<<
 *         if not callable(func):
 *             raise TypeError("Promise hook must be callable.")
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.set_promise_hook", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_19set_promise_hook(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Runtime_19set_promise_hook = {"set_promise_hook", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Runtime_19set_promise_hook, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_19set_promise_hook(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_promise_hook (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_func,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 194, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 194, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_promise_hook", 0) < (0)) __PYX_ERR(0, 194, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_promise_hook", 1, 1, 1, i); __PYX_ERR(0, 194, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 194, __pyx_L3_error)
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_promise_hook", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 194, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.set_promise_hook", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Runtime_18set_promise_hook(((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)__pyx_v_self), __pyx_v_func);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_18set_promise_hook(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_promise_hook", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_7Runtime_set_promise_hook(__pyx_v_self, __pyx_v_func, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.set_promise_hook", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":202
 * 
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.rt != NULL:
 *             JS_FreeRuntime(self.rt)
*/

/* Python wrapper */
static void __pyx_pw_4cyjs_5_cyjs_7Runtime_21__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4cyjs_5_cyjs_7Runtime_21__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_4cyjs_5_cyjs_7Runtime_20__dealloc__(((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4cyjs_5_cyjs_7Runtime_20__dealloc__(struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self) {
  int __pyx_t_1;

  /* "cyjs/_cyjs.pyx":203
 * 
 *     def __dealloc__(self):
 *         if self.rt != NULL:             # <<<<<<<<<<<<<<
 *             JS_FreeRuntime(self.rt)
 * 
*/
  __pyx_t_1 = (__pyx_v_self->rt != NULL);
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":204
 *     def __dealloc__(self):
 *         if self.rt != NULL:
 *             JS_FreeRuntime(self.rt)             # <<<<<<<<<<<<<<
 * 
 * cdef int py_to_atom(JSContext* ctx, object val, JSAtom* at) except -1:
*/
    JS_FreeRuntime(__pyx_v_self->rt);

    /* "cyjs/_cyjs.pyx":203
 * 
 *     def __dealloc__(self):
 *         if self.rt != NULL:             # <<<<<<<<<<<<<<
 *             JS_FreeRuntime(self.rt)
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":202
 * 
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.rt != NULL:
 *             JS_FreeRuntime(self.rt)
*/

  /* function exit code */
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.rt cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_23__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Runtime_23__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Runtime_23__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_23__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Runtime_22__reduce_cython__(((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_22__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.rt cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.rt cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_rt_cannot_be_converted_to_a, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.rt cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.rt cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.rt cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_25__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Runtime_25__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Runtime_25__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Runtime_25__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Runtime_24__setstate_cython__(((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Runtime_24__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.rt cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.rt cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_rt_cannot_be_converted_to_a, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.rt cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.rt cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.Runtime.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":206
 *             JS_FreeRuntime(self.rt)
 * 
 * cdef int py_to_atom(JSContext* ctx, object val, JSAtom* at) except -1:             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     if isinstance(val, Object):
*/

static int __pyx_f_4cyjs_5_cyjs_py_to_atom(JSContext *__pyx_v_ctx, PyObject *__pyx_v_val, JSAtom *__pyx_v_at) {
  Py_buffer __pyx_v_view;
  JSAtom __pyx_v_atom;
  int __pyx_r;
  int __pyx_t_1;
  uint32_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyjs/_cyjs.pyx":208
 * cdef int py_to_atom(JSContext* ctx, object val, JSAtom* at) except -1:
 *     cdef Py_buffer view
 *     if isinstance(val, Object):             # <<<<<<<<<<<<<<
 *         at[0] = JS_ValueToAtom(ctx, (<Object>val).value)
 *     elif isinstance(val, int):
*/
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_val, __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object); 
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":209
 *     cdef Py_buffer view
 *     if isinstance(val, Object):
 *         at[0] = JS_ValueToAtom(ctx, (<Object>val).value)             # <<<<<<<<<<<<<<
 *     elif isinstance(val, int):
 *         at[0] = JS_NewAtomUInt32(ctx, <uint32_t>val)
*/
    (__pyx_v_at[0]) = JS_ValueToAtom(__pyx_v_ctx, ((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_val)->value);

    /* "cyjs/_cyjs.pyx":208
 * cdef int py_to_atom(JSContext* ctx, object val, JSAtom* at) except -1:
 *     cdef Py_buffer view
 *     if isinstance(val, Object):             # <<<<<<<<<<<<<<
 *         at[0] = JS_ValueToAtom(ctx, (<Object>val).value)
 *     elif isinstance(val, int):
*/
    goto __pyx_L3;
  }

  /* "cyjs/_cyjs.pyx":210
 *     if isinstance(val, Object):
 *         at[0] = JS_ValueToAtom(ctx, (<Object>val).value)
 *     elif isinstance(val, int):             # <<<<<<<<<<<<<<
 *         at[0] = JS_NewAtomUInt32(ctx, <uint32_t>val)
 *     else:
*/
  __pyx_t_1 = PyLong_Check(__pyx_v_val); 
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":211
 *         at[0] = JS_ValueToAtom(ctx, (<Object>val).value)
 *     elif isinstance(val, int):
 *         at[0] = JS_NewAtomUInt32(ctx, <uint32_t>val)             # <<<<<<<<<<<<<<
 *     else:
 *         if cyjs_get_buffer(val, &view) < 0:
*/
    __pyx_t_2 = __Pyx_PyLong_As_uint32_t(__pyx_v_val); if (unlikely((__pyx_t_2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 211, __pyx_L1_error)
    (__pyx_v_at[0]) = JS_NewAtomUInt32(__pyx_v_ctx, ((uint32_t)__pyx_t_2));

    /* "cyjs/_cyjs.pyx":210
 *     if isinstance(val, Object):
 *         at[0] = JS_ValueToAtom(ctx, (<Object>val).value)
 *     elif isinstance(val, int):             # <<<<<<<<<<<<<<
 *         at[0] = JS_NewAtomUInt32(ctx, <uint32_t>val)
 *     else:
*/
    goto __pyx_L3;
  }

  /* "cyjs/_cyjs.pyx":213
 *         at[0] = JS_NewAtomUInt32(ctx, <uint32_t>val)
 *     else:
 *         if cyjs_get_buffer(val, &view) < 0:             # <<<<<<<<<<<<<<
 *             return -1
 *         atom = JS_NewAtomLen(ctx, <const char*>view.buf, view.len)
*/
  /*else*/ {
    __pyx_t_3 = cyjs_get_buffer(__pyx_v_val, (&__pyx_v_view)); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 213, __pyx_L1_error)
    __pyx_t_1 = (__pyx_t_3 < 0);
    if (__pyx_t_1) {

      /* "cyjs/_cyjs.pyx":214
 *     else:
 *         if cyjs_get_buffer(val, &view) < 0:
 *             return -1             # <<<<<<<<<<<<<<
 *         atom = JS_NewAtomLen(ctx, <const char*>view.buf, view.len)
 *         cyjs_release_buffer(&view)
*/
      __pyx_r = -1;
      goto __pyx_L0;

      /* "cyjs/_cyjs.pyx":213
 *         at[0] = JS_NewAtomUInt32(ctx, <uint32_t>val)
 *     else:
 *         if cyjs_get_buffer(val, &view) < 0:             # <<<<<<<<<<<<<<
 *             return -1
 *         atom = JS_NewAtomLen(ctx, <const char*>view.buf, view.len)
*/
    }

    /* "cyjs/_cyjs.pyx":215
 *         if cyjs_get_buffer(val, &view) < 0:
 *             return -1
 *         atom = JS_NewAtomLen(ctx, <const char*>view.buf, view.len)             # <<<<<<<<<<<<<<
 *         cyjs_release_buffer(&view)
 *         if atom == 0:
*/
    __pyx_v_atom = JS_NewAtomLen(__pyx_v_ctx, ((char const *)__pyx_v_view.buf), __pyx_v_view.len);

    /* "cyjs/_cyjs.pyx":216
 *             return -1
 *         atom = JS_NewAtomLen(ctx, <const char*>view.buf, view.len)
 *         cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *         if atom == 0:
 *             PyErr_SetObject(JSError, "Failed to convert python value to JSAtom")
*/
    cyjs_release_buffer((&__pyx_v_view));

    /* "cyjs/_cyjs.pyx":217
 *         atom = JS_NewAtomLen(ctx, <const char*>view.buf, view.len)
 *         cyjs_release_buffer(&view)
 *         if atom == 0:             # <<<<<<<<<<<<<<
 *             PyErr_SetObject(JSError, "Failed to convert python value to JSAtom")
 *             return -1
*/
    __pyx_t_1 = (__pyx_v_atom == 0);
    if (__pyx_t_1) {

      /* "cyjs/_cyjs.pyx":218
 *         cyjs_release_buffer(&view)
 *         if atom == 0:
 *             PyErr_SetObject(JSError, "Failed to convert python value to JSAtom")             # <<<<<<<<<<<<<<
 *             return -1
 *         at[0] = atom
*/
      PyErr_SetObject(((PyObject *)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_JSError), __pyx_mstate_global->__pyx_kp_u_Failed_to_convert_python_value_t);

      /* "cyjs/_cyjs.pyx":219
 *         if atom == 0:
 *             PyErr_SetObject(JSError, "Failed to convert python value to JSAtom")
 *             return -1             # <<<<<<<<<<<<<<
 *         at[0] = atom
 *     return 0
*/
      __pyx_r = -1;
      goto __pyx_L0;

      /* "cyjs/_cyjs.pyx":217
 *         atom = JS_NewAtomLen(ctx, <const char*>view.buf, view.len)
 *         cyjs_release_buffer(&view)
 *         if atom == 0:             # <<<<<<<<<<<<<<
 *             PyErr_SetObject(JSError, "Failed to convert python value to JSAtom")
 *             return -1
*/
    }

    /* "cyjs/_cyjs.pyx":220
 *             PyErr_SetObject(JSError, "Failed to convert python value to JSAtom")
 *             return -1
 *         at[0] = atom             # <<<<<<<<<<<<<<
 *     return 0
 * 
*/
    (__pyx_v_at[0]) = __pyx_v_atom;
  }
  __pyx_L3:;

  /* "cyjs/_cyjs.pyx":221
 *             return -1
 *         at[0] = atom
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef object atom_to_py(JSContext* ctx, JSAtom at):
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":206
 *             JS_FreeRuntime(self.rt)
 * 
 * cdef int py_to_atom(JSContext* ctx, object val, JSAtom* at) except -1:             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     if isinstance(val, Object):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.py_to_atom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":223
 *     return 0
 * 
 * cdef object atom_to_py(JSContext* ctx, JSAtom at):             # <<<<<<<<<<<<<<
 *     cdef size_t size
 *     cdef const char* c_str = JS_AtomToCStringLen(ctx, &size, at)
*/

static PyObject *__pyx_f_4cyjs_5_cyjs_atom_to_py(JSContext *__pyx_v_ctx, JSAtom __pyx_v_at) {
  size_t __pyx_v_size;
  char const *__pyx_v_c_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  char const *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("atom_to_py", 0);

  /* "cyjs/_cyjs.pyx":225
 * cdef object atom_to_py(JSContext* ctx, JSAtom at):
 *     cdef size_t size
 *     cdef const char* c_str = JS_AtomToCStringLen(ctx, &size, at)             # <<<<<<<<<<<<<<
 *     try:
 *         return PyUnicode_FromStringAndSize(c_str, <Py_ssize_t>size)
*/
  __pyx_v_c_str = JS_AtomToCStringLen(__pyx_v_ctx, (&__pyx_v_size), __pyx_v_at);

  /* "cyjs/_cyjs.pyx":226
 *     cdef size_t size
 *     cdef const char* c_str = JS_AtomToCStringLen(ctx, &size, at)
 *     try:             # <<<<<<<<<<<<<<
 *         return PyUnicode_FromStringAndSize(c_str, <Py_ssize_t>size)
 *     finally:
*/
  /*try:*/ {

    /* "cyjs/_cyjs.pyx":227
 *     cdef const char* c_str = JS_AtomToCStringLen(ctx, &size, at)
 *     try:
 *         return PyUnicode_FromStringAndSize(c_str, <Py_ssize_t>size)             # <<<<<<<<<<<<<<
 *     finally:
 *         JS_FreeCString(ctx, c_str)
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyUnicode_FromStringAndSize(__pyx_v_c_str, ((Py_ssize_t)__pyx_v_size)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 227, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L3_return;
  }

  /* "cyjs/_cyjs.pyx":229
 *         return PyUnicode_FromStringAndSize(c_str, <Py_ssize_t>size)
 *     finally:
 *         JS_FreeCString(ctx, c_str)             # <<<<<<<<<<<<<<
 * 
 * # TODO: Planned for to_quickjs to speedup certain funtions
*/
  /*finally:*/ {
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
       __Pyx_ExceptionSwap(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      if ( unlikely(__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0)) __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __pyx_t_2 = __pyx_lineno; __pyx_t_3 = __pyx_clineno; __pyx_t_4 = __pyx_filename;
      {
        JS_FreeCString(__pyx_v_ctx, __pyx_v_c_str);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ErrRestore(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0;
      __pyx_lineno = __pyx_t_2; __pyx_clineno = __pyx_t_3; __pyx_filename = __pyx_t_4;
      goto __pyx_L1_error;
    }
    __pyx_L3_return: {
      __pyx_t_10 = __pyx_r;
      __pyx_r = 0;
      JS_FreeCString(__pyx_v_ctx, __pyx_v_c_str);
      __pyx_r = __pyx_t_10;
      __pyx_t_10 = 0;
      goto __pyx_L0;
    }
  }

  /* "cyjs/_cyjs.pyx":223
 *     return 0
 * 
 * cdef object atom_to_py(JSContext* ctx, JSAtom at):             # <<<<<<<<<<<<<<
 *     cdef size_t size
 *     cdef const char* c_str = JS_AtomToCStringLen(ctx, &size, at)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.atom_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":236
 * # this gets used quite a few times hence inlining it felt
 * # like the best choice of apporch
 * cdef inline int fast_parse_json_from_obj(JSContext* ctx, JSValue* val, object obj) noexcept:             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef JSValue value
*/

static CYTHON_INLINE int __pyx_f_4cyjs_5_cyjs_fast_parse_json_from_obj(JSContext *__pyx_v_ctx, JSValue *__pyx_v_val, PyObject *__pyx_v_obj) {
  Py_buffer __pyx_v_view;
  JSValue __pyx_v_value;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyjs/_cyjs.pyx":240
 *     cdef JSValue value
 * 
 *     if cyjs_get_buffer(obj, &view) < 0:             # <<<<<<<<<<<<<<
 *         return -1
 *     value = JS_ParseJSON(ctx, <const char*>view.buf, <size_t>view.len, b"<input>")
*/
  __pyx_t_1 = cyjs_get_buffer(__pyx_v_obj, (&__pyx_v_view)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 240, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 < 0);
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":241
 * 
 *     if cyjs_get_buffer(obj, &view) < 0:
 *         return -1             # <<<<<<<<<<<<<<
 *     value = JS_ParseJSON(ctx, <const char*>view.buf, <size_t>view.len, b"<input>")
 *     cyjs_release_buffer(&view)
*/
    __pyx_r = -1;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":240
 *     cdef JSValue value
 * 
 *     if cyjs_get_buffer(obj, &view) < 0:             # <<<<<<<<<<<<<<
 *         return -1
 *     value = JS_ParseJSON(ctx, <const char*>view.buf, <size_t>view.len, b"<input>")
*/
  }

  /* "cyjs/_cyjs.pyx":242
 *     if cyjs_get_buffer(obj, &view) < 0:
 *         return -1
 *     value = JS_ParseJSON(ctx, <const char*>view.buf, <size_t>view.len, b"<input>")             # <<<<<<<<<<<<<<
 *     cyjs_release_buffer(&view)
 *     if JS_IsException(value):
*/
  __pyx_v_value = JS_ParseJSON(__pyx_v_ctx, ((char const *)__pyx_v_view.buf), ((size_t)__pyx_v_view.len), __pyx_k_input);

  /* "cyjs/_cyjs.pyx":243
 *         return -1
 *     value = JS_ParseJSON(ctx, <const char*>view.buf, <size_t>view.len, b"<input>")
 *     cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *     if JS_IsException(value):
 *         return -1
*/
  cyjs_release_buffer((&__pyx_v_view));

  /* "cyjs/_cyjs.pyx":244
 *     value = JS_ParseJSON(ctx, <const char*>view.buf, <size_t>view.len, b"<input>")
 *     cyjs_release_buffer(&view)
 *     if JS_IsException(value):             # <<<<<<<<<<<<<<
 *         return -1
 *     # same as *val = value in C since were just saying pick up this address please.
*/
  __pyx_t_2 = JS_IsException(__pyx_v_value);
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":245
 *     cyjs_release_buffer(&view)
 *     if JS_IsException(value):
 *         return -1             # <<<<<<<<<<<<<<
 *     # same as *val = value in C since were just saying pick up this address please.
 *     val[0] = value
*/
    __pyx_r = -1;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":244
 *     value = JS_ParseJSON(ctx, <const char*>view.buf, <size_t>view.len, b"<input>")
 *     cyjs_release_buffer(&view)
 *     if JS_IsException(value):             # <<<<<<<<<<<<<<
 *         return -1
 *     # same as *val = value in C since were just saying pick up this address please.
*/
  }

  /* "cyjs/_cyjs.pyx":247
 *         return -1
 *     # same as *val = value in C since were just saying pick up this address please.
 *     val[0] = value             # <<<<<<<<<<<<<<
 *     return 0
 * 
*/
  (__pyx_v_val[0]) = __pyx_v_value;

  /* "cyjs/_cyjs.pyx":248
 *     # same as *val = value in C since were just saying pick up this address please.
 *     val[0] = value
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":236
 * # this gets used quite a few times hence inlining it felt
 * # like the best choice of apporch
 * cdef inline int fast_parse_json_from_obj(JSContext* ctx, JSValue* val, object obj) noexcept:             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef JSValue value
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("cyjs._cyjs.fast_parse_json_from_obj", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":251
 * 
 * 
 * cdef inline int fast_serlize_dict_or_list(JSContext* ctx, JSValue* val, object obj) noexcept:             # <<<<<<<<<<<<<<
 *     try:
 *         # TODO: Try using an external json cython library for this
*/

static CYTHON_INLINE int __pyx_f_4cyjs_5_cyjs_fast_serlize_dict_or_list(JSContext *__pyx_v_ctx, JSValue *__pyx_v_val, PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fast_serlize_dict_or_list", 0);

  /* "cyjs/_cyjs.pyx":252
 * 
 * cdef inline int fast_serlize_dict_or_list(JSContext* ctx, JSValue* val, object obj) noexcept:
 *     try:             # <<<<<<<<<<<<<<
 *         # TODO: Try using an external json cython library for this
 *         # in the future (I think I had a yyjson thingy somewhere
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "cyjs/_cyjs.pyx":257
 *         # and a rewrite of that project to become cython-only might
 *         # do well here)
 *         return fast_parse_json_from_obj(ctx, val, json.dumps(obj))             # <<<<<<<<<<<<<<
 *     except BaseException:
 *         # reraise later...
*/
      __pyx_t_5 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_json); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 257, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_dumps); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 257, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_8 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
        assert(__pyx_t_5);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
        __pyx_t_8 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_obj};
        __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 257, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
      }
      __pyx_r = __pyx_f_4cyjs_5_cyjs_fast_parse_json_from_obj(__pyx_v_ctx, __pyx_v_val, __pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "cyjs/_cyjs.pyx":252
 * 
 * cdef inline int fast_serlize_dict_or_list(JSContext* ctx, JSValue* val, object obj) noexcept:
 *     try:             # <<<<<<<<<<<<<<
 *         # TODO: Try using an external json cython library for this
 *         # in the future (I think I had a yyjson thingy somewhere
*/
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "cyjs/_cyjs.pyx":258
 *         # do well here)
 *         return fast_parse_json_from_obj(ctx, val, json.dumps(obj))
 *     except BaseException:             # <<<<<<<<<<<<<<
 *         # reraise later...
 *         return -1
*/
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_BaseException))));
    if (__pyx_t_9) {
      __Pyx_ErrRestore(0,0,0);

      /* "cyjs/_cyjs.pyx":260
 *     except BaseException:
 *         # reraise later...
 *         return -1             # <<<<<<<<<<<<<<
 * 
 * 
*/
      __pyx_r = -1;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;

    /* "cyjs/_cyjs.pyx":252
 * 
 * cdef inline int fast_serlize_dict_or_list(JSContext* ctx, JSValue* val, object obj) noexcept:
 *     try:             # <<<<<<<<<<<<<<
 *         # TODO: Try using an external json cython library for this
 *         # in the future (I think I had a yyjson thingy somewhere
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "cyjs/_cyjs.pyx":251
 * 
 * 
 * cdef inline int fast_serlize_dict_or_list(JSContext* ctx, JSValue* val, object obj) noexcept:             # <<<<<<<<<<<<<<
 *     try:
 *         # TODO: Try using an external json cython library for this
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_WriteUnraisable("cyjs._cyjs.fast_serlize_dict_or_list", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":263
 * 
 * 
 * cdef int fast_serlize_string(JSContext* ctx, JSValue* val, object obj):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef JSValue value
*/

static int __pyx_f_4cyjs_5_cyjs_fast_serlize_string(JSContext *__pyx_v_ctx, JSValue *__pyx_v_val, PyObject *__pyx_v_obj) {
  Py_buffer __pyx_v_view;
  JSValue __pyx_v_value;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyjs/_cyjs.pyx":267
 *     cdef JSValue value
 * 
 *     if cyjs_get_buffer(obj, &view) < 0:             # <<<<<<<<<<<<<<
 *         return -1
 *     value = JS_NewStringLen(ctx, <const char*>view.buf, <size_t>view.len)
*/
  __pyx_t_1 = cyjs_get_buffer(__pyx_v_obj, (&__pyx_v_view)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 267, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 < 0);
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":268
 * 
 *     if cyjs_get_buffer(obj, &view) < 0:
 *         return -1             # <<<<<<<<<<<<<<
 *     value = JS_NewStringLen(ctx, <const char*>view.buf, <size_t>view.len)
 *     cyjs_release_buffer(&view)
*/
    __pyx_r = -1;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":267
 *     cdef JSValue value
 * 
 *     if cyjs_get_buffer(obj, &view) < 0:             # <<<<<<<<<<<<<<
 *         return -1
 *     value = JS_NewStringLen(ctx, <const char*>view.buf, <size_t>view.len)
*/
  }

  /* "cyjs/_cyjs.pyx":269
 *     if cyjs_get_buffer(obj, &view) < 0:
 *         return -1
 *     value = JS_NewStringLen(ctx, <const char*>view.buf, <size_t>view.len)             # <<<<<<<<<<<<<<
 *     cyjs_release_buffer(&view)
 *     if JS_IsException(value):
*/
  __pyx_v_value = JS_NewStringLen(__pyx_v_ctx, ((char const *)__pyx_v_view.buf), ((size_t)__pyx_v_view.len));

  /* "cyjs/_cyjs.pyx":270
 *         return -1
 *     value = JS_NewStringLen(ctx, <const char*>view.buf, <size_t>view.len)
 *     cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *     if JS_IsException(value):
 *         return -1
*/
  cyjs_release_buffer((&__pyx_v_view));

  /* "cyjs/_cyjs.pyx":271
 *     value = JS_NewStringLen(ctx, <const char*>view.buf, <size_t>view.len)
 *     cyjs_release_buffer(&view)
 *     if JS_IsException(value):             # <<<<<<<<<<<<<<
 *         return -1
 *     val[0] = value
*/
  __pyx_t_2 = JS_IsException(__pyx_v_value);
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":272
 *     cyjs_release_buffer(&view)
 *     if JS_IsException(value):
 *         return -1             # <<<<<<<<<<<<<<
 *     val[0] = value
 *     return 0
*/
    __pyx_r = -1;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":271
 *     value = JS_NewStringLen(ctx, <const char*>view.buf, <size_t>view.len)
 *     cyjs_release_buffer(&view)
 *     if JS_IsException(value):             # <<<<<<<<<<<<<<
 *         return -1
 *     val[0] = value
*/
  }

  /* "cyjs/_cyjs.pyx":273
 *     if JS_IsException(value):
 *         return -1
 *     val[0] = value             # <<<<<<<<<<<<<<
 *     return 0
 * 
*/
  (__pyx_v_val[0]) = __pyx_v_value;

  /* "cyjs/_cyjs.pyx":274
 *         return -1
 *     val[0] = value
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":263
 * 
 * 
 * cdef int fast_serlize_string(JSContext* ctx, JSValue* val, object obj):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef JSValue value
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.fast_serlize_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":277
 * 
 * 
 * cdef inline JSValue py_to_js_exception(JSContext* ctx, object obj):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef JSValue value
*/

static CYTHON_INLINE JSValue __pyx_f_4cyjs_5_cyjs_py_to_js_exception(JSContext *__pyx_v_ctx, PyObject *__pyx_v_obj) {
  Py_buffer __pyx_v_view;
  JSValue __pyx_v_value;
  PyObject *__pyx_v_msg = 0;
  JSValue __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("py_to_js_exception", 0);

  /* "cyjs/_cyjs.pyx":280
 *     cdef Py_buffer view
 *     cdef JSValue value
 *     cdef str msg = PyObject_Str(obj)             # <<<<<<<<<<<<<<
 *     if cyjs_get_buffer(msg, &view) < 0:
 *         return CYJS_ThrowException(ctx, "Unable to create exception from a string object")
*/
  __pyx_t_1 = PyObject_Str(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 280, __pyx_L1_error)
  __pyx_v_msg = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":281
 *     cdef JSValue value
 *     cdef str msg = PyObject_Str(obj)
 *     if cyjs_get_buffer(msg, &view) < 0:             # <<<<<<<<<<<<<<
 *         return CYJS_ThrowException(ctx, "Unable to create exception from a string object")
 *     value = CYJS_ThrowException(ctx, <const char*>view.buf)
*/
  __pyx_t_2 = cyjs_get_buffer(__pyx_v_msg, (&__pyx_v_view)); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 281, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 < 0);
  if (__pyx_t_3) {

    /* "cyjs/_cyjs.pyx":282
 *     cdef str msg = PyObject_Str(obj)
 *     if cyjs_get_buffer(msg, &view) < 0:
 *         return CYJS_ThrowException(ctx, "Unable to create exception from a string object")             # <<<<<<<<<<<<<<
 *     value = CYJS_ThrowException(ctx, <const char*>view.buf)
 *     cyjs_release_buffer(&view)
*/
    __pyx_r = CYJS_ThrowException(__pyx_v_ctx, __pyx_k_Unable_to_create_exception_from);
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":281
 *     cdef JSValue value
 *     cdef str msg = PyObject_Str(obj)
 *     if cyjs_get_buffer(msg, &view) < 0:             # <<<<<<<<<<<<<<
 *         return CYJS_ThrowException(ctx, "Unable to create exception from a string object")
 *     value = CYJS_ThrowException(ctx, <const char*>view.buf)
*/
  }

  /* "cyjs/_cyjs.pyx":283
 *     if cyjs_get_buffer(msg, &view) < 0:
 *         return CYJS_ThrowException(ctx, "Unable to create exception from a string object")
 *     value = CYJS_ThrowException(ctx, <const char*>view.buf)             # <<<<<<<<<<<<<<
 *     cyjs_release_buffer(&view)
 *     return value
*/
  __pyx_v_value = CYJS_ThrowException(__pyx_v_ctx, ((char const *)__pyx_v_view.buf));

  /* "cyjs/_cyjs.pyx":284
 *         return CYJS_ThrowException(ctx, "Unable to create exception from a string object")
 *     value = CYJS_ThrowException(ctx, <const char*>view.buf)
 *     cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *     return value
 * 
*/
  cyjs_release_buffer((&__pyx_v_view));

  /* "cyjs/_cyjs.pyx":285
 *     value = CYJS_ThrowException(ctx, <const char*>view.buf)
 *     cyjs_release_buffer(&view)
 *     return value             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":277
 * 
 * 
 * cdef inline JSValue py_to_js_exception(JSContext* ctx, object obj):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef JSValue value
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.py_to_js_exception", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":290
 * # Returns -1 on failre as a failsafe for C to evacuate safely from
 * # TODO: Optimize under this logic ->
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:             # <<<<<<<<<<<<<<
 *     # Start with the best case scenario
 *     cdef int overflow = 0
*/

static int __pyx_fuse_0__pyx_f_4cyjs_5_cyjs_to_quickjs(CYTHON_UNUSED JSContext *__pyx_v_ctx, JSValue *__pyx_v_val, struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_obj) {
  CYTHON_UNUSED int __pyx_v_overflow;
  int __pyx_r;
  JSValue __pyx_t_1;

  /* "cyjs/_cyjs.pyx":292
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:
 *     # Start with the best case scenario
 *     cdef int overflow = 0             # <<<<<<<<<<<<<<
 *     cdef long ival
 * 
*/
  __pyx_v_overflow = 0;

  /* "cyjs/_cyjs.pyx":301
 *         return 0
 *     elif quickjs_type_t is JSFunction:
 *         val[0] = (<JSFunction>obj).value             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
  __pyx_t_1 = __pyx_v_obj->value;
  (__pyx_v_val[0]) = __pyx_t_1;

  /* "cyjs/_cyjs.pyx":302
 *     elif quickjs_type_t is JSFunction:
 *         val[0] = (<JSFunction>obj).value
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     elif quickjs_type_t is Exception:
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":290
 * # Returns -1 on failre as a failsafe for C to evacuate safely from
 * # TODO: Optimize under this logic ->
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:             # <<<<<<<<<<<<<<
 *     # Start with the best case scenario
 *     cdef int overflow = 0
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static int __pyx_fuse_1__pyx_f_4cyjs_5_cyjs_to_quickjs(CYTHON_UNUSED JSContext *__pyx_v_ctx, JSValue *__pyx_v_val, struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_obj) {
  CYTHON_UNUSED int __pyx_v_overflow;
  int __pyx_r;
  JSValue __pyx_t_1;

  /* "cyjs/_cyjs.pyx":292
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:
 *     # Start with the best case scenario
 *     cdef int overflow = 0             # <<<<<<<<<<<<<<
 *     cdef long ival
 * 
*/
  __pyx_v_overflow = 0;

  /* "cyjs/_cyjs.pyx":298
 *     # https://cython.readthedocs.io/en/latest/src/userguide/fusedtypes.html#type-checking-specializations
 *     if quickjs_type_t is Object:
 *         val[0] = (<Object>obj).value             # <<<<<<<<<<<<<<
 *         return 0
 *     elif quickjs_type_t is JSFunction:
*/
  __pyx_t_1 = __pyx_v_obj->value;
  (__pyx_v_val[0]) = __pyx_t_1;

  /* "cyjs/_cyjs.pyx":299
 *     if quickjs_type_t is Object:
 *         val[0] = (<Object>obj).value
 *         return 0             # <<<<<<<<<<<<<<
 *     elif quickjs_type_t is JSFunction:
 *         val[0] = (<JSFunction>obj).value
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":290
 * # Returns -1 on failre as a failsafe for C to evacuate safely from
 * # TODO: Optimize under this logic ->
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:             # <<<<<<<<<<<<<<
 *     # Start with the best case scenario
 *     cdef int overflow = 0
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static int __pyx_fuse_2__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *__pyx_v_ctx, JSValue *__pyx_v_val, PyObject *__pyx_v_obj) {
  CYTHON_UNUSED int __pyx_v_overflow;
  int __pyx_r;
  JSValue __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyjs/_cyjs.pyx":292
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:
 *     # Start with the best case scenario
 *     cdef int overflow = 0             # <<<<<<<<<<<<<<
 *     cdef long ival
 * 
*/
  __pyx_v_overflow = 0;

  /* "cyjs/_cyjs.pyx":305
 * 
 *     elif quickjs_type_t is Exception:
 *         val[0] = py_to_js_exception(ctx, obj)             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_py_to_js_exception(__pyx_v_ctx, __pyx_v_obj); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 305, __pyx_L1_error)
  (__pyx_v_val[0]) = __pyx_t_1;

  /* "cyjs/_cyjs.pyx":306
 *     elif quickjs_type_t is Exception:
 *         val[0] = py_to_js_exception(ctx, obj)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     elif quickjs_type_t is dict:
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":290
 * # Returns -1 on failre as a failsafe for C to evacuate safely from
 * # TODO: Optimize under this logic ->
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:             # <<<<<<<<<<<<<<
 *     # Start with the best case scenario
 *     cdef int overflow = 0
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("cyjs._cyjs.to_quickjs", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

static int __pyx_fuse_3__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *__pyx_v_ctx, JSValue *__pyx_v_val, PyObject *__pyx_v_obj) {
  CYTHON_UNUSED int __pyx_v_overflow;
  int __pyx_r;

  /* "cyjs/_cyjs.pyx":292
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:
 *     # Start with the best case scenario
 *     cdef int overflow = 0             # <<<<<<<<<<<<<<
 *     cdef long ival
 * 
*/
  __pyx_v_overflow = 0;

  /* "cyjs/_cyjs.pyx":309
 * 
 *     elif quickjs_type_t is dict:
 *         return fast_serlize_dict_or_list(ctx, val, obj)             # <<<<<<<<<<<<<<
 * 
 *     elif quickjs_type_t is list:
*/
  __pyx_r = __pyx_f_4cyjs_5_cyjs_fast_serlize_dict_or_list(__pyx_v_ctx, __pyx_v_val, __pyx_v_obj);
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":290
 * # Returns -1 on failre as a failsafe for C to evacuate safely from
 * # TODO: Optimize under this logic ->
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:             # <<<<<<<<<<<<<<
 *     # Start with the best case scenario
 *     cdef int overflow = 0
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static int __pyx_fuse_4__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *__pyx_v_ctx, JSValue *__pyx_v_val, PyObject *__pyx_v_obj) {
  CYTHON_UNUSED int __pyx_v_overflow;
  int __pyx_r;

  /* "cyjs/_cyjs.pyx":292
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:
 *     # Start with the best case scenario
 *     cdef int overflow = 0             # <<<<<<<<<<<<<<
 *     cdef long ival
 * 
*/
  __pyx_v_overflow = 0;

  /* "cyjs/_cyjs.pyx":312
 * 
 *     elif quickjs_type_t is list:
 *         return fast_serlize_dict_or_list(ctx, val, obj)             # <<<<<<<<<<<<<<
 * 
 *     elif quickjs_type_t is str:
*/
  __pyx_r = __pyx_f_4cyjs_5_cyjs_fast_serlize_dict_or_list(__pyx_v_ctx, __pyx_v_val, __pyx_v_obj);
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":290
 * # Returns -1 on failre as a failsafe for C to evacuate safely from
 * # TODO: Optimize under this logic ->
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:             # <<<<<<<<<<<<<<
 *     # Start with the best case scenario
 *     cdef int overflow = 0
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static int __pyx_fuse_5__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *__pyx_v_ctx, JSValue *__pyx_v_val, PyObject *__pyx_v_obj) {
  CYTHON_UNUSED int __pyx_v_overflow;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyjs/_cyjs.pyx":292
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:
 *     # Start with the best case scenario
 *     cdef int overflow = 0             # <<<<<<<<<<<<<<
 *     cdef long ival
 * 
*/
  __pyx_v_overflow = 0;

  /* "cyjs/_cyjs.pyx":315
 * 
 *     elif quickjs_type_t is str:
 *         return fast_serlize_string(ctx, val, obj)             # <<<<<<<<<<<<<<
 * 
 *     elif quickjs_type_t is int:
*/
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_fast_serlize_string(__pyx_v_ctx, __pyx_v_val, __pyx_v_obj); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 315, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":290
 * # Returns -1 on failre as a failsafe for C to evacuate safely from
 * # TODO: Optimize under this logic ->
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:             # <<<<<<<<<<<<<<
 *     # Start with the best case scenario
 *     cdef int overflow = 0
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("cyjs._cyjs.to_quickjs", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

static int __pyx_fuse_6__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *__pyx_v_ctx, JSValue *__pyx_v_val, int __pyx_v_obj) {
  CYTHON_UNUSED int __pyx_v_overflow;
  int __pyx_r;

  /* "cyjs/_cyjs.pyx":292
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:
 *     # Start with the best case scenario
 *     cdef int overflow = 0             # <<<<<<<<<<<<<<
 *     cdef long ival
 * 
*/
  __pyx_v_overflow = 0;

  /* "cyjs/_cyjs.pyx":318
 * 
 *     elif quickjs_type_t is int:
 *         val[0] = JS_NewInt32(ctx, obj)             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
  (__pyx_v_val[0]) = JS_NewInt32(__pyx_v_ctx, __pyx_v_obj);

  /* "cyjs/_cyjs.pyx":319
 *     elif quickjs_type_t is int:
 *         val[0] = JS_NewInt32(ctx, obj)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     elif quickjs_type_t is bint:
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":290
 * # Returns -1 on failre as a failsafe for C to evacuate safely from
 * # TODO: Optimize under this logic ->
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:             # <<<<<<<<<<<<<<
 *     # Start with the best case scenario
 *     cdef int overflow = 0
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static int __pyx_fuse_7__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *__pyx_v_ctx, JSValue *__pyx_v_val, float __pyx_v_obj) {
  int __pyx_v_overflow;
  long __pyx_v_ival;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  JSValue __pyx_t_3;
  long __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_7to_quickjs", 0);

  /* "cyjs/_cyjs.pyx":292
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:
 *     # Start with the best case scenario
 *     cdef int overflow = 0             # <<<<<<<<<<<<<<
 *     cdef long ival
 * 
*/
  __pyx_v_overflow = 0;

  /* "cyjs/_cyjs.pyx":335
 *         # can have PyObject_CheckBuffer(obj) since these aren't clearly explained.
 * 
 *         if isinstance(obj, Object):             # <<<<<<<<<<<<<<
 *             val[0] = (<Object>obj).value
 *             return 0
*/
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":336
 * 
 *         if isinstance(obj, Object):
 *             val[0] = (<Object>obj).value             # <<<<<<<<<<<<<<
 *             return 0
 *         elif isinstance(obj, JSFunction):
*/
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object)))) __PYX_ERR(0, 336, __pyx_L1_error)
    __pyx_t_3 = ((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_t_1)->value;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_val[0]) = __pyx_t_3;

    /* "cyjs/_cyjs.pyx":337
 *         if isinstance(obj, Object):
 *             val[0] = (<Object>obj).value
 *             return 0             # <<<<<<<<<<<<<<
 *         elif isinstance(obj, JSFunction):
 *             # very quick shortcut
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":335
 *         # can have PyObject_CheckBuffer(obj) since these aren't clearly explained.
 * 
 *         if isinstance(obj, Object):             # <<<<<<<<<<<<<<
 *             val[0] = (<Object>obj).value
 *             return 0
*/
  }

  /* "cyjs/_cyjs.pyx":338
 *             val[0] = (<Object>obj).value
 *             return 0
 *         elif isinstance(obj, JSFunction):             # <<<<<<<<<<<<<<
 *             # very quick shortcut
 *             val[0] = (<JSFunction>obj).value
*/
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_JSFunction); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":340
 *         elif isinstance(obj, JSFunction):
 *             # very quick shortcut
 *             val[0] = (<JSFunction>obj).value             # <<<<<<<<<<<<<<
 *             return 0
 * 
*/
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_JSFunction)))) __PYX_ERR(0, 340, __pyx_L1_error)
    __pyx_t_3 = ((struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)__pyx_t_1)->value;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_val[0]) = __pyx_t_3;

    /* "cyjs/_cyjs.pyx":341
 *             # very quick shortcut
 *             val[0] = (<JSFunction>obj).value
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(obj, Exception):
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":338
 *             val[0] = (<Object>obj).value
 *             return 0
 *         elif isinstance(obj, JSFunction):             # <<<<<<<<<<<<<<
 *             # very quick shortcut
 *             val[0] = (<JSFunction>obj).value
*/
  }

  /* "cyjs/_cyjs.pyx":343
 *             return 0
 * 
 *         elif isinstance(obj, Exception):             # <<<<<<<<<<<<<<
 *             # Convert exception
 *             val[0] = py_to_js_exception(ctx, obj)
*/
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyException_Check(__pyx_t_1); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":345
 *         elif isinstance(obj, Exception):
 *             # Convert exception
 *             val[0] = py_to_js_exception(ctx, obj)             # <<<<<<<<<<<<<<
 *             return 0
 * 
*/
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __pyx_f_4cyjs_5_cyjs_py_to_js_exception(__pyx_v_ctx, __pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 345, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_val[0]) = __pyx_t_3;

    /* "cyjs/_cyjs.pyx":346
 *             # Convert exception
 *             val[0] = py_to_js_exception(ctx, obj)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(obj, bool):
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":343
 *             return 0
 * 
 *         elif isinstance(obj, Exception):             # <<<<<<<<<<<<<<
 *             # Convert exception
 *             val[0] = py_to_js_exception(ctx, obj)
*/
  }

  /* "cyjs/_cyjs.pyx":348
 *             return 0
 * 
 *         elif isinstance(obj, bool):             # <<<<<<<<<<<<<<
 *             val[0] = JS_NewBool(ctx, <bint>obj)
 *             return 0
*/
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyBool_Check(__pyx_t_1); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":349
 * 
 *         elif isinstance(obj, bool):
 *             val[0] = JS_NewBool(ctx, <bint>obj)             # <<<<<<<<<<<<<<
 *             return 0
 * 
*/
    (__pyx_v_val[0]) = JS_NewBool(__pyx_v_ctx, (__pyx_v_obj != 0));

    /* "cyjs/_cyjs.pyx":350
 *         elif isinstance(obj, bool):
 *             val[0] = JS_NewBool(ctx, <bint>obj)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(obj, int):
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":348
 *             return 0
 * 
 *         elif isinstance(obj, bool):             # <<<<<<<<<<<<<<
 *             val[0] = JS_NewBool(ctx, <bint>obj)
 *             return 0
*/
  }

  /* "cyjs/_cyjs.pyx":352
 *             return 0
 * 
 *         elif isinstance(obj, int):             # <<<<<<<<<<<<<<
 *             ival = PyLong_AsLongAndOverflow(obj, &overflow)
 *             if overflow:
*/
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyLong_Check(__pyx_t_1); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":353
 * 
 *         elif isinstance(obj, int):
 *             ival = PyLong_AsLongAndOverflow(obj, &overflow)             # <<<<<<<<<<<<<<
 *             if overflow:
 *                 return fast_parse_json_from_obj(ctx, val, repr(obj))
*/
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyLong_AsLongAndOverflow(__pyx_t_1, (&__pyx_v_overflow)); if (unlikely(__pyx_t_4 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_ival = __pyx_t_4;

    /* "cyjs/_cyjs.pyx":354
 *         elif isinstance(obj, int):
 *             ival = PyLong_AsLongAndOverflow(obj, &overflow)
 *             if overflow:             # <<<<<<<<<<<<<<
 *                 return fast_parse_json_from_obj(ctx, val, repr(obj))
 *             val[0] = JS_NewInt32(ctx, ival)
*/
    __pyx_t_2 = (__pyx_v_overflow != 0);
    if (__pyx_t_2) {

      /* "cyjs/_cyjs.pyx":355
 *             ival = PyLong_AsLongAndOverflow(obj, &overflow)
 *             if overflow:
 *                 return fast_parse_json_from_obj(ctx, val, repr(obj))             # <<<<<<<<<<<<<<
 *             val[0] = JS_NewInt32(ctx, ival)
 *             return 0
*/
      __pyx_t_1 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 355, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_r = __pyx_f_4cyjs_5_cyjs_fast_parse_json_from_obj(__pyx_v_ctx, __pyx_v_val, __pyx_t_5);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "cyjs/_cyjs.pyx":354
 *         elif isinstance(obj, int):
 *             ival = PyLong_AsLongAndOverflow(obj, &overflow)
 *             if overflow:             # <<<<<<<<<<<<<<
 *                 return fast_parse_json_from_obj(ctx, val, repr(obj))
 *             val[0] = JS_NewInt32(ctx, ival)
*/
    }

    /* "cyjs/_cyjs.pyx":356
 *             if overflow:
 *                 return fast_parse_json_from_obj(ctx, val, repr(obj))
 *             val[0] = JS_NewInt32(ctx, ival)             # <<<<<<<<<<<<<<
 *             return 0
 * 
*/
    (__pyx_v_val[0]) = JS_NewInt32(__pyx_v_ctx, __pyx_v_ival);

    /* "cyjs/_cyjs.pyx":357
 *                 return fast_parse_json_from_obj(ctx, val, repr(obj))
 *             val[0] = JS_NewInt32(ctx, ival)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(obj, float):
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":352
 *             return 0
 * 
 *         elif isinstance(obj, int):             # <<<<<<<<<<<<<<
 *             ival = PyLong_AsLongAndOverflow(obj, &overflow)
 *             if overflow:
*/
  }

  /* "cyjs/_cyjs.pyx":359
 *             return 0
 * 
 *         elif isinstance(obj, float):             # <<<<<<<<<<<<<<
 *             val[0] = JS_NewFloat64(ctx, <double>obj)
 *             return 0
*/
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyFloat_Check(__pyx_t_5); 
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":360
 * 
 *         elif isinstance(obj, float):
 *             val[0] = JS_NewFloat64(ctx, <double>obj)             # <<<<<<<<<<<<<<
 *             return 0
 * 
*/
    (__pyx_v_val[0]) = JS_NewFloat64(__pyx_v_ctx, ((double)__pyx_v_obj));

    /* "cyjs/_cyjs.pyx":361
 *         elif isinstance(obj, float):
 *             val[0] = JS_NewFloat64(ctx, <double>obj)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(obj, str):
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":359
 *             return 0
 * 
 *         elif isinstance(obj, float):             # <<<<<<<<<<<<<<
 *             val[0] = JS_NewFloat64(ctx, <double>obj)
 *             return 0
*/
  }

  /* "cyjs/_cyjs.pyx":363
 *             return 0
 * 
 *         elif isinstance(obj, str):             # <<<<<<<<<<<<<<
 *             return fast_serlize_string(ctx, val, obj)
 * 
*/
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyUnicode_Check(__pyx_t_5); 
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":364
 * 
 *         elif isinstance(obj, str):
 *             return fast_serlize_string(ctx, val, obj)             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(obj, (dict, list)):
*/
    __pyx_t_5 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __pyx_f_4cyjs_5_cyjs_fast_serlize_string(__pyx_v_ctx, __pyx_v_val, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 364, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_6;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":363
 *             return 0
 * 
 *         elif isinstance(obj, str):             # <<<<<<<<<<<<<<
 *             return fast_serlize_string(ctx, val, obj)
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":366
 *             return fast_serlize_string(ctx, val, obj)
 * 
 *         elif isinstance(obj, (dict, list)):             # <<<<<<<<<<<<<<
 *             return fast_serlize_dict_or_list(ctx, val, obj)
 * 
*/
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyDict_Check(__pyx_t_5); 
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyList_Check(__pyx_t_5); 
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __pyx_t_7;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":367
 * 
 *         elif isinstance(obj, (dict, list)):
 *             return fast_serlize_dict_or_list(ctx, val, obj)             # <<<<<<<<<<<<<<
 * 
 *         elif PyObject_CheckBuffer(obj):
*/
    __pyx_t_5 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = __pyx_f_4cyjs_5_cyjs_fast_serlize_dict_or_list(__pyx_v_ctx, __pyx_v_val, __pyx_t_5);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":366
 *             return fast_serlize_string(ctx, val, obj)
 * 
 *         elif isinstance(obj, (dict, list)):             # <<<<<<<<<<<<<<
 *             return fast_serlize_dict_or_list(ctx, val, obj)
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":369
 *             return fast_serlize_dict_or_list(ctx, val, obj)
 * 
 *         elif PyObject_CheckBuffer(obj):             # <<<<<<<<<<<<<<
 *             return fast_serlize_string(ctx, val, obj)
 * 
*/
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyObject_CheckBuffer(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":370
 * 
 *         elif PyObject_CheckBuffer(obj):
 *             return fast_serlize_string(ctx, val, obj)             # <<<<<<<<<<<<<<
 * 
 *     PyErr_SetObject(TypeError, f"unknown object {type(obj).__name__!r}")
*/
    __pyx_t_5 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __pyx_f_4cyjs_5_cyjs_fast_serlize_string(__pyx_v_ctx, __pyx_v_val, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_6;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":369
 *             return fast_serlize_dict_or_list(ctx, val, obj)
 * 
 *         elif PyObject_CheckBuffer(obj):             # <<<<<<<<<<<<<<
 *             return fast_serlize_string(ctx, val, obj)
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":372
 *             return fast_serlize_string(ctx, val, obj)
 * 
 *     PyErr_SetObject(TypeError, f"unknown object {type(obj).__name__!r}")             # <<<<<<<<<<<<<<
 *     return -1
 * 
*/
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_obj); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_t_5)), __pyx_mstate_global->__pyx_n_u_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_t_1), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_unknown_object, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyErr_SetObject(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":373
 * 
 *     PyErr_SetObject(TypeError, f"unknown object {type(obj).__name__!r}")
 *     return -1             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = -1;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":290
 * # Returns -1 on failre as a failsafe for C to evacuate safely from
 * # TODO: Optimize under this logic ->
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:             # <<<<<<<<<<<<<<
 *     # Start with the best case scenario
 *     cdef int overflow = 0
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("cyjs._cyjs.to_quickjs", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_fuse_8__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *__pyx_v_ctx, JSValue *__pyx_v_val, PyObject *__pyx_v_obj) {
  int __pyx_v_overflow;
  long __pyx_v_ival;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  JSValue __pyx_t_2;
  long __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  double __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_8to_quickjs", 0);

  /* "cyjs/_cyjs.pyx":292
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:
 *     # Start with the best case scenario
 *     cdef int overflow = 0             # <<<<<<<<<<<<<<
 *     cdef long ival
 * 
*/
  __pyx_v_overflow = 0;

  /* "cyjs/_cyjs.pyx":335
 *         # can have PyObject_CheckBuffer(obj) since these aren't clearly explained.
 * 
 *         if isinstance(obj, Object):             # <<<<<<<<<<<<<<
 *             val[0] = (<Object>obj).value
 *             return 0
*/
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object); 
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":336
 * 
 *         if isinstance(obj, Object):
 *             val[0] = (<Object>obj).value             # <<<<<<<<<<<<<<
 *             return 0
 *         elif isinstance(obj, JSFunction):
*/
    __pyx_t_2 = ((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_obj)->value;
    (__pyx_v_val[0]) = __pyx_t_2;

    /* "cyjs/_cyjs.pyx":337
 *         if isinstance(obj, Object):
 *             val[0] = (<Object>obj).value
 *             return 0             # <<<<<<<<<<<<<<
 *         elif isinstance(obj, JSFunction):
 *             # very quick shortcut
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":335
 *         # can have PyObject_CheckBuffer(obj) since these aren't clearly explained.
 * 
 *         if isinstance(obj, Object):             # <<<<<<<<<<<<<<
 *             val[0] = (<Object>obj).value
 *             return 0
*/
  }

  /* "cyjs/_cyjs.pyx":338
 *             val[0] = (<Object>obj).value
 *             return 0
 *         elif isinstance(obj, JSFunction):             # <<<<<<<<<<<<<<
 *             # very quick shortcut
 *             val[0] = (<JSFunction>obj).value
*/
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_JSFunction); 
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":340
 *         elif isinstance(obj, JSFunction):
 *             # very quick shortcut
 *             val[0] = (<JSFunction>obj).value             # <<<<<<<<<<<<<<
 *             return 0
 * 
*/
    __pyx_t_2 = ((struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)__pyx_v_obj)->value;
    (__pyx_v_val[0]) = __pyx_t_2;

    /* "cyjs/_cyjs.pyx":341
 *             # very quick shortcut
 *             val[0] = (<JSFunction>obj).value
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(obj, Exception):
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":338
 *             val[0] = (<Object>obj).value
 *             return 0
 *         elif isinstance(obj, JSFunction):             # <<<<<<<<<<<<<<
 *             # very quick shortcut
 *             val[0] = (<JSFunction>obj).value
*/
  }

  /* "cyjs/_cyjs.pyx":343
 *             return 0
 * 
 *         elif isinstance(obj, Exception):             # <<<<<<<<<<<<<<
 *             # Convert exception
 *             val[0] = py_to_js_exception(ctx, obj)
*/
  __pyx_t_1 = __Pyx_PyException_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":345
 *         elif isinstance(obj, Exception):
 *             # Convert exception
 *             val[0] = py_to_js_exception(ctx, obj)             # <<<<<<<<<<<<<<
 *             return 0
 * 
*/
    __pyx_t_2 = __pyx_f_4cyjs_5_cyjs_py_to_js_exception(__pyx_v_ctx, __pyx_v_obj); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 345, __pyx_L1_error)
    (__pyx_v_val[0]) = __pyx_t_2;

    /* "cyjs/_cyjs.pyx":346
 *             # Convert exception
 *             val[0] = py_to_js_exception(ctx, obj)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(obj, bool):
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":343
 *             return 0
 * 
 *         elif isinstance(obj, Exception):             # <<<<<<<<<<<<<<
 *             # Convert exception
 *             val[0] = py_to_js_exception(ctx, obj)
*/
  }

  /* "cyjs/_cyjs.pyx":348
 *             return 0
 * 
 *         elif isinstance(obj, bool):             # <<<<<<<<<<<<<<
 *             val[0] = JS_NewBool(ctx, <bint>obj)
 *             return 0
*/
  __pyx_t_1 = PyBool_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":349
 * 
 *         elif isinstance(obj, bool):
 *             val[0] = JS_NewBool(ctx, <bint>obj)             # <<<<<<<<<<<<<<
 *             return 0
 * 
*/
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 349, __pyx_L1_error)
    (__pyx_v_val[0]) = JS_NewBool(__pyx_v_ctx, __pyx_t_1);

    /* "cyjs/_cyjs.pyx":350
 *         elif isinstance(obj, bool):
 *             val[0] = JS_NewBool(ctx, <bint>obj)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(obj, int):
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":348
 *             return 0
 * 
 *         elif isinstance(obj, bool):             # <<<<<<<<<<<<<<
 *             val[0] = JS_NewBool(ctx, <bint>obj)
 *             return 0
*/
  }

  /* "cyjs/_cyjs.pyx":352
 *             return 0
 * 
 *         elif isinstance(obj, int):             # <<<<<<<<<<<<<<
 *             ival = PyLong_AsLongAndOverflow(obj, &overflow)
 *             if overflow:
*/
  __pyx_t_1 = PyLong_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":353
 * 
 *         elif isinstance(obj, int):
 *             ival = PyLong_AsLongAndOverflow(obj, &overflow)             # <<<<<<<<<<<<<<
 *             if overflow:
 *                 return fast_parse_json_from_obj(ctx, val, repr(obj))
*/
    __pyx_t_3 = PyLong_AsLongAndOverflow(__pyx_v_obj, (&__pyx_v_overflow)); if (unlikely(__pyx_t_3 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 353, __pyx_L1_error)
    __pyx_v_ival = __pyx_t_3;

    /* "cyjs/_cyjs.pyx":354
 *         elif isinstance(obj, int):
 *             ival = PyLong_AsLongAndOverflow(obj, &overflow)
 *             if overflow:             # <<<<<<<<<<<<<<
 *                 return fast_parse_json_from_obj(ctx, val, repr(obj))
 *             val[0] = JS_NewInt32(ctx, ival)
*/
    __pyx_t_1 = (__pyx_v_overflow != 0);
    if (__pyx_t_1) {

      /* "cyjs/_cyjs.pyx":355
 *             ival = PyLong_AsLongAndOverflow(obj, &overflow)
 *             if overflow:
 *                 return fast_parse_json_from_obj(ctx, val, repr(obj))             # <<<<<<<<<<<<<<
 *             val[0] = JS_NewInt32(ctx, ival)
 *             return 0
*/
      __pyx_t_4 = PyObject_Repr(__pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 355, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = __pyx_f_4cyjs_5_cyjs_fast_parse_json_from_obj(__pyx_v_ctx, __pyx_v_val, __pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "cyjs/_cyjs.pyx":354
 *         elif isinstance(obj, int):
 *             ival = PyLong_AsLongAndOverflow(obj, &overflow)
 *             if overflow:             # <<<<<<<<<<<<<<
 *                 return fast_parse_json_from_obj(ctx, val, repr(obj))
 *             val[0] = JS_NewInt32(ctx, ival)
*/
    }

    /* "cyjs/_cyjs.pyx":356
 *             if overflow:
 *                 return fast_parse_json_from_obj(ctx, val, repr(obj))
 *             val[0] = JS_NewInt32(ctx, ival)             # <<<<<<<<<<<<<<
 *             return 0
 * 
*/
    (__pyx_v_val[0]) = JS_NewInt32(__pyx_v_ctx, __pyx_v_ival);

    /* "cyjs/_cyjs.pyx":357
 *                 return fast_parse_json_from_obj(ctx, val, repr(obj))
 *             val[0] = JS_NewInt32(ctx, ival)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(obj, float):
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":352
 *             return 0
 * 
 *         elif isinstance(obj, int):             # <<<<<<<<<<<<<<
 *             ival = PyLong_AsLongAndOverflow(obj, &overflow)
 *             if overflow:
*/
  }

  /* "cyjs/_cyjs.pyx":359
 *             return 0
 * 
 *         elif isinstance(obj, float):             # <<<<<<<<<<<<<<
 *             val[0] = JS_NewFloat64(ctx, <double>obj)
 *             return 0
*/
  __pyx_t_1 = PyFloat_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":360
 * 
 *         elif isinstance(obj, float):
 *             val[0] = JS_NewFloat64(ctx, <double>obj)             # <<<<<<<<<<<<<<
 *             return 0
 * 
*/
    __pyx_t_5 = __Pyx_PyFloat_AsDouble(__pyx_v_obj); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 360, __pyx_L1_error)
    (__pyx_v_val[0]) = JS_NewFloat64(__pyx_v_ctx, ((double)__pyx_t_5));

    /* "cyjs/_cyjs.pyx":361
 *         elif isinstance(obj, float):
 *             val[0] = JS_NewFloat64(ctx, <double>obj)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(obj, str):
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":359
 *             return 0
 * 
 *         elif isinstance(obj, float):             # <<<<<<<<<<<<<<
 *             val[0] = JS_NewFloat64(ctx, <double>obj)
 *             return 0
*/
  }

  /* "cyjs/_cyjs.pyx":363
 *             return 0
 * 
 *         elif isinstance(obj, str):             # <<<<<<<<<<<<<<
 *             return fast_serlize_string(ctx, val, obj)
 * 
*/
  __pyx_t_1 = PyUnicode_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":364
 * 
 *         elif isinstance(obj, str):
 *             return fast_serlize_string(ctx, val, obj)             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(obj, (dict, list)):
*/
    __pyx_t_6 = __pyx_f_4cyjs_5_cyjs_fast_serlize_string(__pyx_v_ctx, __pyx_v_val, __pyx_v_obj); if (unlikely(__pyx_t_6 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 364, __pyx_L1_error)
    __pyx_r = __pyx_t_6;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":363
 *             return 0
 * 
 *         elif isinstance(obj, str):             # <<<<<<<<<<<<<<
 *             return fast_serlize_string(ctx, val, obj)
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":366
 *             return fast_serlize_string(ctx, val, obj)
 * 
 *         elif isinstance(obj, (dict, list)):             # <<<<<<<<<<<<<<
 *             return fast_serlize_dict_or_list(ctx, val, obj)
 * 
*/
  __pyx_t_7 = PyDict_Check(__pyx_v_obj); 
  if (!__pyx_t_7) {
  } else {
    __pyx_t_1 = __pyx_t_7;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_7 = PyList_Check(__pyx_v_obj); 
  __pyx_t_1 = __pyx_t_7;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":367
 * 
 *         elif isinstance(obj, (dict, list)):
 *             return fast_serlize_dict_or_list(ctx, val, obj)             # <<<<<<<<<<<<<<
 * 
 *         elif PyObject_CheckBuffer(obj):
*/
    __pyx_r = __pyx_f_4cyjs_5_cyjs_fast_serlize_dict_or_list(__pyx_v_ctx, __pyx_v_val, __pyx_v_obj);
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":366
 *             return fast_serlize_string(ctx, val, obj)
 * 
 *         elif isinstance(obj, (dict, list)):             # <<<<<<<<<<<<<<
 *             return fast_serlize_dict_or_list(ctx, val, obj)
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":369
 *             return fast_serlize_dict_or_list(ctx, val, obj)
 * 
 *         elif PyObject_CheckBuffer(obj):             # <<<<<<<<<<<<<<
 *             return fast_serlize_string(ctx, val, obj)
 * 
*/
  __pyx_t_1 = PyObject_CheckBuffer(__pyx_v_obj);
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":370
 * 
 *         elif PyObject_CheckBuffer(obj):
 *             return fast_serlize_string(ctx, val, obj)             # <<<<<<<<<<<<<<
 * 
 *     PyErr_SetObject(TypeError, f"unknown object {type(obj).__name__!r}")
*/
    __pyx_t_6 = __pyx_f_4cyjs_5_cyjs_fast_serlize_string(__pyx_v_ctx, __pyx_v_val, __pyx_v_obj); if (unlikely(__pyx_t_6 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 370, __pyx_L1_error)
    __pyx_r = __pyx_t_6;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":369
 *             return fast_serlize_dict_or_list(ctx, val, obj)
 * 
 *         elif PyObject_CheckBuffer(obj):             # <<<<<<<<<<<<<<
 *             return fast_serlize_string(ctx, val, obj)
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":372
 *             return fast_serlize_string(ctx, val, obj)
 * 
 *     PyErr_SetObject(TypeError, f"unknown object {type(obj).__name__!r}")             # <<<<<<<<<<<<<<
 *     return -1
 * 
*/
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_obj)), __pyx_mstate_global->__pyx_n_u_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_t_4), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_unknown_object, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  PyErr_SetObject(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "cyjs/_cyjs.pyx":373
 * 
 *     PyErr_SetObject(TypeError, f"unknown object {type(obj).__name__!r}")
 *     return -1             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = -1;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":290
 * # Returns -1 on failre as a failsafe for C to evacuate safely from
 * # TODO: Optimize under this logic ->
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:             # <<<<<<<<<<<<<<
 *     # Start with the best case scenario
 *     cdef int overflow = 0
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_WriteUnraisable("cyjs._cyjs.to_quickjs", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_fuse_9__pyx_f_4cyjs_5_cyjs_to_quickjs(JSContext *__pyx_v_ctx, JSValue *__pyx_v_val, int __pyx_v_obj) {
  CYTHON_UNUSED int __pyx_v_overflow;
  int __pyx_r;

  /* "cyjs/_cyjs.pyx":292
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:
 *     # Start with the best case scenario
 *     cdef int overflow = 0             # <<<<<<<<<<<<<<
 *     cdef long ival
 * 
*/
  __pyx_v_overflow = 0;

  /* "cyjs/_cyjs.pyx":322
 * 
 *     elif quickjs_type_t is bint:
 *         val[0] = JS_NewBool(ctx, <bint>obj)             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
  (__pyx_v_val[0]) = JS_NewBool(__pyx_v_ctx, __pyx_v_obj);

  /* "cyjs/_cyjs.pyx":323
 *     elif quickjs_type_t is bint:
 *         val[0] = JS_NewBool(ctx, <bint>obj)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     else:
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":290
 * # Returns -1 on failre as a failsafe for C to evacuate safely from
 * # TODO: Optimize under this logic ->
 * cdef int to_quickjs(JSContext* ctx, JSValue* val, quickjs_type_t obj) noexcept:             # <<<<<<<<<<<<<<
 *     # Start with the best case scenario
 *     cdef int overflow = 0
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":390
 * # does both incomming and outgoing python and js values
 * # in both directions
 * cdef void arg_parser_init(arg_parser_t* self, JSContext* ctx, JSValue* values, Py_ssize_t size, Py_ssize_t len):             # <<<<<<<<<<<<<<
 *     self.ctx = ctx
 *     self.values = values
*/

static void __pyx_f_4cyjs_5_cyjs_arg_parser_init(struct __pyx_t_4cyjs_5_cyjs_arg_parser_t *__pyx_v_self, JSContext *__pyx_v_ctx, JSValue *__pyx_v_values, Py_ssize_t __pyx_v_size, Py_ssize_t __pyx_v_len) {

  /* "cyjs/_cyjs.pyx":391
 * # in both directions
 * cdef void arg_parser_init(arg_parser_t* self, JSContext* ctx, JSValue* values, Py_ssize_t size, Py_ssize_t len):
 *     self.ctx = ctx             # <<<<<<<<<<<<<<
 *     self.values = values
 *     self.size = size
*/
  __pyx_v_self->ctx = __pyx_v_ctx;

  /* "cyjs/_cyjs.pyx":392
 * cdef void arg_parser_init(arg_parser_t* self, JSContext* ctx, JSValue* values, Py_ssize_t size, Py_ssize_t len):
 *     self.ctx = ctx
 *     self.values = values             # <<<<<<<<<<<<<<
 *     self.size = size
 *     self.len = len
*/
  __pyx_v_self->values = __pyx_v_values;

  /* "cyjs/_cyjs.pyx":393
 *     self.ctx = ctx
 *     self.values = values
 *     self.size = size             # <<<<<<<<<<<<<<
 *     self.len = len
 *     self.heap = 0
*/
  __pyx_v_self->size = __pyx_v_size;

  /* "cyjs/_cyjs.pyx":394
 *     self.values = values
 *     self.size = size
 *     self.len = len             # <<<<<<<<<<<<<<
 *     self.heap = 0
 * 
*/
  __pyx_v_self->len = __pyx_v_len;

  /* "cyjs/_cyjs.pyx":395
 *     self.size = size
 *     self.len = len
 *     self.heap = 0             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_v_self->heap = 0;

  /* "cyjs/_cyjs.pyx":390
 * # does both incomming and outgoing python and js values
 * # in both directions
 * cdef void arg_parser_init(arg_parser_t* self, JSContext* ctx, JSValue* values, Py_ssize_t size, Py_ssize_t len):             # <<<<<<<<<<<<<<
 *     self.ctx = ctx
 *     self.values = values
*/

  /* function exit code */
}

/* "cyjs/_cyjs.pyx":398
 * 
 * 
 * cdef int arg_parser_append(arg_parser_t* self, JSValue val):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t new_size
 *     cdef JSValue* new_values
*/

static int __pyx_f_4cyjs_5_cyjs_arg_parser_append(struct __pyx_t_4cyjs_5_cyjs_arg_parser_t *__pyx_v_self, JSValue __pyx_v_val) {
  Py_ssize_t __pyx_v_new_size;
  JSValue *__pyx_v_new_values;
  int __pyx_r;
  int __pyx_t_1;
  PyObject *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyjs/_cyjs.pyx":401
 *     cdef Py_ssize_t new_size
 *     cdef JSValue* new_values
 *     if self.len >= self.size:             # <<<<<<<<<<<<<<
 *         new_size = self.size * 2
 *         new_values = <JSValue*>js_realloc(self.ctx, self.values, new_size * sizeof(JSValue))
*/
  __pyx_t_1 = (__pyx_v_self->len >= __pyx_v_self->size);
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":402
 *     cdef JSValue* new_values
 *     if self.len >= self.size:
 *         new_size = self.size * 2             # <<<<<<<<<<<<<<
 *         new_values = <JSValue*>js_realloc(self.ctx, self.values, new_size * sizeof(JSValue))
 *         if new_values == NULL:
*/
    __pyx_v_new_size = (__pyx_v_self->size * 2);

    /* "cyjs/_cyjs.pyx":403
 *     if self.len >= self.size:
 *         new_size = self.size * 2
 *         new_values = <JSValue*>js_realloc(self.ctx, self.values, new_size * sizeof(JSValue))             # <<<<<<<<<<<<<<
 *         if new_values == NULL:
 *             PyErr_NoMemory()
*/
    __pyx_v_new_values = ((JSValue *)js_realloc(__pyx_v_self->ctx, __pyx_v_self->values, (__pyx_v_new_size * (sizeof(JSValue)))));

    /* "cyjs/_cyjs.pyx":404
 *         new_size = self.size * 2
 *         new_values = <JSValue*>js_realloc(self.ctx, self.values, new_size * sizeof(JSValue))
 *         if new_values == NULL:             # <<<<<<<<<<<<<<
 *             PyErr_NoMemory()
 *             return -1
*/
    __pyx_t_1 = (__pyx_v_new_values == NULL);
    if (__pyx_t_1) {

      /* "cyjs/_cyjs.pyx":405
 *         new_values = <JSValue*>js_realloc(self.ctx, self.values, new_size * sizeof(JSValue))
 *         if new_values == NULL:
 *             PyErr_NoMemory()             # <<<<<<<<<<<<<<
 *             return -1
 *         self.heap = 1
*/
      __pyx_t_2 = PyErr_NoMemory(); if (unlikely(__pyx_t_2 == ((void *)NULL))) __PYX_ERR(0, 405, __pyx_L1_error)

      /* "cyjs/_cyjs.pyx":406
 *         if new_values == NULL:
 *             PyErr_NoMemory()
 *             return -1             # <<<<<<<<<<<<<<
 *         self.heap = 1
 *         self.values = new_values
*/
      __pyx_r = -1;
      goto __pyx_L0;

      /* "cyjs/_cyjs.pyx":404
 *         new_size = self.size * 2
 *         new_values = <JSValue*>js_realloc(self.ctx, self.values, new_size * sizeof(JSValue))
 *         if new_values == NULL:             # <<<<<<<<<<<<<<
 *             PyErr_NoMemory()
 *             return -1
*/
    }

    /* "cyjs/_cyjs.pyx":407
 *             PyErr_NoMemory()
 *             return -1
 *         self.heap = 1             # <<<<<<<<<<<<<<
 *         self.values = new_values
 *         self.size = new_size
*/
    __pyx_v_self->heap = 1;

    /* "cyjs/_cyjs.pyx":408
 *             return -1
 *         self.heap = 1
 *         self.values = new_values             # <<<<<<<<<<<<<<
 *         self.size = new_size
 * 
*/
    __pyx_v_self->values = __pyx_v_new_values;

    /* "cyjs/_cyjs.pyx":409
 *         self.heap = 1
 *         self.values = new_values
 *         self.size = new_size             # <<<<<<<<<<<<<<
 * 
 *     self.values[self.len] = val
*/
    __pyx_v_self->size = __pyx_v_new_size;

    /* "cyjs/_cyjs.pyx":401
 *     cdef Py_ssize_t new_size
 *     cdef JSValue* new_values
 *     if self.len >= self.size:             # <<<<<<<<<<<<<<
 *         new_size = self.size * 2
 *         new_values = <JSValue*>js_realloc(self.ctx, self.values, new_size * sizeof(JSValue))
*/
  }

  /* "cyjs/_cyjs.pyx":411
 *         self.size = new_size
 * 
 *     self.values[self.len] = val             # <<<<<<<<<<<<<<
 *     self.len += 1
 *     return 0
*/
  (__pyx_v_self->values[__pyx_v_self->len]) = __pyx_v_val;

  /* "cyjs/_cyjs.pyx":412
 * 
 *     self.values[self.len] = val
 *     self.len += 1             # <<<<<<<<<<<<<<
 *     return 0
 * 
*/
  __pyx_v_self->len = (__pyx_v_self->len + 1);

  /* "cyjs/_cyjs.pyx":413
 *     self.values[self.len] = val
 *     self.len += 1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef int arg_parser_pack(arg_parser_t* self, tuple args):
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":398
 * 
 * 
 * cdef int arg_parser_append(arg_parser_t* self, JSValue val):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t new_size
 *     cdef JSValue* new_values
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.arg_parser_append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":415
 *     return 0
 * 
 * cdef int arg_parser_pack(arg_parser_t* self, tuple args):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t size = PyTuple_GET_SIZE(args)
*/

static int __pyx_f_4cyjs_5_cyjs_arg_parser_pack(struct __pyx_t_4cyjs_5_cyjs_arg_parser_t *__pyx_v_self, PyObject *__pyx_v_args) {
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_size;
  JSValue __pyx_v_val;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("arg_parser_pack", 0);

  /* "cyjs/_cyjs.pyx":417
 * cdef int arg_parser_pack(arg_parser_t* self, tuple args):
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t size = PyTuple_GET_SIZE(args)             # <<<<<<<<<<<<<<
 *     cdef JSValue val
 * 
*/
  __pyx_v_size = PyTuple_GET_SIZE(__pyx_v_args);

  /* "cyjs/_cyjs.pyx":420
 *     cdef JSValue val
 * 
 *     for i in range(size):             # <<<<<<<<<<<<<<
 *         if to_quickjs(self.ctx, &val, PyTuple_GET_ITEM(args, i)) < 0:
 *             return -1
*/
  __pyx_t_1 = __pyx_v_size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "cyjs/_cyjs.pyx":421
 * 
 *     for i in range(size):
 *         if to_quickjs(self.ctx, &val, PyTuple_GET_ITEM(args, i)) < 0:             # <<<<<<<<<<<<<<
 *             return -1
 *         if JS_IsException(val):
*/
    __pyx_t_4 = PyTuple_GET_ITEM(__pyx_v_args, __pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = (__pyx_fuse_8__pyx_f_4cyjs_5_cyjs_to_quickjs(__pyx_v_self->ctx, (&__pyx_v_val), __pyx_t_4) < 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {

      /* "cyjs/_cyjs.pyx":422
 *     for i in range(size):
 *         if to_quickjs(self.ctx, &val, PyTuple_GET_ITEM(args, i)) < 0:
 *             return -1             # <<<<<<<<<<<<<<
 *         if JS_IsException(val):
 *             return -1
*/
      __pyx_r = -1;
      goto __pyx_L0;

      /* "cyjs/_cyjs.pyx":421
 * 
 *     for i in range(size):
 *         if to_quickjs(self.ctx, &val, PyTuple_GET_ITEM(args, i)) < 0:             # <<<<<<<<<<<<<<
 *             return -1
 *         if JS_IsException(val):
*/
    }

    /* "cyjs/_cyjs.pyx":423
 *         if to_quickjs(self.ctx, &val, PyTuple_GET_ITEM(args, i)) < 0:
 *             return -1
 *         if JS_IsException(val):             # <<<<<<<<<<<<<<
 *             return -1
 *         if arg_parser_append(self, val) < 0:
*/
    __pyx_t_5 = JS_IsException(__pyx_v_val);
    if (__pyx_t_5) {

      /* "cyjs/_cyjs.pyx":424
 *             return -1
 *         if JS_IsException(val):
 *             return -1             # <<<<<<<<<<<<<<
 *         if arg_parser_append(self, val) < 0:
 *             return -1
*/
      __pyx_r = -1;
      goto __pyx_L0;

      /* "cyjs/_cyjs.pyx":423
 *         if to_quickjs(self.ctx, &val, PyTuple_GET_ITEM(args, i)) < 0:
 *             return -1
 *         if JS_IsException(val):             # <<<<<<<<<<<<<<
 *             return -1
 *         if arg_parser_append(self, val) < 0:
*/
    }

    /* "cyjs/_cyjs.pyx":425
 *         if JS_IsException(val):
 *             return -1
 *         if arg_parser_append(self, val) < 0:             # <<<<<<<<<<<<<<
 *             return -1
 *     return 0
*/
    __pyx_t_6 = __pyx_f_4cyjs_5_cyjs_arg_parser_append(__pyx_v_self, __pyx_v_val); if (unlikely(__pyx_t_6 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 425, __pyx_L1_error)
    __pyx_t_5 = (__pyx_t_6 < 0);
    if (__pyx_t_5) {

      /* "cyjs/_cyjs.pyx":426
 *             return -1
 *         if arg_parser_append(self, val) < 0:
 *             return -1             # <<<<<<<<<<<<<<
 *     return 0
 * 
*/
      __pyx_r = -1;
      goto __pyx_L0;

      /* "cyjs/_cyjs.pyx":425
 *         if JS_IsException(val):
 *             return -1
 *         if arg_parser_append(self, val) < 0:             # <<<<<<<<<<<<<<
 *             return -1
 *     return 0
*/
    }
  }

  /* "cyjs/_cyjs.pyx":427
 *         if arg_parser_append(self, val) < 0:
 *             return -1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * # While being Python error handled this parser does the reverse to pack
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":415
 *     return 0
 * 
 * cdef int arg_parser_pack(arg_parser_t* self, tuple args):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t size = PyTuple_GET_SIZE(args)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.arg_parser_pack", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":431
 * # While being Python error handled this parser does the reverse to pack
 * # when handling incoming objects.
 * cdef inline tuple arg_parser_unpack(arg_parser_t* self):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i
 *     cdef list args = [to_python(self.ctx, JS_DupValue(self.ctx, self.values[i])) for i in range(self.len)]
*/

static CYTHON_INLINE PyObject *__pyx_f_4cyjs_5_cyjs_arg_parser_unpack(struct __pyx_t_4cyjs_5_cyjs_arg_parser_t *__pyx_v_self) {
  PyObject *__pyx_v_args = 0;
  Py_ssize_t __pyx_7genexpr__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("arg_parser_unpack", 0);

  /* "cyjs/_cyjs.pyx":433
 * cdef inline tuple arg_parser_unpack(arg_parser_t* self):
 *     cdef Py_ssize_t i
 *     cdef list args = [to_python(self.ctx, JS_DupValue(self.ctx, self.values[i])) for i in range(self.len)]             # <<<<<<<<<<<<<<
 *     return PyList_AsTuple(args)
 * 
*/
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_v_self->len;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_7genexpr__pyx_v_i = __pyx_t_4;
      __pyx_t_5 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->ctx, JS_DupValue(__pyx_v_self->ctx, (__pyx_v_self->values[__pyx_7genexpr__pyx_v_i]))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 433, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 433, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  } /* exit inner scope */
  __pyx_v_args = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":434
 *     cdef Py_ssize_t i
 *     cdef list args = [to_python(self.ctx, JS_DupValue(self.ctx, self.values[i])) for i in range(self.len)]
 *     return PyList_AsTuple(args)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_AsTuple(__pyx_v_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":431
 * # While being Python error handled this parser does the reverse to pack
 * # when handling incoming objects.
 * cdef inline tuple arg_parser_unpack(arg_parser_t* self):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i
 *     cdef list args = [to_python(self.ctx, JS_DupValue(self.ctx, self.values[i])) for i in range(self.len)]
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cyjs._cyjs.arg_parser_unpack", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":437
 * 
 * 
 * cdef void arg_parser_finish(arg_parser_t* self):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i
 *     # Free them all...
*/

static void __pyx_f_4cyjs_5_cyjs_arg_parser_finish(struct __pyx_t_4cyjs_5_cyjs_arg_parser_t *__pyx_v_self) {
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;

  /* "cyjs/_cyjs.pyx":440
 *     cdef Py_ssize_t i
 *     # Free them all...
 *     for i in range(self.len):             # <<<<<<<<<<<<<<
 *         JS_FreeValue(self.ctx, self.values[i])
 * 
*/
  __pyx_t_1 = __pyx_v_self->len;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "cyjs/_cyjs.pyx":441
 *     # Free them all...
 *     for i in range(self.len):
 *         JS_FreeValue(self.ctx, self.values[i])             # <<<<<<<<<<<<<<
 * 
 *     if self.heap:
*/
    JS_FreeValue(__pyx_v_self->ctx, (__pyx_v_self->values[__pyx_v_i]));
  }

  /* "cyjs/_cyjs.pyx":443
 *         JS_FreeValue(self.ctx, self.values[i])
 * 
 *     if self.heap:             # <<<<<<<<<<<<<<
 *         # if we had ownership over a heap (mostly in python -> js cases)
 *         # free the heap.
*/
  if (__pyx_v_self->heap) {

    /* "cyjs/_cyjs.pyx":446
 *         # if we had ownership over a heap (mostly in python -> js cases)
 *         # free the heap.
 *         js_free(self.ctx, self.values)             # <<<<<<<<<<<<<<
 * 
 * 
*/
    js_free(__pyx_v_self->ctx, __pyx_v_self->values);

    /* "cyjs/_cyjs.pyx":443
 *         JS_FreeValue(self.ctx, self.values[i])
 * 
 *     if self.heap:             # <<<<<<<<<<<<<<
 *         # if we had ownership over a heap (mostly in python -> js cases)
 *         # free the heap.
*/
  }

  /* "cyjs/_cyjs.pyx":437
 * 
 * 
 * cdef void arg_parser_finish(arg_parser_t* self):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i
 *     # Free them all...
*/

  /* function exit code */
}

/* "cyjs/_cyjs.pyx":459
 *         uint32_t len
 * 
 *     def __init__(self, Object obj):             # <<<<<<<<<<<<<<
 *         if JS_GetOwnPropertyNames(obj.ctx, &self.tab, &self.len, obj.value, JS_GPN_STRING_MASK | JS_GPN_ENUM_ONLY) < 0:
 *             if obj.context.has_exception():
*/

/* Python wrapper */
static int __pyx_pw_4cyjs_5_cyjs_6_OView_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4cyjs_5_cyjs_6_OView_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_obj = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_obj,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 459, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 459, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 459, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, i); __PYX_ERR(0, 459, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 459, __pyx_L3_error)
    }
    __pyx_v_obj = ((struct __pyx_obj_4cyjs_5_cyjs_Object *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 459, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs._OView.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_obj), __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object, 1, "obj", 0))) __PYX_ERR(0, 459, __pyx_L1_error)
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6_OView___init__(((struct __pyx_obj_4cyjs_5_cyjs__OView *)__pyx_v_self), __pyx_v_obj);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4cyjs_5_cyjs_6_OView___init__(struct __pyx_obj_4cyjs_5_cyjs__OView *__pyx_v_self, struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  JSContext *__pyx_t_5;
  JSValue __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "cyjs/_cyjs.pyx":460
 * 
 *     def __init__(self, Object obj):
 *         if JS_GetOwnPropertyNames(obj.ctx, &self.tab, &self.len, obj.value, JS_GPN_STRING_MASK | JS_GPN_ENUM_ONLY) < 0:             # <<<<<<<<<<<<<<
 *             if obj.context.has_exception():
 *                 obj.context.raise_exception()
*/
  __pyx_t_1 = (JS_GetOwnPropertyNames(__pyx_v_obj->ctx, (&__pyx_v_self->tab), (&__pyx_v_self->len), __pyx_v_obj->value, (JS_GPN_STRING_MASK | JS_GPN_ENUM_ONLY)) < 0);
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":461
 *     def __init__(self, Object obj):
 *         if JS_GetOwnPropertyNames(obj.ctx, &self.tab, &self.len, obj.value, JS_GPN_STRING_MASK | JS_GPN_ENUM_ONLY) < 0:
 *             if obj.context.has_exception():             # <<<<<<<<<<<<<<
 *                 obj.context.raise_exception()
 *                 raise
*/
    __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_7Context_has_exception(__pyx_v_obj->context); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 461, __pyx_L1_error)
    if (unlikely(__pyx_t_1)) {

      /* "cyjs/_cyjs.pyx":462
 *         if JS_GetOwnPropertyNames(obj.ctx, &self.tab, &self.len, obj.value, JS_GPN_STRING_MASK | JS_GPN_ENUM_ONLY) < 0:
 *             if obj.context.has_exception():
 *                 obj.context.raise_exception()             # <<<<<<<<<<<<<<
 *                 raise
 *             else:
*/
      __pyx_t_2 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_obj->context->__pyx_vtab)->raise_exception(__pyx_v_obj->context); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 462, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "cyjs/_cyjs.pyx":463
 *             if obj.context.has_exception():
 *                 obj.context.raise_exception()
 *                 raise             # <<<<<<<<<<<<<<
 *             else:
 *                 raise RuntimeError("failed to get object's Property names")
*/
      __Pyx_ReraiseException();
      __PYX_ERR(0, 463, __pyx_L1_error)

      /* "cyjs/_cyjs.pyx":461
 *     def __init__(self, Object obj):
 *         if JS_GetOwnPropertyNames(obj.ctx, &self.tab, &self.len, obj.value, JS_GPN_STRING_MASK | JS_GPN_ENUM_ONLY) < 0:
 *             if obj.context.has_exception():             # <<<<<<<<<<<<<<
 *                 obj.context.raise_exception()
 *                 raise
*/
    }

    /* "cyjs/_cyjs.pyx":465
 *                 raise
 *             else:
 *                 raise RuntimeError("failed to get object's Property names")             # <<<<<<<<<<<<<<
 * 
 *         self.obj = obj
*/
    /*else*/ {
      __pyx_t_3 = NULL;
      __pyx_t_4 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_failed_to_get_object_s_Property};
        __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_RuntimeError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 465, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 465, __pyx_L1_error)
    }

    /* "cyjs/_cyjs.pyx":460
 * 
 *     def __init__(self, Object obj):
 *         if JS_GetOwnPropertyNames(obj.ctx, &self.tab, &self.len, obj.value, JS_GPN_STRING_MASK | JS_GPN_ENUM_ONLY) < 0:             # <<<<<<<<<<<<<<
 *             if obj.context.has_exception():
 *                 obj.context.raise_exception()
*/
  }

  /* "cyjs/_cyjs.pyx":467
 *                 raise RuntimeError("failed to get object's Property names")
 * 
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.ctx = obj.ctx
 *         self.value = obj.value
*/
  __Pyx_INCREF((PyObject *)__pyx_v_obj);
  __Pyx_GIVEREF((PyObject *)__pyx_v_obj);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->obj);
  __Pyx_DECREF((PyObject *)__pyx_v_self->obj);
  __pyx_v_self->obj = __pyx_v_obj;

  /* "cyjs/_cyjs.pyx":468
 * 
 *         self.obj = obj
 *         self.ctx = obj.ctx             # <<<<<<<<<<<<<<
 *         self.value = obj.value
 * 
*/
  __pyx_t_5 = __pyx_v_obj->ctx;
  __pyx_v_self->ctx = __pyx_t_5;

  /* "cyjs/_cyjs.pyx":469
 *         self.obj = obj
 *         self.ctx = obj.ctx
 *         self.value = obj.value             # <<<<<<<<<<<<<<
 * 
 *     # Mainly used with items and values not keys
*/
  __pyx_t_6 = __pyx_v_obj->value;
  __pyx_v_self->value = __pyx_t_6;

  /* "cyjs/_cyjs.pyx":459
 *         uint32_t len
 * 
 *     def __init__(self, Object obj):             # <<<<<<<<<<<<<<
 *         if JS_GetOwnPropertyNames(obj.ctx, &self.tab, &self.len, obj.value, JS_GPN_STRING_MASK | JS_GPN_ENUM_ONLY) < 0:
 *             if obj.context.has_exception():
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs._OView.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":474
 *     # it has simillar usage to PyDict_Next
 *     # this will not raise exceptions, so be very careful when using it...
 *     cdef bint cnext(self, size_t* pos, JSAtom* key, JSValue* val):             # <<<<<<<<<<<<<<
 *         cdef JSAtom at
 *         cdef size_t _pos = pos[0]
*/

static int __pyx_f_4cyjs_5_cyjs_6_OView_cnext(struct __pyx_obj_4cyjs_5_cyjs__OView *__pyx_v_self, size_t *__pyx_v_pos, JSAtom *__pyx_v_key, JSValue *__pyx_v_val) {
  JSAtom __pyx_v_at;
  size_t __pyx_v__pos;
  int __pyx_r;
  int __pyx_t_1;
  JSAtom __pyx_t_2;

  /* "cyjs/_cyjs.pyx":476
 *     cdef bint cnext(self, size_t* pos, JSAtom* key, JSValue* val):
 *         cdef JSAtom at
 *         cdef size_t _pos = pos[0]             # <<<<<<<<<<<<<<
 *         if _pos > self.len:
 *             return 0
*/
  __pyx_v__pos = (__pyx_v_pos[0]);

  /* "cyjs/_cyjs.pyx":477
 *         cdef JSAtom at
 *         cdef size_t _pos = pos[0]
 *         if _pos > self.len:             # <<<<<<<<<<<<<<
 *             return 0
 * 
*/
  __pyx_t_1 = (__pyx_v__pos > __pyx_v_self->len);
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":478
 *         cdef size_t _pos = pos[0]
 *         if _pos > self.len:
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         at = self.tab[_pos].atom
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":477
 *         cdef JSAtom at
 *         cdef size_t _pos = pos[0]
 *         if _pos > self.len:             # <<<<<<<<<<<<<<
 *             return 0
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":480
 *             return 0
 * 
 *         at = self.tab[_pos].atom             # <<<<<<<<<<<<<<
 *         # Cython can't do *val so try val[0] instead
 *         val[0] = JS_GetProperty(self.ctx, self.value, at)
*/
  __pyx_t_2 = (__pyx_v_self->tab[__pyx_v__pos]).atom;
  __pyx_v_at = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":482
 *         at = self.tab[_pos].atom
 *         # Cython can't do *val so try val[0] instead
 *         val[0] = JS_GetProperty(self.ctx, self.value, at)             # <<<<<<<<<<<<<<
 *         key[0] = at
 *         pos[0] = _pos + 1
*/
  (__pyx_v_val[0]) = JS_GetProperty(__pyx_v_self->ctx, __pyx_v_self->value, __pyx_v_at);

  /* "cyjs/_cyjs.pyx":483
 *         # Cython can't do *val so try val[0] instead
 *         val[0] = JS_GetProperty(self.ctx, self.value, at)
 *         key[0] = at             # <<<<<<<<<<<<<<
 *         pos[0] = _pos + 1
 *         return 1
*/
  (__pyx_v_key[0]) = __pyx_v_at;

  /* "cyjs/_cyjs.pyx":484
 *         val[0] = JS_GetProperty(self.ctx, self.value, at)
 *         key[0] = at
 *         pos[0] = _pos + 1             # <<<<<<<<<<<<<<
 *         return 1
 * 
*/
  (__pyx_v_pos[0]) = (__pyx_v__pos + 1);

  /* "cyjs/_cyjs.pyx":485
 *         key[0] = at
 *         pos[0] = _pos + 1
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
  __pyx_r = 1;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":474
 *     # it has simillar usage to PyDict_Next
 *     # this will not raise exceptions, so be very careful when using it...
 *     cdef bint cnext(self, size_t* pos, JSAtom* key, JSValue* val):             # <<<<<<<<<<<<<<
 *         cdef JSAtom at
 *         cdef size_t _pos = pos[0]
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":487
 *         return 1
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.tab != NULL:
 *             js_free(self.ctx, self.tab)
*/

/* Python wrapper */
static void __pyx_pw_4cyjs_5_cyjs_6_OView_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4cyjs_5_cyjs_6_OView_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_4cyjs_5_cyjs_6_OView_2__dealloc__(((struct __pyx_obj_4cyjs_5_cyjs__OView *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4cyjs_5_cyjs_6_OView_2__dealloc__(struct __pyx_obj_4cyjs_5_cyjs__OView *__pyx_v_self) {
  int __pyx_t_1;

  /* "cyjs/_cyjs.pyx":488
 * 
 *     def __dealloc__(self):
 *         if self.tab != NULL:             # <<<<<<<<<<<<<<
 *             js_free(self.ctx, self.tab)
 * 
*/
  __pyx_t_1 = (__pyx_v_self->tab != NULL);
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":489
 *     def __dealloc__(self):
 *         if self.tab != NULL:
 *             js_free(self.ctx, self.tab)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
*/
    js_free(__pyx_v_self->ctx, __pyx_v_self->tab);

    /* "cyjs/_cyjs.pyx":488
 * 
 *     def __dealloc__(self):
 *         if self.tab != NULL:             # <<<<<<<<<<<<<<
 *             js_free(self.ctx, self.tab)
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":487
 *         return 1
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.tab != NULL:
 *             js_free(self.ctx, self.tab)
*/

  /* function exit code */
}

/* "cyjs/_cyjs.pyx":491
 *             js_free(self.ctx, self.tab)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.len
 * 
*/

/* Python wrapper */
static Py_ssize_t __pyx_pw_4cyjs_5_cyjs_6_OView_5__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_4cyjs_5_cyjs_6_OView_5__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6_OView_4__len__(((struct __pyx_obj_4cyjs_5_cyjs__OView *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_4cyjs_5_cyjs_6_OView_4__len__(struct __pyx_obj_4cyjs_5_cyjs__OView *__pyx_v_self) {
  Py_ssize_t __pyx_r;

  /* "cyjs/_cyjs.pyx":492
 * 
 *     def __len__(self):
 *         return self.len             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = __pyx_v_self->len;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":491
 *             js_free(self.ctx, self.tab)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.len
 * 
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_6_OView_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_6_OView_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6_OView_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_6_OView_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6_OView_6__reduce_cython__(((struct __pyx_obj_4cyjs_5_cyjs__OView *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6_OView_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs__OView *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ctx_self_tab_self_value_can, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs._OView.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_6_OView_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_6_OView_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6_OView_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_6_OView_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs._OView.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6_OView_8__setstate_cython__(((struct __pyx_obj_4cyjs_5_cyjs__OView *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6_OView_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs__OView *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ctx_self_tab_self_value_can, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs._OView.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_4cyjs_5_cyjs_16_ObjectItemsView_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "cyjs/_cyjs.pyx":496
 * 
 * cdef class _ObjectItemsView(_OView):
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef JSAtom key
 *         cdef JSValue val
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_16_ObjectItemsView_1__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_16_ObjectItemsView_1__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_16_ObjectItemsView___iter__(((struct __pyx_obj_4cyjs_5_cyjs__ObjectItemsView *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_16_ObjectItemsView___iter__(struct __pyx_obj_4cyjs_5_cyjs__ObjectItemsView *__pyx_v_self) {
  struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__ *)__pyx_tp_new_4cyjs_5_cyjs___pyx_scope_struct____iter__(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct____iter__, __pyx_mstate_global->__pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 496, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_4cyjs_5_cyjs_16_ObjectItemsView_2generator, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0]), (PyObject *) __pyx_cur_scope, __pyx_mstate_global->__pyx_n_u_iter, __pyx_mstate_global->__pyx_n_u_ObjectItemsView___iter, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs); if (unlikely(!gen)) __PYX_ERR(0, 496, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs._ObjectItemsView.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_4cyjs_5_cyjs_16_ObjectItemsView_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__ *__pyx_cur_scope = ((struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(__pyx_sent_value != Py_None)) {
    if (unlikely(__pyx_sent_value)) PyErr_SetString(PyExc_TypeError, "can't send non-None value to a just-started generator");
    __PYX_ERR(0, 496, __pyx_L1_error)
  }

  /* "cyjs/_cyjs.pyx":499
 *         cdef JSAtom key
 *         cdef JSValue val
 *         cdef size_t pos = 0             # <<<<<<<<<<<<<<
 * 
 *         while self.cnext(&pos, &key, &val):
*/
  __pyx_cur_scope->__pyx_v_pos = 0;

  /* "cyjs/_cyjs.pyx":501
 *         cdef size_t pos = 0
 * 
 *         while self.cnext(&pos, &key, &val):             # <<<<<<<<<<<<<<
 *             yield (atom_to_py(self.ctx, key), to_python(self.ctx, val))
 * 
*/
  while (1) {
    __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs__ObjectItemsView *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.cnext(((struct __pyx_obj_4cyjs_5_cyjs__OView *)__pyx_cur_scope->__pyx_v_self), (&__pyx_cur_scope->__pyx_v_pos), (&__pyx_cur_scope->__pyx_v_key), (&__pyx_cur_scope->__pyx_v_val)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 501, __pyx_L1_error)
    if (!__pyx_t_1) break;

    /* "cyjs/_cyjs.pyx":502
 * 
 *         while self.cnext(&pos, &key, &val):
 *             yield (atom_to_py(self.ctx, key), to_python(self.ctx, val))             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_t_2 = __pyx_f_4cyjs_5_cyjs_atom_to_py(__pyx_cur_scope->__pyx_v_self->__pyx_base.ctx, __pyx_cur_scope->__pyx_v_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_cur_scope->__pyx_v_self->__pyx_base.ctx, __pyx_cur_scope->__pyx_v_val); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2) != (0)) __PYX_ERR(0, 502, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3) != (0)) __PYX_ERR(0, 502, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 502, __pyx_L1_error)
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "cyjs/_cyjs.pyx":496
 * 
 * cdef class _ObjectItemsView(_OView):
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef JSAtom key
 *         cdef JSValue val
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  if (__Pyx_PyErr_Occurred()) {
    __Pyx_Generator_Replace_StopIteration(0);
    __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  }
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_16_ObjectItemsView_4__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_16_ObjectItemsView_4__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_16_ObjectItemsView_4__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_16_ObjectItemsView_4__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_16_ObjectItemsView_3__reduce_cython__(((struct __pyx_obj_4cyjs_5_cyjs__ObjectItemsView *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_16_ObjectItemsView_3__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs__ObjectItemsView *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ctx_self_tab_self_value_can, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs._ObjectItemsView.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_16_ObjectItemsView_6__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_16_ObjectItemsView_6__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_16_ObjectItemsView_6__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_16_ObjectItemsView_6__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs._ObjectItemsView.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_16_ObjectItemsView_5__setstate_cython__(((struct __pyx_obj_4cyjs_5_cyjs__ObjectItemsView *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_16_ObjectItemsView_5__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs__ObjectItemsView *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ctx_self_tab_self_value_can, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs._ObjectItemsView.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_4cyjs_5_cyjs_15_ObjectKeysView_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "cyjs/_cyjs.pyx":506
 * 
 * cdef class _ObjectKeysView(_OView):
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self.len):
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_15_ObjectKeysView_1__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_15_ObjectKeysView_1__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_15_ObjectKeysView___iter__(((struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_15_ObjectKeysView___iter__(struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView *__pyx_v_self) {
  struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ *)__pyx_tp_new_4cyjs_5_cyjs___pyx_scope_struct_1___iter__(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_1___iter__, __pyx_mstate_global->__pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 506, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_4cyjs_5_cyjs_15_ObjectKeysView_2generator1, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1]), (PyObject *) __pyx_cur_scope, __pyx_mstate_global->__pyx_n_u_iter, __pyx_mstate_global->__pyx_n_u_ObjectKeysView___iter, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs); if (unlikely(!gen)) __PYX_ERR(0, 506, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs._ObjectKeysView.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_4cyjs_5_cyjs_15_ObjectKeysView_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ *__pyx_cur_scope = ((struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  uint32_t __pyx_t_1;
  uint32_t __pyx_t_2;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(__pyx_sent_value != Py_None)) {
    if (unlikely(__pyx_sent_value)) PyErr_SetString(PyExc_TypeError, "can't send non-None value to a just-started generator");
    __PYX_ERR(0, 506, __pyx_L1_error)
  }

  /* "cyjs/_cyjs.pyx":508
 *     def __iter__(self):
 *         cdef size_t i
 *         for i in range(self.len):             # <<<<<<<<<<<<<<
 *             yield atom_to_py(self.ctx, self.tab[i].atom)
 * 
*/
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_self->__pyx_base.len;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_cur_scope->__pyx_v_i = __pyx_t_3;

    /* "cyjs/_cyjs.pyx":509
 *         cdef size_t i
 *         for i in range(self.len):
 *             yield atom_to_py(self.ctx, self.tab[i].atom)             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, object key):
*/
    __pyx_t_4 = __pyx_f_4cyjs_5_cyjs_atom_to_py(__pyx_cur_scope->__pyx_v_self->__pyx_base.ctx, (__pyx_cur_scope->__pyx_v_self->__pyx_base.tab[__pyx_cur_scope->__pyx_v_i]).atom); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 509, __pyx_L1_error)
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "cyjs/_cyjs.pyx":506
 * 
 * cdef class _ObjectKeysView(_OView):
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self.len):
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  if (__Pyx_PyErr_Occurred()) {
    __Pyx_Generator_Replace_StopIteration(0);
    __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  }
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":511
 *             yield atom_to_py(self.ctx, self.tab[i].atom)
 * 
 *     def __contains__(self, object key):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef JSAtom at
*/

/* Python wrapper */
static int __pyx_pw_4cyjs_5_cyjs_15_ObjectKeysView_4__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static int __pyx_pw_4cyjs_5_cyjs_15_ObjectKeysView_4__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_15_ObjectKeysView_3__contains__(((struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4cyjs_5_cyjs_15_ObjectKeysView_3__contains__(struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView *__pyx_v_self, PyObject *__pyx_v_key) {
  size_t __pyx_v_i;
  JSAtom __pyx_v_at;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  uint32_t __pyx_t_3;
  uint32_t __pyx_t_4;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyjs/_cyjs.pyx":514
 *         cdef size_t i
 *         cdef JSAtom at
 *         if py_to_atom(self.ctx, key, &at) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_py_to_atom(__pyx_v_self->__pyx_base.ctx, __pyx_v_key, (&__pyx_v_at)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 514, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":515
 *         cdef JSAtom at
 *         if py_to_atom(self.ctx, key, &at) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.len):
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 515, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":514
 *         cdef size_t i
 *         cdef JSAtom at
 *         if py_to_atom(self.ctx, key, &at) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":517
 *             raise
 * 
 *         for i in range(self.len):             # <<<<<<<<<<<<<<
 *             if self.tab[i].atom == at:
 *                 return True
*/
  __pyx_t_3 = __pyx_v_self->__pyx_base.len;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "cyjs/_cyjs.pyx":518
 * 
 *         for i in range(self.len):
 *             if self.tab[i].atom == at:             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
*/
    __pyx_t_2 = ((__pyx_v_self->__pyx_base.tab[__pyx_v_i]).atom == __pyx_v_at);
    if (__pyx_t_2) {

      /* "cyjs/_cyjs.pyx":519
 *         for i in range(self.len):
 *             if self.tab[i].atom == at:
 *                 return True             # <<<<<<<<<<<<<<
 *         return False
 * 
*/
      __pyx_r = 1;
      goto __pyx_L0;

      /* "cyjs/_cyjs.pyx":518
 * 
 *         for i in range(self.len):
 *             if self.tab[i].atom == at:             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
*/
    }
  }

  /* "cyjs/_cyjs.pyx":520
 *             if self.tab[i].atom == at:
 *                 return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 * cdef class _ObjectValuesView(_OView):
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":511
 *             yield atom_to_py(self.ctx, self.tab[i].atom)
 * 
 *     def __contains__(self, object key):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef JSAtom at
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs._ObjectKeysView.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_15_ObjectKeysView_6__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_15_ObjectKeysView_6__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_15_ObjectKeysView_6__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_15_ObjectKeysView_6__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_15_ObjectKeysView_5__reduce_cython__(((struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_15_ObjectKeysView_5__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ctx_self_tab_self_value_can, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs._ObjectKeysView.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_15_ObjectKeysView_8__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_15_ObjectKeysView_8__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_15_ObjectKeysView_8__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_15_ObjectKeysView_8__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs._ObjectKeysView.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_15_ObjectKeysView_7__setstate_cython__(((struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_15_ObjectKeysView_7__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ctx_self_tab_self_value_can, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs._ObjectKeysView.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_4cyjs_5_cyjs_17_ObjectValuesView_2generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "cyjs/_cyjs.pyx":523
 * 
 * cdef class _ObjectValuesView(_OView):
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef JSAtom key
 *         cdef JSValue val
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_17_ObjectValuesView_1__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_17_ObjectValuesView_1__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_17_ObjectValuesView___iter__(((struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_17_ObjectValuesView___iter__(struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView *__pyx_v_self) {
  struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ *)__pyx_tp_new_4cyjs_5_cyjs___pyx_scope_struct_2___iter__(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_2___iter__, __pyx_mstate_global->__pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 523, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_4cyjs_5_cyjs_17_ObjectValuesView_2generator2, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[2]), (PyObject *) __pyx_cur_scope, __pyx_mstate_global->__pyx_n_u_iter, __pyx_mstate_global->__pyx_n_u_ObjectValuesView___iter, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs); if (unlikely(!gen)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs._ObjectValuesView.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_4cyjs_5_cyjs_17_ObjectValuesView_2generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ *__pyx_cur_scope = ((struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(__pyx_sent_value != Py_None)) {
    if (unlikely(__pyx_sent_value)) PyErr_SetString(PyExc_TypeError, "can't send non-None value to a just-started generator");
    __PYX_ERR(0, 523, __pyx_L1_error)
  }

  /* "cyjs/_cyjs.pyx":526
 *         cdef JSAtom key
 *         cdef JSValue val
 *         cdef size_t pos = 0             # <<<<<<<<<<<<<<
 * 
 *         while self.cnext(&pos, &key, &val):
*/
  __pyx_cur_scope->__pyx_v_pos = 0;

  /* "cyjs/_cyjs.pyx":528
 *         cdef size_t pos = 0
 * 
 *         while self.cnext(&pos, &key, &val):             # <<<<<<<<<<<<<<
 *             yield to_python(self.ctx, val)
 * 
*/
  while (1) {
    __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs__ObjectValuesView *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.cnext(((struct __pyx_obj_4cyjs_5_cyjs__OView *)__pyx_cur_scope->__pyx_v_self), (&__pyx_cur_scope->__pyx_v_pos), (&__pyx_cur_scope->__pyx_v_key), (&__pyx_cur_scope->__pyx_v_val)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 528, __pyx_L1_error)
    if (!__pyx_t_1) break;

    /* "cyjs/_cyjs.pyx":529
 * 
 *         while self.cnext(&pos, &key, &val):
 *             yield to_python(self.ctx, val)             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, object value):
*/
    __pyx_t_2 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_cur_scope->__pyx_v_self->__pyx_base.ctx, __pyx_cur_scope->__pyx_v_val); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 529, __pyx_L1_error)
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "cyjs/_cyjs.pyx":523
 * 
 * cdef class _ObjectValuesView(_OView):
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef JSAtom key
 *         cdef JSValue val
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  if (__Pyx_PyErr_Occurred()) {
    __Pyx_Generator_Replace_StopIteration(0);
    __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  }
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":531
 *             yield to_python(self.ctx, val)
 * 
 *     def __contains__(self, object value):             # <<<<<<<<<<<<<<
 *         cdef JSAtom key
 *         cdef JSValue val
*/

/* Python wrapper */
static int __pyx_pw_4cyjs_5_cyjs_17_ObjectValuesView_4__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4cyjs_5_cyjs_17_ObjectValuesView_4__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_17_ObjectValuesView_3__contains__(((struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4cyjs_5_cyjs_17_ObjectValuesView_3__contains__(struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView *__pyx_v_self, PyObject *__pyx_v_value) {
  JSAtom __pyx_v_key;
  JSValue __pyx_v_val;
  size_t __pyx_v_pos;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "cyjs/_cyjs.pyx":534
 *         cdef JSAtom key
 *         cdef JSValue val
 *         cdef size_t pos = 0             # <<<<<<<<<<<<<<
 * 
 *         while self.cnext(&pos, &key, &val):
*/
  __pyx_v_pos = 0;

  /* "cyjs/_cyjs.pyx":536
 *         cdef size_t pos = 0
 * 
 *         while self.cnext(&pos, &key, &val):             # <<<<<<<<<<<<<<
 *             if value == to_python(self.ctx, val):
 *                 return True
*/
  while (1) {
    __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs__ObjectValuesView *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.cnext(((struct __pyx_obj_4cyjs_5_cyjs__OView *)__pyx_v_self), (&__pyx_v_pos), (&__pyx_v_key), (&__pyx_v_val)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 536, __pyx_L1_error)
    if (!__pyx_t_1) break;

    /* "cyjs/_cyjs.pyx":537
 * 
 *         while self.cnext(&pos, &key, &val):
 *             if value == to_python(self.ctx, val):             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
*/
    __pyx_t_2 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->__pyx_base.ctx, __pyx_v_val); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 537, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_value, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 537, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 537, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {

      /* "cyjs/_cyjs.pyx":538
 *         while self.cnext(&pos, &key, &val):
 *             if value == to_python(self.ctx, val):
 *                 return True             # <<<<<<<<<<<<<<
 *         return False
 * 
*/
      __pyx_r = 1;
      goto __pyx_L0;

      /* "cyjs/_cyjs.pyx":537
 * 
 *         while self.cnext(&pos, &key, &val):
 *             if value == to_python(self.ctx, val):             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
*/
    }
  }

  /* "cyjs/_cyjs.pyx":539
 *             if value == to_python(self.ctx, val):
 *                 return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":531
 *             yield to_python(self.ctx, val)
 * 
 *     def __contains__(self, object value):             # <<<<<<<<<<<<<<
 *         cdef JSAtom key
 *         cdef JSValue val
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs._ObjectValuesView.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_17_ObjectValuesView_6__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_17_ObjectValuesView_6__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_17_ObjectValuesView_6__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_17_ObjectValuesView_6__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_17_ObjectValuesView_5__reduce_cython__(((struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_17_ObjectValuesView_5__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ctx_self_tab_self_value_can, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs._ObjectValuesView.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_17_ObjectValuesView_8__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_17_ObjectValuesView_8__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_17_ObjectValuesView_8__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_17_ObjectValuesView_8__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs._ObjectValuesView.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_17_ObjectValuesView_7__setstate_cython__(((struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_17_ObjectValuesView_7__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ctx_self_tab_self_value_can, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs._ObjectValuesView.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":545
 * cdef class Object:
 * 
 *     cdef void init(self, Context ctx, JSValue value):             # <<<<<<<<<<<<<<
 *         self.context = ctx
 *         self.value = value
*/

static void __pyx_f_4cyjs_5_cyjs_6Object_init(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_ctx, JSValue __pyx_v_value) {
  __Pyx_RefNannyDeclarations
  JSContext *__pyx_t_1;
  __Pyx_RefNannySetupContext("init", 0);

  /* "cyjs/_cyjs.pyx":546
 * 
 *     cdef void init(self, Context ctx, JSValue value):
 *         self.context = ctx             # <<<<<<<<<<<<<<
 *         self.value = value
 *         self.ctx = ctx.ctx
*/
  __Pyx_INCREF((PyObject *)__pyx_v_ctx);
  __Pyx_GIVEREF((PyObject *)__pyx_v_ctx);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->context);
  __Pyx_DECREF((PyObject *)__pyx_v_self->context);
  __pyx_v_self->context = __pyx_v_ctx;

  /* "cyjs/_cyjs.pyx":547
 *     cdef void init(self, Context ctx, JSValue value):
 *         self.context = ctx
 *         self.value = value             # <<<<<<<<<<<<<<
 *         self.ctx = ctx.ctx
 *         # make tag a readonly attribute instead of a property so we can get quicker access to it.
*/
  __pyx_v_self->value = __pyx_v_value;

  /* "cyjs/_cyjs.pyx":548
 *         self.context = ctx
 *         self.value = value
 *         self.ctx = ctx.ctx             # <<<<<<<<<<<<<<
 *         # make tag a readonly attribute instead of a property so we can get quicker access to it.
 *         self.tag = JS_VALUE_GET_TAG(self.value)
*/
  __pyx_t_1 = __pyx_v_ctx->ctx;
  __pyx_v_self->ctx = __pyx_t_1;

  /* "cyjs/_cyjs.pyx":550
 *         self.ctx = ctx.ctx
 *         # make tag a readonly attribute instead of a property so we can get quicker access to it.
 *         self.tag = JS_VALUE_GET_TAG(self.value)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bytes to_json(self):
*/
  __pyx_v_self->tag = JS_VALUE_GET_TAG(__pyx_v_self->value);

  /* "cyjs/_cyjs.pyx":545
 * cdef class Object:
 * 
 *     cdef void init(self, Context ctx, JSValue value):             # <<<<<<<<<<<<<<
 *         self.context = ctx
 *         self.value = value
*/

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cyjs/_cyjs.pyx":552
 *         self.tag = JS_VALUE_GET_TAG(self.value)
 * 
 *     cpdef bytes to_json(self):             # <<<<<<<<<<<<<<
 *         """
 *         Useful when debugging or handling unknown js to py conversions
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_1to_json(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_6Object_to_json(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, int __pyx_skip_dispatch) {
  CYTHON_UNUSED JSValue __pyx_v_v;
  size_t __pyx_v_size;
  char const *__pyx_v_data;
  PyObject *__pyx_v_ret = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_json", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_to_json); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_6Object_1to_json)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 552, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        if (!(likely(PyBytes_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_2))) __PYX_ERR(0, 552, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":558
 *         or msgspec are advised.
 *         """
 *         cdef JSValue v = JS_JSONStringify(self.ctx, self.value, JS_UNDEFINED, JS_UNDEFINED)             # <<<<<<<<<<<<<<
 *         cdef size_t size
 *         cdef const char* data = JS_ToCStringLen(self.ctx, &size, self.value)
*/
  __pyx_v_v = JS_JSONStringify(__pyx_v_self->ctx, __pyx_v_self->value, JS_UNDEFINED, JS_UNDEFINED);

  /* "cyjs/_cyjs.pyx":560
 *         cdef JSValue v = JS_JSONStringify(self.ctx, self.value, JS_UNDEFINED, JS_UNDEFINED)
 *         cdef size_t size
 *         cdef const char* data = JS_ToCStringLen(self.ctx, &size, self.value)             # <<<<<<<<<<<<<<
 *         cdef bytes ret = PyBytes_FromStringAndSize(data, size)
 *         JS_FreeCString(self.ctx, data)
*/
  __pyx_v_data = JS_ToCStringLen(__pyx_v_self->ctx, (&__pyx_v_size), __pyx_v_self->value);

  /* "cyjs/_cyjs.pyx":561
 *         cdef size_t size
 *         cdef const char* data = JS_ToCStringLen(self.ctx, &size, self.value)
 *         cdef bytes ret = PyBytes_FromStringAndSize(data, size)             # <<<<<<<<<<<<<<
 *         JS_FreeCString(self.ctx, data)
 *         return ret
*/
  __pyx_t_1 = PyBytes_FromStringAndSize(__pyx_v_data, __pyx_v_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ret = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":562
 *         cdef const char* data = JS_ToCStringLen(self.ctx, &size, self.value)
 *         cdef bytes ret = PyBytes_FromStringAndSize(data, size)
 *         JS_FreeCString(self.ctx, data)             # <<<<<<<<<<<<<<
 *         return ret
 * 
*/
  JS_FreeCString(__pyx_v_self->ctx, __pyx_v_data);

  /* "cyjs/_cyjs.pyx":563
 *         cdef bytes ret = PyBytes_FromStringAndSize(data, size)
 *         JS_FreeCString(self.ctx, data)
 *         return ret             # <<<<<<<<<<<<<<
 * 
 *     # serves for debugging type tags
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":552
 *         self.tag = JS_VALUE_GET_TAG(self.value)
 * 
 *     cpdef bytes to_json(self):             # <<<<<<<<<<<<<<
 *         """
 *         Useful when debugging or handling unknown js to py conversions
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Object.to_json", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_1to_json(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4cyjs_5_cyjs_6Object_to_json, "\n        Useful when debugging or handling unknown js to py conversions\n        NOTE: for best results, using third party libraries like orjson\n        or msgspec are advised.\n        ");
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_6Object_1to_json = {"to_json", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_1to_json, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4cyjs_5_cyjs_6Object_to_json};
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_1to_json(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_json (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("to_json", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("to_json", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6Object_to_json(((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_to_json(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_json", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_6Object_to_json(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Object.to_json", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":567
 *     # serves for debugging type tags
 * 
 *     def get(self, key):             # <<<<<<<<<<<<<<
 * 
 *         cdef JSValue v
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_3get(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_6Object_3get = {"get", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_3get, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_3get(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 567, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 567, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get", 0) < (0)) __PYX_ERR(0, 567, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get", 1, 1, 1, i); __PYX_ERR(0, 567, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 567, __pyx_L3_error)
    }
    __pyx_v_key = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 567, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Object.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6Object_2get(((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_self), __pyx_v_key);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_2get(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, PyObject *__pyx_v_key) {
  JSValue __pyx_v_v;
  JSAtom __pyx_v_atom;
  JSContext *__pyx_v_ctx;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  JSContext *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 0);

  /* "cyjs/_cyjs.pyx":571
 *         cdef JSValue v
 *         cdef JSAtom atom
 *         cdef JSContext* ctx = self.ctx             # <<<<<<<<<<<<<<
 * 
 *         # raise exception if hit
*/
  __pyx_t_1 = __pyx_v_self->ctx;
  __pyx_v_ctx = __pyx_t_1;

  /* "cyjs/_cyjs.pyx":574
 * 
 *         # raise exception if hit
 *         if py_to_atom(ctx, key, &atom) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_2 = __pyx_f_4cyjs_5_cyjs_py_to_atom(__pyx_v_ctx, __pyx_v_key, (&__pyx_v_atom)); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 574, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 < 0);
  if (unlikely(__pyx_t_3)) {

    /* "cyjs/_cyjs.pyx":575
 *         # raise exception if hit
 *         if py_to_atom(ctx, key, &atom) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         v = JS_GetProperty(ctx, self.value, atom)
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 575, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":574
 * 
 *         # raise exception if hit
 *         if py_to_atom(ctx, key, &atom) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":577
 *             raise
 * 
 *         v = JS_GetProperty(ctx, self.value, atom)             # <<<<<<<<<<<<<<
 * 
 *         # Dup JS_Value because to_python can and will Free information
*/
  __pyx_v_v = JS_GetProperty(__pyx_v_ctx, __pyx_v_self->value, __pyx_v_atom);

  /* "cyjs/_cyjs.pyx":580
 * 
 *         # Dup JS_Value because to_python can and will Free information
 *         return to_python(ctx, JS_DupValue(ctx, v))             # <<<<<<<<<<<<<<
 * 
 *     def set(self, object key, object value):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_ctx, JS_DupValue(__pyx_v_ctx, __pyx_v_v)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":567
 *     # serves for debugging type tags
 * 
 *     def get(self, key):             # <<<<<<<<<<<<<<
 * 
 *         cdef JSValue v
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Object.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":582
 *         return to_python(ctx, JS_DupValue(ctx, v))
 * 
 *     def set(self, object key, object value):             # <<<<<<<<<<<<<<
 *         cdef JSContext* ctx = self.ctx
 *         cdef JSValue js_value
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_5set(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_6Object_5set = {"set", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_5set, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_5set(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,&__pyx_mstate_global->__pyx_n_u_value,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 582, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 582, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 582, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set", 0) < (0)) __PYX_ERR(0, 582, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set", 1, 2, 2, i); __PYX_ERR(0, 582, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 582, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 582, __pyx_L3_error)
    }
    __pyx_v_key = values[0];
    __pyx_v_value = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 582, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Object.set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6Object_4set(((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_self), __pyx_v_key, __pyx_v_value);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_4set(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  JSContext *__pyx_v_ctx;
  JSValue __pyx_v_js_value;
  JSAtom __pyx_v_atom;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  JSContext *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7[3];
  PyObject *__pyx_t_8 = NULL;
  size_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set", 0);

  /* "cyjs/_cyjs.pyx":583
 * 
 *     def set(self, object key, object value):
 *         cdef JSContext* ctx = self.ctx             # <<<<<<<<<<<<<<
 *         cdef JSValue js_value
 *         cdef JSAtom atom
*/
  __pyx_t_1 = __pyx_v_self->ctx;
  __pyx_v_ctx = __pyx_t_1;

  /* "cyjs/_cyjs.pyx":587
 *         cdef JSAtom atom
 * 
 *         if py_to_atom(ctx, key, &atom) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_2 = __pyx_f_4cyjs_5_cyjs_py_to_atom(__pyx_v_ctx, __pyx_v_key, (&__pyx_v_atom)); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 587, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 < 0);
  if (unlikely(__pyx_t_3)) {

    /* "cyjs/_cyjs.pyx":588
 * 
 *         if py_to_atom(ctx, key, &atom) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         if to_quickjs(ctx, &js_value, value) < 0:
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 588, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":587
 *         cdef JSAtom atom
 * 
 *         if py_to_atom(ctx, key, &atom) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":590
 *             raise
 * 
 *         if to_quickjs(ctx, &js_value, value) < 0:             # <<<<<<<<<<<<<<
 *             self.context.raise_exception()
 *             # raising the second time will be for if python is at fault for this.
*/
  __pyx_t_3 = (__pyx_fuse_8__pyx_f_4cyjs_5_cyjs_to_quickjs(__pyx_v_ctx, (&__pyx_v_js_value), __pyx_v_value) < 0);
  if (unlikely(__pyx_t_3)) {

    /* "cyjs/_cyjs.pyx":591
 * 
 *         if to_quickjs(ctx, &js_value, value) < 0:
 *             self.context.raise_exception()             # <<<<<<<<<<<<<<
 *             # raising the second time will be for if python is at fault for this.
 *             raise
*/
    __pyx_t_4 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->context->__pyx_vtab)->raise_exception(__pyx_v_self->context); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "cyjs/_cyjs.pyx":593
 *             self.context.raise_exception()
 *             # raising the second time will be for if python is at fault for this.
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         if JS_IsException(js_value):
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 593, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":590
 *             raise
 * 
 *         if to_quickjs(ctx, &js_value, value) < 0:             # <<<<<<<<<<<<<<
 *             self.context.raise_exception()
 *             # raising the second time will be for if python is at fault for this.
*/
  }

  /* "cyjs/_cyjs.pyx":595
 *             raise
 * 
 *         if JS_IsException(js_value):             # <<<<<<<<<<<<<<
 *             self.context.raise_exception()
 *             # raising the second time will be for if python is at fault for this.
*/
  __pyx_t_3 = JS_IsException(__pyx_v_js_value);
  if (unlikely(__pyx_t_3)) {

    /* "cyjs/_cyjs.pyx":596
 * 
 *         if JS_IsException(js_value):
 *             self.context.raise_exception()             # <<<<<<<<<<<<<<
 *             # raising the second time will be for if python is at fault for this.
 *             raise
*/
    __pyx_t_4 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->context->__pyx_vtab)->raise_exception(__pyx_v_self->context); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "cyjs/_cyjs.pyx":598
 *             self.context.raise_exception()
 *             # raising the second time will be for if python is at fault for this.
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         if JS_SetProperty(ctx, self.value, atom, js_value) < 0:
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 598, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":595
 *             raise
 * 
 *         if JS_IsException(js_value):             # <<<<<<<<<<<<<<
 *             self.context.raise_exception()
 *             # raising the second time will be for if python is at fault for this.
*/
  }

  /* "cyjs/_cyjs.pyx":600
 *             raise
 * 
 *         if JS_SetProperty(ctx, self.value, atom, js_value) < 0:             # <<<<<<<<<<<<<<
 *             raise AttributeError(f"Could not set {key} to Object")
 * 
*/
  __pyx_t_3 = (JS_SetProperty(__pyx_v_ctx, __pyx_v_self->value, __pyx_v_atom, __pyx_v_js_value) < 0);
  if (unlikely(__pyx_t_3)) {

    /* "cyjs/_cyjs.pyx":601
 * 
 *         if JS_SetProperty(ctx, self.value, atom, js_value) < 0:
 *             raise AttributeError(f"Could not set {key} to Object")             # <<<<<<<<<<<<<<
 * 
 *     def __call__(self, *args):
*/
    __pyx_t_5 = NULL;
    __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_v_key, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7[0] = __pyx_mstate_global->__pyx_kp_u_Could_not_set;
    __pyx_t_7[1] = __pyx_t_6;
    __pyx_t_7[2] = __pyx_mstate_global->__pyx_kp_u_to_Object;
    __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_7, 3, 14 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6) + 10, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6));
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_9 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_8};
      __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AttributeError)), __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 601, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 601, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":600
 *             raise
 * 
 *         if JS_SetProperty(ctx, self.value, atom, js_value) < 0:             # <<<<<<<<<<<<<<
 *             raise AttributeError(f"Could not set {key} to Object")
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":582
 *         return to_python(ctx, JS_DupValue(ctx, v))
 * 
 *     def set(self, object key, object value):             # <<<<<<<<<<<<<<
 *         cdef JSContext* ctx = self.ctx
 *         cdef JSValue js_value
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("cyjs._cyjs.Object.set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":603
 *             raise AttributeError(f"Could not set {key} to Object")
 * 
 *     def __call__(self, *args):             # <<<<<<<<<<<<<<
 *         cdef arg_parser_t p
 *         cdef JSValue values[10]
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_7__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_7__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__call__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__call__", __pyx_kwds); return NULL;}
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6Object_6__call__(((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_self), __pyx_v_args);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_6__call__(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, PyObject *__pyx_v_args) {
  struct __pyx_t_4cyjs_5_cyjs_arg_parser_t __pyx_v_p;
  JSValue __pyx_v_values[10];
  JSValue __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  char const *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__call__", 0);

  /* "cyjs/_cyjs.pyx":608
 *         cdef JSValue ret
 *         # we will have 10 free slots with a start position of 0
 *         arg_parser_init(&p, self.ctx, values, 10, 0)             # <<<<<<<<<<<<<<
 *         try:
 *             if arg_parser_pack(&p, args) < 0:
*/
  __pyx_f_4cyjs_5_cyjs_arg_parser_init((&__pyx_v_p), __pyx_v_self->ctx, __pyx_v_values, 10, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 608, __pyx_L1_error)

  /* "cyjs/_cyjs.pyx":609
 *         # we will have 10 free slots with a start position of 0
 *         arg_parser_init(&p, self.ctx, values, 10, 0)
 *         try:             # <<<<<<<<<<<<<<
 *             if arg_parser_pack(&p, args) < 0:
 *                 # see if it's QuickJs's Fault otherwise the Fault falls to python
*/
  /*try:*/ {

    /* "cyjs/_cyjs.pyx":610
 *         arg_parser_init(&p, self.ctx, values, 10, 0)
 *         try:
 *             if arg_parser_pack(&p, args) < 0:             # <<<<<<<<<<<<<<
 *                 # see if it's QuickJs's Fault otherwise the Fault falls to python
 *                 self.context.raise_exception()
*/
    __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_arg_parser_pack((&__pyx_v_p), __pyx_v_args); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 610, __pyx_L4_error)
    __pyx_t_2 = (__pyx_t_1 < 0);
    if (unlikely(__pyx_t_2)) {

      /* "cyjs/_cyjs.pyx":612
 *             if arg_parser_pack(&p, args) < 0:
 *                 # see if it's QuickJs's Fault otherwise the Fault falls to python
 *                 self.context.raise_exception()             # <<<<<<<<<<<<<<
 *                 # raising the second time will be for if python is at fault for this.
 *                 raise
*/
      __pyx_t_3 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->context->__pyx_vtab)->raise_exception(__pyx_v_self->context); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 612, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "cyjs/_cyjs.pyx":614
 *                 self.context.raise_exception()
 *                 # raising the second time will be for if python is at fault for this.
 *                 raise             # <<<<<<<<<<<<<<
 * 
 *             ret = JS_Call(self.ctx, self.value, JS_NULL, p.len, p.values)
*/
      __Pyx_ReraiseException();
      __PYX_ERR(0, 614, __pyx_L4_error)

      /* "cyjs/_cyjs.pyx":610
 *         arg_parser_init(&p, self.ctx, values, 10, 0)
 *         try:
 *             if arg_parser_pack(&p, args) < 0:             # <<<<<<<<<<<<<<
 *                 # see if it's QuickJs's Fault otherwise the Fault falls to python
 *                 self.context.raise_exception()
*/
    }

    /* "cyjs/_cyjs.pyx":616
 *                 raise
 * 
 *             ret = JS_Call(self.ctx, self.value, JS_NULL, p.len, p.values)             # <<<<<<<<<<<<<<
 * 
 *             if JS_IsException(ret):
*/
    __pyx_v_ret = JS_Call(__pyx_v_self->ctx, __pyx_v_self->value, JS_NULL, __pyx_v_p.len, __pyx_v_p.values);

    /* "cyjs/_cyjs.pyx":618
 *             ret = JS_Call(self.ctx, self.value, JS_NULL, p.len, p.values)
 * 
 *             if JS_IsException(ret):             # <<<<<<<<<<<<<<
 *                 self.context.raise_exception()
 *                 # this should almost never occur but if it does do this
*/
    __pyx_t_2 = JS_IsException(__pyx_v_ret);
    if (unlikely(__pyx_t_2)) {

      /* "cyjs/_cyjs.pyx":619
 * 
 *             if JS_IsException(ret):
 *                 self.context.raise_exception()             # <<<<<<<<<<<<<<
 *                 # this should almost never occur but if it does do this
 *                 raise JSError("raised exception without it being Quickjs's own fault")
*/
      __pyx_t_3 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->context->__pyx_vtab)->raise_exception(__pyx_v_self->context); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 619, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "cyjs/_cyjs.pyx":621
 *                 self.context.raise_exception()
 *                 # this should almost never occur but if it does do this
 *                 raise JSError("raised exception without it being Quickjs's own fault")             # <<<<<<<<<<<<<<
 *             return to_python(self.ctx, ret)
 * 
*/
      __pyx_t_4 = NULL;
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_raised_exception_without_it_bein};
        __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_JSError, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 621, __pyx_L4_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_3);
      }
      __Pyx_Raise(((PyObject *)__pyx_t_3), 0, 0, 0);
      __Pyx_DECREF((PyObject *)__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 621, __pyx_L4_error)

      /* "cyjs/_cyjs.pyx":618
 *             ret = JS_Call(self.ctx, self.value, JS_NULL, p.len, p.values)
 * 
 *             if JS_IsException(ret):             # <<<<<<<<<<<<<<
 *                 self.context.raise_exception()
 *                 # this should almost never occur but if it does do this
*/
    }

    /* "cyjs/_cyjs.pyx":622
 *                 # this should almost never occur but if it does do this
 *                 raise JSError("raised exception without it being Quickjs's own fault")
 *             return to_python(self.ctx, ret)             # <<<<<<<<<<<<<<
 * 
 *         finally:
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->ctx, __pyx_v_ret); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L3_return;
  }

  /* "cyjs/_cyjs.pyx":626
 *         finally:
 *             # Cleanup
 *             arg_parser_finish(&p)             # <<<<<<<<<<<<<<
 * 
 *     def invoke(self, object func, *args):
*/
  /*finally:*/ {
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
       __Pyx_ExceptionSwap(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      if ( unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __pyx_t_1 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_7 = __pyx_filename;
      {
        __pyx_f_4cyjs_5_cyjs_arg_parser_finish((&__pyx_v_p)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 626, __pyx_L9_error)
      }
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ErrRestore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __pyx_lineno = __pyx_t_1; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_7;
      goto __pyx_L1_error;
      __pyx_L9_error:;
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L3_return: {
      __pyx_t_13 = __pyx_r;
      __pyx_r = 0;
      __pyx_f_4cyjs_5_cyjs_arg_parser_finish((&__pyx_v_p)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 626, __pyx_L1_error)
      __pyx_r = __pyx_t_13;
      __pyx_t_13 = 0;
      goto __pyx_L0;
    }
  }

  /* "cyjs/_cyjs.pyx":603
 *             raise AttributeError(f"Could not set {key} to Object")
 * 
 *     def __call__(self, *args):             # <<<<<<<<<<<<<<
 *         cdef arg_parser_t p
 *         cdef JSValue values[10]
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Object.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":628
 *             arg_parser_finish(&p)
 * 
 *     def invoke(self, object func, *args):             # <<<<<<<<<<<<<<
 *         cdef JSAtom atom
 *         cdef JSContext* ctx = self.ctx
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_9invoke(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_6Object_9invoke = {"invoke", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_9invoke, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_9invoke(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_args = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("invoke (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_args = __Pyx_ArgsSlice_FASTCALL(__pyx_args, 1, __pyx_nargs);
  if (unlikely(!__pyx_v_args)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_v_args);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_func,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 628, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        default:
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 628, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      const Py_ssize_t used_pos_args = (kwd_pos_args < 1) ? kwd_pos_args : 1;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, used_pos_args, __pyx_kwds_len, "invoke", 0) < (0)) __PYX_ERR(0, 628, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("invoke", 0, 1, 1, i); __PYX_ERR(0, 628, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs < 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 628, __pyx_L3_error)
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("invoke", 0, 1, 1, __pyx_nargs); __PYX_ERR(0, 628, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_AddTraceback("cyjs._cyjs.Object.invoke", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6Object_8invoke(((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_self), __pyx_v_func, __pyx_v_args);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_8invoke(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_args) {
  JSAtom __pyx_v_atom;
  JSContext *__pyx_v_ctx;
  JSValue __pyx_v_ret;
  JSValue __pyx_v_js_args[10];
  struct __pyx_t_4cyjs_5_cyjs_arg_parser_t __pyx_v_p;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  JSContext *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("invoke", 0);

  /* "cyjs/_cyjs.pyx":630
 *     def invoke(self, object func, *args):
 *         cdef JSAtom atom
 *         cdef JSContext* ctx = self.ctx             # <<<<<<<<<<<<<<
 *         cdef JSValue ret
 *         cdef JSValue js_args[10]
*/
  __pyx_t_1 = __pyx_v_self->ctx;
  __pyx_v_ctx = __pyx_t_1;

  /* "cyjs/_cyjs.pyx":635
 *         cdef arg_parser_t p
 * 
 *         if py_to_atom(ctx, func, &atom) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_2 = __pyx_f_4cyjs_5_cyjs_py_to_atom(__pyx_v_ctx, __pyx_v_func, (&__pyx_v_atom)); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 635, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 < 0);
  if (unlikely(__pyx_t_3)) {

    /* "cyjs/_cyjs.pyx":636
 * 
 *         if py_to_atom(ctx, func, &atom) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         arg_parser_init(&p, self.ctx, js_args, 10, 0)
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 636, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":635
 *         cdef arg_parser_t p
 * 
 *         if py_to_atom(ctx, func, &atom) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":638
 *             raise
 * 
 *         arg_parser_init(&p, self.ctx, js_args, 10, 0)             # <<<<<<<<<<<<<<
 *         try:
 *             if arg_parser_pack(&p, args) < 0:
*/
  __pyx_f_4cyjs_5_cyjs_arg_parser_init((&__pyx_v_p), __pyx_v_self->ctx, __pyx_v_js_args, 10, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 638, __pyx_L1_error)

  /* "cyjs/_cyjs.pyx":639
 * 
 *         arg_parser_init(&p, self.ctx, js_args, 10, 0)
 *         try:             # <<<<<<<<<<<<<<
 *             if arg_parser_pack(&p, args) < 0:
 *                 raise
*/
  /*try:*/ {

    /* "cyjs/_cyjs.pyx":640
 *         arg_parser_init(&p, self.ctx, js_args, 10, 0)
 *         try:
 *             if arg_parser_pack(&p, args) < 0:             # <<<<<<<<<<<<<<
 *                 raise
 *             ret = JS_Invoke(ctx, self.value, atom, p.len, p.values)
*/
    __pyx_t_2 = __pyx_f_4cyjs_5_cyjs_arg_parser_pack((&__pyx_v_p), __pyx_v_args); if (unlikely(__pyx_t_2 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 640, __pyx_L5_error)
    __pyx_t_3 = (__pyx_t_2 < 0);
    if (unlikely(__pyx_t_3)) {

      /* "cyjs/_cyjs.pyx":641
 *         try:
 *             if arg_parser_pack(&p, args) < 0:
 *                 raise             # <<<<<<<<<<<<<<
 *             ret = JS_Invoke(ctx, self.value, atom, p.len, p.values)
 *             return to_python(ctx, ret)
*/
      __Pyx_ReraiseException();
      __PYX_ERR(0, 641, __pyx_L5_error)

      /* "cyjs/_cyjs.pyx":640
 *         arg_parser_init(&p, self.ctx, js_args, 10, 0)
 *         try:
 *             if arg_parser_pack(&p, args) < 0:             # <<<<<<<<<<<<<<
 *                 raise
 *             ret = JS_Invoke(ctx, self.value, atom, p.len, p.values)
*/
    }

    /* "cyjs/_cyjs.pyx":642
 *             if arg_parser_pack(&p, args) < 0:
 *                 raise
 *             ret = JS_Invoke(ctx, self.value, atom, p.len, p.values)             # <<<<<<<<<<<<<<
 *             return to_python(ctx, ret)
 *         finally:
*/
    __pyx_v_ret = JS_Invoke(__pyx_v_ctx, __pyx_v_self->value, __pyx_v_atom, __pyx_v_p.len, __pyx_v_p.values);

    /* "cyjs/_cyjs.pyx":643
 *                 raise
 *             ret = JS_Invoke(ctx, self.value, atom, p.len, p.values)
 *             return to_python(ctx, ret)             # <<<<<<<<<<<<<<
 *         finally:
 *             arg_parser_finish(&p)
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_ctx, __pyx_v_ret); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 643, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_return;
  }

  /* "cyjs/_cyjs.pyx":645
 *             return to_python(ctx, ret)
 *         finally:
 *             arg_parser_finish(&p)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  /*finally:*/ {
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
       __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if ( unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __pyx_t_2 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;
      {
        __pyx_f_4cyjs_5_cyjs_arg_parser_finish((&__pyx_v_p)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 645, __pyx_L9_error)
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __pyx_lineno = __pyx_t_2; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;
      goto __pyx_L1_error;
      __pyx_L9_error:;
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L4_return: {
      __pyx_t_12 = __pyx_r;
      __pyx_r = 0;
      __pyx_f_4cyjs_5_cyjs_arg_parser_finish((&__pyx_v_p)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 645, __pyx_L1_error)
      __pyx_r = __pyx_t_12;
      __pyx_t_12 = 0;
      goto __pyx_L0;
    }
  }

  /* "cyjs/_cyjs.pyx":628
 *             arg_parser_finish(&p)
 * 
 *     def invoke(self, object func, *args):             # <<<<<<<<<<<<<<
 *         cdef JSAtom atom
 *         cdef JSContext* ctx = self.ctx
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Object.invoke", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":648
 * 
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         JS_FreeValue(self.ctx, self.value)
 * 
*/

/* Python wrapper */
static void __pyx_pw_4cyjs_5_cyjs_6Object_11__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4cyjs_5_cyjs_6Object_11__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_4cyjs_5_cyjs_6Object_10__dealloc__(((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4cyjs_5_cyjs_6Object_10__dealloc__(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self) {

  /* "cyjs/_cyjs.pyx":649
 * 
 *     def __dealloc__(self):
 *         JS_FreeValue(self.ctx, self.value)             # <<<<<<<<<<<<<<
 * 
 *     # Mapping attributes can help with debugging attributes
*/
  JS_FreeValue(__pyx_v_self->ctx, __pyx_v_self->value);

  /* "cyjs/_cyjs.pyx":648
 * 
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         JS_FreeValue(self.ctx, self.value)
 * 
*/

  /* function exit code */
}

/* "cyjs/_cyjs.pyx":652
 * 
 *     # Mapping attributes can help with debugging attributes
 *     def items(self):             # <<<<<<<<<<<<<<
 *         return _ObjectItemsView(self)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_13items(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_6Object_13items = {"items", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_13items, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_13items(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("items (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("items", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("items", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6Object_12items(((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_12items(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("items", 0);

  /* "cyjs/_cyjs.pyx":653
 *     # Mapping attributes can help with debugging attributes
 *     def items(self):
 *         return _ObjectItemsView(self)             # <<<<<<<<<<<<<<
 * 
 *     def values(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 653, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_r = ((PyObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":652
 * 
 *     # Mapping attributes can help with debugging attributes
 *     def items(self):             # <<<<<<<<<<<<<<
 *         return _ObjectItemsView(self)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cyjs._cyjs.Object.items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":655
 *         return _ObjectItemsView(self)
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         return _ObjectValuesView(self)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_15values(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_6Object_15values = {"values", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_15values, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_15values(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("values (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("values", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("values", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6Object_14values(((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_14values(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("values", 0);

  /* "cyjs/_cyjs.pyx":656
 * 
 *     def values(self):
 *         return _ObjectValuesView(self)             # <<<<<<<<<<<<<<
 * 
 *     def keys(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 656, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_r = ((PyObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":655
 *         return _ObjectItemsView(self)
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         return _ObjectValuesView(self)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cyjs._cyjs.Object.values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":658
 *         return _ObjectValuesView(self)
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         return _ObjectKeysView(self)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_17keys(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_6Object_17keys = {"keys", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_17keys, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_17keys(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keys (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("keys", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("keys", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6Object_16keys(((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_16keys(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("keys", 0);

  /* "cyjs/_cyjs.pyx":659
 * 
 *     def keys(self):
 *         return _ObjectKeysView(self)             # <<<<<<<<<<<<<<
 * 
 *     # eval_this functions (proxied for the sake of easier access)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 659, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_r = ((PyObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":658
 *         return _ObjectValuesView(self)
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         return _ObjectKeysView(self)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cyjs._cyjs.Object.keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":663
 *     # eval_this functions (proxied for the sake of easier access)
 * 
 *     cpdef object eval(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_19eval(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_6Object_eval(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, PyObject *__pyx_v_code, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_6Object_eval *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":666
 *         self,
 *         object code,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":667
 *         object code,
 *         object filename = None,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":668
 *         object filename = None,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":669
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         return self.context.ceval_this(code, self, filename, False, strict, backtrace_barrier, promise)
*/
  int __pyx_v_promise = ((int)0);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  struct __pyx_fuse_1__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eval", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_strict = __pyx_optional_args->strict;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_promise = __pyx_optional_args->promise;
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":663
 *     # eval_this functions (proxied for the sake of easier access)
 * 
 *     cpdef object eval(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_eval); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 663, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_6Object_19eval)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_strict); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 663, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_backtrace_barrier); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 663, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_promise); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 663, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_8 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[6] = {__pyx_t_3, __pyx_v_code, __pyx_v_filename, __pyx_t_5, __pyx_t_6, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_8, (6-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 663, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":671
 *         bint promise = False
 *     ):
 *         return self.context.ceval_this(code, self, filename, False, strict, backtrace_barrier, promise)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object eval_module(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_9.__pyx_n = 5;
  __pyx_t_9.filename = __pyx_v_filename;
  __pyx_t_9.module = 0;
  __pyx_t_9.strict = __pyx_v_strict;
  __pyx_t_9.backtrace_barrier = __pyx_v_backtrace_barrier;
  __pyx_t_9.promise = __pyx_v_promise;
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->context->__pyx_vtab)->__pyx_fuse_1ceval_this(__pyx_v_self->context, __pyx_v_code, __pyx_v_self, &__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":663
 *     # eval_this functions (proxied for the sake of easier access)
 * 
 *     cpdef object eval(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cyjs._cyjs.Object.eval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_19eval(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_6Object_19eval = {"eval", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_19eval, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_19eval(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_code = 0;
  PyObject *__pyx_v_filename = 0;
  int __pyx_v_strict;
  int __pyx_v_backtrace_barrier;
  int __pyx_v_promise;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("eval (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_code,&__pyx_mstate_global->__pyx_n_u_filename,&__pyx_mstate_global->__pyx_n_u_strict,&__pyx_mstate_global->__pyx_n_u_backtrace_barrier,&__pyx_mstate_global->__pyx_n_u_promise,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 663, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 663, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 663, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 663, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 663, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 663, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "eval", 0) < (0)) __PYX_ERR(0, 663, __pyx_L3_error)

      /* "cyjs/_cyjs.pyx":666
 *         self,
 *         object code,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("eval", 0, 1, 5, i); __PYX_ERR(0, 663, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 663, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 663, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 663, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 663, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 663, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_code = values[0];
    __pyx_v_filename = values[1];
    if (values[2]) {
      __pyx_v_strict = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_strict == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 667, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":667
 *         object code,
 *         object filename = None,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
      __pyx_v_strict = ((int)0);
    }
    if (values[3]) {
      __pyx_v_backtrace_barrier = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_backtrace_barrier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 668, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":668
 *         object filename = None,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
      __pyx_v_backtrace_barrier = ((int)0);
    }
    if (values[4]) {
      __pyx_v_promise = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_promise == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 669, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":669
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         return self.context.ceval_this(code, self, filename, False, strict, backtrace_barrier, promise)
*/
      __pyx_v_promise = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("eval", 0, 1, 5, __pyx_nargs); __PYX_ERR(0, 663, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Object.eval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6Object_18eval(((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_self), __pyx_v_code, __pyx_v_filename, __pyx_v_strict, __pyx_v_backtrace_barrier, __pyx_v_promise);

  /* "cyjs/_cyjs.pyx":663
 *     # eval_this functions (proxied for the sake of easier access)
 * 
 *     cpdef object eval(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_18eval(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_filename, int __pyx_v_strict, int __pyx_v_backtrace_barrier, int __pyx_v_promise) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4cyjs_5_cyjs_6Object_eval __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eval", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 4;
  __pyx_t_2.filename = __pyx_v_filename;
  __pyx_t_2.strict = __pyx_v_strict;
  __pyx_t_2.backtrace_barrier = __pyx_v_backtrace_barrier;
  __pyx_t_2.promise = __pyx_v_promise;
  __pyx_t_1 = __pyx_vtabptr_4cyjs_5_cyjs_Object->eval(__pyx_v_self, __pyx_v_code, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Object.eval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":673
 *         return self.context.ceval_this(code, self, filename, False, strict, backtrace_barrier, promise)
 * 
 *     cpdef object eval_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_21eval_module(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_6Object_eval_module(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, PyObject *__pyx_v_code, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_6Object_eval_module *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":676
 *         self,
 *         object code,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":677
 *         object code,
 *         object filename = None,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":678
 *         object filename = None,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":679
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         return self.context.ceval_this(code, self, filename, True, strict, backtrace_barrier, promise)
*/
  int __pyx_v_promise = ((int)0);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  struct __pyx_fuse_1__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eval_module", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_strict = __pyx_optional_args->strict;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_promise = __pyx_optional_args->promise;
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":673
 *         return self.context.ceval_this(code, self, filename, False, strict, backtrace_barrier, promise)
 * 
 *     cpdef object eval_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_eval_module); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 673, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_6Object_21eval_module)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_strict); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 673, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_backtrace_barrier); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 673, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_promise); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 673, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_8 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[6] = {__pyx_t_3, __pyx_v_code, __pyx_v_filename, __pyx_t_5, __pyx_t_6, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_8, (6-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 673, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":681
 *         bint promise = False
 *     ):
 *         return self.context.ceval_this(code, self, filename, True, strict, backtrace_barrier, promise)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_9.__pyx_n = 5;
  __pyx_t_9.filename = __pyx_v_filename;
  __pyx_t_9.module = 1;
  __pyx_t_9.strict = __pyx_v_strict;
  __pyx_t_9.backtrace_barrier = __pyx_v_backtrace_barrier;
  __pyx_t_9.promise = __pyx_v_promise;
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->context->__pyx_vtab)->__pyx_fuse_1ceval_this(__pyx_v_self->context, __pyx_v_code, __pyx_v_self, &__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":673
 *         return self.context.ceval_this(code, self, filename, False, strict, backtrace_barrier, promise)
 * 
 *     cpdef object eval_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cyjs._cyjs.Object.eval_module", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_21eval_module(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_6Object_21eval_module = {"eval_module", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_21eval_module, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_21eval_module(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_code = 0;
  PyObject *__pyx_v_filename = 0;
  int __pyx_v_strict;
  int __pyx_v_backtrace_barrier;
  int __pyx_v_promise;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("eval_module (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_code,&__pyx_mstate_global->__pyx_n_u_filename,&__pyx_mstate_global->__pyx_n_u_strict,&__pyx_mstate_global->__pyx_n_u_backtrace_barrier,&__pyx_mstate_global->__pyx_n_u_promise,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 673, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 673, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 673, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 673, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 673, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 673, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "eval_module", 0) < (0)) __PYX_ERR(0, 673, __pyx_L3_error)

      /* "cyjs/_cyjs.pyx":676
 *         self,
 *         object code,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("eval_module", 0, 1, 5, i); __PYX_ERR(0, 673, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 673, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 673, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 673, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 673, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 673, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_code = values[0];
    __pyx_v_filename = values[1];
    if (values[2]) {
      __pyx_v_strict = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_strict == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":677
 *         object code,
 *         object filename = None,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
      __pyx_v_strict = ((int)0);
    }
    if (values[3]) {
      __pyx_v_backtrace_barrier = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_backtrace_barrier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 678, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":678
 *         object filename = None,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
      __pyx_v_backtrace_barrier = ((int)0);
    }
    if (values[4]) {
      __pyx_v_promise = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_promise == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 679, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":679
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         return self.context.ceval_this(code, self, filename, True, strict, backtrace_barrier, promise)
*/
      __pyx_v_promise = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("eval_module", 0, 1, 5, __pyx_nargs); __PYX_ERR(0, 673, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Object.eval_module", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6Object_20eval_module(((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_self), __pyx_v_code, __pyx_v_filename, __pyx_v_strict, __pyx_v_backtrace_barrier, __pyx_v_promise);

  /* "cyjs/_cyjs.pyx":673
 *         return self.context.ceval_this(code, self, filename, False, strict, backtrace_barrier, promise)
 * 
 *     cpdef object eval_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_20eval_module(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_filename, int __pyx_v_strict, int __pyx_v_backtrace_barrier, int __pyx_v_promise) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4cyjs_5_cyjs_6Object_eval_module __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eval_module", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 4;
  __pyx_t_2.filename = __pyx_v_filename;
  __pyx_t_2.strict = __pyx_v_strict;
  __pyx_t_2.backtrace_barrier = __pyx_v_backtrace_barrier;
  __pyx_t_2.promise = __pyx_v_promise;
  __pyx_t_1 = __pyx_vtabptr_4cyjs_5_cyjs_Object->eval_module(__pyx_v_self, __pyx_v_code, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Object.eval_module", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":149
 * cdef class Object:
 *     cdef:
 *         readonly Context context             # <<<<<<<<<<<<<<
 *         readonly int32_t tag
 *         JSValue value
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_7context_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_7context_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6Object_7context___get__(((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_7context___get__(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->context);
  __pyx_r = ((PyObject *)__pyx_v_self->context);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":150
 *     cdef:
 *         readonly Context context
 *         readonly int32_t tag             # <<<<<<<<<<<<<<
 *         JSValue value
 *         JSContext* ctx
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_3tag_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_3tag_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6Object_3tag___get__(((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_3tag___get__(struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int32_t(__pyx_v_self->tag); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Object.tag.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_23__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_6Object_23__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_23__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_23__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6Object_22__reduce_cython__(((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_22__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ctx_self_value_cannot_be_co, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.Object.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_25__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_6Object_25__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_25__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_6Object_25__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Object.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6Object_24__setstate_cython__(((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6Object_24__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ctx_self_value_cannot_be_co, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.Object.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":688
 * # way to obtain it elsewhere using a INCREF (on creation) and DECREF (on finalization) respectively...
 * 
 * cdef JSValue on_cclosure_callback(JSContext *ctx,             # <<<<<<<<<<<<<<
 *     JSValue this_val,
 *     int argc, JSValue *argv,
*/

static JSValue __pyx_f_4cyjs_5_cyjs_on_cclosure_callback(JSContext *__pyx_v_ctx, JSValue __pyx_v_this_val, int __pyx_v_argc, JSValue *__pyx_v_argv, int __pyx_v_magic, void *__pyx_v_func_data) {
  struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_jsfunc = 0;
  JSValue __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_RefNannySetupContext("on_cclosure_callback", 0);

  /* "cyjs/_cyjs.pyx":692
 *     int argc, JSValue *argv,
 *     int magic, void *func_data) noexcept with gil:
 *     cdef JSFunction jsfunc = <JSFunction>(func_data)             # <<<<<<<<<<<<<<
 *     return jsfunc.call_js(ctx, this_val, argc, argv, magic)
 * 
*/
  __pyx_t_1 = ((PyObject *)__pyx_v_func_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_jsfunc = ((struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":693
 *     int magic, void *func_data) noexcept with gil:
 *     cdef JSFunction jsfunc = <JSFunction>(func_data)
 *     return jsfunc.call_js(ctx, this_val, argc, argv, magic)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_JSFunction *)__pyx_v_jsfunc->__pyx_vtab)->call_js(__pyx_v_jsfunc, __pyx_v_ctx, __pyx_v_this_val, __pyx_v_argc, __pyx_v_argv, __pyx_v_magic);
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":688
 * # way to obtain it elsewhere using a INCREF (on creation) and DECREF (on finalization) respectively...
 * 
 * cdef JSValue on_cclosure_callback(JSContext *ctx,             # <<<<<<<<<<<<<<
 *     JSValue this_val,
 *     int argc, JSValue *argv,
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_jsfunc);
  __Pyx_RefNannyFinishContext();
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":698
 * # TODO: (Vizonex) Check if using INCREF and DECREF Causes memory leaks
 * # Everyone Else: Feel free to throw an issue if this is casuing memory leaks on your end.
 * cdef void on_cclosure_opaque_finalize(void *opaque) noexcept with gil:             # <<<<<<<<<<<<<<
 *     cdef JSFunction jsfunc = <JSFunction>opaque
 *     # Deref it as it possibly means we finished needing it within Quickjs
*/

static void __pyx_f_4cyjs_5_cyjs_on_cclosure_opaque_finalize(void *__pyx_v_opaque) {
  struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_jsfunc = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_RefNannySetupContext("on_cclosure_opaque_finalize", 0);

  /* "cyjs/_cyjs.pyx":699
 * # Everyone Else: Feel free to throw an issue if this is casuing memory leaks on your end.
 * cdef void on_cclosure_opaque_finalize(void *opaque) noexcept with gil:
 *     cdef JSFunction jsfunc = <JSFunction>opaque             # <<<<<<<<<<<<<<
 *     # Deref it as it possibly means we finished needing it within Quickjs
 *     Py_XDECREF(jsfunc)
*/
  __pyx_t_1 = ((PyObject *)__pyx_v_opaque);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_jsfunc = ((struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":701
 *     cdef JSFunction jsfunc = <JSFunction>opaque
 *     # Deref it as it possibly means we finished needing it within Quickjs
 *     Py_XDECREF(jsfunc)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  Py_XDECREF(((PyObject *)__pyx_v_jsfunc));

  /* "cyjs/_cyjs.pyx":698
 * # TODO: (Vizonex) Check if using INCREF and DECREF Causes memory leaks
 * # Everyone Else: Feel free to throw an issue if this is casuing memory leaks on your end.
 * cdef void on_cclosure_opaque_finalize(void *opaque) noexcept with gil:             # <<<<<<<<<<<<<<
 *     cdef JSFunction jsfunc = <JSFunction>opaque
 *     # Deref it as it possibly means we finished needing it within Quickjs
*/

  /* function exit code */
  __Pyx_XDECREF((PyObject *)__pyx_v_jsfunc);
  __Pyx_RefNannyFinishContext();
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
}

/* "cyjs/_cyjs.pyx":714
 * 
 *     # Converter for JSFunction
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef JSFunction new(Context context, object func):
 *         cdef JSFunction self = JSFunction.__new__(JSFunction)
*/

static struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_f_4cyjs_5_cyjs_10JSFunction_new(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_context, PyObject *__pyx_v_func) {
  struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_self = 0;
  struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  JSContext *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 0);

  /* "cyjs/_cyjs.pyx":716
 *     @staticmethod
 *     cdef JSFunction new(Context context, object func):
 *         cdef JSFunction self = JSFunction.__new__(JSFunction)             # <<<<<<<<<<<<<<
 *         self.context = context
 *         self.ctx = context.ctx
*/
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_4cyjs_5_cyjs_JSFunction(((PyTypeObject *)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_JSFunction), __pyx_mstate_global->__pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_self = ((struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":717
 *     cdef JSFunction new(Context context, object func):
 *         cdef JSFunction self = JSFunction.__new__(JSFunction)
 *         self.context = context             # <<<<<<<<<<<<<<
 *         self.ctx = context.ctx
 *         # NOTE: self.value comes in a bit later...
*/
  __Pyx_INCREF((PyObject *)__pyx_v_context);
  __Pyx_GIVEREF((PyObject *)__pyx_v_context);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->context);
  __Pyx_DECREF((PyObject *)__pyx_v_self->context);
  __pyx_v_self->context = __pyx_v_context;

  /* "cyjs/_cyjs.pyx":718
 *         cdef JSFunction self = JSFunction.__new__(JSFunction)
 *         self.context = context
 *         self.ctx = context.ctx             # <<<<<<<<<<<<<<
 *         # NOTE: self.value comes in a bit later...
 *         self.func = func
*/
  __pyx_t_2 = __pyx_v_context->ctx;
  __pyx_v_self->ctx = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":720
 *         self.ctx = context.ctx
 *         # NOTE: self.value comes in a bit later...
 *         self.func = func             # <<<<<<<<<<<<<<
 *         return self
 * 
*/
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->func);
  __Pyx_DECREF(__pyx_v_self->func);
  __pyx_v_self->func = __pyx_v_func;

  /* "cyjs/_cyjs.pyx":721
 *         # NOTE: self.value comes in a bit later...
 *         self.func = func
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":714
 * 
 *     # Converter for JSFunction
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef JSFunction new(Context context, object func):
 *         cdef JSFunction self = JSFunction.__new__(JSFunction)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.JSFunction.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":723
 *         return self
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         JS_FreeValue(self.ctx, self.value)
 * 
*/

/* Python wrapper */
static void __pyx_pw_4cyjs_5_cyjs_10JSFunction_1__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4cyjs_5_cyjs_10JSFunction_1__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_4cyjs_5_cyjs_10JSFunction___dealloc__(((struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4cyjs_5_cyjs_10JSFunction___dealloc__(struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_self) {

  /* "cyjs/_cyjs.pyx":724
 * 
 *     def __dealloc__(self):
 *         JS_FreeValue(self.ctx, self.value)             # <<<<<<<<<<<<<<
 * 
 *     # quick access shortcut for calling JSFunction as if it were a python function.
*/
  JS_FreeValue(__pyx_v_self->ctx, __pyx_v_self->value);

  /* "cyjs/_cyjs.pyx":723
 *         return self
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         JS_FreeValue(self.ctx, self.value)
 * 
*/

  /* function exit code */
}

/* "cyjs/_cyjs.pyx":727
 * 
 *     # quick access shortcut for calling JSFunction as if it were a python function.
 *     def __call__(self, *args, **kwds):             # <<<<<<<<<<<<<<
 *         return self.func.__call__(self, *args, **kwds)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_10JSFunction_3__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_10JSFunction_3__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwds = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__call__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (__pyx_kwds_len > 0) {
    if (unlikely(__Pyx_CheckKeywordStrings("__call__", __pyx_kwds) == -1)) return NULL;
    __pyx_v_kwds = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwds)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwds);
  } else {
    __pyx_v_kwds = PyDict_New();
    if (unlikely(!__pyx_v_kwds)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwds);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_10JSFunction_2__call__(((struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)__pyx_v_self), __pyx_v_args, __pyx_v_kwds);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwds);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_10JSFunction_2__call__(struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwds) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__call__", 0);

  /* "cyjs/_cyjs.pyx":728
 *     # quick access shortcut for calling JSFunction as if it were a python function.
 *     def __call__(self, *args, **kwds):
 *         return self.func.__call__(self, *args, **kwds)             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->func, __pyx_mstate_global->__pyx_n_u_call); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self)) != (0)) __PYX_ERR(0, 728, __pyx_L1_error);
  __pyx_t_3 = PyNumber_Add(__pyx_t_2, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwds); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":727
 * 
 *     # quick access shortcut for calling JSFunction as if it were a python function.
 *     def __call__(self, *args, **kwds):             # <<<<<<<<<<<<<<
 *         return self.func.__call__(self, *args, **kwds)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.JSFunction.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":730
 *         return self.func.__call__(self, *args, **kwds)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def object(self):
 *         """Obtains the value as a Javascript Object that could in theory be manipulated
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_10JSFunction_6object_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_10JSFunction_6object_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_10JSFunction_6object___get__(((struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_10JSFunction_6object___get__(struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "cyjs/_cyjs.pyx":734
 *         """Obtains the value as a Javascript Object that could in theory be manipulated
 *         NOTE: that this calls upon JS_DupValue to prevent derefing the values"""
 *         return jsv_to_object(self.ctx, JS_DupValue(self.ctx, self.value))             # <<<<<<<<<<<<<<
 * 
 *     cdef JSValue call_js(self, JSContext *ctx, JSValue this_val, int argc, JSValue *argv, int magic) noexcept:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_4cyjs_5_cyjs_jsv_to_object(__pyx_v_self->ctx, JS_DupValue(__pyx_v_self->ctx, __pyx_v_self->value))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":730
 *         return self.func.__call__(self, *args, **kwds)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def object(self):
 *         """Obtains the value as a Javascript Object that could in theory be manipulated
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.JSFunction.object.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":736
 *         return jsv_to_object(self.ctx, JS_DupValue(self.ctx, self.value))
 * 
 *     cdef JSValue call_js(self, JSContext *ctx, JSValue this_val, int argc, JSValue *argv, int magic) noexcept:             # <<<<<<<<<<<<<<
 *         cdef arg_parser_t parser
 *         cdef tuple args
*/

static JSValue __pyx_f_4cyjs_5_cyjs_10JSFunction_call_js(struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_self, JSContext *__pyx_v_ctx, CYTHON_UNUSED JSValue __pyx_v_this_val, int __pyx_v_argc, JSValue *__pyx_v_argv, CYTHON_UNUSED int __pyx_v_magic) {
  struct __pyx_t_4cyjs_5_cyjs_arg_parser_t __pyx_v_parser;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_ret = 0;
  JSValue __pyx_v_js_ret;
  PyObject *__pyx_v_e = NULL;
  JSValue __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  JSValue __pyx_t_9;
  int __pyx_t_10;
  char const *__pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("call_js", 0);

  /* "cyjs/_cyjs.pyx":741
 *         cdef object ret
 *         cdef JSValue js_ret
 *         arg_parser_init(&parser, ctx, argv, argc, argc)             # <<<<<<<<<<<<<<
 * 
 *         try:
*/
  __pyx_f_4cyjs_5_cyjs_arg_parser_init((&__pyx_v_parser), __pyx_v_ctx, __pyx_v_argv, __pyx_v_argc, __pyx_v_argc); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 741, __pyx_L1_error)

  /* "cyjs/_cyjs.pyx":743
 *         arg_parser_init(&parser, ctx, argv, argc, argc)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             args = arg_parser_unpack(&parser)
 *             ret = PyObject_CallObject(self.func, args)
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "cyjs/_cyjs.pyx":744
 * 
 *         try:
 *             args = arg_parser_unpack(&parser)             # <<<<<<<<<<<<<<
 *             ret = PyObject_CallObject(self.func, args)
 *             arg_parser_finish(&parser)
*/
      __pyx_t_4 = __pyx_f_4cyjs_5_cyjs_arg_parser_unpack((&__pyx_v_parser)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 744, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_args = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "cyjs/_cyjs.pyx":745
 *         try:
 *             args = arg_parser_unpack(&parser)
 *             ret = PyObject_CallObject(self.func, args)             # <<<<<<<<<<<<<<
 *             arg_parser_finish(&parser)
 * 
*/
      __pyx_t_4 = __pyx_v_self->func;
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_5 = PyObject_CallObject(__pyx_t_4, __pyx_v_args); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 745, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_ret = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "cyjs/_cyjs.pyx":746
 *             args = arg_parser_unpack(&parser)
 *             ret = PyObject_CallObject(self.func, args)
 *             arg_parser_finish(&parser)             # <<<<<<<<<<<<<<
 * 
 *             if to_quickjs(ctx, &js_ret, ret) < 0:
*/
      __pyx_f_4cyjs_5_cyjs_arg_parser_finish((&__pyx_v_parser)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 746, __pyx_L3_error)

      /* "cyjs/_cyjs.pyx":748
 *             arg_parser_finish(&parser)
 * 
 *             if to_quickjs(ctx, &js_ret, ret) < 0:             # <<<<<<<<<<<<<<
 *                 raise
 *             return js_ret
*/
      __pyx_t_6 = (__pyx_fuse_8__pyx_f_4cyjs_5_cyjs_to_quickjs(__pyx_v_ctx, (&__pyx_v_js_ret), __pyx_v_ret) < 0);
      if (unlikely(__pyx_t_6)) {

        /* "cyjs/_cyjs.pyx":749
 * 
 *             if to_quickjs(ctx, &js_ret, ret) < 0:
 *                 raise             # <<<<<<<<<<<<<<
 *             return js_ret
 * 
*/
        __Pyx_ReraiseException();
        __PYX_ERR(0, 749, __pyx_L3_error)

        /* "cyjs/_cyjs.pyx":748
 *             arg_parser_finish(&parser)
 * 
 *             if to_quickjs(ctx, &js_ret, ret) < 0:             # <<<<<<<<<<<<<<
 *                 raise
 *             return js_ret
*/
      }

      /* "cyjs/_cyjs.pyx":750
 *             if to_quickjs(ctx, &js_ret, ret) < 0:
 *                 raise
 *             return js_ret             # <<<<<<<<<<<<<<
 * 
 *         except Exception as e:
*/
      __pyx_r = __pyx_v_js_ret;
      goto __pyx_L7_try_return;

      /* "cyjs/_cyjs.pyx":743
 *         arg_parser_init(&parser, ctx, argv, argc, argc)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             args = arg_parser_unpack(&parser)
 *             ret = PyObject_CallObject(self.func, args)
*/
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "cyjs/_cyjs.pyx":752
 *             return js_ret
 * 
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             # set callback exception so python can throw it later...
 *             self.context._cb_exception = e
*/
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_Exception))));
    if (__pyx_t_7) {
      __Pyx_AddTraceback("cyjs._cyjs.JSFunction.call_js", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_8) < 0) __PYX_ERR(0, 752, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_v_e = __pyx_t_4;
      /*try:*/ {

        /* "cyjs/_cyjs.pyx":754
 *         except Exception as e:
 *             # set callback exception so python can throw it later...
 *             self.context._cb_exception = e             # <<<<<<<<<<<<<<
 *             arg_parser_finish(&parser)
 *             return py_to_js_exception(ctx, e)
*/
        __Pyx_INCREF(__pyx_v_e);
        __Pyx_GIVEREF(__pyx_v_e);
        __Pyx_GOTREF(__pyx_v_self->context->_cb_exception);
        __Pyx_DECREF(__pyx_v_self->context->_cb_exception);
        __pyx_v_self->context->_cb_exception = __pyx_v_e;

        /* "cyjs/_cyjs.pyx":755
 *             # set callback exception so python can throw it later...
 *             self.context._cb_exception = e
 *             arg_parser_finish(&parser)             # <<<<<<<<<<<<<<
 *             return py_to_js_exception(ctx, e)
 * 
*/
        __pyx_f_4cyjs_5_cyjs_arg_parser_finish((&__pyx_v_parser)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 755, __pyx_L15_error)

        /* "cyjs/_cyjs.pyx":756
 *             self.context._cb_exception = e
 *             arg_parser_finish(&parser)
 *             return py_to_js_exception(ctx, e)             # <<<<<<<<<<<<<<
 * 
 * 
*/
        __pyx_t_9 = __pyx_f_4cyjs_5_cyjs_py_to_js_exception(__pyx_v_ctx, __pyx_v_e); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 756, __pyx_L15_error)
        __pyx_r = __pyx_t_9;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L14_return;
      }

      /* "cyjs/_cyjs.pyx":752
 *             return js_ret
 * 
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             # set callback exception so python can throw it later...
 *             self.context._cb_exception = e
*/
      /*finally:*/ {
        __pyx_L15_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
           __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
          if ( unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __pyx_t_7 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_11 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
          }
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_ErrRestore(__pyx_t_12, __pyx_t_13, __pyx_t_14);
          __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
          __pyx_lineno = __pyx_t_7; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_11;
          goto __pyx_L5_except_error;
        }
        __pyx_L14_return: {
          __pyx_t_9 = __pyx_r;
          __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
          __pyx_r = __pyx_t_9;
          goto __pyx_L6_except_return;
        }
      }
    }
    goto __pyx_L5_except_error;

    /* "cyjs/_cyjs.pyx":743
 *         arg_parser_init(&parser, ctx, argv, argc, argc)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             args = arg_parser_unpack(&parser)
 *             ret = PyObject_CallObject(self.func, args)
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "cyjs/_cyjs.pyx":736
 *         return jsv_to_object(self.ctx, JS_DupValue(self.ctx, self.value))
 * 
 *     cdef JSValue call_js(self, JSContext *ctx, JSValue this_val, int argc, JSValue *argv, int magic) noexcept:             # <<<<<<<<<<<<<<
 *         cdef arg_parser_t parser
 *         cdef tuple args
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_WriteUnraisable("cyjs._cyjs.JSFunction.call_js", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":185
 *     cdef:
 *         # access to python Quickjs-ng Context
 *         readonly Context context             # <<<<<<<<<<<<<<
 *         JSContext* ctx # quick access to C JSContext
 *         JSValue value
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_10JSFunction_7context_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_10JSFunction_7context_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_10JSFunction_7context___get__(((struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_10JSFunction_7context___get__(struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->context);
  __pyx_r = ((PyObject *)__pyx_v_self->context);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_10JSFunction_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_10JSFunction_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_10JSFunction_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_10JSFunction_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_10JSFunction_4__reduce_cython__(((struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_10JSFunction_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ctx_self_value_cannot_be_co, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.JSFunction.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_10JSFunction_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_10JSFunction_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_10JSFunction_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_10JSFunction_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.JSFunction.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_10JSFunction_6__setstate_cython__(((struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_10JSFunction_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ctx_self_value_cannot_be_co, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.JSFunction.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":777
 * 
 * cdef class Context: # type: ignore
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *         Runtime runtime = Runtime(),
 *         bint base_objects = True,
*/

/* Python wrapper */
static int __pyx_pw_4cyjs_5_cyjs_7Context_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4cyjs_5_cyjs_7Context_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_runtime = 0;
  int __pyx_v_base_objects;
  int __pyx_v_date;
  int __pyx_v_intrinsic_eval;
  int __pyx_v_regexp_compiler;
  int __pyx_v_regexp;
  int __pyx_v_json;
  int __pyx_v_proxy;
  int __pyx_v_map_set;
  int __pyx_v_typed_arrays;
  int __pyx_v_bigint;
  int __pyx_v_weak_ref;
  int __pyx_v_performance;
  int __pyx_v_dom_exception;
  int __pyx_v_promise;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[15] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_runtime,&__pyx_mstate_global->__pyx_n_u_base_objects,&__pyx_mstate_global->__pyx_n_u_date,&__pyx_mstate_global->__pyx_n_u_intrinsic_eval,&__pyx_mstate_global->__pyx_n_u_regexp_compiler,&__pyx_mstate_global->__pyx_n_u_regexp,&__pyx_mstate_global->__pyx_n_u_json,&__pyx_mstate_global->__pyx_n_u_proxy,&__pyx_mstate_global->__pyx_n_u_map_set,&__pyx_mstate_global->__pyx_n_u_typed_arrays,&__pyx_mstate_global->__pyx_n_u_bigint,&__pyx_mstate_global->__pyx_n_u_weak_ref,&__pyx_mstate_global->__pyx_n_u_performance,&__pyx_mstate_global->__pyx_n_u_dom_exception,&__pyx_mstate_global->__pyx_n_u_promise,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 777, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case 15:
        values[14] = __Pyx_ArgRef_VARARGS(__pyx_args, 14);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[14])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 14:
        values[13] = __Pyx_ArgRef_VARARGS(__pyx_args, 13);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[13])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 13:
        values[12] = __Pyx_ArgRef_VARARGS(__pyx_args, 12);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[12])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 12:
        values[11] = __Pyx_ArgRef_VARARGS(__pyx_args, 11);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[11])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 11:
        values[10] = __Pyx_ArgRef_VARARGS(__pyx_args, 10);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 10:
        values[9] = __Pyx_ArgRef_VARARGS(__pyx_args, 9);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  9:
        values[8] = __Pyx_ArgRef_VARARGS(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 777, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef((PyObject *)__pyx_mstate_global->__pyx_k_);
    } else {
      switch (__pyx_nargs) {
        case 15:
        values[14] = __Pyx_ArgRef_VARARGS(__pyx_args, 14);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[14])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 14:
        values[13] = __Pyx_ArgRef_VARARGS(__pyx_args, 13);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[13])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 13:
        values[12] = __Pyx_ArgRef_VARARGS(__pyx_args, 12);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[12])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 12:
        values[11] = __Pyx_ArgRef_VARARGS(__pyx_args, 11);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[11])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 11:
        values[10] = __Pyx_ArgRef_VARARGS(__pyx_args, 10);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 10:
        values[9] = __Pyx_ArgRef_VARARGS(__pyx_args, 9);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  9:
        values[8] = __Pyx_ArgRef_VARARGS(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 777, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef((PyObject *)__pyx_mstate_global->__pyx_k_);
    }
    __pyx_v_runtime = ((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)values[0]);
    if (values[1]) {
      __pyx_v_base_objects = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_base_objects == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 779, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":779
 *     def __init__(self,
 *         Runtime runtime = Runtime(),
 *         bint base_objects = True,             # <<<<<<<<<<<<<<
 *         bint date = True,
 *         bint intrinsic_eval = True,
*/
      __pyx_v_base_objects = ((int)1);
    }
    if (values[2]) {
      __pyx_v_date = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_date == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 780, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":780
 *         Runtime runtime = Runtime(),
 *         bint base_objects = True,
 *         bint date = True,             # <<<<<<<<<<<<<<
 *         bint intrinsic_eval = True,
 *         bint regexp_compiler = True,
*/
      __pyx_v_date = ((int)1);
    }
    if (values[3]) {
      __pyx_v_intrinsic_eval = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_intrinsic_eval == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 781, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":781
 *         bint base_objects = True,
 *         bint date = True,
 *         bint intrinsic_eval = True,             # <<<<<<<<<<<<<<
 *         bint regexp_compiler = True,
 *         bint regexp = True,
*/
      __pyx_v_intrinsic_eval = ((int)1);
    }
    if (values[4]) {
      __pyx_v_regexp_compiler = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_regexp_compiler == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 782, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":782
 *         bint date = True,
 *         bint intrinsic_eval = True,
 *         bint regexp_compiler = True,             # <<<<<<<<<<<<<<
 *         bint regexp = True,
 *         bint json = True,
*/
      __pyx_v_regexp_compiler = ((int)1);
    }
    if (values[5]) {
      __pyx_v_regexp = __Pyx_PyObject_IsTrue(values[5]); if (unlikely((__pyx_v_regexp == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 783, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":783
 *         bint intrinsic_eval = True,
 *         bint regexp_compiler = True,
 *         bint regexp = True,             # <<<<<<<<<<<<<<
 *         bint json = True,
 *         bint proxy = True,
*/
      __pyx_v_regexp = ((int)1);
    }
    if (values[6]) {
      __pyx_v_json = __Pyx_PyObject_IsTrue(values[6]); if (unlikely((__pyx_v_json == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 784, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":784
 *         bint regexp_compiler = True,
 *         bint regexp = True,
 *         bint json = True,             # <<<<<<<<<<<<<<
 *         bint proxy = True,
 *         bint map_set = True,
*/
      __pyx_v_json = ((int)1);
    }
    if (values[7]) {
      __pyx_v_proxy = __Pyx_PyObject_IsTrue(values[7]); if (unlikely((__pyx_v_proxy == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 785, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":785
 *         bint regexp = True,
 *         bint json = True,
 *         bint proxy = True,             # <<<<<<<<<<<<<<
 *         bint map_set = True,
 *         bint typed_arrays = True,
*/
      __pyx_v_proxy = ((int)1);
    }
    if (values[8]) {
      __pyx_v_map_set = __Pyx_PyObject_IsTrue(values[8]); if (unlikely((__pyx_v_map_set == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 786, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":786
 *         bint json = True,
 *         bint proxy = True,
 *         bint map_set = True,             # <<<<<<<<<<<<<<
 *         bint typed_arrays = True,
 *         bint bigint = True,
*/
      __pyx_v_map_set = ((int)1);
    }
    if (values[9]) {
      __pyx_v_typed_arrays = __Pyx_PyObject_IsTrue(values[9]); if (unlikely((__pyx_v_typed_arrays == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 787, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":787
 *         bint proxy = True,
 *         bint map_set = True,
 *         bint typed_arrays = True,             # <<<<<<<<<<<<<<
 *         bint bigint = True,
 *         bint weak_ref = True,
*/
      __pyx_v_typed_arrays = ((int)1);
    }
    if (values[10]) {
      __pyx_v_bigint = __Pyx_PyObject_IsTrue(values[10]); if (unlikely((__pyx_v_bigint == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 788, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":788
 *         bint map_set = True,
 *         bint typed_arrays = True,
 *         bint bigint = True,             # <<<<<<<<<<<<<<
 *         bint weak_ref = True,
 *         bint performance = True,
*/
      __pyx_v_bigint = ((int)1);
    }
    if (values[11]) {
      __pyx_v_weak_ref = __Pyx_PyObject_IsTrue(values[11]); if (unlikely((__pyx_v_weak_ref == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 789, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":789
 *         bint typed_arrays = True,
 *         bint bigint = True,
 *         bint weak_ref = True,             # <<<<<<<<<<<<<<
 *         bint performance = True,
 *         bint dom_exception = True,
*/
      __pyx_v_weak_ref = ((int)1);
    }
    if (values[12]) {
      __pyx_v_performance = __Pyx_PyObject_IsTrue(values[12]); if (unlikely((__pyx_v_performance == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 790, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":790
 *         bint bigint = True,
 *         bint weak_ref = True,
 *         bint performance = True,             # <<<<<<<<<<<<<<
 *         bint dom_exception = True,
 *         bint promise = True
*/
      __pyx_v_performance = ((int)1);
    }
    if (values[13]) {
      __pyx_v_dom_exception = __Pyx_PyObject_IsTrue(values[13]); if (unlikely((__pyx_v_dom_exception == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 791, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":791
 *         bint weak_ref = True,
 *         bint performance = True,
 *         bint dom_exception = True,             # <<<<<<<<<<<<<<
 *         bint promise = True
 *     ) -> None:
*/
      __pyx_v_dom_exception = ((int)1);
    }
    if (values[14]) {
      __pyx_v_promise = __Pyx_PyObject_IsTrue(values[14]); if (unlikely((__pyx_v_promise == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 792, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":792
 *         bint performance = True,
 *         bint dom_exception = True,
 *         bint promise = True             # <<<<<<<<<<<<<<
 *     ) -> None:
 * 
*/
      __pyx_v_promise = ((int)1);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 15, __pyx_nargs); __PYX_ERR(0, 777, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Context.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_runtime), __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime, 1, "runtime", 0))) __PYX_ERR(0, 778, __pyx_L1_error)
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Context___init__(((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_v_self), __pyx_v_runtime, __pyx_v_base_objects, __pyx_v_date, __pyx_v_intrinsic_eval, __pyx_v_regexp_compiler, __pyx_v_regexp, __pyx_v_json, __pyx_v_proxy, __pyx_v_map_set, __pyx_v_typed_arrays, __pyx_v_bigint, __pyx_v_weak_ref, __pyx_v_performance, __pyx_v_dom_exception, __pyx_v_promise);

  /* "cyjs/_cyjs.pyx":777
 * 
 * cdef class Context: # type: ignore
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *         Runtime runtime = Runtime(),
 *         bint base_objects = True,
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4cyjs_5_cyjs_7Context___init__(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, struct __pyx_obj_4cyjs_5_cyjs_Runtime *__pyx_v_runtime, int __pyx_v_base_objects, int __pyx_v_date, int __pyx_v_intrinsic_eval, int __pyx_v_regexp_compiler, int __pyx_v_regexp, int __pyx_v_json, int __pyx_v_proxy, int __pyx_v_map_set, int __pyx_v_typed_arrays, int __pyx_v_bigint, int __pyx_v_weak_ref, int __pyx_v_performance, int __pyx_v_dom_exception, int __pyx_v_promise) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  JSRuntime *__pyx_t_1;
  JSContext *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "cyjs/_cyjs.pyx":795
 *     ) -> None:
 * 
 *         self.rt = runtime.rt             # <<<<<<<<<<<<<<
 *         # TODO: Enable Raw Settings also...
 *         self.ctx = runtime.new_context()
*/
  __pyx_t_1 = __pyx_v_runtime->rt;
  __pyx_v_self->rt = __pyx_t_1;

  /* "cyjs/_cyjs.pyx":797
 *         self.rt = runtime.rt
 *         # TODO: Enable Raw Settings also...
 *         self.ctx = runtime.new_context()             # <<<<<<<<<<<<<<
 *         JS_SetContextOpaque(self.ctx, <void*>self)
 *         CYJS_InitalizeSettings(
*/
  __pyx_t_2 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Runtime *)__pyx_v_runtime->__pyx_vtab)->new_context(__pyx_v_runtime); if (unlikely(__pyx_t_2 == ((void *)NULL))) __PYX_ERR(0, 797, __pyx_L1_error)
  __pyx_v_self->ctx = __pyx_t_2;

  /* "cyjs/_cyjs.pyx":798
 *         # TODO: Enable Raw Settings also...
 *         self.ctx = runtime.new_context()
 *         JS_SetContextOpaque(self.ctx, <void*>self)             # <<<<<<<<<<<<<<
 *         CYJS_InitalizeSettings(
 *             self.ctx,
*/
  JS_SetContextOpaque(__pyx_v_self->ctx, ((void *)__pyx_v_self));

  /* "cyjs/_cyjs.pyx":799
 *         self.ctx = runtime.new_context()
 *         JS_SetContextOpaque(self.ctx, <void*>self)
 *         CYJS_InitalizeSettings(             # <<<<<<<<<<<<<<
 *             self.ctx,
 *             base_objects,
*/
  CYJS_InitalizeSettings(__pyx_v_self->ctx, __pyx_v_base_objects, __pyx_v_date, __pyx_v_intrinsic_eval, __pyx_v_regexp_compiler, __pyx_v_regexp, __pyx_v_json, __pyx_v_proxy, __pyx_v_map_set, __pyx_v_typed_arrays, __pyx_v_bigint, __pyx_v_weak_ref, __pyx_v_performance, __pyx_v_dom_exception, __pyx_v_promise);

  /* "cyjs/_cyjs.pyx":818
 * 
 *         # we don't access the parent (unless on rare occations but we keep incase of gc or other cases)
 *         self.runtime = runtime             # <<<<<<<<<<<<<<
 *         # Collection exception if seen in a void callback
 *         self._cb_exception = None
*/
  __Pyx_INCREF((PyObject *)__pyx_v_runtime);
  __Pyx_GIVEREF((PyObject *)__pyx_v_runtime);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->runtime);
  __Pyx_DECREF((PyObject *)__pyx_v_self->runtime);
  __pyx_v_self->runtime = __pyx_v_runtime;

  /* "cyjs/_cyjs.pyx":820
 *         self.runtime = runtime
 *         # Collection exception if seen in a void callback
 *         self._cb_exception = None             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_cb_exception);
  __Pyx_DECREF(__pyx_v_self->_cb_exception);
  __pyx_v_self->_cb_exception = Py_None;

  /* "cyjs/_cyjs.pyx":777
 * 
 * cdef class Context: # type: ignore
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *         Runtime runtime = Runtime(),
 *         bint base_objects = True,
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.Context.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":831
 * 
 *     # raises exception if one is seen.
 *     cdef object raise_exception(self):             # <<<<<<<<<<<<<<
 *         cdef object exc
 *         if self._cb_exception is not None:
*/

static PyObject *__pyx_f_4cyjs_5_cyjs_7Context_raise_exception(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self) {
  PyObject *__pyx_v_exc = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  JSValue __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("raise_exception", 0);

  /* "cyjs/_cyjs.pyx":833
 *     cdef object raise_exception(self):
 *         cdef object exc
 *         if self._cb_exception is not None:             # <<<<<<<<<<<<<<
 *             # Let python exception take over first if seen. it's more accurate than the JSError would be
 *             exc = self._cb_exception
*/
  __pyx_t_1 = (__pyx_v_self->_cb_exception != Py_None);
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":835
 *         if self._cb_exception is not None:
 *             # Let python exception take over first if seen. it's more accurate than the JSError would be
 *             exc = self._cb_exception             # <<<<<<<<<<<<<<
 *             self._cb_exception = None
 *             # Cleanup javascript verison of the given exception since python's
*/
    __pyx_t_2 = __pyx_v_self->_cb_exception;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_exc = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "cyjs/_cyjs.pyx":836
 *             # Let python exception take over first if seen. it's more accurate than the JSError would be
 *             exc = self._cb_exception
 *             self._cb_exception = None             # <<<<<<<<<<<<<<
 *             # Cleanup javascript verison of the given exception since python's
 *             # exception would do a better job explaining the problem.
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_cb_exception);
    __Pyx_DECREF(__pyx_v_self->_cb_exception);
    __pyx_v_self->_cb_exception = Py_None;

    /* "cyjs/_cyjs.pyx":839
 *             # Cleanup javascript verison of the given exception since python's
 *             # exception would do a better job explaining the problem.
 *             if JS_HasException(self.ctx):             # <<<<<<<<<<<<<<
 *                 JS_FreeValue(self.ctx, JS_GetException(self.ctx))
 *             raise exc
*/
    __pyx_t_1 = JS_HasException(__pyx_v_self->ctx);
    if (__pyx_t_1) {

      /* "cyjs/_cyjs.pyx":840
 *             # exception would do a better job explaining the problem.
 *             if JS_HasException(self.ctx):
 *                 JS_FreeValue(self.ctx, JS_GetException(self.ctx))             # <<<<<<<<<<<<<<
 *             raise exc
 * 
*/
      JS_FreeValue(__pyx_v_self->ctx, JS_GetException(__pyx_v_self->ctx));

      /* "cyjs/_cyjs.pyx":839
 *             # Cleanup javascript verison of the given exception since python's
 *             # exception would do a better job explaining the problem.
 *             if JS_HasException(self.ctx):             # <<<<<<<<<<<<<<
 *                 JS_FreeValue(self.ctx, JS_GetException(self.ctx))
 *             raise exc
*/
    }

    /* "cyjs/_cyjs.pyx":841
 *             if JS_HasException(self.ctx):
 *                 JS_FreeValue(self.ctx, JS_GetException(self.ctx))
 *             raise exc             # <<<<<<<<<<<<<<
 * 
 *         if JS_HasException(self.ctx):
*/
    __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
    __PYX_ERR(0, 841, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":833
 *     cdef object raise_exception(self):
 *         cdef object exc
 *         if self._cb_exception is not None:             # <<<<<<<<<<<<<<
 *             # Let python exception take over first if seen. it's more accurate than the JSError would be
 *             exc = self._cb_exception
*/
  }

  /* "cyjs/_cyjs.pyx":843
 *             raise exc
 * 
 *         if JS_HasException(self.ctx):             # <<<<<<<<<<<<<<
 *             raise JSError.new(self.ctx, self.get_exception())
 * 
*/
  __pyx_t_1 = JS_HasException(__pyx_v_self->ctx);
  if (unlikely(__pyx_t_1)) {

    /* "cyjs/_cyjs.pyx":844
 * 
 *         if JS_HasException(self.ctx):
 *             raise JSError.new(self.ctx, self.get_exception())             # <<<<<<<<<<<<<<
 * 
 *     # function setup was inspired by rquickjs's EvalOptions struct
*/
    __pyx_t_3 = __pyx_f_4cyjs_5_cyjs_7Context_get_exception(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 844, __pyx_L1_error)
    __pyx_t_2 = ((PyObject *)__pyx_f_4cyjs_5_cyjs_7JSError_new(__pyx_v_self->ctx, __pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 844, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":843
 *             raise exc
 * 
 *         if JS_HasException(self.ctx):             # <<<<<<<<<<<<<<
 *             raise JSError.new(self.ctx, self.get_exception())
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":831
 * 
 *     # raises exception if one is seen.
 *     cdef object raise_exception(self):             # <<<<<<<<<<<<<<
 *         cdef object exc
 *         if self._cb_exception is not None:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cyjs._cyjs.Context.raise_exception", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":847
 * 
 *     # function setup was inspired by rquickjs's EvalOptions struct
 *     cdef object ceval(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

static PyObject *__pyx_f_4cyjs_5_cyjs_7Context_ceval(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_ceval *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":850
 *         self,
 *         object code,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         # Cython & Python have global as a keyword so a workaround was made
 *         bint module = True,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":852
 *         object filename = None,
 *         # Cython & Python have global as a keyword so a workaround was made
 *         bint module = True,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  int __pyx_v_module = ((int)1);

  /* "cyjs/_cyjs.pyx":853
 *         # Cython & Python have global as a keyword so a workaround was made
 *         bint module = True,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":854
 *         bint module = True,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":855
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         """evaluates javascript code in c"""
*/
  int __pyx_v_promise = ((int)0);
  Py_buffer __pyx_v_view;
  Py_buffer __pyx_v_fs_view;
  JSValue __pyx_v_val;
  int __pyx_v_flags;
  PyObject *__pyx_v_fs = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ceval", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_module = __pyx_optional_args->module;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_strict = __pyx_optional_args->strict;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_promise = __pyx_optional_args->promise;
            }
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":860
 *         cdef Py_buffer view, fs_view
 *         cdef JSValue val
 *         cdef int flags = 0             # <<<<<<<<<<<<<<
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
*/
  __pyx_v_flags = 0;

  /* "cyjs/_cyjs.pyx":861
 *         cdef JSValue val
 *         cdef int flags = 0
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"             # <<<<<<<<<<<<<<
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_filename); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 861, __pyx_L1_error)
  if (__pyx_t_2) {
    __pyx_t_3 = CYJS_FSConvert(__pyx_v_filename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b_input);
    __pyx_t_1 = __pyx_mstate_global->__pyx_kp_b_input;
  }
  __pyx_v_fs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":863
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_code, (&__pyx_v_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 863, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":864
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise             # <<<<<<<<<<<<<<
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 864, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":863
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  }

  /* "cyjs/_cyjs.pyx":865
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_fs, (&__pyx_v_fs_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 865, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":866
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
    cyjs_release_buffer((&__pyx_v_view));

    /* "cyjs/_cyjs.pyx":867
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 867, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":865
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  }

  /* "cyjs/_cyjs.pyx":869
 *             raise
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL             # <<<<<<<<<<<<<<
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
*/
  if (__pyx_v_module) {
    __pyx_t_4 = JS_EVAL_TYPE_MODULE;
  } else {
    __pyx_t_4 = JS_EVAL_TYPE_GLOBAL;
  }
  __pyx_v_flags = __pyx_t_4;

  /* "cyjs/_cyjs.pyx":870
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  if (__pyx_v_strict) {

    /* "cyjs/_cyjs.pyx":871
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT             # <<<<<<<<<<<<<<
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_STRICT);

    /* "cyjs/_cyjs.pyx":870
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  }

  /* "cyjs/_cyjs.pyx":872
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  if (__pyx_v_backtrace_barrier) {

    /* "cyjs/_cyjs.pyx":873
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER             # <<<<<<<<<<<<<<
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_BACKTRACE_BARRIER);

    /* "cyjs/_cyjs.pyx":872
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  }

  /* "cyjs/_cyjs.pyx":874
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  if (__pyx_v_promise) {

    /* "cyjs/_cyjs.pyx":875
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_ASYNC);

    /* "cyjs/_cyjs.pyx":874
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":878
 * 
 * 
 *         val = JS_Eval(             # <<<<<<<<<<<<<<
 *             self.ctx,
 *             <const char*>view.buf,
*/
  __pyx_v_val = JS_Eval(__pyx_v_self->ctx, ((char const *)__pyx_v_view.buf), ((size_t)__pyx_v_view.len), ((char const *)__pyx_v_fs_view.buf), __pyx_v_flags);

  /* "cyjs/_cyjs.pyx":886
 *         )
 * 
 *         cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *         cyjs_release_buffer(&fs_view)
 * 
*/
  cyjs_release_buffer((&__pyx_v_view));

  /* "cyjs/_cyjs.pyx":887
 * 
 *         cyjs_release_buffer(&view)
 *         cyjs_release_buffer(&fs_view)             # <<<<<<<<<<<<<<
 * 
 *         self.raise_exception()
*/
  cyjs_release_buffer((&__pyx_v_fs_view));

  /* "cyjs/_cyjs.pyx":889
 *         cyjs_release_buffer(&fs_view)
 * 
 *         self.raise_exception()             # <<<<<<<<<<<<<<
 * 
 *         return to_python(self.ctx, val)
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_vtab)->raise_exception(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":891
 *         self.raise_exception()
 * 
 *         return to_python(self.ctx, val)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object eval(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->ctx, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":847
 * 
 *     # function setup was inspired by rquickjs's EvalOptions struct
 *     cdef object ceval(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.Context.ceval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":893
 *         return to_python(self.ctx, val)
 * 
 *     cpdef object eval(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_3eval(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Context_eval(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":896
 *         self,
 *         object code,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":897
 *         object code,
 *         object filename = None,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":898
 *         object filename = None,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":899
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         return self.ceval(code, filename, module=False, strict=strict, backtrace_barrier=backtrace_barrier, promise=promise)
*/
  int __pyx_v_promise = ((int)0);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  struct __pyx_opt_args_4cyjs_5_cyjs_7Context_ceval __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eval", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_strict = __pyx_optional_args->strict;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_promise = __pyx_optional_args->promise;
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":893
 *         return to_python(self.ctx, val)
 * 
 *     cpdef object eval(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Context &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_eval); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Context_3eval)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_strict); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 893, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_backtrace_barrier); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 893, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_promise); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 893, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_8 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[6] = {__pyx_t_3, __pyx_v_code, __pyx_v_filename, __pyx_t_5, __pyx_t_6, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_8, (6-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 893, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":901
 *         bint promise = False
 *     ):
 *         return self.ceval(code, filename, module=False, strict=strict, backtrace_barrier=backtrace_barrier, promise=promise)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object eval_module(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_9.__pyx_n = 5;
  __pyx_t_9.filename = __pyx_v_filename;
  __pyx_t_9.module = 0;
  __pyx_t_9.strict = __pyx_v_strict;
  __pyx_t_9.backtrace_barrier = __pyx_v_backtrace_barrier;
  __pyx_t_9.promise = __pyx_v_promise;
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_vtab)->ceval(__pyx_v_self, __pyx_v_code, &__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":893
 *         return to_python(self.ctx, val)
 * 
 *     cpdef object eval(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cyjs._cyjs.Context.eval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_3eval(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Context_3eval = {"eval", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_3eval, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_3eval(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_code = 0;
  PyObject *__pyx_v_filename = 0;
  int __pyx_v_strict;
  int __pyx_v_backtrace_barrier;
  int __pyx_v_promise;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("eval (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_code,&__pyx_mstate_global->__pyx_n_u_filename,&__pyx_mstate_global->__pyx_n_u_strict,&__pyx_mstate_global->__pyx_n_u_backtrace_barrier,&__pyx_mstate_global->__pyx_n_u_promise,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 893, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 893, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 893, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 893, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 893, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 893, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "eval", 0) < (0)) __PYX_ERR(0, 893, __pyx_L3_error)

      /* "cyjs/_cyjs.pyx":896
 *         self,
 *         object code,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("eval", 0, 1, 5, i); __PYX_ERR(0, 893, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 893, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 893, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 893, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 893, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 893, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_code = values[0];
    __pyx_v_filename = values[1];
    if (values[2]) {
      __pyx_v_strict = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_strict == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 897, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":897
 *         object code,
 *         object filename = None,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
      __pyx_v_strict = ((int)0);
    }
    if (values[3]) {
      __pyx_v_backtrace_barrier = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_backtrace_barrier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 898, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":898
 *         object filename = None,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
      __pyx_v_backtrace_barrier = ((int)0);
    }
    if (values[4]) {
      __pyx_v_promise = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_promise == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 899, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":899
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         return self.ceval(code, filename, module=False, strict=strict, backtrace_barrier=backtrace_barrier, promise=promise)
*/
      __pyx_v_promise = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("eval", 0, 1, 5, __pyx_nargs); __PYX_ERR(0, 893, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Context.eval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Context_2eval(((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_v_self), __pyx_v_code, __pyx_v_filename, __pyx_v_strict, __pyx_v_backtrace_barrier, __pyx_v_promise);

  /* "cyjs/_cyjs.pyx":893
 *         return to_python(self.ctx, val)
 * 
 *     cpdef object eval(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_2eval(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_filename, int __pyx_v_strict, int __pyx_v_backtrace_barrier, int __pyx_v_promise) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eval", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 4;
  __pyx_t_2.filename = __pyx_v_filename;
  __pyx_t_2.strict = __pyx_v_strict;
  __pyx_t_2.backtrace_barrier = __pyx_v_backtrace_barrier;
  __pyx_t_2.promise = __pyx_v_promise;
  __pyx_t_1 = __pyx_vtabptr_4cyjs_5_cyjs_Context->eval(__pyx_v_self, __pyx_v_code, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Context.eval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":903
 *         return self.ceval(code, filename, module=False, strict=strict, backtrace_barrier=backtrace_barrier, promise=promise)
 * 
 *     cpdef object eval_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_5eval_module(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Context_eval_module(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_module *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":906
 *         self,
 *         object code,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":907
 *         object code,
 *         object filename = None,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":908
 *         object filename = None,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":909
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         return self.ceval(code, filename, module=True, strict=strict, backtrace_barrier=backtrace_barrier, promise=promise)
*/
  int __pyx_v_promise = ((int)0);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  struct __pyx_opt_args_4cyjs_5_cyjs_7Context_ceval __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eval_module", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_strict = __pyx_optional_args->strict;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_promise = __pyx_optional_args->promise;
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":903
 *         return self.ceval(code, filename, module=False, strict=strict, backtrace_barrier=backtrace_barrier, promise=promise)
 * 
 *     cpdef object eval_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Context &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_eval_module); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 903, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Context_5eval_module)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_strict); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 903, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_backtrace_barrier); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 903, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_promise); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 903, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_8 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[6] = {__pyx_t_3, __pyx_v_code, __pyx_v_filename, __pyx_t_5, __pyx_t_6, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_8, (6-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 903, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":911
 *         bint promise = False
 *     ):
 *         return self.ceval(code, filename, module=True, strict=strict, backtrace_barrier=backtrace_barrier, promise=promise)             # <<<<<<<<<<<<<<
 * 
 *     # TODO: Rename to global_object and make it into a @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_9.__pyx_n = 5;
  __pyx_t_9.filename = __pyx_v_filename;
  __pyx_t_9.module = 1;
  __pyx_t_9.strict = __pyx_v_strict;
  __pyx_t_9.backtrace_barrier = __pyx_v_backtrace_barrier;
  __pyx_t_9.promise = __pyx_v_promise;
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_vtab)->ceval(__pyx_v_self, __pyx_v_code, &__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":903
 *         return self.ceval(code, filename, module=False, strict=strict, backtrace_barrier=backtrace_barrier, promise=promise)
 * 
 *     cpdef object eval_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cyjs._cyjs.Context.eval_module", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_5eval_module(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Context_5eval_module = {"eval_module", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_5eval_module, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_5eval_module(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_code = 0;
  PyObject *__pyx_v_filename = 0;
  int __pyx_v_strict;
  int __pyx_v_backtrace_barrier;
  int __pyx_v_promise;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("eval_module (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_code,&__pyx_mstate_global->__pyx_n_u_filename,&__pyx_mstate_global->__pyx_n_u_strict,&__pyx_mstate_global->__pyx_n_u_backtrace_barrier,&__pyx_mstate_global->__pyx_n_u_promise,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 903, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 903, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 903, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 903, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 903, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 903, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "eval_module", 0) < (0)) __PYX_ERR(0, 903, __pyx_L3_error)

      /* "cyjs/_cyjs.pyx":906
 *         self,
 *         object code,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("eval_module", 0, 1, 5, i); __PYX_ERR(0, 903, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 903, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 903, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 903, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 903, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 903, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_code = values[0];
    __pyx_v_filename = values[1];
    if (values[2]) {
      __pyx_v_strict = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_strict == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 907, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":907
 *         object code,
 *         object filename = None,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
      __pyx_v_strict = ((int)0);
    }
    if (values[3]) {
      __pyx_v_backtrace_barrier = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_backtrace_barrier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 908, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":908
 *         object filename = None,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
      __pyx_v_backtrace_barrier = ((int)0);
    }
    if (values[4]) {
      __pyx_v_promise = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_promise == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 909, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":909
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         return self.ceval(code, filename, module=True, strict=strict, backtrace_barrier=backtrace_barrier, promise=promise)
*/
      __pyx_v_promise = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("eval_module", 0, 1, 5, __pyx_nargs); __PYX_ERR(0, 903, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Context.eval_module", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Context_4eval_module(((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_v_self), __pyx_v_code, __pyx_v_filename, __pyx_v_strict, __pyx_v_backtrace_barrier, __pyx_v_promise);

  /* "cyjs/_cyjs.pyx":903
 *         return self.ceval(code, filename, module=False, strict=strict, backtrace_barrier=backtrace_barrier, promise=promise)
 * 
 *     cpdef object eval_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_4eval_module(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_filename, int __pyx_v_strict, int __pyx_v_backtrace_barrier, int __pyx_v_promise) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_module __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eval_module", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 4;
  __pyx_t_2.filename = __pyx_v_filename;
  __pyx_t_2.strict = __pyx_v_strict;
  __pyx_t_2.backtrace_barrier = __pyx_v_backtrace_barrier;
  __pyx_t_2.promise = __pyx_v_promise;
  __pyx_t_1 = __pyx_vtabptr_4cyjs_5_cyjs_Context->eval_module(__pyx_v_self, __pyx_v_code, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Context.eval_module", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":916
 *     # Otherwise Run JS_GetGlobalObject immediately for duping
 *     # and free on __dealloc__ instead.
 *     def get_global(self):             # <<<<<<<<<<<<<<
 *         return to_python(self.ctx, JS_GetGlobalObject(self.ctx))
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_7get_global(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Context_7get_global = {"get_global", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_7get_global, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_7get_global(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_global (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("get_global", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("get_global", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Context_6get_global(((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_6get_global(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_global", 0);

  /* "cyjs/_cyjs.pyx":917
 *     # and free on __dealloc__ instead.
 *     def get_global(self):
 *         return to_python(self.ctx, JS_GetGlobalObject(self.ctx))             # <<<<<<<<<<<<<<
 * 
 *     # also inspired from seeing rquickjs (rust) and retired python-quickjs
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->ctx, JS_GetGlobalObject(__pyx_v_self->ctx)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":916
 *     # Otherwise Run JS_GetGlobalObject immediately for duping
 *     # and free on __dealloc__ instead.
 *     def get_global(self):             # <<<<<<<<<<<<<<
 *         return to_python(self.ctx, JS_GetGlobalObject(self.ctx))
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Context.get_global", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":920
 * 
 *     # also inspired from seeing rquickjs (rust) and retired python-quickjs
 *     def json_parse(self, object json):             # <<<<<<<<<<<<<<
 *         cdef JSValue ret
 *         cdef Py_buffer view
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_9json_parse(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Context_9json_parse = {"json_parse", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_9json_parse, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_9json_parse(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_json = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("json_parse (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_json,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 920, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 920, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "json_parse", 0) < (0)) __PYX_ERR(0, 920, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("json_parse", 1, 1, 1, i); __PYX_ERR(0, 920, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 920, __pyx_L3_error)
    }
    __pyx_v_json = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("json_parse", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 920, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Context.json_parse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Context_8json_parse(((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_v_self), __pyx_v_json);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_8json_parse(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_json) {
  JSValue __pyx_v_ret;
  Py_buffer __pyx_v_view;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("json_parse", 0);

  /* "cyjs/_cyjs.pyx":923
 *         cdef JSValue ret
 *         cdef Py_buffer view
 *         if cyjs_get_buffer(json, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         ret = JS_ParseJSON(self.ctx, <const char*>view.buf, view.len, b"<input>")
*/
  __pyx_t_1 = cyjs_get_buffer(__pyx_v_json, (&__pyx_v_view)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 923, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":924
 *         cdef Py_buffer view
 *         if cyjs_get_buffer(json, &view) < 0:
 *             raise             # <<<<<<<<<<<<<<
 *         ret = JS_ParseJSON(self.ctx, <const char*>view.buf, view.len, b"<input>")
 *         cyjs_release_buffer(&view)
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 924, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":923
 *         cdef JSValue ret
 *         cdef Py_buffer view
 *         if cyjs_get_buffer(json, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         ret = JS_ParseJSON(self.ctx, <const char*>view.buf, view.len, b"<input>")
*/
  }

  /* "cyjs/_cyjs.pyx":925
 *         if cyjs_get_buffer(json, &view) < 0:
 *             raise
 *         ret = JS_ParseJSON(self.ctx, <const char*>view.buf, view.len, b"<input>")             # <<<<<<<<<<<<<<
 *         cyjs_release_buffer(&view)
 *         # it will be transformed to a Object as needed instead of typical python
*/
  __pyx_v_ret = JS_ParseJSON(__pyx_v_self->ctx, ((char const *)__pyx_v_view.buf), __pyx_v_view.len, __pyx_k_input);

  /* "cyjs/_cyjs.pyx":926
 *             raise
 *         ret = JS_ParseJSON(self.ctx, <const char*>view.buf, view.len, b"<input>")
 *         cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *         # it will be transformed to a Object as needed instead of typical python
 *         return to_python(self.ctx, ret)
*/
  cyjs_release_buffer((&__pyx_v_view));

  /* "cyjs/_cyjs.pyx":928
 *         cyjs_release_buffer(&view)
 *         # it will be transformed to a Object as needed instead of typical python
 *         return to_python(self.ctx, ret)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->ctx, __pyx_v_ret); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":920
 * 
 *     # also inspired from seeing rquickjs (rust) and retired python-quickjs
 *     def json_parse(self, object json):             # <<<<<<<<<<<<<<
 *         cdef JSValue ret
 *         cdef Py_buffer view
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.Context.json_parse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":936
 *     # pretty much supporting
 *     # def get(self, name: bytes | bytearray | array.array | memoryview | str)
 *     def get(self, object name):             # <<<<<<<<<<<<<<
 *         """Implements a Shortcut for converting a global object to a python object and setting a value to utilize
 *         off of."""
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_11get(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4cyjs_5_cyjs_7Context_10get, "Implements a Shortcut for converting a global object to a python object and setting a value to utilize\n        off of.");
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Context_11get = {"get", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_11get, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4cyjs_5_cyjs_7Context_10get};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_11get(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_name = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_name_2,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 936, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 936, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get", 0) < (0)) __PYX_ERR(0, 936, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get", 1, 1, 1, i); __PYX_ERR(0, 936, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 936, __pyx_L3_error)
    }
    __pyx_v_name = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 936, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Context.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Context_10get(((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_10get(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_name) {
  JSValue __pyx_v_glob;
  JSValue __pyx_v_val;
  JSAtom __pyx_v_atom;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 0);

  /* "cyjs/_cyjs.pyx":941
 *         cdef JSValue glob, val
 *         cdef JSAtom atom
 *         glob = JS_GetGlobalObject(self.ctx)             # <<<<<<<<<<<<<<
 *         if py_to_atom(self.ctx, name, &atom) < 0:
 *             raise
*/
  __pyx_v_glob = JS_GetGlobalObject(__pyx_v_self->ctx);

  /* "cyjs/_cyjs.pyx":942
 *         cdef JSAtom atom
 *         glob = JS_GetGlobalObject(self.ctx)
 *         if py_to_atom(self.ctx, name, &atom) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         val = JS_GetProperty(self.ctx, glob, atom)
*/
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_py_to_atom(__pyx_v_self->ctx, __pyx_v_name, (&__pyx_v_atom)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 942, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":943
 *         glob = JS_GetGlobalObject(self.ctx)
 *         if py_to_atom(self.ctx, name, &atom) < 0:
 *             raise             # <<<<<<<<<<<<<<
 *         val = JS_GetProperty(self.ctx, glob, atom)
 *         JS_FreeValue(self.ctx, glob)
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 943, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":942
 *         cdef JSAtom atom
 *         glob = JS_GetGlobalObject(self.ctx)
 *         if py_to_atom(self.ctx, name, &atom) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         val = JS_GetProperty(self.ctx, glob, atom)
*/
  }

  /* "cyjs/_cyjs.pyx":944
 *         if py_to_atom(self.ctx, name, &atom) < 0:
 *             raise
 *         val = JS_GetProperty(self.ctx, glob, atom)             # <<<<<<<<<<<<<<
 *         JS_FreeValue(self.ctx, glob)
 *         return to_python(self.ctx, val)
*/
  __pyx_v_val = JS_GetProperty(__pyx_v_self->ctx, __pyx_v_glob, __pyx_v_atom);

  /* "cyjs/_cyjs.pyx":945
 *             raise
 *         val = JS_GetProperty(self.ctx, glob, atom)
 *         JS_FreeValue(self.ctx, glob)             # <<<<<<<<<<<<<<
 *         return to_python(self.ctx, val)
 * 
*/
  JS_FreeValue(__pyx_v_self->ctx, __pyx_v_glob);

  /* "cyjs/_cyjs.pyx":946
 *         val = JS_GetProperty(self.ctx, glob, atom)
 *         JS_FreeValue(self.ctx, glob)
 *         return to_python(self.ctx, val)             # <<<<<<<<<<<<<<
 * 
 *     def set(self, object name, object item):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->ctx, __pyx_v_val); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":936
 *     # pretty much supporting
 *     # def get(self, name: bytes | bytearray | array.array | memoryview | str)
 *     def get(self, object name):             # <<<<<<<<<<<<<<
 *         """Implements a Shortcut for converting a global object to a python object and setting a value to utilize
 *         off of."""
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.Context.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":948
 *         return to_python(self.ctx, val)
 * 
 *     def set(self, object name, object item):             # <<<<<<<<<<<<<<
 *         """Sets an item to the current globalThis object"""
 *         cdef JSValue value
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_13set(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4cyjs_5_cyjs_7Context_12set, "Sets an item to the current globalThis object");
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Context_13set = {"set", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_13set, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4cyjs_5_cyjs_7Context_12set};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_13set(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_item = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_name_2,&__pyx_mstate_global->__pyx_n_u_item,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 948, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 948, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 948, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set", 0) < (0)) __PYX_ERR(0, 948, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set", 1, 2, 2, i); __PYX_ERR(0, 948, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 948, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 948, __pyx_L3_error)
    }
    __pyx_v_name = values[0];
    __pyx_v_item = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 948, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Context.set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Context_12set(((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_v_self), __pyx_v_name, __pyx_v_item);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_12set(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_item) {
  JSValue __pyx_v_value;
  JSAtom __pyx_v_atom;
  JSValue __pyx_v_glob;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  char const *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set", 0);

  /* "cyjs/_cyjs.pyx":954
 *         cdef JSValue glob
 * 
 *         if py_to_atom(self.ctx, name, &atom) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if to_quickjs(self.ctx, &value, item) < 0:
*/
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_py_to_atom(__pyx_v_self->ctx, __pyx_v_name, (&__pyx_v_atom)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 954, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":955
 * 
 *         if py_to_atom(self.ctx, name, &atom) < 0:
 *             raise             # <<<<<<<<<<<<<<
 *         if to_quickjs(self.ctx, &value, item) < 0:
 *             raise
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 955, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":954
 *         cdef JSValue glob
 * 
 *         if py_to_atom(self.ctx, name, &atom) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if to_quickjs(self.ctx, &value, item) < 0:
*/
  }

  /* "cyjs/_cyjs.pyx":956
 *         if py_to_atom(self.ctx, name, &atom) < 0:
 *             raise
 *         if to_quickjs(self.ctx, &value, item) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         glob = JS_GetGlobalObject(self.ctx)
*/
  __pyx_t_2 = (__pyx_fuse_8__pyx_f_4cyjs_5_cyjs_to_quickjs(__pyx_v_self->ctx, (&__pyx_v_value), __pyx_v_item) < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":957
 *             raise
 *         if to_quickjs(self.ctx, &value, item) < 0:
 *             raise             # <<<<<<<<<<<<<<
 *         glob = JS_GetGlobalObject(self.ctx)
 *         try:
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 957, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":956
 *         if py_to_atom(self.ctx, name, &atom) < 0:
 *             raise
 *         if to_quickjs(self.ctx, &value, item) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         glob = JS_GetGlobalObject(self.ctx)
*/
  }

  /* "cyjs/_cyjs.pyx":958
 *         if to_quickjs(self.ctx, &value, item) < 0:
 *             raise
 *         glob = JS_GetGlobalObject(self.ctx)             # <<<<<<<<<<<<<<
 *         try:
 *             if JS_SetProperty(self.ctx, glob, atom, value) < 0:
*/
  __pyx_v_glob = JS_GetGlobalObject(__pyx_v_self->ctx);

  /* "cyjs/_cyjs.pyx":959
 *             raise
 *         glob = JS_GetGlobalObject(self.ctx)
 *         try:             # <<<<<<<<<<<<<<
 *             if JS_SetProperty(self.ctx, glob, atom, value) < 0:
 *                 raise TypeError("Failed setting the variable.")
*/
  /*try:*/ {

    /* "cyjs/_cyjs.pyx":960
 *         glob = JS_GetGlobalObject(self.ctx)
 *         try:
 *             if JS_SetProperty(self.ctx, glob, atom, value) < 0:             # <<<<<<<<<<<<<<
 *                 raise TypeError("Failed setting the variable.")
 *         finally:
*/
    __pyx_t_2 = (JS_SetProperty(__pyx_v_self->ctx, __pyx_v_glob, __pyx_v_atom, __pyx_v_value) < 0);
    if (unlikely(__pyx_t_2)) {

      /* "cyjs/_cyjs.pyx":961
 *         try:
 *             if JS_SetProperty(self.ctx, glob, atom, value) < 0:
 *                 raise TypeError("Failed setting the variable.")             # <<<<<<<<<<<<<<
 *         finally:
 *             JS_FreeValue(self.ctx, glob)
*/
      __pyx_t_4 = NULL;
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Failed_setting_the_variable};
        __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 961, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 961, __pyx_L6_error)

      /* "cyjs/_cyjs.pyx":960
 *         glob = JS_GetGlobalObject(self.ctx)
 *         try:
 *             if JS_SetProperty(self.ctx, glob, atom, value) < 0:             # <<<<<<<<<<<<<<
 *                 raise TypeError("Failed setting the variable.")
 *         finally:
*/
    }
  }

  /* "cyjs/_cyjs.pyx":963
 *                 raise TypeError("Failed setting the variable.")
 *         finally:
 *             JS_FreeValue(self.ctx, glob)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  /*finally:*/ {
    /*normal exit:*/{
      JS_FreeValue(__pyx_v_self->ctx, __pyx_v_glob);
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
       __Pyx_ExceptionSwap(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      if ( unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __pyx_t_1 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_7 = __pyx_filename;
      {
        JS_FreeValue(__pyx_v_self->ctx, __pyx_v_glob);
      }
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ErrRestore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __pyx_lineno = __pyx_t_1; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_7;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "cyjs/_cyjs.pyx":948
 *         return to_python(self.ctx, val)
 * 
 *     def set(self, object name, object item):             # <<<<<<<<<<<<<<
 *         """Sets an item to the current globalThis object"""
 *         cdef JSValue value
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Context.set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":966
 * 
 * 
 *     cpdef JSFunction add_function(             # <<<<<<<<<<<<<<
 *         self,
 *         object func,
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_15add_function(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_f_4cyjs_5_cyjs_7Context_add_function(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_func, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_add_function *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":969
 *         self,
 *         object func,
 *         object name = None,             # <<<<<<<<<<<<<<
 *         # incase python function contains dynamic amounts of arguments,
 *         # for example: "def func(*args):..."
*/
  PyObject *__pyx_v_name = ((PyObject *)Py_None);
  int __pyx_v_count = ((int)-1);
  int __pyx_v_magic = ((int)0);
  PyObject *__pyx_v__name = 0;
  Py_buffer __pyx_v_view;
  int __pyx_v_argc;
  JSValue __pyx_v_closure;
  struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_js_func = 0;
  struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_function", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_name = __pyx_optional_args->name;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_count = __pyx_optional_args->count;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_magic = __pyx_optional_args->magic;
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":966
 * 
 * 
 *     cpdef JSFunction add_function(             # <<<<<<<<<<<<<<
 *         self,
 *         object func,
*/
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Context &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_add_function); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Context_15add_function)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = __Pyx_PyLong_From_int(__pyx_v_count); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 966, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyLong_From_int(__pyx_v_magic); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 966, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_7 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[5] = {__pyx_t_3, __pyx_v_func, __pyx_v_name, __pyx_t_5, __pyx_t_6};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_7, (5-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 966, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_JSFunction))))) __PYX_ERR(0, 966, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":998
 *         cdef JSValue closure
 *         cdef JSFunction js_func
 *         if not callable(func):             # <<<<<<<<<<<<<<
 *             raise TypeError("function must be callable.")
 * 
*/
  __pyx_t_8 = __Pyx_PyCallable_Check(__pyx_v_func); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 998, __pyx_L1_error)
  __pyx_t_9 = (!__pyx_t_8);
  if (unlikely(__pyx_t_9)) {

    /* "cyjs/_cyjs.pyx":999
 *         cdef JSFunction js_func
 *         if not callable(func):
 *             raise TypeError("function must be callable.")             # <<<<<<<<<<<<<<
 * 
 *         _name = name or func.__name__
*/
    __pyx_t_2 = NULL;
    __pyx_t_7 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_function_must_be_callable};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 999, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 999, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":998
 *         cdef JSValue closure
 *         cdef JSFunction js_func
 *         if not callable(func):             # <<<<<<<<<<<<<<
 *             raise TypeError("function must be callable.")
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1001
 *             raise TypeError("function must be callable.")
 * 
 *         _name = name or func.__name__             # <<<<<<<<<<<<<<
 * 
 *         # Were bindly guessing but in the future, a smarter technique should be gone about
*/
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_name); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 1001, __pyx_L1_error)
  if (!__pyx_t_9) {
  } else {
    __Pyx_INCREF(__pyx_v_name);
    __pyx_t_1 = __pyx_v_name;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_mstate_global->__pyx_n_u_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_L4_bool_binop_done:;
  __pyx_v__name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1006
 *         # Otherwise I'll expand this number to the very maximum number of arguments
 *         # as python can accept dynamic numbers without question at times...
 *         argc = <int>len(func.__annotations__) if (count < 0) else count             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_9 = (__pyx_v_count < 0);
  if (__pyx_t_9) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_mstate_global->__pyx_n_u_annotations); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1006, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_10 = ((int)__pyx_t_11);
  } else {
    __pyx_t_10 = __pyx_v_count;
  }
  __pyx_v_argc = __pyx_t_10;

  /* "cyjs/_cyjs.pyx":1009
 * 
 * 
 *         if cyjs_get_buffer(_name, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         try:
*/
  __pyx_t_10 = cyjs_get_buffer(__pyx_v__name, (&__pyx_v_view)); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1009, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_10 < 0);
  if (unlikely(__pyx_t_9)) {

    /* "cyjs/_cyjs.pyx":1010
 * 
 *         if cyjs_get_buffer(_name, &view) < 0:
 *             raise             # <<<<<<<<<<<<<<
 *         try:
 * 
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1010, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1009
 * 
 * 
 *         if cyjs_get_buffer(_name, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         try:
*/
  }

  /* "cyjs/_cyjs.pyx":1011
 *         if cyjs_get_buffer(_name, &view) < 0:
 *             raise
 *         try:             # <<<<<<<<<<<<<<
 * 
 *             js_func = JSFunction.new(self, func)
*/
  /*try:*/ {

    /* "cyjs/_cyjs.pyx":1013
 *         try:
 * 
 *             js_func = JSFunction.new(self, func)             # <<<<<<<<<<<<<<
 *             closure = JS_NewCClosure(self.ctx, on_cclosure_callback, <const char*>view.buf, on_cclosure_opaque_finalize, argc, magic, <void*>js_func)
 *             if JS_IsException(closure):
*/
    __pyx_t_1 = ((PyObject *)__pyx_f_4cyjs_5_cyjs_10JSFunction_new(__pyx_v_self, __pyx_v_func)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1013, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_js_func = ((struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "cyjs/_cyjs.pyx":1014
 * 
 *             js_func = JSFunction.new(self, func)
 *             closure = JS_NewCClosure(self.ctx, on_cclosure_callback, <const char*>view.buf, on_cclosure_opaque_finalize, argc, magic, <void*>js_func)             # <<<<<<<<<<<<<<
 *             if JS_IsException(closure):
 *                 self.raise_exception()
*/
    __pyx_v_closure = JS_NewCClosure(__pyx_v_self->ctx, __pyx_f_4cyjs_5_cyjs_on_cclosure_callback, ((char const *)__pyx_v_view.buf), __pyx_f_4cyjs_5_cyjs_on_cclosure_opaque_finalize, __pyx_v_argc, __pyx_v_magic, ((void *)__pyx_v_js_func));

    /* "cyjs/_cyjs.pyx":1015
 *             js_func = JSFunction.new(self, func)
 *             closure = JS_NewCClosure(self.ctx, on_cclosure_callback, <const char*>view.buf, on_cclosure_opaque_finalize, argc, magic, <void*>js_func)
 *             if JS_IsException(closure):             # <<<<<<<<<<<<<<
 *                 self.raise_exception()
 *                 raise
*/
    __pyx_t_9 = JS_IsException(__pyx_v_closure);
    if (unlikely(__pyx_t_9)) {

      /* "cyjs/_cyjs.pyx":1016
 *             closure = JS_NewCClosure(self.ctx, on_cclosure_callback, <const char*>view.buf, on_cclosure_opaque_finalize, argc, magic, <void*>js_func)
 *             if JS_IsException(closure):
 *                 self.raise_exception()             # <<<<<<<<<<<<<<
 *                 raise
 *             # Have CClosure hold onto it's own reference of js_func if all was successfully implemented
*/
      __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_vtab)->raise_exception(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1016, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "cyjs/_cyjs.pyx":1017
 *             if JS_IsException(closure):
 *                 self.raise_exception()
 *                 raise             # <<<<<<<<<<<<<<
 *             # Have CClosure hold onto it's own reference of js_func if all was successfully implemented
 *             Py_XINCREF(js_func)
*/
      __Pyx_ReraiseException();
      __PYX_ERR(0, 1017, __pyx_L8_error)

      /* "cyjs/_cyjs.pyx":1015
 *             js_func = JSFunction.new(self, func)
 *             closure = JS_NewCClosure(self.ctx, on_cclosure_callback, <const char*>view.buf, on_cclosure_opaque_finalize, argc, magic, <void*>js_func)
 *             if JS_IsException(closure):             # <<<<<<<<<<<<<<
 *                 self.raise_exception()
 *                 raise
*/
    }

    /* "cyjs/_cyjs.pyx":1019
 *                 raise
 *             # Have CClosure hold onto it's own reference of js_func if all was successfully implemented
 *             Py_XINCREF(js_func)             # <<<<<<<<<<<<<<
 *             js_func.value = closure
 *             return js_func
*/
    Py_XINCREF(((PyObject *)__pyx_v_js_func));

    /* "cyjs/_cyjs.pyx":1020
 *             # Have CClosure hold onto it's own reference of js_func if all was successfully implemented
 *             Py_XINCREF(js_func)
 *             js_func.value = closure             # <<<<<<<<<<<<<<
 *             return js_func
 * 
*/
    __pyx_v_js_func->value = __pyx_v_closure;

    /* "cyjs/_cyjs.pyx":1021
 *             Py_XINCREF(js_func)
 *             js_func.value = closure
 *             return js_func             # <<<<<<<<<<<<<<
 * 
 *         finally:
*/
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_js_func);
    __pyx_r = __pyx_v_js_func;
    goto __pyx_L7_return;
  }

  /* "cyjs/_cyjs.pyx":1024
 * 
 *         finally:
 *             cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 * 
 *     cdef object ceval_this(
*/
  /*finally:*/ {
    __pyx_L8_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
       __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
      if ( unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_10 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
      {
        cyjs_release_buffer((&__pyx_v_view));
      }
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_XGIVEREF(__pyx_t_18);
      __Pyx_XGIVEREF(__pyx_t_19);
      __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
      __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
      goto __pyx_L1_error;
    }
    __pyx_L7_return: {
      __pyx_t_20 = __pyx_r;
      __pyx_r = 0;
      cyjs_release_buffer((&__pyx_v_view));
      __pyx_r = __pyx_t_20;
      __pyx_t_20 = 0;
      goto __pyx_L0;
    }
  }

  /* "cyjs/_cyjs.pyx":966
 * 
 * 
 *     cpdef JSFunction add_function(             # <<<<<<<<<<<<<<
 *         self,
 *         object func,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cyjs._cyjs.Context.add_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__name);
  __Pyx_XDECREF((PyObject *)__pyx_v_js_func);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_15add_function(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4cyjs_5_cyjs_7Context_14add_function, "adds a python function to quickjs using \n        `JS_NewCClosure` since Quickjs-ng doesn't have \n        a good way for bidning python fucntions well yet\n        \n        :param func: the python function to invoke with \n            quickjs note: that it may not pass along keyword arguments `**kw`\n        :param name: an alternate name to provide the given function\n            defaults to obtaining function's name from func.__name__\n        :param count: a number of arguments to provide, if count is less than\n            zero (0), then functions's __annotations__ will be used to figure that out\n                just know that it may fail at inspection since the inspect library \n                is not utilized for the sake of speed.\n        :param magic:\n            this value is not passed to the python function and is set to zero\n            unless internally used with QuickJS this value might be deprecated in the future\n        \n        ");
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Context_15add_function = {"add_function", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_15add_function, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4cyjs_5_cyjs_7Context_14add_function};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_15add_function(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_name = 0;
  int __pyx_v_count;
  int __pyx_v_magic;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_function (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_func,&__pyx_mstate_global->__pyx_n_u_name_2,&__pyx_mstate_global->__pyx_n_u_count,&__pyx_mstate_global->__pyx_n_u_magic,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 966, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 966, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 966, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 966, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 966, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "add_function", 0) < (0)) __PYX_ERR(0, 966, __pyx_L3_error)

      /* "cyjs/_cyjs.pyx":969
 *         self,
 *         object func,
 *         object name = None,             # <<<<<<<<<<<<<<
 *         # incase python function contains dynamic amounts of arguments,
 *         # for example: "def func(*args):..."
*/
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("add_function", 0, 1, 4, i); __PYX_ERR(0, 966, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 966, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 966, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 966, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 966, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_func = values[0];
    __pyx_v_name = values[1];
    if (values[2]) {
      __pyx_v_count = __Pyx_PyLong_As_int(values[2]); if (unlikely((__pyx_v_count == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 972, __pyx_L3_error)
    } else {
      __pyx_v_count = ((int)-1);
    }
    if (values[3]) {
      __pyx_v_magic = __Pyx_PyLong_As_int(values[3]); if (unlikely((__pyx_v_magic == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 974, __pyx_L3_error)
    } else {
      __pyx_v_magic = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_function", 0, 1, 4, __pyx_nargs); __PYX_ERR(0, 966, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Context.add_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Context_14add_function(((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_v_self), __pyx_v_func, __pyx_v_name, __pyx_v_count, __pyx_v_magic);

  /* "cyjs/_cyjs.pyx":966
 * 
 * 
 *     cpdef JSFunction add_function(             # <<<<<<<<<<<<<<
 *         self,
 *         object func,
*/

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_14add_function(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_name, int __pyx_v_count, int __pyx_v_magic) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4cyjs_5_cyjs_7Context_add_function __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_function", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 3;
  __pyx_t_2.name = __pyx_v_name;
  __pyx_t_2.count = __pyx_v_count;
  __pyx_t_2.magic = __pyx_v_magic;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_4cyjs_5_cyjs_Context->add_function(__pyx_v_self, __pyx_v_func, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Context.add_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1026
 *             cyjs_release_buffer(&view)
 * 
 *     cdef object ceval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

static PyObject *__pyx_fuse_0__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, struct __pyx_obj_4cyjs_5_cyjs_JSFunction *__pyx_v_this, struct __pyx_fuse_0__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":1030
 *         object code,
 *         quickjs_type_t this,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint module = False,
 *         bint strict = False,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":1031
 *         quickjs_type_t this,
 *         object filename = None,
 *         bint module = False,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  int __pyx_v_module = ((int)0);

  /* "cyjs/_cyjs.pyx":1032
 *         object filename = None,
 *         bint module = False,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":1033
 *         bint module = False,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":1034
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         """evaluates javascript code in from another object"""
*/
  int __pyx_v_promise = ((int)0);
  Py_buffer __pyx_v_view;
  Py_buffer __pyx_v_fs_view;
  JSValue __pyx_v_val;
  JSValue __pyx_v_this_val;
  int __pyx_v_flags;
  PyObject *__pyx_v_fs = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_0ceval_this", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_module = __pyx_optional_args->module;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_strict = __pyx_optional_args->strict;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_promise = __pyx_optional_args->promise;
            }
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":1039
 *         cdef Py_buffer view, fs_view
 *         cdef JSValue val, this_val
 *         cdef int flags = 0             # <<<<<<<<<<<<<<
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
*/
  __pyx_v_flags = 0;

  /* "cyjs/_cyjs.pyx":1040
 *         cdef JSValue val, this_val
 *         cdef int flags = 0
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"             # <<<<<<<<<<<<<<
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_filename); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1040, __pyx_L1_error)
  if (__pyx_t_2) {
    __pyx_t_3 = CYJS_FSConvert(__pyx_v_filename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b_input);
    __pyx_t_1 = __pyx_mstate_global->__pyx_kp_b_input;
  }
  __pyx_v_fs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_code, (&__pyx_v_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1042, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1043
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise             # <<<<<<<<<<<<<<
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1043, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  }

  /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_fs, (&__pyx_v_fs_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1044, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1045
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
    cyjs_release_buffer((&__pyx_v_view));

    /* "cyjs/_cyjs.pyx":1046
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1046, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  }

  /* "cyjs/_cyjs.pyx":1048
 *             raise
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL             # <<<<<<<<<<<<<<
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
*/
  if (__pyx_v_module) {
    __pyx_t_4 = JS_EVAL_TYPE_MODULE;
  } else {
    __pyx_t_4 = JS_EVAL_TYPE_GLOBAL;
  }
  __pyx_v_flags = __pyx_t_4;

  /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  if (__pyx_v_strict) {

    /* "cyjs/_cyjs.pyx":1050
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT             # <<<<<<<<<<<<<<
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_STRICT);

    /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  }

  /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  if (__pyx_v_backtrace_barrier) {

    /* "cyjs/_cyjs.pyx":1052
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER             # <<<<<<<<<<<<<<
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_BACKTRACE_BARRIER);

    /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  }

  /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  if (__pyx_v_promise) {

    /* "cyjs/_cyjs.pyx":1054
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC             # <<<<<<<<<<<<<<
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_ASYNC);

    /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_2 = (__pyx_fuse_0__pyx_f_4cyjs_5_cyjs_to_quickjs(__pyx_v_self->ctx, (&__pyx_v_this_val), __pyx_v_this) < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1057
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         val = JS_EvalThis(
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1057, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1059
 *             raise
 * 
 *         val = JS_EvalThis(             # <<<<<<<<<<<<<<
 *             self.ctx,
 *             this_val,
*/
  __pyx_v_val = JS_EvalThis(__pyx_v_self->ctx, __pyx_v_this_val, ((char const *)__pyx_v_view.buf), ((size_t)__pyx_v_view.len), ((char const *)__pyx_v_fs_view.buf), __pyx_v_flags);

  /* "cyjs/_cyjs.pyx":1068
 *         )
 * 
 *         cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *         cyjs_release_buffer(&fs_view)
 * 
*/
  cyjs_release_buffer((&__pyx_v_view));

  /* "cyjs/_cyjs.pyx":1069
 * 
 *         cyjs_release_buffer(&view)
 *         cyjs_release_buffer(&fs_view)             # <<<<<<<<<<<<<<
 * 
 *         self.raise_exception()
*/
  cyjs_release_buffer((&__pyx_v_fs_view));

  /* "cyjs/_cyjs.pyx":1071
 *         cyjs_release_buffer(&fs_view)
 * 
 *         self.raise_exception()             # <<<<<<<<<<<<<<
 * 
 *         return to_python(self.ctx, val)
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_vtab)->raise_exception(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1073
 *         self.raise_exception()
 * 
 *         return to_python(self.ctx, val)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object eval_this(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->ctx, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1026
 *             cyjs_release_buffer(&view)
 * 
 *     cdef object ceval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.Context.ceval_this", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_fuse_1__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_this, struct __pyx_fuse_1__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":1030
 *         object code,
 *         quickjs_type_t this,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint module = False,
 *         bint strict = False,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":1031
 *         quickjs_type_t this,
 *         object filename = None,
 *         bint module = False,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  int __pyx_v_module = ((int)0);

  /* "cyjs/_cyjs.pyx":1032
 *         object filename = None,
 *         bint module = False,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":1033
 *         bint module = False,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":1034
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         """evaluates javascript code in from another object"""
*/
  int __pyx_v_promise = ((int)0);
  Py_buffer __pyx_v_view;
  Py_buffer __pyx_v_fs_view;
  JSValue __pyx_v_val;
  JSValue __pyx_v_this_val;
  int __pyx_v_flags;
  PyObject *__pyx_v_fs = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_1ceval_this", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_module = __pyx_optional_args->module;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_strict = __pyx_optional_args->strict;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_promise = __pyx_optional_args->promise;
            }
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":1039
 *         cdef Py_buffer view, fs_view
 *         cdef JSValue val, this_val
 *         cdef int flags = 0             # <<<<<<<<<<<<<<
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
*/
  __pyx_v_flags = 0;

  /* "cyjs/_cyjs.pyx":1040
 *         cdef JSValue val, this_val
 *         cdef int flags = 0
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"             # <<<<<<<<<<<<<<
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_filename); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1040, __pyx_L1_error)
  if (__pyx_t_2) {
    __pyx_t_3 = CYJS_FSConvert(__pyx_v_filename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b_input);
    __pyx_t_1 = __pyx_mstate_global->__pyx_kp_b_input;
  }
  __pyx_v_fs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_code, (&__pyx_v_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1042, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1043
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise             # <<<<<<<<<<<<<<
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1043, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  }

  /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_fs, (&__pyx_v_fs_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1044, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1045
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
    cyjs_release_buffer((&__pyx_v_view));

    /* "cyjs/_cyjs.pyx":1046
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1046, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  }

  /* "cyjs/_cyjs.pyx":1048
 *             raise
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL             # <<<<<<<<<<<<<<
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
*/
  if (__pyx_v_module) {
    __pyx_t_4 = JS_EVAL_TYPE_MODULE;
  } else {
    __pyx_t_4 = JS_EVAL_TYPE_GLOBAL;
  }
  __pyx_v_flags = __pyx_t_4;

  /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  if (__pyx_v_strict) {

    /* "cyjs/_cyjs.pyx":1050
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT             # <<<<<<<<<<<<<<
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_STRICT);

    /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  }

  /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  if (__pyx_v_backtrace_barrier) {

    /* "cyjs/_cyjs.pyx":1052
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER             # <<<<<<<<<<<<<<
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_BACKTRACE_BARRIER);

    /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  }

  /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  if (__pyx_v_promise) {

    /* "cyjs/_cyjs.pyx":1054
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC             # <<<<<<<<<<<<<<
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_ASYNC);

    /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_2 = (__pyx_fuse_1__pyx_f_4cyjs_5_cyjs_to_quickjs(__pyx_v_self->ctx, (&__pyx_v_this_val), __pyx_v_this) < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1057
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         val = JS_EvalThis(
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1057, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1059
 *             raise
 * 
 *         val = JS_EvalThis(             # <<<<<<<<<<<<<<
 *             self.ctx,
 *             this_val,
*/
  __pyx_v_val = JS_EvalThis(__pyx_v_self->ctx, __pyx_v_this_val, ((char const *)__pyx_v_view.buf), ((size_t)__pyx_v_view.len), ((char const *)__pyx_v_fs_view.buf), __pyx_v_flags);

  /* "cyjs/_cyjs.pyx":1068
 *         )
 * 
 *         cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *         cyjs_release_buffer(&fs_view)
 * 
*/
  cyjs_release_buffer((&__pyx_v_view));

  /* "cyjs/_cyjs.pyx":1069
 * 
 *         cyjs_release_buffer(&view)
 *         cyjs_release_buffer(&fs_view)             # <<<<<<<<<<<<<<
 * 
 *         self.raise_exception()
*/
  cyjs_release_buffer((&__pyx_v_fs_view));

  /* "cyjs/_cyjs.pyx":1071
 *         cyjs_release_buffer(&fs_view)
 * 
 *         self.raise_exception()             # <<<<<<<<<<<<<<
 * 
 *         return to_python(self.ctx, val)
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_vtab)->raise_exception(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1073
 *         self.raise_exception()
 * 
 *         return to_python(self.ctx, val)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object eval_this(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->ctx, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1026
 *             cyjs_release_buffer(&view)
 * 
 *     cdef object ceval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.Context.ceval_this", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_fuse_2__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, struct __pyx_fuse_2__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":1030
 *         object code,
 *         quickjs_type_t this,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint module = False,
 *         bint strict = False,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":1031
 *         quickjs_type_t this,
 *         object filename = None,
 *         bint module = False,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  int __pyx_v_module = ((int)0);

  /* "cyjs/_cyjs.pyx":1032
 *         object filename = None,
 *         bint module = False,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":1033
 *         bint module = False,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":1034
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         """evaluates javascript code in from another object"""
*/
  int __pyx_v_promise = ((int)0);
  Py_buffer __pyx_v_view;
  Py_buffer __pyx_v_fs_view;
  JSValue __pyx_v_val;
  JSValue __pyx_v_this_val;
  int __pyx_v_flags;
  PyObject *__pyx_v_fs = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_2ceval_this", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_module = __pyx_optional_args->module;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_strict = __pyx_optional_args->strict;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_promise = __pyx_optional_args->promise;
            }
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":1039
 *         cdef Py_buffer view, fs_view
 *         cdef JSValue val, this_val
 *         cdef int flags = 0             # <<<<<<<<<<<<<<
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
*/
  __pyx_v_flags = 0;

  /* "cyjs/_cyjs.pyx":1040
 *         cdef JSValue val, this_val
 *         cdef int flags = 0
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"             # <<<<<<<<<<<<<<
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_filename); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1040, __pyx_L1_error)
  if (__pyx_t_2) {
    __pyx_t_3 = CYJS_FSConvert(__pyx_v_filename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b_input);
    __pyx_t_1 = __pyx_mstate_global->__pyx_kp_b_input;
  }
  __pyx_v_fs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_code, (&__pyx_v_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1042, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1043
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise             # <<<<<<<<<<<<<<
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1043, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  }

  /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_fs, (&__pyx_v_fs_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1044, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1045
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
    cyjs_release_buffer((&__pyx_v_view));

    /* "cyjs/_cyjs.pyx":1046
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1046, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  }

  /* "cyjs/_cyjs.pyx":1048
 *             raise
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL             # <<<<<<<<<<<<<<
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
*/
  if (__pyx_v_module) {
    __pyx_t_4 = JS_EVAL_TYPE_MODULE;
  } else {
    __pyx_t_4 = JS_EVAL_TYPE_GLOBAL;
  }
  __pyx_v_flags = __pyx_t_4;

  /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  if (__pyx_v_strict) {

    /* "cyjs/_cyjs.pyx":1050
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT             # <<<<<<<<<<<<<<
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_STRICT);

    /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  }

  /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  if (__pyx_v_backtrace_barrier) {

    /* "cyjs/_cyjs.pyx":1052
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER             # <<<<<<<<<<<<<<
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_BACKTRACE_BARRIER);

    /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  }

  /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  if (__pyx_v_promise) {

    /* "cyjs/_cyjs.pyx":1054
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC             # <<<<<<<<<<<<<<
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_ASYNC);

    /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_2 = (__pyx_fuse_2__pyx_f_4cyjs_5_cyjs_to_quickjs(__pyx_v_self->ctx, (&__pyx_v_this_val), __pyx_v_this) < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1057
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         val = JS_EvalThis(
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1057, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1059
 *             raise
 * 
 *         val = JS_EvalThis(             # <<<<<<<<<<<<<<
 *             self.ctx,
 *             this_val,
*/
  __pyx_v_val = JS_EvalThis(__pyx_v_self->ctx, __pyx_v_this_val, ((char const *)__pyx_v_view.buf), ((size_t)__pyx_v_view.len), ((char const *)__pyx_v_fs_view.buf), __pyx_v_flags);

  /* "cyjs/_cyjs.pyx":1068
 *         )
 * 
 *         cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *         cyjs_release_buffer(&fs_view)
 * 
*/
  cyjs_release_buffer((&__pyx_v_view));

  /* "cyjs/_cyjs.pyx":1069
 * 
 *         cyjs_release_buffer(&view)
 *         cyjs_release_buffer(&fs_view)             # <<<<<<<<<<<<<<
 * 
 *         self.raise_exception()
*/
  cyjs_release_buffer((&__pyx_v_fs_view));

  /* "cyjs/_cyjs.pyx":1071
 *         cyjs_release_buffer(&fs_view)
 * 
 *         self.raise_exception()             # <<<<<<<<<<<<<<
 * 
 *         return to_python(self.ctx, val)
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_vtab)->raise_exception(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1073
 *         self.raise_exception()
 * 
 *         return to_python(self.ctx, val)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object eval_this(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->ctx, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1026
 *             cyjs_release_buffer(&view)
 * 
 *     cdef object ceval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.Context.ceval_this", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_fuse_3__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, struct __pyx_fuse_3__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":1030
 *         object code,
 *         quickjs_type_t this,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint module = False,
 *         bint strict = False,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":1031
 *         quickjs_type_t this,
 *         object filename = None,
 *         bint module = False,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  int __pyx_v_module = ((int)0);

  /* "cyjs/_cyjs.pyx":1032
 *         object filename = None,
 *         bint module = False,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":1033
 *         bint module = False,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":1034
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         """evaluates javascript code in from another object"""
*/
  int __pyx_v_promise = ((int)0);
  Py_buffer __pyx_v_view;
  Py_buffer __pyx_v_fs_view;
  JSValue __pyx_v_val;
  JSValue __pyx_v_this_val;
  int __pyx_v_flags;
  PyObject *__pyx_v_fs = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_3ceval_this", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_module = __pyx_optional_args->module;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_strict = __pyx_optional_args->strict;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_promise = __pyx_optional_args->promise;
            }
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":1039
 *         cdef Py_buffer view, fs_view
 *         cdef JSValue val, this_val
 *         cdef int flags = 0             # <<<<<<<<<<<<<<
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
*/
  __pyx_v_flags = 0;

  /* "cyjs/_cyjs.pyx":1040
 *         cdef JSValue val, this_val
 *         cdef int flags = 0
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"             # <<<<<<<<<<<<<<
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_filename); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1040, __pyx_L1_error)
  if (__pyx_t_2) {
    __pyx_t_3 = CYJS_FSConvert(__pyx_v_filename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b_input);
    __pyx_t_1 = __pyx_mstate_global->__pyx_kp_b_input;
  }
  __pyx_v_fs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_code, (&__pyx_v_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1042, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1043
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise             # <<<<<<<<<<<<<<
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1043, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  }

  /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_fs, (&__pyx_v_fs_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1044, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1045
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
    cyjs_release_buffer((&__pyx_v_view));

    /* "cyjs/_cyjs.pyx":1046
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1046, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  }

  /* "cyjs/_cyjs.pyx":1048
 *             raise
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL             # <<<<<<<<<<<<<<
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
*/
  if (__pyx_v_module) {
    __pyx_t_4 = JS_EVAL_TYPE_MODULE;
  } else {
    __pyx_t_4 = JS_EVAL_TYPE_GLOBAL;
  }
  __pyx_v_flags = __pyx_t_4;

  /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  if (__pyx_v_strict) {

    /* "cyjs/_cyjs.pyx":1050
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT             # <<<<<<<<<<<<<<
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_STRICT);

    /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  }

  /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  if (__pyx_v_backtrace_barrier) {

    /* "cyjs/_cyjs.pyx":1052
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER             # <<<<<<<<<<<<<<
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_BACKTRACE_BARRIER);

    /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  }

  /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  if (__pyx_v_promise) {

    /* "cyjs/_cyjs.pyx":1054
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC             # <<<<<<<<<<<<<<
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_ASYNC);

    /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_2 = (__pyx_fuse_3__pyx_f_4cyjs_5_cyjs_to_quickjs(__pyx_v_self->ctx, (&__pyx_v_this_val), __pyx_v_this) < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1057
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         val = JS_EvalThis(
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1057, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1059
 *             raise
 * 
 *         val = JS_EvalThis(             # <<<<<<<<<<<<<<
 *             self.ctx,
 *             this_val,
*/
  __pyx_v_val = JS_EvalThis(__pyx_v_self->ctx, __pyx_v_this_val, ((char const *)__pyx_v_view.buf), ((size_t)__pyx_v_view.len), ((char const *)__pyx_v_fs_view.buf), __pyx_v_flags);

  /* "cyjs/_cyjs.pyx":1068
 *         )
 * 
 *         cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *         cyjs_release_buffer(&fs_view)
 * 
*/
  cyjs_release_buffer((&__pyx_v_view));

  /* "cyjs/_cyjs.pyx":1069
 * 
 *         cyjs_release_buffer(&view)
 *         cyjs_release_buffer(&fs_view)             # <<<<<<<<<<<<<<
 * 
 *         self.raise_exception()
*/
  cyjs_release_buffer((&__pyx_v_fs_view));

  /* "cyjs/_cyjs.pyx":1071
 *         cyjs_release_buffer(&fs_view)
 * 
 *         self.raise_exception()             # <<<<<<<<<<<<<<
 * 
 *         return to_python(self.ctx, val)
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_vtab)->raise_exception(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1073
 *         self.raise_exception()
 * 
 *         return to_python(self.ctx, val)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object eval_this(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->ctx, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1026
 *             cyjs_release_buffer(&view)
 * 
 *     cdef object ceval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.Context.ceval_this", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_fuse_4__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, struct __pyx_fuse_4__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":1030
 *         object code,
 *         quickjs_type_t this,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint module = False,
 *         bint strict = False,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":1031
 *         quickjs_type_t this,
 *         object filename = None,
 *         bint module = False,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  int __pyx_v_module = ((int)0);

  /* "cyjs/_cyjs.pyx":1032
 *         object filename = None,
 *         bint module = False,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":1033
 *         bint module = False,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":1034
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         """evaluates javascript code in from another object"""
*/
  int __pyx_v_promise = ((int)0);
  Py_buffer __pyx_v_view;
  Py_buffer __pyx_v_fs_view;
  JSValue __pyx_v_val;
  JSValue __pyx_v_this_val;
  int __pyx_v_flags;
  PyObject *__pyx_v_fs = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_4ceval_this", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_module = __pyx_optional_args->module;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_strict = __pyx_optional_args->strict;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_promise = __pyx_optional_args->promise;
            }
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":1039
 *         cdef Py_buffer view, fs_view
 *         cdef JSValue val, this_val
 *         cdef int flags = 0             # <<<<<<<<<<<<<<
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
*/
  __pyx_v_flags = 0;

  /* "cyjs/_cyjs.pyx":1040
 *         cdef JSValue val, this_val
 *         cdef int flags = 0
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"             # <<<<<<<<<<<<<<
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_filename); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1040, __pyx_L1_error)
  if (__pyx_t_2) {
    __pyx_t_3 = CYJS_FSConvert(__pyx_v_filename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b_input);
    __pyx_t_1 = __pyx_mstate_global->__pyx_kp_b_input;
  }
  __pyx_v_fs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_code, (&__pyx_v_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1042, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1043
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise             # <<<<<<<<<<<<<<
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1043, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  }

  /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_fs, (&__pyx_v_fs_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1044, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1045
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
    cyjs_release_buffer((&__pyx_v_view));

    /* "cyjs/_cyjs.pyx":1046
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1046, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  }

  /* "cyjs/_cyjs.pyx":1048
 *             raise
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL             # <<<<<<<<<<<<<<
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
*/
  if (__pyx_v_module) {
    __pyx_t_4 = JS_EVAL_TYPE_MODULE;
  } else {
    __pyx_t_4 = JS_EVAL_TYPE_GLOBAL;
  }
  __pyx_v_flags = __pyx_t_4;

  /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  if (__pyx_v_strict) {

    /* "cyjs/_cyjs.pyx":1050
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT             # <<<<<<<<<<<<<<
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_STRICT);

    /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  }

  /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  if (__pyx_v_backtrace_barrier) {

    /* "cyjs/_cyjs.pyx":1052
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER             # <<<<<<<<<<<<<<
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_BACKTRACE_BARRIER);

    /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  }

  /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  if (__pyx_v_promise) {

    /* "cyjs/_cyjs.pyx":1054
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC             # <<<<<<<<<<<<<<
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_ASYNC);

    /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_2 = (__pyx_fuse_4__pyx_f_4cyjs_5_cyjs_to_quickjs(__pyx_v_self->ctx, (&__pyx_v_this_val), __pyx_v_this) < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1057
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         val = JS_EvalThis(
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1057, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1059
 *             raise
 * 
 *         val = JS_EvalThis(             # <<<<<<<<<<<<<<
 *             self.ctx,
 *             this_val,
*/
  __pyx_v_val = JS_EvalThis(__pyx_v_self->ctx, __pyx_v_this_val, ((char const *)__pyx_v_view.buf), ((size_t)__pyx_v_view.len), ((char const *)__pyx_v_fs_view.buf), __pyx_v_flags);

  /* "cyjs/_cyjs.pyx":1068
 *         )
 * 
 *         cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *         cyjs_release_buffer(&fs_view)
 * 
*/
  cyjs_release_buffer((&__pyx_v_view));

  /* "cyjs/_cyjs.pyx":1069
 * 
 *         cyjs_release_buffer(&view)
 *         cyjs_release_buffer(&fs_view)             # <<<<<<<<<<<<<<
 * 
 *         self.raise_exception()
*/
  cyjs_release_buffer((&__pyx_v_fs_view));

  /* "cyjs/_cyjs.pyx":1071
 *         cyjs_release_buffer(&fs_view)
 * 
 *         self.raise_exception()             # <<<<<<<<<<<<<<
 * 
 *         return to_python(self.ctx, val)
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_vtab)->raise_exception(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1073
 *         self.raise_exception()
 * 
 *         return to_python(self.ctx, val)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object eval_this(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->ctx, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1026
 *             cyjs_release_buffer(&view)
 * 
 *     cdef object ceval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.Context.ceval_this", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_fuse_5__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, struct __pyx_fuse_5__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":1030
 *         object code,
 *         quickjs_type_t this,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint module = False,
 *         bint strict = False,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":1031
 *         quickjs_type_t this,
 *         object filename = None,
 *         bint module = False,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  int __pyx_v_module = ((int)0);

  /* "cyjs/_cyjs.pyx":1032
 *         object filename = None,
 *         bint module = False,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":1033
 *         bint module = False,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":1034
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         """evaluates javascript code in from another object"""
*/
  int __pyx_v_promise = ((int)0);
  Py_buffer __pyx_v_view;
  Py_buffer __pyx_v_fs_view;
  JSValue __pyx_v_val;
  JSValue __pyx_v_this_val;
  int __pyx_v_flags;
  PyObject *__pyx_v_fs = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_5ceval_this", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_module = __pyx_optional_args->module;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_strict = __pyx_optional_args->strict;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_promise = __pyx_optional_args->promise;
            }
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":1039
 *         cdef Py_buffer view, fs_view
 *         cdef JSValue val, this_val
 *         cdef int flags = 0             # <<<<<<<<<<<<<<
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
*/
  __pyx_v_flags = 0;

  /* "cyjs/_cyjs.pyx":1040
 *         cdef JSValue val, this_val
 *         cdef int flags = 0
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"             # <<<<<<<<<<<<<<
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_filename); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1040, __pyx_L1_error)
  if (__pyx_t_2) {
    __pyx_t_3 = CYJS_FSConvert(__pyx_v_filename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b_input);
    __pyx_t_1 = __pyx_mstate_global->__pyx_kp_b_input;
  }
  __pyx_v_fs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_code, (&__pyx_v_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1042, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1043
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise             # <<<<<<<<<<<<<<
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1043, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  }

  /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_fs, (&__pyx_v_fs_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1044, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1045
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
    cyjs_release_buffer((&__pyx_v_view));

    /* "cyjs/_cyjs.pyx":1046
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1046, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  }

  /* "cyjs/_cyjs.pyx":1048
 *             raise
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL             # <<<<<<<<<<<<<<
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
*/
  if (__pyx_v_module) {
    __pyx_t_4 = JS_EVAL_TYPE_MODULE;
  } else {
    __pyx_t_4 = JS_EVAL_TYPE_GLOBAL;
  }
  __pyx_v_flags = __pyx_t_4;

  /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  if (__pyx_v_strict) {

    /* "cyjs/_cyjs.pyx":1050
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT             # <<<<<<<<<<<<<<
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_STRICT);

    /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  }

  /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  if (__pyx_v_backtrace_barrier) {

    /* "cyjs/_cyjs.pyx":1052
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER             # <<<<<<<<<<<<<<
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_BACKTRACE_BARRIER);

    /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  }

  /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  if (__pyx_v_promise) {

    /* "cyjs/_cyjs.pyx":1054
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC             # <<<<<<<<<<<<<<
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_ASYNC);

    /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_2 = (__pyx_fuse_5__pyx_f_4cyjs_5_cyjs_to_quickjs(__pyx_v_self->ctx, (&__pyx_v_this_val), __pyx_v_this) < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1057
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         val = JS_EvalThis(
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1057, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1059
 *             raise
 * 
 *         val = JS_EvalThis(             # <<<<<<<<<<<<<<
 *             self.ctx,
 *             this_val,
*/
  __pyx_v_val = JS_EvalThis(__pyx_v_self->ctx, __pyx_v_this_val, ((char const *)__pyx_v_view.buf), ((size_t)__pyx_v_view.len), ((char const *)__pyx_v_fs_view.buf), __pyx_v_flags);

  /* "cyjs/_cyjs.pyx":1068
 *         )
 * 
 *         cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *         cyjs_release_buffer(&fs_view)
 * 
*/
  cyjs_release_buffer((&__pyx_v_view));

  /* "cyjs/_cyjs.pyx":1069
 * 
 *         cyjs_release_buffer(&view)
 *         cyjs_release_buffer(&fs_view)             # <<<<<<<<<<<<<<
 * 
 *         self.raise_exception()
*/
  cyjs_release_buffer((&__pyx_v_fs_view));

  /* "cyjs/_cyjs.pyx":1071
 *         cyjs_release_buffer(&fs_view)
 * 
 *         self.raise_exception()             # <<<<<<<<<<<<<<
 * 
 *         return to_python(self.ctx, val)
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_vtab)->raise_exception(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1073
 *         self.raise_exception()
 * 
 *         return to_python(self.ctx, val)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object eval_this(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->ctx, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1026
 *             cyjs_release_buffer(&view)
 * 
 *     cdef object ceval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.Context.ceval_this", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_fuse_6__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, int __pyx_v_this, struct __pyx_fuse_6__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":1030
 *         object code,
 *         quickjs_type_t this,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint module = False,
 *         bint strict = False,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":1031
 *         quickjs_type_t this,
 *         object filename = None,
 *         bint module = False,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  int __pyx_v_module = ((int)0);

  /* "cyjs/_cyjs.pyx":1032
 *         object filename = None,
 *         bint module = False,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":1033
 *         bint module = False,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":1034
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         """evaluates javascript code in from another object"""
*/
  int __pyx_v_promise = ((int)0);
  Py_buffer __pyx_v_view;
  Py_buffer __pyx_v_fs_view;
  JSValue __pyx_v_val;
  JSValue __pyx_v_this_val;
  int __pyx_v_flags;
  PyObject *__pyx_v_fs = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_6ceval_this", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_module = __pyx_optional_args->module;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_strict = __pyx_optional_args->strict;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_promise = __pyx_optional_args->promise;
            }
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":1039
 *         cdef Py_buffer view, fs_view
 *         cdef JSValue val, this_val
 *         cdef int flags = 0             # <<<<<<<<<<<<<<
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
*/
  __pyx_v_flags = 0;

  /* "cyjs/_cyjs.pyx":1040
 *         cdef JSValue val, this_val
 *         cdef int flags = 0
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"             # <<<<<<<<<<<<<<
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_filename); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1040, __pyx_L1_error)
  if (__pyx_t_2) {
    __pyx_t_3 = CYJS_FSConvert(__pyx_v_filename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b_input);
    __pyx_t_1 = __pyx_mstate_global->__pyx_kp_b_input;
  }
  __pyx_v_fs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_code, (&__pyx_v_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1042, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1043
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise             # <<<<<<<<<<<<<<
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1043, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  }

  /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_fs, (&__pyx_v_fs_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1044, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1045
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
    cyjs_release_buffer((&__pyx_v_view));

    /* "cyjs/_cyjs.pyx":1046
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1046, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  }

  /* "cyjs/_cyjs.pyx":1048
 *             raise
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL             # <<<<<<<<<<<<<<
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
*/
  if (__pyx_v_module) {
    __pyx_t_4 = JS_EVAL_TYPE_MODULE;
  } else {
    __pyx_t_4 = JS_EVAL_TYPE_GLOBAL;
  }
  __pyx_v_flags = __pyx_t_4;

  /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  if (__pyx_v_strict) {

    /* "cyjs/_cyjs.pyx":1050
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT             # <<<<<<<<<<<<<<
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_STRICT);

    /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  }

  /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  if (__pyx_v_backtrace_barrier) {

    /* "cyjs/_cyjs.pyx":1052
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER             # <<<<<<<<<<<<<<
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_BACKTRACE_BARRIER);

    /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  }

  /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  if (__pyx_v_promise) {

    /* "cyjs/_cyjs.pyx":1054
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC             # <<<<<<<<<<<<<<
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_ASYNC);

    /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_2 = (__pyx_fuse_6__pyx_f_4cyjs_5_cyjs_to_quickjs(__pyx_v_self->ctx, (&__pyx_v_this_val), __pyx_v_this) < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1057
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         val = JS_EvalThis(
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1057, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1059
 *             raise
 * 
 *         val = JS_EvalThis(             # <<<<<<<<<<<<<<
 *             self.ctx,
 *             this_val,
*/
  __pyx_v_val = JS_EvalThis(__pyx_v_self->ctx, __pyx_v_this_val, ((char const *)__pyx_v_view.buf), ((size_t)__pyx_v_view.len), ((char const *)__pyx_v_fs_view.buf), __pyx_v_flags);

  /* "cyjs/_cyjs.pyx":1068
 *         )
 * 
 *         cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *         cyjs_release_buffer(&fs_view)
 * 
*/
  cyjs_release_buffer((&__pyx_v_view));

  /* "cyjs/_cyjs.pyx":1069
 * 
 *         cyjs_release_buffer(&view)
 *         cyjs_release_buffer(&fs_view)             # <<<<<<<<<<<<<<
 * 
 *         self.raise_exception()
*/
  cyjs_release_buffer((&__pyx_v_fs_view));

  /* "cyjs/_cyjs.pyx":1071
 *         cyjs_release_buffer(&fs_view)
 * 
 *         self.raise_exception()             # <<<<<<<<<<<<<<
 * 
 *         return to_python(self.ctx, val)
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_vtab)->raise_exception(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1073
 *         self.raise_exception()
 * 
 *         return to_python(self.ctx, val)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object eval_this(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->ctx, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1026
 *             cyjs_release_buffer(&view)
 * 
 *     cdef object ceval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.Context.ceval_this", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_fuse_7__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, float __pyx_v_this, struct __pyx_fuse_7__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":1030
 *         object code,
 *         quickjs_type_t this,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint module = False,
 *         bint strict = False,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":1031
 *         quickjs_type_t this,
 *         object filename = None,
 *         bint module = False,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  int __pyx_v_module = ((int)0);

  /* "cyjs/_cyjs.pyx":1032
 *         object filename = None,
 *         bint module = False,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":1033
 *         bint module = False,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":1034
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         """evaluates javascript code in from another object"""
*/
  int __pyx_v_promise = ((int)0);
  Py_buffer __pyx_v_view;
  Py_buffer __pyx_v_fs_view;
  JSValue __pyx_v_val;
  JSValue __pyx_v_this_val;
  int __pyx_v_flags;
  PyObject *__pyx_v_fs = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_7ceval_this", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_module = __pyx_optional_args->module;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_strict = __pyx_optional_args->strict;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_promise = __pyx_optional_args->promise;
            }
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":1039
 *         cdef Py_buffer view, fs_view
 *         cdef JSValue val, this_val
 *         cdef int flags = 0             # <<<<<<<<<<<<<<
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
*/
  __pyx_v_flags = 0;

  /* "cyjs/_cyjs.pyx":1040
 *         cdef JSValue val, this_val
 *         cdef int flags = 0
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"             # <<<<<<<<<<<<<<
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_filename); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1040, __pyx_L1_error)
  if (__pyx_t_2) {
    __pyx_t_3 = CYJS_FSConvert(__pyx_v_filename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b_input);
    __pyx_t_1 = __pyx_mstate_global->__pyx_kp_b_input;
  }
  __pyx_v_fs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_code, (&__pyx_v_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1042, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1043
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise             # <<<<<<<<<<<<<<
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1043, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  }

  /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_fs, (&__pyx_v_fs_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1044, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1045
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
    cyjs_release_buffer((&__pyx_v_view));

    /* "cyjs/_cyjs.pyx":1046
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1046, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  }

  /* "cyjs/_cyjs.pyx":1048
 *             raise
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL             # <<<<<<<<<<<<<<
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
*/
  if (__pyx_v_module) {
    __pyx_t_4 = JS_EVAL_TYPE_MODULE;
  } else {
    __pyx_t_4 = JS_EVAL_TYPE_GLOBAL;
  }
  __pyx_v_flags = __pyx_t_4;

  /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  if (__pyx_v_strict) {

    /* "cyjs/_cyjs.pyx":1050
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT             # <<<<<<<<<<<<<<
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_STRICT);

    /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  }

  /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  if (__pyx_v_backtrace_barrier) {

    /* "cyjs/_cyjs.pyx":1052
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER             # <<<<<<<<<<<<<<
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_BACKTRACE_BARRIER);

    /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  }

  /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  if (__pyx_v_promise) {

    /* "cyjs/_cyjs.pyx":1054
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC             # <<<<<<<<<<<<<<
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_ASYNC);

    /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_2 = (__pyx_fuse_7__pyx_f_4cyjs_5_cyjs_to_quickjs(__pyx_v_self->ctx, (&__pyx_v_this_val), __pyx_v_this) < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1057
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         val = JS_EvalThis(
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1057, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1059
 *             raise
 * 
 *         val = JS_EvalThis(             # <<<<<<<<<<<<<<
 *             self.ctx,
 *             this_val,
*/
  __pyx_v_val = JS_EvalThis(__pyx_v_self->ctx, __pyx_v_this_val, ((char const *)__pyx_v_view.buf), ((size_t)__pyx_v_view.len), ((char const *)__pyx_v_fs_view.buf), __pyx_v_flags);

  /* "cyjs/_cyjs.pyx":1068
 *         )
 * 
 *         cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *         cyjs_release_buffer(&fs_view)
 * 
*/
  cyjs_release_buffer((&__pyx_v_view));

  /* "cyjs/_cyjs.pyx":1069
 * 
 *         cyjs_release_buffer(&view)
 *         cyjs_release_buffer(&fs_view)             # <<<<<<<<<<<<<<
 * 
 *         self.raise_exception()
*/
  cyjs_release_buffer((&__pyx_v_fs_view));

  /* "cyjs/_cyjs.pyx":1071
 *         cyjs_release_buffer(&fs_view)
 * 
 *         self.raise_exception()             # <<<<<<<<<<<<<<
 * 
 *         return to_python(self.ctx, val)
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_vtab)->raise_exception(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1073
 *         self.raise_exception()
 * 
 *         return to_python(self.ctx, val)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object eval_this(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->ctx, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1026
 *             cyjs_release_buffer(&view)
 * 
 *     cdef object ceval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.Context.ceval_this", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_fuse_8__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, struct __pyx_fuse_8__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":1030
 *         object code,
 *         quickjs_type_t this,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint module = False,
 *         bint strict = False,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":1031
 *         quickjs_type_t this,
 *         object filename = None,
 *         bint module = False,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  int __pyx_v_module = ((int)0);

  /* "cyjs/_cyjs.pyx":1032
 *         object filename = None,
 *         bint module = False,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":1033
 *         bint module = False,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":1034
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         """evaluates javascript code in from another object"""
*/
  int __pyx_v_promise = ((int)0);
  Py_buffer __pyx_v_view;
  Py_buffer __pyx_v_fs_view;
  JSValue __pyx_v_val;
  JSValue __pyx_v_this_val;
  int __pyx_v_flags;
  PyObject *__pyx_v_fs = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_8ceval_this", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_module = __pyx_optional_args->module;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_strict = __pyx_optional_args->strict;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_promise = __pyx_optional_args->promise;
            }
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":1039
 *         cdef Py_buffer view, fs_view
 *         cdef JSValue val, this_val
 *         cdef int flags = 0             # <<<<<<<<<<<<<<
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
*/
  __pyx_v_flags = 0;

  /* "cyjs/_cyjs.pyx":1040
 *         cdef JSValue val, this_val
 *         cdef int flags = 0
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"             # <<<<<<<<<<<<<<
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_filename); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1040, __pyx_L1_error)
  if (__pyx_t_2) {
    __pyx_t_3 = CYJS_FSConvert(__pyx_v_filename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b_input);
    __pyx_t_1 = __pyx_mstate_global->__pyx_kp_b_input;
  }
  __pyx_v_fs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_code, (&__pyx_v_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1042, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1043
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise             # <<<<<<<<<<<<<<
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1043, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  }

  /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_fs, (&__pyx_v_fs_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1044, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1045
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
    cyjs_release_buffer((&__pyx_v_view));

    /* "cyjs/_cyjs.pyx":1046
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1046, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  }

  /* "cyjs/_cyjs.pyx":1048
 *             raise
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL             # <<<<<<<<<<<<<<
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
*/
  if (__pyx_v_module) {
    __pyx_t_4 = JS_EVAL_TYPE_MODULE;
  } else {
    __pyx_t_4 = JS_EVAL_TYPE_GLOBAL;
  }
  __pyx_v_flags = __pyx_t_4;

  /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  if (__pyx_v_strict) {

    /* "cyjs/_cyjs.pyx":1050
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT             # <<<<<<<<<<<<<<
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_STRICT);

    /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  }

  /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  if (__pyx_v_backtrace_barrier) {

    /* "cyjs/_cyjs.pyx":1052
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER             # <<<<<<<<<<<<<<
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_BACKTRACE_BARRIER);

    /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  }

  /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  if (__pyx_v_promise) {

    /* "cyjs/_cyjs.pyx":1054
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC             # <<<<<<<<<<<<<<
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_ASYNC);

    /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_2 = (__pyx_fuse_8__pyx_f_4cyjs_5_cyjs_to_quickjs(__pyx_v_self->ctx, (&__pyx_v_this_val), __pyx_v_this) < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1057
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         val = JS_EvalThis(
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1057, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1059
 *             raise
 * 
 *         val = JS_EvalThis(             # <<<<<<<<<<<<<<
 *             self.ctx,
 *             this_val,
*/
  __pyx_v_val = JS_EvalThis(__pyx_v_self->ctx, __pyx_v_this_val, ((char const *)__pyx_v_view.buf), ((size_t)__pyx_v_view.len), ((char const *)__pyx_v_fs_view.buf), __pyx_v_flags);

  /* "cyjs/_cyjs.pyx":1068
 *         )
 * 
 *         cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *         cyjs_release_buffer(&fs_view)
 * 
*/
  cyjs_release_buffer((&__pyx_v_view));

  /* "cyjs/_cyjs.pyx":1069
 * 
 *         cyjs_release_buffer(&view)
 *         cyjs_release_buffer(&fs_view)             # <<<<<<<<<<<<<<
 * 
 *         self.raise_exception()
*/
  cyjs_release_buffer((&__pyx_v_fs_view));

  /* "cyjs/_cyjs.pyx":1071
 *         cyjs_release_buffer(&fs_view)
 * 
 *         self.raise_exception()             # <<<<<<<<<<<<<<
 * 
 *         return to_python(self.ctx, val)
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_vtab)->raise_exception(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1073
 *         self.raise_exception()
 * 
 *         return to_python(self.ctx, val)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object eval_this(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->ctx, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1026
 *             cyjs_release_buffer(&view)
 * 
 *     cdef object ceval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.Context.ceval_this", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_fuse_9__pyx_f_4cyjs_5_cyjs_7Context_ceval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, int __pyx_v_this, struct __pyx_fuse_9__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":1030
 *         object code,
 *         quickjs_type_t this,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint module = False,
 *         bint strict = False,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":1031
 *         quickjs_type_t this,
 *         object filename = None,
 *         bint module = False,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  int __pyx_v_module = ((int)0);

  /* "cyjs/_cyjs.pyx":1032
 *         object filename = None,
 *         bint module = False,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":1033
 *         bint module = False,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":1034
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         """evaluates javascript code in from another object"""
*/
  int __pyx_v_promise = ((int)0);
  Py_buffer __pyx_v_view;
  Py_buffer __pyx_v_fs_view;
  JSValue __pyx_v_val;
  JSValue __pyx_v_this_val;
  int __pyx_v_flags;
  PyObject *__pyx_v_fs = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_9ceval_this", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_module = __pyx_optional_args->module;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_strict = __pyx_optional_args->strict;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_promise = __pyx_optional_args->promise;
            }
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":1039
 *         cdef Py_buffer view, fs_view
 *         cdef JSValue val, this_val
 *         cdef int flags = 0             # <<<<<<<<<<<<<<
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
*/
  __pyx_v_flags = 0;

  /* "cyjs/_cyjs.pyx":1040
 *         cdef JSValue val, this_val
 *         cdef int flags = 0
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"             # <<<<<<<<<<<<<<
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_filename); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1040, __pyx_L1_error)
  if (__pyx_t_2) {
    __pyx_t_3 = CYJS_FSConvert(__pyx_v_filename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b_input);
    __pyx_t_1 = __pyx_mstate_global->__pyx_kp_b_input;
  }
  __pyx_v_fs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_code, (&__pyx_v_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1042, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1043
 * 
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise             # <<<<<<<<<<<<<<
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1043, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1042
 *         cdef object fs =  CYJS_FSConvert(filename) if filename else b"<input>"
 * 
 *         if cyjs_get_buffer(code, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
*/
  }

  /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  __pyx_t_4 = cyjs_get_buffer(__pyx_v_fs, (&__pyx_v_fs_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1044, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1045
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
    cyjs_release_buffer((&__pyx_v_view));

    /* "cyjs/_cyjs.pyx":1046
 *         if cyjs_get_buffer(fs, &fs_view) < 0:
 *             cyjs_release_buffer(&view)
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1046, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1044
 *         if cyjs_get_buffer(code, &view) < 0:
 *             raise
 *         if cyjs_get_buffer(fs, &fs_view) < 0:             # <<<<<<<<<<<<<<
 *             cyjs_release_buffer(&view)
 *             raise
*/
  }

  /* "cyjs/_cyjs.pyx":1048
 *             raise
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL             # <<<<<<<<<<<<<<
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
*/
  if (__pyx_v_module) {
    __pyx_t_4 = JS_EVAL_TYPE_MODULE;
  } else {
    __pyx_t_4 = JS_EVAL_TYPE_GLOBAL;
  }
  __pyx_v_flags = __pyx_t_4;

  /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  if (__pyx_v_strict) {

    /* "cyjs/_cyjs.pyx":1050
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT             # <<<<<<<<<<<<<<
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_STRICT);

    /* "cyjs/_cyjs.pyx":1049
 * 
 *         flags = JS_EVAL_TYPE_MODULE if module else JS_EVAL_TYPE_GLOBAL
 *         if strict:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
*/
  }

  /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  if (__pyx_v_backtrace_barrier) {

    /* "cyjs/_cyjs.pyx":1052
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER             # <<<<<<<<<<<<<<
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_BACKTRACE_BARRIER);

    /* "cyjs/_cyjs.pyx":1051
 *         if strict:
 *             flags |= JS_EVAL_FLAG_STRICT
 *         if backtrace_barrier:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
*/
  }

  /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  if (__pyx_v_promise) {

    /* "cyjs/_cyjs.pyx":1054
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:
 *             flags |= JS_EVAL_FLAG_ASYNC             # <<<<<<<<<<<<<<
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
*/
    __pyx_v_flags = (__pyx_v_flags | JS_EVAL_FLAG_ASYNC);

    /* "cyjs/_cyjs.pyx":1053
 *         if backtrace_barrier:
 *             flags |= JS_EVAL_FLAG_BACKTRACE_BARRIER
 *         if promise:             # <<<<<<<<<<<<<<
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_2 = (__pyx_fuse_9__pyx_f_4cyjs_5_cyjs_to_quickjs(__pyx_v_self->ctx, (&__pyx_v_this_val), __pyx_v_this) < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyjs/_cyjs.pyx":1057
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         val = JS_EvalThis(
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1057, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1056
 *             flags |= JS_EVAL_FLAG_ASYNC
 * 
 *         if to_quickjs(self.ctx, &this_val, this) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1059
 *             raise
 * 
 *         val = JS_EvalThis(             # <<<<<<<<<<<<<<
 *             self.ctx,
 *             this_val,
*/
  __pyx_v_val = JS_EvalThis(__pyx_v_self->ctx, __pyx_v_this_val, ((char const *)__pyx_v_view.buf), ((size_t)__pyx_v_view.len), ((char const *)__pyx_v_fs_view.buf), __pyx_v_flags);

  /* "cyjs/_cyjs.pyx":1068
 *         )
 * 
 *         cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *         cyjs_release_buffer(&fs_view)
 * 
*/
  cyjs_release_buffer((&__pyx_v_view));

  /* "cyjs/_cyjs.pyx":1069
 * 
 *         cyjs_release_buffer(&view)
 *         cyjs_release_buffer(&fs_view)             # <<<<<<<<<<<<<<
 * 
 *         self.raise_exception()
*/
  cyjs_release_buffer((&__pyx_v_fs_view));

  /* "cyjs/_cyjs.pyx":1071
 *         cyjs_release_buffer(&fs_view)
 * 
 *         self.raise_exception()             # <<<<<<<<<<<<<<
 * 
 *         return to_python(self.ctx, val)
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_vtab)->raise_exception(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1073
 *         self.raise_exception()
 * 
 *         return to_python(self.ctx, val)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object eval_this(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_self->ctx, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1026
 *             cyjs_release_buffer(&view)
 * 
 *     cdef object ceval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.Context.ceval_this", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1075
 *         return to_python(self.ctx, val)
 * 
 *     cpdef object eval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_17eval_this(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Context_eval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_this *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":1079
 *         object code,
 *         object this,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":1080
 *         object this,
 *         object filename = None,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":1081
 *         object filename = None,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":1082
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         return self.ceval_this(code, this, filename, False, strict, backtrace_barrier, promise)
*/
  int __pyx_v_promise = ((int)0);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  struct __pyx_fuse_8__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eval_this", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_strict = __pyx_optional_args->strict;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_promise = __pyx_optional_args->promise;
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":1075
 *         return to_python(self.ctx, val)
 * 
 *     cpdef object eval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Context &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_eval_this); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Context_17eval_this)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_strict); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1075, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_backtrace_barrier); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1075, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_promise); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1075, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_8 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[7] = {__pyx_t_3, __pyx_v_code, __pyx_v_this, __pyx_v_filename, __pyx_t_5, __pyx_t_6, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_8, (7-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1075, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":1084
 *         bint promise = False
 *     ):
 *         return self.ceval_this(code, this, filename, False, strict, backtrace_barrier, promise)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object eval_this_with_module(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_9.__pyx_n = 5;
  __pyx_t_9.filename = __pyx_v_filename;
  __pyx_t_9.module = 0;
  __pyx_t_9.strict = __pyx_v_strict;
  __pyx_t_9.backtrace_barrier = __pyx_v_backtrace_barrier;
  __pyx_t_9.promise = __pyx_v_promise;
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_vtab)->__pyx_fuse_8ceval_this(__pyx_v_self, __pyx_v_code, __pyx_v_this, &__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1075
 *         return to_python(self.ctx, val)
 * 
 *     cpdef object eval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cyjs._cyjs.Context.eval_this", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_17eval_this(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Context_17eval_this = {"eval_this", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_17eval_this, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_17eval_this(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_code = 0;
  PyObject *__pyx_v_this = 0;
  PyObject *__pyx_v_filename = 0;
  int __pyx_v_strict;
  int __pyx_v_backtrace_barrier;
  int __pyx_v_promise;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("eval_this (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_code,&__pyx_mstate_global->__pyx_n_u_this,&__pyx_mstate_global->__pyx_n_u_filename,&__pyx_mstate_global->__pyx_n_u_strict,&__pyx_mstate_global->__pyx_n_u_backtrace_barrier,&__pyx_mstate_global->__pyx_n_u_promise,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1075, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 1075, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1075, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1075, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1075, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1075, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1075, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "eval_this", 0) < (0)) __PYX_ERR(0, 1075, __pyx_L3_error)

      /* "cyjs/_cyjs.pyx":1079
 *         object code,
 *         object this,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("eval_this", 0, 2, 6, i); __PYX_ERR(0, 1075, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 1075, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1075, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1075, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1075, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1075, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1075, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_code = values[0];
    __pyx_v_this = values[1];
    __pyx_v_filename = values[2];
    if (values[3]) {
      __pyx_v_strict = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_strict == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1080, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":1080
 *         object this,
 *         object filename = None,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
      __pyx_v_strict = ((int)0);
    }
    if (values[4]) {
      __pyx_v_backtrace_barrier = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_backtrace_barrier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1081, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":1081
 *         object filename = None,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
      __pyx_v_backtrace_barrier = ((int)0);
    }
    if (values[5]) {
      __pyx_v_promise = __Pyx_PyObject_IsTrue(values[5]); if (unlikely((__pyx_v_promise == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1082, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":1082
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         return self.ceval_this(code, this, filename, False, strict, backtrace_barrier, promise)
*/
      __pyx_v_promise = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("eval_this", 0, 2, 6, __pyx_nargs); __PYX_ERR(0, 1075, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Context.eval_this", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Context_16eval_this(((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_v_self), __pyx_v_code, __pyx_v_this, __pyx_v_filename, __pyx_v_strict, __pyx_v_backtrace_barrier, __pyx_v_promise);

  /* "cyjs/_cyjs.pyx":1075
 *         return to_python(self.ctx, val)
 * 
 *     cpdef object eval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_16eval_this(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, PyObject *__pyx_v_filename, int __pyx_v_strict, int __pyx_v_backtrace_barrier, int __pyx_v_promise) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_this __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eval_this", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 4;
  __pyx_t_2.filename = __pyx_v_filename;
  __pyx_t_2.strict = __pyx_v_strict;
  __pyx_t_2.backtrace_barrier = __pyx_v_backtrace_barrier;
  __pyx_t_2.promise = __pyx_v_promise;
  __pyx_t_1 = __pyx_vtabptr_4cyjs_5_cyjs_Context->eval_this(__pyx_v_self, __pyx_v_code, __pyx_v_this, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Context.eval_this", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1086
 *         return self.ceval_this(code, this, filename, False, strict, backtrace_barrier, promise)
 * 
 *     cpdef object eval_this_with_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_19eval_this_with_module(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Context_eval_this_with_module(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_this_with_module *__pyx_optional_args) {

  /* "cyjs/_cyjs.pyx":1090
 *         object code,
 *         object this,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
  PyObject *__pyx_v_filename = ((PyObject *)Py_None);

  /* "cyjs/_cyjs.pyx":1091
 *         object this,
 *         object filename = None,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
  int __pyx_v_strict = ((int)0);

  /* "cyjs/_cyjs.pyx":1092
 *         object filename = None,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
  int __pyx_v_backtrace_barrier = ((int)0);

  /* "cyjs/_cyjs.pyx":1093
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         return self.ceval_this(code, this, filename, True, strict, backtrace_barrier, promise)
*/
  int __pyx_v_promise = ((int)0);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  struct __pyx_fuse_8__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eval_this_with_module", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_filename = __pyx_optional_args->filename;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_strict = __pyx_optional_args->strict;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_backtrace_barrier = __pyx_optional_args->backtrace_barrier;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_promise = __pyx_optional_args->promise;
          }
        }
      }
    }
  }

  /* "cyjs/_cyjs.pyx":1086
 *         return self.ceval_this(code, this, filename, False, strict, backtrace_barrier, promise)
 * 
 *     cpdef object eval_this_with_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Context &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_eval_this_with_module); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1086, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Context_19eval_this_with_module)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_strict); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1086, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_backtrace_barrier); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1086, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_promise); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1086, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_8 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[7] = {__pyx_t_3, __pyx_v_code, __pyx_v_this, __pyx_v_filename, __pyx_t_5, __pyx_t_6, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_8, (7-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1086, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":1095
 *         bint promise = False
 *     ):
 *         return self.ceval_this(code, this, filename, True, strict, backtrace_barrier, promise)             # <<<<<<<<<<<<<<
 * 
 *     # TODO: Soon as I figure out how to make callbacks and promises work...
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_9.__pyx_n = 5;
  __pyx_t_9.filename = __pyx_v_filename;
  __pyx_t_9.module = 1;
  __pyx_t_9.strict = __pyx_v_strict;
  __pyx_t_9.backtrace_barrier = __pyx_v_backtrace_barrier;
  __pyx_t_9.promise = __pyx_v_promise;
  __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_vtab)->__pyx_fuse_8ceval_this(__pyx_v_self, __pyx_v_code, __pyx_v_this, &__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1086
 *         return self.ceval_this(code, this, filename, False, strict, backtrace_barrier, promise)
 * 
 *     cpdef object eval_this_with_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cyjs._cyjs.Context.eval_this_with_module", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_19eval_this_with_module(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Context_19eval_this_with_module = {"eval_this_with_module", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_19eval_this_with_module, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_19eval_this_with_module(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_code = 0;
  PyObject *__pyx_v_this = 0;
  PyObject *__pyx_v_filename = 0;
  int __pyx_v_strict;
  int __pyx_v_backtrace_barrier;
  int __pyx_v_promise;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("eval_this_with_module (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_code,&__pyx_mstate_global->__pyx_n_u_this,&__pyx_mstate_global->__pyx_n_u_filename,&__pyx_mstate_global->__pyx_n_u_strict,&__pyx_mstate_global->__pyx_n_u_backtrace_barrier,&__pyx_mstate_global->__pyx_n_u_promise,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1086, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 1086, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1086, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1086, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1086, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1086, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1086, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "eval_this_with_module", 0) < (0)) __PYX_ERR(0, 1086, __pyx_L3_error)

      /* "cyjs/_cyjs.pyx":1090
 *         object code,
 *         object this,
 *         object filename = None,             # <<<<<<<<<<<<<<
 *         bint strict = False,
 *         bint backtrace_barrier = False,
*/
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("eval_this_with_module", 0, 2, 6, i); __PYX_ERR(0, 1086, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 1086, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1086, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1086, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1086, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1086, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1086, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_code = values[0];
    __pyx_v_this = values[1];
    __pyx_v_filename = values[2];
    if (values[3]) {
      __pyx_v_strict = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_strict == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1091, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":1091
 *         object this,
 *         object filename = None,
 *         bint strict = False,             # <<<<<<<<<<<<<<
 *         bint backtrace_barrier = False,
 *         bint promise = False
*/
      __pyx_v_strict = ((int)0);
    }
    if (values[4]) {
      __pyx_v_backtrace_barrier = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_backtrace_barrier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1092, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":1092
 *         object filename = None,
 *         bint strict = False,
 *         bint backtrace_barrier = False,             # <<<<<<<<<<<<<<
 *         bint promise = False
 *     ):
*/
      __pyx_v_backtrace_barrier = ((int)0);
    }
    if (values[5]) {
      __pyx_v_promise = __Pyx_PyObject_IsTrue(values[5]); if (unlikely((__pyx_v_promise == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1093, __pyx_L3_error)
    } else {

      /* "cyjs/_cyjs.pyx":1093
 *         bint strict = False,
 *         bint backtrace_barrier = False,
 *         bint promise = False             # <<<<<<<<<<<<<<
 *     ):
 *         return self.ceval_this(code, this, filename, True, strict, backtrace_barrier, promise)
*/
      __pyx_v_promise = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("eval_this_with_module", 0, 2, 6, __pyx_nargs); __PYX_ERR(0, 1086, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Context.eval_this_with_module", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Context_18eval_this_with_module(((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_v_self), __pyx_v_code, __pyx_v_this, __pyx_v_filename, __pyx_v_strict, __pyx_v_backtrace_barrier, __pyx_v_promise);

  /* "cyjs/_cyjs.pyx":1086
 *         return self.ceval_this(code, this, filename, False, strict, backtrace_barrier, promise)
 * 
 *     cpdef object eval_this_with_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_18eval_this_with_module(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_this, PyObject *__pyx_v_filename, int __pyx_v_strict, int __pyx_v_backtrace_barrier, int __pyx_v_promise) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_this_with_module __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eval_this_with_module", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 4;
  __pyx_t_2.filename = __pyx_v_filename;
  __pyx_t_2.strict = __pyx_v_strict;
  __pyx_t_2.backtrace_barrier = __pyx_v_backtrace_barrier;
  __pyx_t_2.promise = __pyx_v_promise;
  __pyx_t_1 = __pyx_vtabptr_4cyjs_5_cyjs_Context->eval_this_with_module(__pyx_v_self, __pyx_v_code, __pyx_v_this, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Context.eval_this_with_module", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pxd":213
 *     cdef:
 *         # public runtime for python access to value (can't be deleted)
 *         readonly Runtime runtime             # <<<<<<<<<<<<<<
 *         JSRuntime* rt
 *         JSContext* ctx
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_7runtime_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_7runtime_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Context_7runtime___get__(((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_7runtime___get__(struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->runtime);
  __pyx_r = ((PyObject *)__pyx_v_self->runtime);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.rt cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_21__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Context_21__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_21__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_21__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Context_20__reduce_cython__(((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_20__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.rt cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ctx,self.rt cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ctx_self_rt_cannot_be_conve, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.rt cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.Context.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.rt cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.rt cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_23__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Context_23__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_23__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Context_23__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Context.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Context_22__setstate_cython__(((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Context_22__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_Context *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.ctx,self.rt cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ctx,self.rt cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ctx_self_rt_cannot_be_conve, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.rt cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.rt cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.Context.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1105
 * 
 * 
 * cdef int py_to_js_str(JSContext* ctx, object obj, JSValue* value):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef JSValue val
*/

static int __pyx_f_4cyjs_5_cyjs_py_to_js_str(JSContext *__pyx_v_ctx, PyObject *__pyx_v_obj, JSValue *__pyx_v_value) {
  Py_buffer __pyx_v_view;
  JSValue __pyx_v_val;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyjs/_cyjs.pyx":1108
 *     cdef Py_buffer view
 *     cdef JSValue val
 *     if cyjs_get_buffer(obj, &view) < 0:             # <<<<<<<<<<<<<<
 *         return -1
 *     val = JS_NewStringLen(ctx, <const char*>view.buf, <size_t>view.len)
*/
  __pyx_t_1 = cyjs_get_buffer(__pyx_v_obj, (&__pyx_v_view)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 1108, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 < 0);
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":1109
 *     cdef JSValue val
 *     if cyjs_get_buffer(obj, &view) < 0:
 *         return -1             # <<<<<<<<<<<<<<
 *     val = JS_NewStringLen(ctx, <const char*>view.buf, <size_t>view.len)
 *     cyjs_release_buffer(&view)
*/
    __pyx_r = -1;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":1108
 *     cdef Py_buffer view
 *     cdef JSValue val
 *     if cyjs_get_buffer(obj, &view) < 0:             # <<<<<<<<<<<<<<
 *         return -1
 *     val = JS_NewStringLen(ctx, <const char*>view.buf, <size_t>view.len)
*/
  }

  /* "cyjs/_cyjs.pyx":1110
 *     if cyjs_get_buffer(obj, &view) < 0:
 *         return -1
 *     val = JS_NewStringLen(ctx, <const char*>view.buf, <size_t>view.len)             # <<<<<<<<<<<<<<
 *     cyjs_release_buffer(&view)
 *     value[0] = val
*/
  __pyx_v_val = JS_NewStringLen(__pyx_v_ctx, ((char const *)__pyx_v_view.buf), ((size_t)__pyx_v_view.len));

  /* "cyjs/_cyjs.pyx":1111
 *         return -1
 *     val = JS_NewStringLen(ctx, <const char*>view.buf, <size_t>view.len)
 *     cyjs_release_buffer(&view)             # <<<<<<<<<<<<<<
 *     value[0] = val
 *     return 0
*/
  cyjs_release_buffer((&__pyx_v_view));

  /* "cyjs/_cyjs.pyx":1112
 *     val = JS_NewStringLen(ctx, <const char*>view.buf, <size_t>view.len)
 *     cyjs_release_buffer(&view)
 *     value[0] = val             # <<<<<<<<<<<<<<
 *     return 0
 * 
*/
  (__pyx_v_value[0]) = __pyx_v_val;

  /* "cyjs/_cyjs.pyx":1113
 *     cyjs_release_buffer(&view)
 *     value[0] = val
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1105
 * 
 * 
 * cdef int py_to_js_str(JSContext* ctx, object obj, JSValue* value):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef JSValue val
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.py_to_js_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1120
 * 
 * 
 * cdef object jsv_to_str(JSContext* ctx, JSValue value):             # <<<<<<<<<<<<<<
 *     cdef size_t len
 *     cdef const char* cstr = JS_ToCStringLen(ctx, &len , value)
*/

static PyObject *__pyx_f_4cyjs_5_cyjs_jsv_to_str(JSContext *__pyx_v_ctx, JSValue __pyx_v_value) {
  size_t __pyx_v_len;
  char const *__pyx_v_cstr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  char const *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("jsv_to_str", 0);

  /* "cyjs/_cyjs.pyx":1122
 * cdef object jsv_to_str(JSContext* ctx, JSValue value):
 *     cdef size_t len
 *     cdef const char* cstr = JS_ToCStringLen(ctx, &len , value)             # <<<<<<<<<<<<<<
 *     try:
 *         return PyUnicode_FromStringAndSize(cstr, <Py_ssize_t>len)
*/
  __pyx_v_cstr = JS_ToCStringLen(__pyx_v_ctx, (&__pyx_v_len), __pyx_v_value);

  /* "cyjs/_cyjs.pyx":1123
 *     cdef size_t len
 *     cdef const char* cstr = JS_ToCStringLen(ctx, &len , value)
 *     try:             # <<<<<<<<<<<<<<
 *         return PyUnicode_FromStringAndSize(cstr, <Py_ssize_t>len)
 *     finally:
*/
  /*try:*/ {

    /* "cyjs/_cyjs.pyx":1124
 *     cdef const char* cstr = JS_ToCStringLen(ctx, &len , value)
 *     try:
 *         return PyUnicode_FromStringAndSize(cstr, <Py_ssize_t>len)             # <<<<<<<<<<<<<<
 *     finally:
 *         JS_FreeCString(ctx, cstr)
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyUnicode_FromStringAndSize(__pyx_v_cstr, ((Py_ssize_t)__pyx_v_len)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1124, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L3_return;
  }

  /* "cyjs/_cyjs.pyx":1126
 *         return PyUnicode_FromStringAndSize(cstr, <Py_ssize_t>len)
 *     finally:
 *         JS_FreeCString(ctx, cstr)             # <<<<<<<<<<<<<<
 * 
 * cdef object jsv_to_big_int(JSContext* ctx, JSValue value):
*/
  /*finally:*/ {
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
       __Pyx_ExceptionSwap(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      if ( unlikely(__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0)) __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __pyx_t_2 = __pyx_lineno; __pyx_t_3 = __pyx_clineno; __pyx_t_4 = __pyx_filename;
      {
        JS_FreeCString(__pyx_v_ctx, __pyx_v_cstr);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ErrRestore(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0;
      __pyx_lineno = __pyx_t_2; __pyx_clineno = __pyx_t_3; __pyx_filename = __pyx_t_4;
      goto __pyx_L1_error;
    }
    __pyx_L3_return: {
      __pyx_t_10 = __pyx_r;
      __pyx_r = 0;
      JS_FreeCString(__pyx_v_ctx, __pyx_v_cstr);
      __pyx_r = __pyx_t_10;
      __pyx_t_10 = 0;
      goto __pyx_L0;
    }
  }

  /* "cyjs/_cyjs.pyx":1120
 * 
 * 
 * cdef object jsv_to_str(JSContext* ctx, JSValue value):             # <<<<<<<<<<<<<<
 *     cdef size_t len
 *     cdef const char* cstr = JS_ToCStringLen(ctx, &len , value)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.jsv_to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1128
 *         JS_FreeCString(ctx, cstr)
 * 
 * cdef object jsv_to_big_int(JSContext* ctx, JSValue value):             # <<<<<<<<<<<<<<
 *     cdef const char* cstr = JS_ToCString(ctx, value)
 *     try:
*/

static PyObject *__pyx_f_4cyjs_5_cyjs_jsv_to_big_int(JSContext *__pyx_v_ctx, JSValue __pyx_v_value) {
  char const *__pyx_v_cstr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  char const *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("jsv_to_big_int", 0);

  /* "cyjs/_cyjs.pyx":1129
 * 
 * cdef object jsv_to_big_int(JSContext* ctx, JSValue value):
 *     cdef const char* cstr = JS_ToCString(ctx, value)             # <<<<<<<<<<<<<<
 *     try:
 *         return PyLong_FromString(cstr, NULL, 10)
*/
  __pyx_v_cstr = JS_ToCString(__pyx_v_ctx, __pyx_v_value);

  /* "cyjs/_cyjs.pyx":1130
 * cdef object jsv_to_big_int(JSContext* ctx, JSValue value):
 *     cdef const char* cstr = JS_ToCString(ctx, value)
 *     try:             # <<<<<<<<<<<<<<
 *         return PyLong_FromString(cstr, NULL, 10)
 *     finally:
*/
  /*try:*/ {

    /* "cyjs/_cyjs.pyx":1131
 *     cdef const char* cstr = JS_ToCString(ctx, value)
 *     try:
 *         return PyLong_FromString(cstr, NULL, 10)             # <<<<<<<<<<<<<<
 *     finally:
 *         JS_FreeCString(ctx, cstr)
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyLong_FromString(__pyx_v_cstr, NULL, 10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1131, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L3_return;
  }

  /* "cyjs/_cyjs.pyx":1133
 *         return PyLong_FromString(cstr, NULL, 10)
 *     finally:
 *         JS_FreeCString(ctx, cstr)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  /*finally:*/ {
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
       __Pyx_ExceptionSwap(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      if ( unlikely(__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0)) __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __pyx_t_2 = __pyx_lineno; __pyx_t_3 = __pyx_clineno; __pyx_t_4 = __pyx_filename;
      {
        JS_FreeCString(__pyx_v_ctx, __pyx_v_cstr);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ErrRestore(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0;
      __pyx_lineno = __pyx_t_2; __pyx_clineno = __pyx_t_3; __pyx_filename = __pyx_t_4;
      goto __pyx_L1_error;
    }
    __pyx_L3_return: {
      __pyx_t_10 = __pyx_r;
      __pyx_r = 0;
      JS_FreeCString(__pyx_v_ctx, __pyx_v_cstr);
      __pyx_r = __pyx_t_10;
      __pyx_t_10 = 0;
      goto __pyx_L0;
    }
  }

  /* "cyjs/_cyjs.pyx":1128
 *         JS_FreeCString(ctx, cstr)
 * 
 * cdef object jsv_to_big_int(JSContext* ctx, JSValue value):             # <<<<<<<<<<<<<<
 *     cdef const char* cstr = JS_ToCString(ctx, value)
 *     try:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.jsv_to_big_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_14CancelledError_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_14CancelledError_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_14CancelledError_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_14CancelledError_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_14CancelledError___reduce_cython__(((struct __pyx_obj_4cyjs_5_cyjs_CancelledError *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_14CancelledError___reduce_cython__(struct __pyx_obj_4cyjs_5_cyjs_CancelledError *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_empty_tuple);
  __pyx_v_state = __pyx_mstate_global->__pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(1, 7, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_CancelledError, (type(self), 0xe3b0c44, None), state
*/
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CancelledError, (type(self), 0xe3b0c44, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_CancelledError, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_CancelledError, (type(self), 0xe3b0c44, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_CancelledError); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_238750788) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CancelledError, (type(self), 0xe3b0c44, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_CancelledError, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_CancelledError, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CancelledError__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_CancelledError); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_238750788) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cyjs._cyjs.CancelledError.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CancelledError, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CancelledError__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_14CancelledError_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_14CancelledError_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_14CancelledError_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_14CancelledError_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.CancelledError.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_14CancelledError_2__setstate_cython__(((struct __pyx_obj_4cyjs_5_cyjs_CancelledError *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_14CancelledError_2__setstate_cython__(struct __pyx_obj_4cyjs_5_cyjs_CancelledError *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_CancelledError, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CancelledError__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(1, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_4cyjs_5_cyjs___pyx_unpickle_CancelledError__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CancelledError, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CancelledError__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cyjs._cyjs.CancelledError.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_17InvalidStateError_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_17InvalidStateError_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_17InvalidStateError_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_17InvalidStateError_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_17InvalidStateError___reduce_cython__(((struct __pyx_obj_4cyjs_5_cyjs_InvalidStateError *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_17InvalidStateError___reduce_cython__(struct __pyx_obj_4cyjs_5_cyjs_InvalidStateError *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_empty_tuple);
  __pyx_v_state = __pyx_mstate_global->__pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(1, 7, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_InvalidStateError, (type(self), 0xe3b0c44, None), state
*/
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InvalidStateError, (type(self), 0xe3b0c44, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_InvalidStateError, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_InvalidStateError, (type(self), 0xe3b0c44, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_InvalidStateError); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_238750788) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InvalidStateError, (type(self), 0xe3b0c44, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_InvalidStateError, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_InvalidStateError, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InvalidStateError__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_InvalidStateError); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_238750788) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cyjs._cyjs.InvalidStateError.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InvalidStateError, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InvalidStateError__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_17InvalidStateError_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_17InvalidStateError_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_17InvalidStateError_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_17InvalidStateError_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.InvalidStateError.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_17InvalidStateError_2__setstate_cython__(((struct __pyx_obj_4cyjs_5_cyjs_InvalidStateError *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_17InvalidStateError_2__setstate_cython__(struct __pyx_obj_4cyjs_5_cyjs_InvalidStateError *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_InvalidStateError, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InvalidStateError__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(1, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_4cyjs_5_cyjs___pyx_unpickle_InvalidStateError__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InvalidStateError, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InvalidStateError__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cyjs._cyjs.InvalidStateError.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1151
 *         bint completed
 * 
 *     cpdef object add_done_callback(self, object fn):             # <<<<<<<<<<<<<<
 *         """Attaches a callable callback when promise finishes or raises an exception"""
 *         if not callable(fn):
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_1add_done_callback(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Promise_add_done_callback(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, PyObject *__pyx_v_fn, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_done_callback", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Promise &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_add_done_callback); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1151, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Promise_1add_done_callback)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_fn};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1151, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":1153
 *     cpdef object add_done_callback(self, object fn):
 *         """Attaches a callable callback when promise finishes or raises an exception"""
 *         if not callable(fn):             # <<<<<<<<<<<<<<
 *             raise TypeError("done callbacks must be callable")
 * 
*/
  __pyx_t_6 = __Pyx_PyCallable_Check(__pyx_v_fn); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1153, __pyx_L1_error)
  __pyx_t_7 = (!__pyx_t_6);
  if (unlikely(__pyx_t_7)) {

    /* "cyjs/_cyjs.pyx":1154
 *         """Attaches a callable callback when promise finishes or raises an exception"""
 *         if not callable(fn):
 *             raise TypeError("done callbacks must be callable")             # <<<<<<<<<<<<<<
 * 
 *         if JS_PromiseState(self.ctx, self.value) == JS_PROMISE_PENDING:
*/
    __pyx_t_2 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_done_callbacks_must_be_callable};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1154, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1153
 *     cpdef object add_done_callback(self, object fn):
 *         """Attaches a callable callback when promise finishes or raises an exception"""
 *         if not callable(fn):             # <<<<<<<<<<<<<<
 *             raise TypeError("done callbacks must be callable")
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1156
 *             raise TypeError("done callbacks must be callable")
 * 
 *         if JS_PromiseState(self.ctx, self.value) == JS_PROMISE_PENDING:             # <<<<<<<<<<<<<<
 *             self._callbacks.append(fn)
 *         else:
*/
  __pyx_t_7 = (JS_PromiseState(__pyx_v_self->__pyx_base.ctx, __pyx_v_self->__pyx_base.value) == JS_PROMISE_PENDING);
  if (__pyx_t_7) {

    /* "cyjs/_cyjs.pyx":1157
 * 
 *         if JS_PromiseState(self.ctx, self.value) == JS_PROMISE_PENDING:
 *             self._callbacks.append(fn)             # <<<<<<<<<<<<<<
 *         else:
 *             fn(self)
*/
    if (unlikely(__pyx_v_self->_callbacks == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
      __PYX_ERR(0, 1157, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_self->_callbacks, __pyx_v_fn); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1157, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1156
 *             raise TypeError("done callbacks must be callable")
 * 
 *         if JS_PromiseState(self.ctx, self.value) == JS_PROMISE_PENDING:             # <<<<<<<<<<<<<<
 *             self._callbacks.append(fn)
 *         else:
*/
    goto __pyx_L4;
  }

  /* "cyjs/_cyjs.pyx":1159
 *             self._callbacks.append(fn)
 *         else:
 *             fn(self)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object exception(self):
*/
  /*else*/ {
    __pyx_t_2 = NULL;
    __Pyx_INCREF(__pyx_v_fn);
    __pyx_t_4 = __pyx_v_fn; 
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, ((PyObject *)__pyx_v_self)};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1159, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L4:;

  /* "cyjs/_cyjs.pyx":1151
 *         bint completed
 * 
 *     cpdef object add_done_callback(self, object fn):             # <<<<<<<<<<<<<<
 *         """Attaches a callable callback when promise finishes or raises an exception"""
 *         if not callable(fn):
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Promise.add_done_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_1add_done_callback(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4cyjs_5_cyjs_7Promise_add_done_callback, "Attaches a callable callback when promise finishes or raises an exception");
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Promise_1add_done_callback = {"add_done_callback", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Promise_1add_done_callback, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4cyjs_5_cyjs_7Promise_add_done_callback};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_1add_done_callback(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_fn = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_done_callback (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_fn,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1151, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1151, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "add_done_callback", 0) < (0)) __PYX_ERR(0, 1151, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("add_done_callback", 1, 1, 1, i); __PYX_ERR(0, 1151, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1151, __pyx_L3_error)
    }
    __pyx_v_fn = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_done_callback", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1151, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Promise.add_done_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Promise_add_done_callback(((struct __pyx_obj_4cyjs_5_cyjs_Promise *)__pyx_v_self), __pyx_v_fn);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Promise_add_done_callback(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, PyObject *__pyx_v_fn) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_done_callback", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_7Promise_add_done_callback(__pyx_v_self, __pyx_v_fn, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Promise.add_done_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1161
 *             fn(self)
 * 
 *     cpdef object exception(self):             # <<<<<<<<<<<<<<
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)
 *         if state == JS_PROMISE_REJECTED:
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_3exception(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Promise_exception(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, int __pyx_skip_dispatch) {
  JSPromiseStateEnum __pyx_v_state;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("exception", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Promise &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_exception); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1161, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Promise_3exception)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1161, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":1162
 * 
 *     cpdef object exception(self):
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)             # <<<<<<<<<<<<<<
 *         if state == JS_PROMISE_REJECTED:
 *             raise CancelledError()
*/
  __pyx_v_state = JS_PromiseState(__pyx_v_self->__pyx_base.ctx, __pyx_v_self->__pyx_base.value);

  /* "cyjs/_cyjs.pyx":1163
 *     cpdef object exception(self):
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)
 *         if state == JS_PROMISE_REJECTED:             # <<<<<<<<<<<<<<
 *             raise CancelledError()
 *         elif state == JS_PROMISE_PENDING:
*/
  switch (__pyx_v_state) {
    case JS_PROMISE_REJECTED:

    /* "cyjs/_cyjs.pyx":1164
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)
 *         if state == JS_PROMISE_REJECTED:
 *             raise CancelledError()             # <<<<<<<<<<<<<<
 *         elif state == JS_PROMISE_PENDING:
 *             raise InvalidStateError('Result is not ready.')
*/
    __pyx_t_2 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_CancelledError, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1164, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_1);
    }
    __Pyx_Raise(((PyObject *)__pyx_t_1), 0, 0, 0);
    __Pyx_DECREF((PyObject *)__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1164, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1163
 *     cpdef object exception(self):
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)
 *         if state == JS_PROMISE_REJECTED:             # <<<<<<<<<<<<<<
 *             raise CancelledError()
 *         elif state == JS_PROMISE_PENDING:
*/
    break;
    case JS_PROMISE_PENDING:

    /* "cyjs/_cyjs.pyx":1166
 *             raise CancelledError()
 *         elif state == JS_PROMISE_PENDING:
 *             raise InvalidStateError('Result is not ready.')             # <<<<<<<<<<<<<<
 *         return self._exception
 * 
*/
    __pyx_t_2 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Result_is_not_ready};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1166, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_1);
    }
    __Pyx_Raise(((PyObject *)__pyx_t_1), 0, 0, 0);
    __Pyx_DECREF((PyObject *)__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1166, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1165
 *         if state == JS_PROMISE_REJECTED:
 *             raise CancelledError()
 *         elif state == JS_PROMISE_PENDING:             # <<<<<<<<<<<<<<
 *             raise InvalidStateError('Result is not ready.')
 *         return self._exception
*/
    break;
    default: break;
  }

  /* "cyjs/_cyjs.pyx":1167
 *         elif state == JS_PROMISE_PENDING:
 *             raise InvalidStateError('Result is not ready.')
 *         return self._exception             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint done(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_exception);
  __pyx_r = __pyx_v_self->_exception;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1161
 *             fn(self)
 * 
 *     cpdef object exception(self):             # <<<<<<<<<<<<<<
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)
 *         if state == JS_PROMISE_REJECTED:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Promise.exception", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_3exception(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Promise_3exception = {"exception", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Promise_3exception, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_3exception(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exception (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("exception", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("exception", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Promise_2exception(((struct __pyx_obj_4cyjs_5_cyjs_Promise *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Promise_2exception(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("exception", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_7Promise_exception(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Promise.exception", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1169
 *         return self._exception
 * 
 *     cpdef bint done(self):             # <<<<<<<<<<<<<<
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)
 *         return state == JS_PROMISE_FULFILLED or state == JS_PROMISE_REJECTED
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_5done(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_4cyjs_5_cyjs_7Promise_done(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, int __pyx_skip_dispatch) {
  JSPromiseStateEnum __pyx_v_state;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("done", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Promise &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_done); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1169, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Promise_5done)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1169, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1169, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":1170
 * 
 *     cpdef bint done(self):
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)             # <<<<<<<<<<<<<<
 *         return state == JS_PROMISE_FULFILLED or state == JS_PROMISE_REJECTED
 * 
*/
  __pyx_v_state = JS_PromiseState(__pyx_v_self->__pyx_base.ctx, __pyx_v_self->__pyx_base.value);

  /* "cyjs/_cyjs.pyx":1171
 *     cpdef bint done(self):
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)
 *         return state == JS_PROMISE_FULFILLED or state == JS_PROMISE_REJECTED             # <<<<<<<<<<<<<<
 * 
 *     # TODO: Figure out how to reject incomming future objects...
*/
  switch (__pyx_v_state) {
    case JS_PROMISE_FULFILLED:
    case JS_PROMISE_REJECTED:
    __pyx_t_6 = 1;
    break;
    default:
    __pyx_t_6 = 0;
    break;
  }
  __pyx_r = __pyx_t_6;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1169
 *         return self._exception
 * 
 *     cpdef bint done(self):             # <<<<<<<<<<<<<<
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)
 *         return state == JS_PROMISE_FULFILLED or state == JS_PROMISE_REJECTED
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Promise.done", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_5done(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Promise_5done = {"done", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Promise_5done, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_5done(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("done (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("done", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("done", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Promise_4done(((struct __pyx_obj_4cyjs_5_cyjs_Promise *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Promise_4done(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("done", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_7Promise_done(__pyx_v_self, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1169, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cyjs._cyjs.Promise.done", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1176
 * 
 * 
 *     cpdef Py_ssize_t remove_done_callback(self, object fn):             # <<<<<<<<<<<<<<
 *         """Remove all instances of a callback from the "call when done" list.
 * 
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_7remove_done_callback(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static Py_ssize_t __pyx_f_4cyjs_5_cyjs_7Promise_remove_done_callback(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, PyObject *__pyx_v_fn, int __pyx_skip_dispatch) {
  Py_ssize_t __pyx_v_removed_count;
  PyObject *__pyx_v_filtered_callbacks = 0;
  PyObject *__pyx_8genexpr1__pyx_v_f = NULL;
  PyObject *__pyx_8genexpr1__pyx_v_ctx = NULL;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("remove_done_callback", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Promise &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_remove_done_callback); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Promise_7remove_done_callback)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_fn};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1176, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_t_2); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1176, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":1182
 *         """
 *         cdef Py_ssize_t removed_count
 *         cdef list filtered_callbacks = [(f, ctx)             # <<<<<<<<<<<<<<
 *                               for (f, ctx) in self._callbacks
 *                               if f != fn]
*/
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1182, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "cyjs/_cyjs.pyx":1183
 *         cdef Py_ssize_t removed_count
 *         cdef list filtered_callbacks = [(f, ctx)
 *                               for (f, ctx) in self._callbacks             # <<<<<<<<<<<<<<
 *                               if f != fn]
 *         removed_count = <Py_ssize_t>(len(self._callbacks) - len(filtered_callbacks))
*/
    if (unlikely(__pyx_v_self->_callbacks == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 1183, __pyx_L5_error)
    }
    __pyx_t_2 = __pyx_v_self->_callbacks; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_6 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1183, __pyx_L5_error)
        #endif
        if (__pyx_t_6 >= __pyx_temp) break;
      }
      __pyx_t_4 = __Pyx_PyList_GetItemRefFast(__pyx_t_2, __pyx_t_6, __Pyx_ReferenceSharing_OwnStrongReference);
      ++__pyx_t_6;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1183, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1183, __pyx_L5_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0);
          __Pyx_INCREF(__pyx_t_3);
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1);
          __Pyx_INCREF(__pyx_t_7);
        } else {
          __pyx_t_3 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1183, __pyx_L5_error)
          __Pyx_XGOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1183, __pyx_L5_error)
          __Pyx_XGOTREF(__pyx_t_7);
        }
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1183, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1183, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_8 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1183, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_9 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_8);
        index = 0; __pyx_t_3 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_3)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_3);
        index = 1; __pyx_t_7 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_7);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 2) < (0)) __PYX_ERR(0, 1183, __pyx_L5_error)
        __pyx_t_9 = NULL;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L9_unpacking_done;
        __pyx_L8_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1183, __pyx_L5_error)
        __pyx_L9_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_f, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_ctx, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "cyjs/_cyjs.pyx":1184
 *         cdef list filtered_callbacks = [(f, ctx)
 *                               for (f, ctx) in self._callbacks
 *                               if f != fn]             # <<<<<<<<<<<<<<
 *         removed_count = <Py_ssize_t>(len(self._callbacks) - len(filtered_callbacks))
 *         if removed_count:
*/
      __pyx_t_4 = PyObject_RichCompare(__pyx_8genexpr1__pyx_v_f, __pyx_v_fn, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1184, __pyx_L5_error)
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1184, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_10) {

        /* "cyjs/_cyjs.pyx":1182
 *         """
 *         cdef Py_ssize_t removed_count
 *         cdef list filtered_callbacks = [(f, ctx)             # <<<<<<<<<<<<<<
 *                               for (f, ctx) in self._callbacks
 *                               if f != fn]
*/
        __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1182, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_8genexpr1__pyx_v_f);
        __Pyx_GIVEREF(__pyx_8genexpr1__pyx_v_f);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_8genexpr1__pyx_v_f) != (0)) __PYX_ERR(0, 1182, __pyx_L5_error);
        __Pyx_INCREF(__pyx_8genexpr1__pyx_v_ctx);
        __Pyx_GIVEREF(__pyx_8genexpr1__pyx_v_ctx);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_8genexpr1__pyx_v_ctx) != (0)) __PYX_ERR(0, 1182, __pyx_L5_error);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 1182, __pyx_L5_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "cyjs/_cyjs.pyx":1184
 *         cdef list filtered_callbacks = [(f, ctx)
 *                               for (f, ctx) in self._callbacks
 *                               if f != fn]             # <<<<<<<<<<<<<<
 *         removed_count = <Py_ssize_t>(len(self._callbacks) - len(filtered_callbacks))
 *         if removed_count:
*/
      }

      /* "cyjs/_cyjs.pyx":1183
 *         cdef Py_ssize_t removed_count
 *         cdef list filtered_callbacks = [(f, ctx)
 *                               for (f, ctx) in self._callbacks             # <<<<<<<<<<<<<<
 *                               if f != fn]
 *         removed_count = <Py_ssize_t>(len(self._callbacks) - len(filtered_callbacks))
*/
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_ctx); __pyx_8genexpr1__pyx_v_ctx = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_f); __pyx_8genexpr1__pyx_v_f = 0;
    goto __pyx_L12_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_ctx); __pyx_8genexpr1__pyx_v_ctx = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_f); __pyx_8genexpr1__pyx_v_f = 0;
    goto __pyx_L1_error;
    __pyx_L12_exit_scope:;
  } /* exit inner scope */
  __pyx_v_filtered_callbacks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1185
 *                               for (f, ctx) in self._callbacks
 *                               if f != fn]
 *         removed_count = <Py_ssize_t>(len(self._callbacks) - len(filtered_callbacks))             # <<<<<<<<<<<<<<
 *         if removed_count:
 *             self._callbacks[:] = filtered_callbacks
*/
  __pyx_t_1 = __pyx_v_self->_callbacks;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1185, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1185, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = __Pyx_PyList_GET_SIZE(__pyx_v_filtered_callbacks); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1185, __pyx_L1_error)
  __pyx_v_removed_count = ((Py_ssize_t)(__pyx_t_6 - __pyx_t_11));

  /* "cyjs/_cyjs.pyx":1186
 *                               if f != fn]
 *         removed_count = <Py_ssize_t>(len(self._callbacks) - len(filtered_callbacks))
 *         if removed_count:             # <<<<<<<<<<<<<<
 *             self._callbacks[:] = filtered_callbacks
 *         return removed_count
*/
  __pyx_t_10 = (__pyx_v_removed_count != 0);
  if (__pyx_t_10) {

    /* "cyjs/_cyjs.pyx":1187
 *         removed_count = <Py_ssize_t>(len(self._callbacks) - len(filtered_callbacks))
 *         if removed_count:
 *             self._callbacks[:] = filtered_callbacks             # <<<<<<<<<<<<<<
 *         return removed_count
 * 
*/
    if (unlikely(__pyx_v_self->_callbacks == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 1187, __pyx_L1_error)
    }
    if (__Pyx_PyObject_SetSlice(__pyx_v_self->_callbacks, __pyx_v_filtered_callbacks, 0, 0, NULL, NULL, NULL, 0, 0, 1) < (0)) __PYX_ERR(0, 1187, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1186
 *                               if f != fn]
 *         removed_count = <Py_ssize_t>(len(self._callbacks) - len(filtered_callbacks))
 *         if removed_count:             # <<<<<<<<<<<<<<
 *             self._callbacks[:] = filtered_callbacks
 *         return removed_count
*/
  }

  /* "cyjs/_cyjs.pyx":1188
 *         if removed_count:
 *             self._callbacks[:] = filtered_callbacks
 *         return removed_count             # <<<<<<<<<<<<<<
 * 
 *     cpdef object result(self):
*/
  __pyx_r = __pyx_v_removed_count;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1176
 * 
 * 
 *     cpdef Py_ssize_t remove_done_callback(self, object fn):             # <<<<<<<<<<<<<<
 *         """Remove all instances of a callback from the "call when done" list.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("cyjs._cyjs.Promise.remove_done_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_filtered_callbacks);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_f);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_ctx);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_7remove_done_callback(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4cyjs_5_cyjs_7Promise_6remove_done_callback, "Remove all instances of a callback from the \"call when done\" list.\n\n        Returns the number of callbacks removed.\n        ");
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Promise_7remove_done_callback = {"remove_done_callback", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Promise_7remove_done_callback, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4cyjs_5_cyjs_7Promise_6remove_done_callback};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_7remove_done_callback(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_fn = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove_done_callback (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_fn,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1176, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1176, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "remove_done_callback", 0) < (0)) __PYX_ERR(0, 1176, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("remove_done_callback", 1, 1, 1, i); __PYX_ERR(0, 1176, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1176, __pyx_L3_error)
    }
    __pyx_v_fn = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("remove_done_callback", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1176, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Promise.remove_done_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Promise_6remove_done_callback(((struct __pyx_obj_4cyjs_5_cyjs_Promise *)__pyx_v_self), __pyx_v_fn);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Promise_6remove_done_callback(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, PyObject *__pyx_v_fn) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("remove_done_callback", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_7Promise_remove_done_callback(__pyx_v_self, __pyx_v_fn, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1176, __pyx_L1_error)
  __pyx_t_2 = PyLong_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cyjs._cyjs.Promise.remove_done_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1190
 *         return removed_count
 * 
 *     cpdef object result(self):             # <<<<<<<<<<<<<<
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)
 *         if state == JS_PROMISE_REJECTED:
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_9result(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Promise_result(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, int __pyx_skip_dispatch) {
  JSPromiseStateEnum __pyx_v_state;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("result", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Promise &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_result); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Promise_9result)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1190, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":1191
 * 
 *     cpdef object result(self):
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)             # <<<<<<<<<<<<<<
 *         if state == JS_PROMISE_REJECTED:
 *             raise CancelledError()
*/
  __pyx_v_state = JS_PromiseState(__pyx_v_self->__pyx_base.ctx, __pyx_v_self->__pyx_base.value);

  /* "cyjs/_cyjs.pyx":1192
 *     cpdef object result(self):
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)
 *         if state == JS_PROMISE_REJECTED:             # <<<<<<<<<<<<<<
 *             raise CancelledError()
 *         if state == JS_PROMISE_FULFILLED:
*/
  __pyx_t_6 = (__pyx_v_state == JS_PROMISE_REJECTED);
  if (unlikely(__pyx_t_6)) {

    /* "cyjs/_cyjs.pyx":1193
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)
 *         if state == JS_PROMISE_REJECTED:
 *             raise CancelledError()             # <<<<<<<<<<<<<<
 *         if state == JS_PROMISE_FULFILLED:
 *             if self._exception is not None:
*/
    __pyx_t_2 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_CancelledError, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1193, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_1);
    }
    __Pyx_Raise(((PyObject *)__pyx_t_1), 0, 0, 0);
    __Pyx_DECREF((PyObject *)__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1193, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1192
 *     cpdef object result(self):
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)
 *         if state == JS_PROMISE_REJECTED:             # <<<<<<<<<<<<<<
 *             raise CancelledError()
 *         if state == JS_PROMISE_FULFILLED:
*/
  }

  /* "cyjs/_cyjs.pyx":1194
 *         if state == JS_PROMISE_REJECTED:
 *             raise CancelledError()
 *         if state == JS_PROMISE_FULFILLED:             # <<<<<<<<<<<<<<
 *             if self._exception is not None:
 *                 raise self._exception
*/
  __pyx_t_6 = (__pyx_v_state == JS_PROMISE_FULFILLED);
  if (likely(__pyx_t_6)) {

    /* "cyjs/_cyjs.pyx":1195
 *             raise CancelledError()
 *         if state == JS_PROMISE_FULFILLED:
 *             if self._exception is not None:             # <<<<<<<<<<<<<<
 *                 raise self._exception
 *             return self._result
*/
    __pyx_t_6 = (__pyx_v_self->_exception != Py_None);
    if (unlikely(__pyx_t_6)) {

      /* "cyjs/_cyjs.pyx":1196
 *         if state == JS_PROMISE_FULFILLED:
 *             if self._exception is not None:
 *                 raise self._exception             # <<<<<<<<<<<<<<
 *             return self._result
 *         else:
*/
      __Pyx_Raise(__pyx_v_self->_exception, 0, 0, 0);
      __PYX_ERR(0, 1196, __pyx_L1_error)

      /* "cyjs/_cyjs.pyx":1195
 *             raise CancelledError()
 *         if state == JS_PROMISE_FULFILLED:
 *             if self._exception is not None:             # <<<<<<<<<<<<<<
 *                 raise self._exception
 *             return self._result
*/
    }

    /* "cyjs/_cyjs.pyx":1197
 *             if self._exception is not None:
 *                 raise self._exception
 *             return self._result             # <<<<<<<<<<<<<<
 *         else:
 *             raise InvalidStateError('Result is not ready.')
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_self->_result);
    __pyx_r = __pyx_v_self->_result;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":1194
 *         if state == JS_PROMISE_REJECTED:
 *             raise CancelledError()
 *         if state == JS_PROMISE_FULFILLED:             # <<<<<<<<<<<<<<
 *             if self._exception is not None:
 *                 raise self._exception
*/
  }

  /* "cyjs/_cyjs.pyx":1199
 *             return self._result
 *         else:
 *             raise InvalidStateError('Result is not ready.')             # <<<<<<<<<<<<<<
 * 
 *     cdef object complete(self):
*/
  /*else*/ {
    __pyx_t_2 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Result_is_not_ready};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1199, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_1);
    }
    __Pyx_Raise(((PyObject *)__pyx_t_1), 0, 0, 0);
    __Pyx_DECREF((PyObject *)__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1199, __pyx_L1_error)
  }

  /* "cyjs/_cyjs.pyx":1190
 *         return removed_count
 * 
 *     cpdef object result(self):             # <<<<<<<<<<<<<<
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)
 *         if state == JS_PROMISE_REJECTED:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Promise.result", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_9result(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Promise_9result = {"result", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Promise_9result, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_9result(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("result (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("result", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("result", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Promise_8result(((struct __pyx_obj_4cyjs_5_cyjs_Promise *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Promise_8result(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("result", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_7Promise_result(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Promise.result", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1201
 *             raise InvalidStateError('Result is not ready.')
 * 
 *     cdef object complete(self):             # <<<<<<<<<<<<<<
 *         cdef JSContext* ctx = self.ctx
 *         # check if we can skip this step...
*/

static PyObject *__pyx_f_4cyjs_5_cyjs_7Promise_complete(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self) {
  JSContext *__pyx_v_ctx;
  JSValue __pyx_v_v;
  PyObject *__pyx_v_cb = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  JSContext *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("complete", 0);

  /* "cyjs/_cyjs.pyx":1202
 * 
 *     cdef object complete(self):
 *         cdef JSContext* ctx = self.ctx             # <<<<<<<<<<<<<<
 *         # check if we can skip this step...
 *         if self.completed == True:
*/
  __pyx_t_1 = __pyx_v_self->__pyx_base.ctx;
  __pyx_v_ctx = __pyx_t_1;

  /* "cyjs/_cyjs.pyx":1204
 *         cdef JSContext* ctx = self.ctx
 *         # check if we can skip this step...
 *         if self.completed == True:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
  __pyx_t_2 = (__pyx_v_self->completed == 1);
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":1205
 *         # check if we can skip this step...
 *         if self.completed == True:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         v = JS_PromiseResult(ctx, self.value)
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":1204
 *         cdef JSContext* ctx = self.ctx
 *         # check if we can skip this step...
 *         if self.completed == True:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":1207
 *             return
 * 
 *         v = JS_PromiseResult(ctx, self.value)             # <<<<<<<<<<<<<<
 *         if JS_IsException(v):
 *             self._exception = JSError.new(ctx, v)
*/
  __pyx_v_v = JS_PromiseResult(__pyx_v_ctx, __pyx_v_self->__pyx_base.value);

  /* "cyjs/_cyjs.pyx":1208
 * 
 *         v = JS_PromiseResult(ctx, self.value)
 *         if JS_IsException(v):             # <<<<<<<<<<<<<<
 *             self._exception = JSError.new(ctx, v)
 *         else:
*/
  __pyx_t_2 = JS_IsException(__pyx_v_v);
  if (__pyx_t_2) {

    /* "cyjs/_cyjs.pyx":1209
 *         v = JS_PromiseResult(ctx, self.value)
 *         if JS_IsException(v):
 *             self._exception = JSError.new(ctx, v)             # <<<<<<<<<<<<<<
 *         else:
 *             self._result = to_python(ctx, v)
*/
    __pyx_t_3 = ((PyObject *)__pyx_f_4cyjs_5_cyjs_7JSError_new(__pyx_v_ctx, __pyx_v_v)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->_exception);
    __Pyx_DECREF(__pyx_v_self->_exception);
    __pyx_v_self->_exception = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cyjs/_cyjs.pyx":1208
 * 
 *         v = JS_PromiseResult(ctx, self.value)
 *         if JS_IsException(v):             # <<<<<<<<<<<<<<
 *             self._exception = JSError.new(ctx, v)
 *         else:
*/
    goto __pyx_L4;
  }

  /* "cyjs/_cyjs.pyx":1211
 *             self._exception = JSError.new(ctx, v)
 *         else:
 *             self._result = to_python(ctx, v)             # <<<<<<<<<<<<<<
 *         for cb in self._callbacks:
 *             cb(self)
*/
  /*else*/ {
    __pyx_t_3 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_ctx, __pyx_v_v); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1211, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->_result);
    __Pyx_DECREF(__pyx_v_self->_result);
    __pyx_v_self->_result = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_L4:;

  /* "cyjs/_cyjs.pyx":1212
 *         else:
 *             self._result = to_python(ctx, v)
 *         for cb in self._callbacks:             # <<<<<<<<<<<<<<
 *             cb(self)
 *         self.completed = True
*/
  if (unlikely(__pyx_v_self->_callbacks == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1212, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_v_self->_callbacks; __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
      #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1212, __pyx_L1_error)
      #endif
      if (__pyx_t_4 >= __pyx_temp) break;
    }
    __pyx_t_5 = __Pyx_PyList_GetItemRefFast(__pyx_t_3, __pyx_t_4, __Pyx_ReferenceSharing_OwnStrongReference);
    ++__pyx_t_4;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_cb, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "cyjs/_cyjs.pyx":1213
 *             self._result = to_python(ctx, v)
 *         for cb in self._callbacks:
 *             cb(self)             # <<<<<<<<<<<<<<
 *         self.completed = True
 * 
*/
    __pyx_t_6 = NULL;
    __Pyx_INCREF(__pyx_v_cb);
    __pyx_t_7 = __pyx_v_cb; 
    __pyx_t_8 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      assert(__pyx_t_6);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
      __pyx_t_8 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, ((PyObject *)__pyx_v_self)};
      __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1213, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "cyjs/_cyjs.pyx":1212
 *         else:
 *             self._result = to_python(ctx, v)
 *         for cb in self._callbacks:             # <<<<<<<<<<<<<<
 *             cb(self)
 *         self.completed = True
*/
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "cyjs/_cyjs.pyx":1214
 *         for cb in self._callbacks:
 *             cb(self)
 *         self.completed = True             # <<<<<<<<<<<<<<
 * 
 *     # TODO: might be able to utilize __await__ and __iter__ for polling
*/
  __pyx_v_self->completed = 1;

  /* "cyjs/_cyjs.pyx":1201
 *             raise InvalidStateError('Result is not ready.')
 * 
 *     cdef object complete(self):             # <<<<<<<<<<<<<<
 *         cdef JSContext* ctx = self.ctx
 *         # check if we can skip this step...
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cyjs._cyjs.Promise.complete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cb);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1218
 *     # TODO: might be able to utilize __await__ and __iter__ for polling
 * 
 *     cpdef object poll(self):             # <<<<<<<<<<<<<<
 *         """Polls QuickJS Eventloop a single cycle while attempting
 *         to wait for this Promise to complete"""
*/

static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_11poll(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_4cyjs_5_cyjs_7Promise_poll(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, int __pyx_skip_dispatch) {
  JSContext *__pyx_v_ctx;
  JSRuntime *__pyx_v_rt;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  JSContext *__pyx_t_6;
  JSRuntime *__pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("poll", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Promise &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_poll); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_4cyjs_5_cyjs_7Promise_11poll)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1218, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyjs/_cyjs.pyx":1222
 *         to wait for this Promise to complete"""
 *         cdef JSValue v
 *         cdef JSContext* ctx = self.ctx             # <<<<<<<<<<<<<<
 *         cdef JSRuntime* rt = self.context.rt
 * 
*/
  __pyx_t_6 = __pyx_v_self->__pyx_base.ctx;
  __pyx_v_ctx = __pyx_t_6;

  /* "cyjs/_cyjs.pyx":1223
 *         cdef JSValue v
 *         cdef JSContext* ctx = self.ctx
 *         cdef JSRuntime* rt = self.context.rt             # <<<<<<<<<<<<<<
 * 
 *         if self.done():
*/
  __pyx_t_7 = __pyx_v_self->__pyx_base.context->rt;
  __pyx_v_rt = __pyx_t_7;

  /* "cyjs/_cyjs.pyx":1225
 *         cdef JSRuntime* rt = self.context.rt
 * 
 *         if self.done():             # <<<<<<<<<<<<<<
 *             self.complete()
 *             return True
*/
  __pyx_t_8 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Promise *)__pyx_v_self->__pyx_base.__pyx_vtab)->done(__pyx_v_self, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1225, __pyx_L1_error)
  if (__pyx_t_8) {

    /* "cyjs/_cyjs.pyx":1226
 * 
 *         if self.done():
 *             self.complete()             # <<<<<<<<<<<<<<
 *             return True
 * 
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Promise *)__pyx_v_self->__pyx_base.__pyx_vtab)->complete(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "cyjs/_cyjs.pyx":1227
 *         if self.done():
 *             self.complete()
 *             return True             # <<<<<<<<<<<<<<
 * 
 *         if JS_IsJobPending(rt):
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":1225
 *         cdef JSRuntime* rt = self.context.rt
 * 
 *         if self.done():             # <<<<<<<<<<<<<<
 *             self.complete()
 *             return True
*/
  }

  /* "cyjs/_cyjs.pyx":1229
 *             return True
 * 
 *         if JS_IsJobPending(rt):             # <<<<<<<<<<<<<<
 *             if JS_ExecutePendingJob(rt, &ctx) < 0:
 *                 self.context.raise_exception()
*/
  __pyx_t_8 = JS_IsJobPending(__pyx_v_rt);
  if (__pyx_t_8) {

    /* "cyjs/_cyjs.pyx":1230
 * 
 *         if JS_IsJobPending(rt):
 *             if JS_ExecutePendingJob(rt, &ctx) < 0:             # <<<<<<<<<<<<<<
 *                 self.context.raise_exception()
 * 
*/
    __pyx_t_8 = (JS_ExecutePendingJob(__pyx_v_rt, (&__pyx_v_ctx)) < 0);
    if (__pyx_t_8) {

      /* "cyjs/_cyjs.pyx":1231
 *         if JS_IsJobPending(rt):
 *             if JS_ExecutePendingJob(rt, &ctx) < 0:
 *                 self.context.raise_exception()             # <<<<<<<<<<<<<<
 * 
 *         if self.done():
*/
      __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)__pyx_v_self->__pyx_base.context->__pyx_vtab)->raise_exception(__pyx_v_self->__pyx_base.context); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1231, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "cyjs/_cyjs.pyx":1230
 * 
 *         if JS_IsJobPending(rt):
 *             if JS_ExecutePendingJob(rt, &ctx) < 0:             # <<<<<<<<<<<<<<
 *                 self.context.raise_exception()
 * 
*/
    }

    /* "cyjs/_cyjs.pyx":1229
 *             return True
 * 
 *         if JS_IsJobPending(rt):             # <<<<<<<<<<<<<<
 *             if JS_ExecutePendingJob(rt, &ctx) < 0:
 *                 self.context.raise_exception()
*/
  }

  /* "cyjs/_cyjs.pyx":1233
 *                 self.context.raise_exception()
 * 
 *         if self.done():             # <<<<<<<<<<<<<<
 *             self.complete()
 *             return True
*/
  __pyx_t_8 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Promise *)__pyx_v_self->__pyx_base.__pyx_vtab)->done(__pyx_v_self, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1233, __pyx_L1_error)
  if (__pyx_t_8) {

    /* "cyjs/_cyjs.pyx":1234
 * 
 *         if self.done():
 *             self.complete()             # <<<<<<<<<<<<<<
 *             return True
 *         else:
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Promise *)__pyx_v_self->__pyx_base.__pyx_vtab)->complete(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "cyjs/_cyjs.pyx":1235
 *         if self.done():
 *             self.complete()
 *             return True             # <<<<<<<<<<<<<<
 *         else:
 *             return False
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":1233
 *                 self.context.raise_exception()
 * 
 *         if self.done():             # <<<<<<<<<<<<<<
 *             self.complete()
 *             return True
*/
  }

  /* "cyjs/_cyjs.pyx":1237
 *             return True
 *         else:
 *             return False             # <<<<<<<<<<<<<<
 * 
 * 
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;
  }

  /* "cyjs/_cyjs.pyx":1218
 *     # TODO: might be able to utilize __await__ and __iter__ for polling
 * 
 *     cpdef object poll(self):             # <<<<<<<<<<<<<<
 *         """Polls QuickJS Eventloop a single cycle while attempting
 *         to wait for this Promise to complete"""
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyjs._cyjs.Promise.poll", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_11poll(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4cyjs_5_cyjs_7Promise_10poll, "Polls QuickJS Eventloop a single cycle while attempting\n        to wait for this Promise to complete");
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Promise_11poll = {"poll", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Promise_11poll, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4cyjs_5_cyjs_7Promise_10poll};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_11poll(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("poll (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("poll", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("poll", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Promise_10poll(((struct __pyx_obj_4cyjs_5_cyjs_Promise *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Promise_10poll(struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("poll", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_7Promise_poll(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.Promise.poll", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_13__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Promise_13__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Promise_13__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_13__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Promise_12__reduce_cython__(((struct __pyx_obj_4cyjs_5_cyjs_Promise *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Promise_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ctx_self_value_cannot_be_co, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.Promise.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_15__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7Promise_15__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Promise_15__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7Promise_15__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.Promise.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_7Promise_14__setstate_cython__(((struct __pyx_obj_4cyjs_5_cyjs_Promise *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_7Promise_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ctx_self_value_cannot_be_co, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.Promise.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_9NODEFAULT_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_9NODEFAULT_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_9NODEFAULT_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_9NODEFAULT_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_9NODEFAULT___reduce_cython__(((struct __pyx_obj_4cyjs_5_cyjs_NODEFAULT *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_9NODEFAULT___reduce_cython__(struct __pyx_obj_4cyjs_5_cyjs_NODEFAULT *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_empty_tuple);
  __pyx_v_state = __pyx_mstate_global->__pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(1, 7, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_NODEFAULT, (type(self), 0xe3b0c44, None), state
*/
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_NODEFAULT, (type(self), 0xe3b0c44, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_NODEFAULT, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_NODEFAULT, (type(self), 0xe3b0c44, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_NODEFAULT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_238750788) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_NODEFAULT, (type(self), 0xe3b0c44, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_NODEFAULT, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_NODEFAULT, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_NODEFAULT__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_NODEFAULT); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_238750788) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cyjs._cyjs.NODEFAULT.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NODEFAULT, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NODEFAULT__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_9NODEFAULT_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_9NODEFAULT_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_9NODEFAULT_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_9NODEFAULT_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.NODEFAULT.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_9NODEFAULT_2__setstate_cython__(((struct __pyx_obj_4cyjs_5_cyjs_NODEFAULT *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_9NODEFAULT_2__setstate_cython__(struct __pyx_obj_4cyjs_5_cyjs_NODEFAULT *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_NODEFAULT, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_NODEFAULT__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(1, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_4cyjs_5_cyjs___pyx_unpickle_NODEFAULT__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NODEFAULT, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NODEFAULT__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cyjs._cyjs.NODEFAULT.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1247
 *     pass
 * 
 * cdef Object jsv_to_object(JSContext* ctx, JSValue value):             # <<<<<<<<<<<<<<
 *     cdef Object v = Object.__new__(Object)
 *     v.init(<Context>JS_GetContextOpaque(ctx), value)
*/

static struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_f_4cyjs_5_cyjs_jsv_to_object(JSContext *__pyx_v_ctx, JSValue __pyx_v_value) {
  struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_v_v = 0;
  struct __pyx_obj_4cyjs_5_cyjs_Object *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  void *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("jsv_to_object", 0);

  /* "cyjs/_cyjs.pyx":1248
 * 
 * cdef Object jsv_to_object(JSContext* ctx, JSValue value):
 *     cdef Object v = Object.__new__(Object)             # <<<<<<<<<<<<<<
 *     v.init(<Context>JS_GetContextOpaque(ctx), value)
 *     return v
*/
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_4cyjs_5_cyjs_Object(((PyTypeObject *)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object), __pyx_mstate_global->__pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_v = ((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1249
 * cdef Object jsv_to_object(JSContext* ctx, JSValue value):
 *     cdef Object v = Object.__new__(Object)
 *     v.init(<Context>JS_GetContextOpaque(ctx), value)             # <<<<<<<<<<<<<<
 *     return v
 * 
*/
  __pyx_t_2 = JS_GetContextOpaque(__pyx_v_ctx);
  ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Object *)__pyx_v_v->__pyx_vtab)->init(__pyx_v_v, ((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_t_2), __pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1249, __pyx_L1_error)

  /* "cyjs/_cyjs.pyx":1250
 *     cdef Object v = Object.__new__(Object)
 *     v.init(<Context>JS_GetContextOpaque(ctx), value)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * cdef Promise jsv_to_promise(JSContext* ctx, JSValue value):
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_v);
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1247
 *     pass
 * 
 * cdef Object jsv_to_object(JSContext* ctx, JSValue value):             # <<<<<<<<<<<<<<
 *     cdef Object v = Object.__new__(Object)
 *     v.init(<Context>JS_GetContextOpaque(ctx), value)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.jsv_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_v);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1252
 *     return v
 * 
 * cdef Promise jsv_to_promise(JSContext* ctx, JSValue value):             # <<<<<<<<<<<<<<
 *     cdef Promise v = Promise.__new__(Promise)
 *     v.init(<Context>JS_GetContextOpaque(ctx), value)
*/

static struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_f_4cyjs_5_cyjs_jsv_to_promise(JSContext *__pyx_v_ctx, JSValue __pyx_v_value) {
  struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_v_v = 0;
  struct __pyx_obj_4cyjs_5_cyjs_Promise *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  void *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("jsv_to_promise", 0);

  /* "cyjs/_cyjs.pyx":1253
 * 
 * cdef Promise jsv_to_promise(JSContext* ctx, JSValue value):
 *     cdef Promise v = Promise.__new__(Promise)             # <<<<<<<<<<<<<<
 *     v.init(<Context>JS_GetContextOpaque(ctx), value)
 *     v._callbacks = []
*/
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_4cyjs_5_cyjs_Promise(((PyTypeObject *)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Promise), __pyx_mstate_global->__pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1253, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_v = ((struct __pyx_obj_4cyjs_5_cyjs_Promise *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1254
 * cdef Promise jsv_to_promise(JSContext* ctx, JSValue value):
 *     cdef Promise v = Promise.__new__(Promise)
 *     v.init(<Context>JS_GetContextOpaque(ctx), value)             # <<<<<<<<<<<<<<
 *     v._callbacks = []
 *     v._result = NODEFAULT()
*/
  __pyx_t_2 = JS_GetContextOpaque(__pyx_v_ctx);
  ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Promise *)__pyx_v_v->__pyx_base.__pyx_vtab)->__pyx_base.init(((struct __pyx_obj_4cyjs_5_cyjs_Object *)__pyx_v_v), ((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_t_2), __pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1254, __pyx_L1_error)

  /* "cyjs/_cyjs.pyx":1255
 *     cdef Promise v = Promise.__new__(Promise)
 *     v.init(<Context>JS_GetContextOpaque(ctx), value)
 *     v._callbacks = []             # <<<<<<<<<<<<<<
 *     v._result = NODEFAULT()
 *     v._exception = None
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_v->_callbacks);
  __Pyx_DECREF(__pyx_v_v->_callbacks);
  __pyx_v_v->_callbacks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1256
 *     v.init(<Context>JS_GetContextOpaque(ctx), value)
 *     v._callbacks = []
 *     v._result = NODEFAULT()             # <<<<<<<<<<<<<<
 *     v._exception = None
 *     v.completed = False
*/
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_NODEFAULT, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1256, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __Pyx_GIVEREF((PyObject *)__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_v->_result);
  __Pyx_DECREF(__pyx_v_v->_result);
  __pyx_v_v->_result = ((PyObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyjs/_cyjs.pyx":1257
 *     v._callbacks = []
 *     v._result = NODEFAULT()
 *     v._exception = None             # <<<<<<<<<<<<<<
 *     v.completed = False
 *     return v
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_v->_exception);
  __Pyx_DECREF(__pyx_v_v->_exception);
  __pyx_v_v->_exception = Py_None;

  /* "cyjs/_cyjs.pyx":1258
 *     v._result = NODEFAULT()
 *     v._exception = None
 *     v.completed = False             # <<<<<<<<<<<<<<
 *     return v
 * 
*/
  __pyx_v_v->completed = 0;

  /* "cyjs/_cyjs.pyx":1259
 *     v._exception = None
 *     v.completed = False
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_v);
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1252
 *     return v
 * 
 * cdef Promise jsv_to_promise(JSContext* ctx, JSValue value):             # <<<<<<<<<<<<<<
 *     cdef Promise v = Promise.__new__(Promise)
 *     v.init(<Context>JS_GetContextOpaque(ctx), value)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.jsv_to_promise", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_v);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1262
 * 
 * 
 * cdef object to_python(JSContext* ctx, JSValue value):             # <<<<<<<<<<<<<<
 *     cdef int32_t tag = JS_VALUE_GET_TAG(value)
 *     cdef object ret
*/

static PyObject *__pyx_f_4cyjs_5_cyjs_to_python(JSContext *__pyx_v_ctx, JSValue __pyx_v_value) {
  int32_t __pyx_v_tag;
  PyObject *__pyx_v_ret = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  void *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_python", 0);

  /* "cyjs/_cyjs.pyx":1263
 * 
 * cdef object to_python(JSContext* ctx, JSValue value):
 *     cdef int32_t tag = JS_VALUE_GET_TAG(value)             # <<<<<<<<<<<<<<
 *     cdef object ret
 * 
*/
  __pyx_v_tag = JS_VALUE_GET_TAG(__pyx_v_value);

  /* "cyjs/_cyjs.pyx":1266
 *     cdef object ret
 * 
 *     if tag == JS_TAG_EXCEPTION:             # <<<<<<<<<<<<<<
 *         (<Context>JS_GetContextOpaque(ctx)).raise_exception()
 *         raise
*/
  __pyx_t_1 = (__pyx_v_tag == JS_TAG_EXCEPTION);
  if (unlikely(__pyx_t_1)) {

    /* "cyjs/_cyjs.pyx":1267
 * 
 *     if tag == JS_TAG_EXCEPTION:
 *         (<Context>JS_GetContextOpaque(ctx)).raise_exception()             # <<<<<<<<<<<<<<
 *         raise
 * 
*/
    __pyx_t_2 = JS_GetContextOpaque(__pyx_v_ctx);
    __pyx_t_3 = ((struct __pyx_vtabstruct_4cyjs_5_cyjs_Context *)((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_t_2)->__pyx_vtab)->raise_exception(((struct __pyx_obj_4cyjs_5_cyjs_Context *)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "cyjs/_cyjs.pyx":1268
 *     if tag == JS_TAG_EXCEPTION:
 *         (<Context>JS_GetContextOpaque(ctx)).raise_exception()
 *         raise             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 1268, __pyx_L1_error)

    /* "cyjs/_cyjs.pyx":1266
 *     cdef object ret
 * 
 *     if tag == JS_TAG_EXCEPTION:             # <<<<<<<<<<<<<<
 *         (<Context>JS_GetContextOpaque(ctx)).raise_exception()
 *         raise
*/
  }

  /* "cyjs/_cyjs.pyx":1271
 * 
 * 
 *     elif tag == JS_TAG_MODULE or tag == JS_TAG_OBJECT or tag == JS_TAG_SYMBOL:             # <<<<<<<<<<<<<<
 *         if JS_IsPromise(value):
 *             return jsv_to_promise(ctx, value)
*/
  switch (__pyx_v_tag) {
    case JS_TAG_MODULE:
    case JS_TAG_OBJECT:
    case JS_TAG_SYMBOL:
    __pyx_t_1 = 1;
    break;
    default:
    __pyx_t_1 = 0;
    break;
  }
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":1272
 * 
 *     elif tag == JS_TAG_MODULE or tag == JS_TAG_OBJECT or tag == JS_TAG_SYMBOL:
 *         if JS_IsPromise(value):             # <<<<<<<<<<<<<<
 *             return jsv_to_promise(ctx, value)
 *         return jsv_to_object(ctx, value)
*/
    __pyx_t_1 = JS_IsPromise(__pyx_v_value);
    if (__pyx_t_1) {

      /* "cyjs/_cyjs.pyx":1273
 *     elif tag == JS_TAG_MODULE or tag == JS_TAG_OBJECT or tag == JS_TAG_SYMBOL:
 *         if JS_IsPromise(value):
 *             return jsv_to_promise(ctx, value)             # <<<<<<<<<<<<<<
 *         return jsv_to_object(ctx, value)
 *         # pass
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = ((PyObject *)__pyx_f_4cyjs_5_cyjs_jsv_to_promise(__pyx_v_ctx, __pyx_v_value)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "cyjs/_cyjs.pyx":1272
 * 
 *     elif tag == JS_TAG_MODULE or tag == JS_TAG_OBJECT or tag == JS_TAG_SYMBOL:
 *         if JS_IsPromise(value):             # <<<<<<<<<<<<<<
 *             return jsv_to_promise(ctx, value)
 *         return jsv_to_object(ctx, value)
*/
    }

    /* "cyjs/_cyjs.pyx":1274
 *         if JS_IsPromise(value):
 *             return jsv_to_promise(ctx, value)
 *         return jsv_to_object(ctx, value)             # <<<<<<<<<<<<<<
 *         # pass
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((PyObject *)__pyx_f_4cyjs_5_cyjs_jsv_to_object(__pyx_v_ctx, __pyx_v_value)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":1271
 * 
 * 
 *     elif tag == JS_TAG_MODULE or tag == JS_TAG_OBJECT or tag == JS_TAG_SYMBOL:             # <<<<<<<<<<<<<<
 *         if JS_IsPromise(value):
 *             return jsv_to_promise(ctx, value)
*/
  }

  /* "cyjs/_cyjs.pyx":1277
 *         # pass
 * 
 *     elif tag == JS_TAG_INT:             # <<<<<<<<<<<<<<
 *         ret = JS_VALUE_GET_INT(value)
 *         JS_FreeValue(ctx, value)
*/
  __pyx_t_1 = (__pyx_v_tag == JS_TAG_INT);
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":1278
 * 
 *     elif tag == JS_TAG_INT:
 *         ret = JS_VALUE_GET_INT(value)             # <<<<<<<<<<<<<<
 *         JS_FreeValue(ctx, value)
 *         return ret
*/
    __pyx_t_3 = __Pyx_PyLong_From_int32_t(JS_VALUE_GET_INT(__pyx_v_value)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_ret = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cyjs/_cyjs.pyx":1279
 *     elif tag == JS_TAG_INT:
 *         ret = JS_VALUE_GET_INT(value)
 *         JS_FreeValue(ctx, value)             # <<<<<<<<<<<<<<
 *         return ret
 *     elif tag == JS_TAG_BIG_INT:
*/
    JS_FreeValue(__pyx_v_ctx, __pyx_v_value);

    /* "cyjs/_cyjs.pyx":1280
 *         ret = JS_VALUE_GET_INT(value)
 *         JS_FreeValue(ctx, value)
 *         return ret             # <<<<<<<<<<<<<<
 *     elif tag == JS_TAG_BIG_INT:
 *         ret = jsv_to_big_int(ctx, value)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_ret);
    __pyx_r = __pyx_v_ret;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":1277
 *         # pass
 * 
 *     elif tag == JS_TAG_INT:             # <<<<<<<<<<<<<<
 *         ret = JS_VALUE_GET_INT(value)
 *         JS_FreeValue(ctx, value)
*/
  }

  /* "cyjs/_cyjs.pyx":1281
 *         JS_FreeValue(ctx, value)
 *         return ret
 *     elif tag == JS_TAG_BIG_INT:             # <<<<<<<<<<<<<<
 *         ret = jsv_to_big_int(ctx, value)
 *         JS_FreeValue(ctx, value)
*/
  __pyx_t_1 = (__pyx_v_tag == JS_TAG_BIG_INT);
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":1282
 *         return ret
 *     elif tag == JS_TAG_BIG_INT:
 *         ret = jsv_to_big_int(ctx, value)             # <<<<<<<<<<<<<<
 *         JS_FreeValue(ctx, value)
 *         return ret
*/
    __pyx_t_3 = __pyx_f_4cyjs_5_cyjs_jsv_to_big_int(__pyx_v_ctx, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_ret = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cyjs/_cyjs.pyx":1283
 *     elif tag == JS_TAG_BIG_INT:
 *         ret = jsv_to_big_int(ctx, value)
 *         JS_FreeValue(ctx, value)             # <<<<<<<<<<<<<<
 *         return ret
 *     elif tag == JS_TAG_NULL and tag == JS_TAG_UNDEFINED:
*/
    JS_FreeValue(__pyx_v_ctx, __pyx_v_value);

    /* "cyjs/_cyjs.pyx":1284
 *         ret = jsv_to_big_int(ctx, value)
 *         JS_FreeValue(ctx, value)
 *         return ret             # <<<<<<<<<<<<<<
 *     elif tag == JS_TAG_NULL and tag == JS_TAG_UNDEFINED:
 *         JS_FreeValue(ctx, value)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_ret);
    __pyx_r = __pyx_v_ret;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":1281
 *         JS_FreeValue(ctx, value)
 *         return ret
 *     elif tag == JS_TAG_BIG_INT:             # <<<<<<<<<<<<<<
 *         ret = jsv_to_big_int(ctx, value)
 *         JS_FreeValue(ctx, value)
*/
  }

  /* "cyjs/_cyjs.pyx":1285
 *         JS_FreeValue(ctx, value)
 *         return ret
 *     elif tag == JS_TAG_NULL and tag == JS_TAG_UNDEFINED:             # <<<<<<<<<<<<<<
 *         JS_FreeValue(ctx, value)
 *         return None
*/
  switch (__pyx_v_tag) {
    case JS_TAG_NULL:
    case JS_TAG_UNDEFINED:
    __pyx_t_1 = 1;
    break;
    default:
    __pyx_t_1 = 0;
    break;
  }
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":1286
 *         return ret
 *     elif tag == JS_TAG_NULL and tag == JS_TAG_UNDEFINED:
 *         JS_FreeValue(ctx, value)             # <<<<<<<<<<<<<<
 *         return None
 *     elif tag == JS_TAG_STRING:
*/
    JS_FreeValue(__pyx_v_ctx, __pyx_v_value);

    /* "cyjs/_cyjs.pyx":1287
 *     elif tag == JS_TAG_NULL and tag == JS_TAG_UNDEFINED:
 *         JS_FreeValue(ctx, value)
 *         return None             # <<<<<<<<<<<<<<
 *     elif tag == JS_TAG_STRING:
 *         ret = jsv_to_str(ctx, value)
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":1285
 *         JS_FreeValue(ctx, value)
 *         return ret
 *     elif tag == JS_TAG_NULL and tag == JS_TAG_UNDEFINED:             # <<<<<<<<<<<<<<
 *         JS_FreeValue(ctx, value)
 *         return None
*/
  }

  /* "cyjs/_cyjs.pyx":1288
 *         JS_FreeValue(ctx, value)
 *         return None
 *     elif tag == JS_TAG_STRING:             # <<<<<<<<<<<<<<
 *         ret = jsv_to_str(ctx, value)
 *         JS_FreeValue(ctx, value)
*/
  __pyx_t_1 = (__pyx_v_tag == JS_TAG_STRING);
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":1289
 *         return None
 *     elif tag == JS_TAG_STRING:
 *         ret = jsv_to_str(ctx, value)             # <<<<<<<<<<<<<<
 *         JS_FreeValue(ctx, value)
 *         return ret
*/
    __pyx_t_3 = __pyx_f_4cyjs_5_cyjs_jsv_to_str(__pyx_v_ctx, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_ret = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cyjs/_cyjs.pyx":1290
 *     elif tag == JS_TAG_STRING:
 *         ret = jsv_to_str(ctx, value)
 *         JS_FreeValue(ctx, value)             # <<<<<<<<<<<<<<
 *         return ret
 *     elif tag == JS_TAG_FLOAT64:
*/
    JS_FreeValue(__pyx_v_ctx, __pyx_v_value);

    /* "cyjs/_cyjs.pyx":1291
 *         ret = jsv_to_str(ctx, value)
 *         JS_FreeValue(ctx, value)
 *         return ret             # <<<<<<<<<<<<<<
 *     elif tag == JS_TAG_FLOAT64:
 *         ret = JS_VALUE_GET_FLOAT64(value)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_ret);
    __pyx_r = __pyx_v_ret;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":1288
 *         JS_FreeValue(ctx, value)
 *         return None
 *     elif tag == JS_TAG_STRING:             # <<<<<<<<<<<<<<
 *         ret = jsv_to_str(ctx, value)
 *         JS_FreeValue(ctx, value)
*/
  }

  /* "cyjs/_cyjs.pyx":1292
 *         JS_FreeValue(ctx, value)
 *         return ret
 *     elif tag == JS_TAG_FLOAT64:             # <<<<<<<<<<<<<<
 *         ret = JS_VALUE_GET_FLOAT64(value)
 *         JS_FreeValue(ctx, value)
*/
  __pyx_t_1 = (__pyx_v_tag == JS_TAG_FLOAT64);
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":1293
 *         return ret
 *     elif tag == JS_TAG_FLOAT64:
 *         ret = JS_VALUE_GET_FLOAT64(value)             # <<<<<<<<<<<<<<
 *         JS_FreeValue(ctx, value)
 *         return ret
*/
    __pyx_t_3 = PyFloat_FromDouble(JS_VALUE_GET_FLOAT64(__pyx_v_value)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_ret = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cyjs/_cyjs.pyx":1294
 *     elif tag == JS_TAG_FLOAT64:
 *         ret = JS_VALUE_GET_FLOAT64(value)
 *         JS_FreeValue(ctx, value)             # <<<<<<<<<<<<<<
 *         return ret
 *     elif tag == JS_TAG_BOOL:
*/
    JS_FreeValue(__pyx_v_ctx, __pyx_v_value);

    /* "cyjs/_cyjs.pyx":1295
 *         ret = JS_VALUE_GET_FLOAT64(value)
 *         JS_FreeValue(ctx, value)
 *         return ret             # <<<<<<<<<<<<<<
 *     elif tag == JS_TAG_BOOL:
 *         ret = PyBool_FromLong(JS_VALUE_GET_INT(value))
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_ret);
    __pyx_r = __pyx_v_ret;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":1292
 *         JS_FreeValue(ctx, value)
 *         return ret
 *     elif tag == JS_TAG_FLOAT64:             # <<<<<<<<<<<<<<
 *         ret = JS_VALUE_GET_FLOAT64(value)
 *         JS_FreeValue(ctx, value)
*/
  }

  /* "cyjs/_cyjs.pyx":1296
 *         JS_FreeValue(ctx, value)
 *         return ret
 *     elif tag == JS_TAG_BOOL:             # <<<<<<<<<<<<<<
 *         ret = PyBool_FromLong(JS_VALUE_GET_INT(value))
 *         JS_FreeValue(ctx, value)
*/
  __pyx_t_1 = (__pyx_v_tag == JS_TAG_BOOL);
  if (__pyx_t_1) {

    /* "cyjs/_cyjs.pyx":1297
 *         return ret
 *     elif tag == JS_TAG_BOOL:
 *         ret = PyBool_FromLong(JS_VALUE_GET_INT(value))             # <<<<<<<<<<<<<<
 *         JS_FreeValue(ctx, value)
 *         return ret
*/
    __pyx_t_3 = PyBool_FromLong(JS_VALUE_GET_INT(__pyx_v_value)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_ret = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cyjs/_cyjs.pyx":1298
 *     elif tag == JS_TAG_BOOL:
 *         ret = PyBool_FromLong(JS_VALUE_GET_INT(value))
 *         JS_FreeValue(ctx, value)             # <<<<<<<<<<<<<<
 *         return ret
 * 
*/
    JS_FreeValue(__pyx_v_ctx, __pyx_v_value);

    /* "cyjs/_cyjs.pyx":1299
 *         ret = PyBool_FromLong(JS_VALUE_GET_INT(value))
 *         JS_FreeValue(ctx, value)
 *         return ret             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_ret);
    __pyx_r = __pyx_v_ret;
    goto __pyx_L0;

    /* "cyjs/_cyjs.pyx":1296
 *         JS_FreeValue(ctx, value)
 *         return ret
 *     elif tag == JS_TAG_BOOL:             # <<<<<<<<<<<<<<
 *         ret = PyBool_FromLong(JS_VALUE_GET_INT(value))
 *         JS_FreeValue(ctx, value)
*/
  }

  /* "cyjs/_cyjs.pyx":1303
 * 
 *     # Unimplemented (Hold onto value so we can free it up later)
 *     return jsv_to_object(ctx, value)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = ((PyObject *)__pyx_f_4cyjs_5_cyjs_jsv_to_object(__pyx_v_ctx, __pyx_v_value)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1262
 * 
 * 
 * cdef object to_python(JSContext* ctx, JSValue value):             # <<<<<<<<<<<<<<
 *     cdef int32_t tag = JS_VALUE_GET_TAG(value)
 *     cdef object ret
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.to_python", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyjs/_cyjs.pyx":1306
 * 
 * 
 * cdef inline object to_python_with_copy(JSContext* ctx, JSValue value):             # <<<<<<<<<<<<<<
 *     """Copies jsvalue before converting it to python useful in cases where
 *     where JS_DupValue is needed"""
*/

static CYTHON_INLINE PyObject *__pyx_f_4cyjs_5_cyjs_to_python_with_copy(JSContext *__pyx_v_ctx, JSValue __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_python_with_copy", 0);

  /* "cyjs/_cyjs.pyx":1309
 *     """Copies jsvalue before converting it to python useful in cases where
 *     where JS_DupValue is needed"""
 *     return to_python(ctx, JS_DupValue(ctx, value))             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cyjs_5_cyjs_to_python(__pyx_v_ctx, JS_DupValue(__pyx_v_ctx, __pyx_v_value)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyjs/_cyjs.pyx":1306
 * 
 * 
 * cdef inline object to_python_with_copy(JSContext* ctx, JSValue value):             # <<<<<<<<<<<<<<
 *     """Copies jsvalue before converting it to python useful in cases where
 *     where JS_DupValue is needed"""
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.to_python_with_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_JSError(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_1__pyx_unpickle_JSError(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_1__pyx_unpickle_JSError = {"__pyx_unpickle_JSError", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_1__pyx_unpickle_JSError, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_1__pyx_unpickle_JSError(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_JSError (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_JSError", 0) < (0)) __PYX_ERR(1, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_JSError", 1, 3, 3, i); __PYX_ERR(1, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_JSError", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_JSError", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_4cyjs_5_cyjs___pyx_unpickle_JSError(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs___pyx_unpickle_JSError(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_JSError", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_JSError(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')             # <<<<<<<<<<<<<<
 *     __pyx_result = JSError.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, __pyx_k__2); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
 *     __pyx_result = JSError.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_JSError__set_state(<JSError> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_JSError);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
 *     __pyx_result = JSError.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_JSError__set_state(<JSError> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = JSError.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_JSError__set_state(<JSError> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_JSError__set_state(JSError __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(1, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_4cyjs_5_cyjs___pyx_unpickle_JSError__set_state(((struct __pyx_obj_4cyjs_5_cyjs_JSError *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
 *     __pyx_result = JSError.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_JSError__set_state(<JSError> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_JSError__set_state(<JSError> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_JSError__set_state(JSError __pyx_result, __pyx_state: tuple):
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 0)
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_JSError(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_JSError", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_JSError__set_state(<JSError> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_JSError__set_state(JSError __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 0)
*/

static PyObject *__pyx_f_4cyjs_5_cyjs___pyx_unpickle_JSError__set_state(struct __pyx_obj_4cyjs_5_cyjs_JSError *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_JSError__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_JSError__set_state(JSError __pyx_result, __pyx_state: tuple):
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 0)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 0); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_JSError__set_state(<JSError> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_JSError__set_state(JSError __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 0)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_JSError__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_MemoryUsage(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe4999e7, 0x79fea0b, 0x65b687e, b'array_count, atom_count, atom_size, binary_object_count, binary_object_size, fast_array_count, fast_array_elements, js_func_code_size, js_func_count, js_func_pc2line_count, js_func_size, malloc_count, malloc_limit, malloc_size, memory_used_count, memory_used_size, obj_count, obj_size, prop_count, prop_size, shape_count, shape_size, str_count, str_size')
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_3__pyx_unpickle_MemoryUsage(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_3__pyx_unpickle_MemoryUsage = {"__pyx_unpickle_MemoryUsage", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_3__pyx_unpickle_MemoryUsage, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_3__pyx_unpickle_MemoryUsage(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_MemoryUsage (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_MemoryUsage", 0) < (0)) __PYX_ERR(1, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MemoryUsage", 1, 3, 3, i); __PYX_ERR(1, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MemoryUsage", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_MemoryUsage", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_2__pyx_unpickle_MemoryUsage(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_2__pyx_unpickle_MemoryUsage(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_MemoryUsage", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_MemoryUsage(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe4999e7, 0x79fea0b, 0x65b687e, b'array_count, atom_count, atom_size, binary_object_count, binary_object_size, fast_array_count, fast_array_elements, js_func_code_size, js_func_count, js_func_pc2line_count, js_func_size, malloc_count, malloc_limit, malloc_size, memory_used_count, memory_used_size, obj_count, obj_size, prop_count, prop_size, shape_count, shape_size, str_count, str_size')             # <<<<<<<<<<<<<<
 *     __pyx_result = MemoryUsage.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0xe4999e7, 0x79fea0b, 0x65b687e, __pyx_k_array_count_atom_count_atom_size); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe4999e7, 0x79fea0b, 0x65b687e, b'array_count, atom_count, atom_size, binary_object_count, binary_object_size, fast_array_count, fast_array_elements, js_func_code_size, js_func_count, js_func_pc2line_count, js_func_size, malloc_count, malloc_limit, malloc_size, memory_used_count, memory_used_size, obj_count, obj_size, prop_count, prop_size, shape_count, shape_size, str_count, str_size')
 *     __pyx_result = MemoryUsage.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MemoryUsage__set_state(<MemoryUsage> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe4999e7, 0x79fea0b, 0x65b687e, b'array_count, atom_count, atom_size, binary_object_count, binary_object_size, fast_array_count, fast_array_elements, js_func_code_size, js_func_count, js_func_pc2line_count, js_func_size, malloc_count, malloc_limit, malloc_size, memory_used_count, memory_used_size, obj_count, obj_size, prop_count, prop_size, shape_count, shape_size, str_count, str_size')
 *     __pyx_result = MemoryUsage.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_MemoryUsage__set_state(<MemoryUsage> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = MemoryUsage.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MemoryUsage__set_state(<MemoryUsage> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_MemoryUsage__set_state(MemoryUsage __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(1, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_4cyjs_5_cyjs___pyx_unpickle_MemoryUsage__set_state(((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe4999e7, 0x79fea0b, 0x65b687e, b'array_count, atom_count, atom_size, binary_object_count, binary_object_size, fast_array_count, fast_array_elements, js_func_code_size, js_func_count, js_func_pc2line_count, js_func_size, malloc_count, malloc_limit, malloc_size, memory_used_count, memory_used_size, obj_count, obj_size, prop_count, prop_size, shape_count, shape_size, str_count, str_size')
 *     __pyx_result = MemoryUsage.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_MemoryUsage__set_state(<MemoryUsage> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MemoryUsage__set_state(<MemoryUsage> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_MemoryUsage__set_state(MemoryUsage __pyx_result, __pyx_state: tuple):
 *     __pyx_result.array_count = __pyx_state[0]; __pyx_result.atom_count = __pyx_state[1]; __pyx_result.atom_size = __pyx_state[2]; __pyx_result.binary_object_count = __pyx_state[3]; __pyx_result.binary_object_size = __pyx_state[4]; __pyx_result.fast_array_count = __pyx_state[5]; __pyx_result.fast_array_elements = __pyx_state[6]; __pyx_result.js_func_code_size = __pyx_state[7]; __pyx_result.js_func_count = __pyx_state[8]; __pyx_result.js_func_pc2line_count = __pyx_state[9]; __pyx_result.js_func_size = __pyx_state[10]; __pyx_result.malloc_count = __pyx_state[11]; __pyx_result.malloc_limit = __pyx_state[12]; __pyx_result.malloc_size = __pyx_state[13]; __pyx_result.memory_used_count = __pyx_state[14]; __pyx_result.memory_used_size = __pyx_state[15]; __pyx_result.obj_count = __pyx_state[16]; __pyx_result.obj_size = __pyx_state[17]; __pyx_result.prop_count = __pyx_state[18]; __pyx_result.prop_size = __pyx_state[19]; __pyx_result.shape_count = __pyx_state[20]; __pyx_result.shape_size = __pyx_state[21]; __pyx_result.str_count = __pyx_state[22]; __pyx_result.str_size = __pyx_state[23]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_MemoryUsage(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe4999e7, 0x79fea0b, 0x65b687e, b'array_count, atom_count, atom_size, binary_object_count, binary_object_size, fast_array_count, fast_array_elements, js_func_code_size, js_func_count, js_func_pc2line_count, js_func_size, malloc_count, malloc_limit, malloc_size, memory_used_count, memory_used_size, obj_count, obj_size, prop_count, prop_size, shape_count, shape_size, str_count, str_size')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_MemoryUsage", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_MemoryUsage__set_state(<MemoryUsage> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_MemoryUsage__set_state(MemoryUsage __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result.array_count = __pyx_state[0]; __pyx_result.atom_count = __pyx_state[1]; __pyx_result.atom_size = __pyx_state[2]; __pyx_result.binary_object_count = __pyx_state[3]; __pyx_result.binary_object_size = __pyx_state[4]; __pyx_result.fast_array_count = __pyx_state[5]; __pyx_result.fast_array_elements = __pyx_state[6]; __pyx_result.js_func_code_size = __pyx_state[7]; __pyx_result.js_func_count = __pyx_state[8]; __pyx_result.js_func_pc2line_count = __pyx_state[9]; __pyx_result.js_func_size = __pyx_state[10]; __pyx_result.malloc_count = __pyx_state[11]; __pyx_result.malloc_limit = __pyx_state[12]; __pyx_result.malloc_size = __pyx_state[13]; __pyx_result.memory_used_count = __pyx_state[14]; __pyx_result.memory_used_size = __pyx_state[15]; __pyx_result.obj_count = __pyx_state[16]; __pyx_result.obj_size = __pyx_state[17]; __pyx_result.prop_count = __pyx_state[18]; __pyx_result.prop_size = __pyx_state[19]; __pyx_result.shape_count = __pyx_state[20]; __pyx_result.shape_size = __pyx_state[21]; __pyx_result.str_count = __pyx_state[22]; __pyx_result.str_size = __pyx_state[23]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 24)
*/

static PyObject *__pyx_f_4cyjs_5_cyjs___pyx_unpickle_MemoryUsage__set_state(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int64_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_MemoryUsage__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_MemoryUsage__set_state(MemoryUsage __pyx_result, __pyx_state: tuple):
 *     __pyx_result.array_count = __pyx_state[0]; __pyx_result.atom_count = __pyx_state[1]; __pyx_result.atom_size = __pyx_state[2]; __pyx_result.binary_object_count = __pyx_state[3]; __pyx_result.binary_object_size = __pyx_state[4]; __pyx_result.fast_array_count = __pyx_state[5]; __pyx_result.fast_array_elements = __pyx_state[6]; __pyx_result.js_func_code_size = __pyx_state[7]; __pyx_result.js_func_count = __pyx_state[8]; __pyx_result.js_func_pc2line_count = __pyx_state[9]; __pyx_result.js_func_size = __pyx_state[10]; __pyx_result.malloc_count = __pyx_state[11]; __pyx_result.malloc_limit = __pyx_state[12]; __pyx_result.malloc_size = __pyx_state[13]; __pyx_result.memory_used_count = __pyx_state[14]; __pyx_result.memory_used_size = __pyx_state[15]; __pyx_result.obj_count = __pyx_state[16]; __pyx_result.obj_size = __pyx_state[17]; __pyx_result.prop_count = __pyx_state[18]; __pyx_result.prop_size = __pyx_state[19]; __pyx_result.shape_count = __pyx_state[20]; __pyx_result.shape_size = __pyx_state[21]; __pyx_result.str_count = __pyx_state[22]; __pyx_result.str_size = __pyx_state[23]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 24)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->array_count = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->atom_count = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->atom_size = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->binary_object_count = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->binary_object_size = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->fast_array_count = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->fast_array_elements = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->js_func_code_size = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->js_func_count = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->js_func_pc2line_count = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->js_func_size = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->malloc_count = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->malloc_limit = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->malloc_size = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->memory_used_count = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->memory_used_size = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->obj_count = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 17, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->obj_size = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 18, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->prop_count = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 19, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->prop_size = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 20, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->shape_count = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 21, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->shape_size = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 22, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->str_count = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 23, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->str_size = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_MemoryUsage__set_state(MemoryUsage __pyx_result, __pyx_state: tuple):
 *     __pyx_result.array_count = __pyx_state[0]; __pyx_result.atom_count = __pyx_state[1]; __pyx_result.atom_size = __pyx_state[2]; __pyx_result.binary_object_count = __pyx_state[3]; __pyx_result.binary_object_size = __pyx_state[4]; __pyx_result.fast_array_count = __pyx_state[5]; __pyx_result.fast_array_elements = __pyx_state[6]; __pyx_result.js_func_code_size = __pyx_state[7]; __pyx_result.js_func_count = __pyx_state[8]; __pyx_result.js_func_pc2line_count = __pyx_state[9]; __pyx_result.js_func_size = __pyx_state[10]; __pyx_result.malloc_count = __pyx_state[11]; __pyx_result.malloc_limit = __pyx_state[12]; __pyx_result.malloc_size = __pyx_state[13]; __pyx_result.memory_used_count = __pyx_state[14]; __pyx_result.memory_used_size = __pyx_state[15]; __pyx_result.obj_count = __pyx_state[16]; __pyx_result.obj_size = __pyx_state[17]; __pyx_result.prop_count = __pyx_state[18]; __pyx_result.prop_size = __pyx_state[19]; __pyx_result.shape_count = __pyx_state[20]; __pyx_result.shape_size = __pyx_state[21]; __pyx_result.str_count = __pyx_state[22]; __pyx_result.str_size = __pyx_state[23]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 24)             # <<<<<<<<<<<<<<
*/
  __pyx_t_3 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 24); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_MemoryUsage__set_state(<MemoryUsage> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_MemoryUsage__set_state(MemoryUsage __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result.array_count = __pyx_state[0]; __pyx_result.atom_count = __pyx_state[1]; __pyx_result.atom_size = __pyx_state[2]; __pyx_result.binary_object_count = __pyx_state[3]; __pyx_result.binary_object_size = __pyx_state[4]; __pyx_result.fast_array_count = __pyx_state[5]; __pyx_result.fast_array_elements = __pyx_state[6]; __pyx_result.js_func_code_size = __pyx_state[7]; __pyx_result.js_func_count = __pyx_state[8]; __pyx_result.js_func_pc2line_count = __pyx_state[9]; __pyx_result.js_func_size = __pyx_state[10]; __pyx_result.malloc_count = __pyx_state[11]; __pyx_result.malloc_limit = __pyx_state[12]; __pyx_result.malloc_size = __pyx_state[13]; __pyx_result.memory_used_count = __pyx_state[14]; __pyx_result.memory_used_size = __pyx_state[15]; __pyx_result.obj_count = __pyx_state[16]; __pyx_result.obj_size = __pyx_state[17]; __pyx_result.prop_count = __pyx_state[18]; __pyx_result.prop_size = __pyx_state[19]; __pyx_result.shape_count = __pyx_state[20]; __pyx_result.shape_size = __pyx_state[21]; __pyx_result.str_count = __pyx_state[22]; __pyx_result.str_size = __pyx_state[23]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 24)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_MemoryUsage__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_PromiseHook(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x84cfe3b, 0x3581cd1, 0x33c4952, b'func, rt')
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_5__pyx_unpickle_PromiseHook(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_5__pyx_unpickle_PromiseHook = {"__pyx_unpickle_PromiseHook", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_5__pyx_unpickle_PromiseHook, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_5__pyx_unpickle_PromiseHook(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromiseHook (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_PromiseHook", 0) < (0)) __PYX_ERR(1, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromiseHook", 1, 3, 3, i); __PYX_ERR(1, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromiseHook", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_PromiseHook", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_4__pyx_unpickle_PromiseHook(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_4__pyx_unpickle_PromiseHook(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromiseHook", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_PromiseHook(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x84cfe3b, 0x3581cd1, 0x33c4952, b'func, rt')             # <<<<<<<<<<<<<<
 *     __pyx_result = PromiseHook.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0x84cfe3b, 0x3581cd1, 0x33c4952, __pyx_k_func_rt); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x84cfe3b, 0x3581cd1, 0x33c4952, b'func, rt')
 *     __pyx_result = PromiseHook.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromiseHook__set_state(<PromiseHook> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_PromiseHook);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x84cfe3b, 0x3581cd1, 0x33c4952, b'func, rt')
 *     __pyx_result = PromiseHook.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PromiseHook__set_state(<PromiseHook> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = PromiseHook.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromiseHook__set_state(<PromiseHook> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PromiseHook__set_state(PromiseHook __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(1, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_4cyjs_5_cyjs___pyx_unpickle_PromiseHook__set_state(((struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x84cfe3b, 0x3581cd1, 0x33c4952, b'func, rt')
 *     __pyx_result = PromiseHook.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PromiseHook__set_state(<PromiseHook> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromiseHook__set_state(<PromiseHook> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PromiseHook__set_state(PromiseHook __pyx_result, __pyx_state: tuple):
 *     __pyx_result.func = __pyx_state[0]; __pyx_result.rt = __pyx_state[1]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_PromiseHook(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x84cfe3b, 0x3581cd1, 0x33c4952, b'func, rt')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_PromiseHook", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PromiseHook__set_state(<PromiseHook> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PromiseHook__set_state(PromiseHook __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result.func = __pyx_state[0]; __pyx_result.rt = __pyx_state[1]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 2)
*/

static PyObject *__pyx_f_4cyjs_5_cyjs___pyx_unpickle_PromiseHook__set_state(struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromiseHook__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PromiseHook__set_state(PromiseHook __pyx_result, __pyx_state: tuple):
 *     __pyx_result.func = __pyx_state[0]; __pyx_result.rt = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 2)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->func);
  __Pyx_DECREF(__pyx_v___pyx_result->func);
  __pyx_v___pyx_result->func = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->rt);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->rt);
  __pyx_v___pyx_result->rt = ((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_PromiseHook__set_state(PromiseHook __pyx_result, __pyx_state: tuple):
 *     __pyx_result.func = __pyx_state[0]; __pyx_result.rt = __pyx_state[1]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 2)             # <<<<<<<<<<<<<<
*/
  __pyx_t_2 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 2); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_PromiseHook__set_state(<PromiseHook> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PromiseHook__set_state(PromiseHook __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result.func = __pyx_state[0]; __pyx_result.rt = __pyx_state[1]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 2)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_PromiseHook__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_CancelledError(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_7__pyx_unpickle_CancelledError(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_7__pyx_unpickle_CancelledError = {"__pyx_unpickle_CancelledError", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7__pyx_unpickle_CancelledError, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_7__pyx_unpickle_CancelledError(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_CancelledError (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_CancelledError", 0) < (0)) __PYX_ERR(1, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CancelledError", 1, 3, 3, i); __PYX_ERR(1, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CancelledError", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_CancelledError", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_6__pyx_unpickle_CancelledError(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_6__pyx_unpickle_CancelledError(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CancelledError", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_CancelledError(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')             # <<<<<<<<<<<<<<
 *     __pyx_result = CancelledError.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, __pyx_k__2); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
 *     __pyx_result = CancelledError.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CancelledError__set_state(<CancelledError> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_CancelledError);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
 *     __pyx_result = CancelledError.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CancelledError__set_state(<CancelledError> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = CancelledError.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CancelledError__set_state(<CancelledError> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_CancelledError__set_state(CancelledError __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(1, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_4cyjs_5_cyjs___pyx_unpickle_CancelledError__set_state(((struct __pyx_obj_4cyjs_5_cyjs_CancelledError *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
 *     __pyx_result = CancelledError.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CancelledError__set_state(<CancelledError> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CancelledError__set_state(<CancelledError> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_CancelledError__set_state(CancelledError __pyx_result, __pyx_state: tuple):
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 0)
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_CancelledError(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_CancelledError", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_CancelledError__set_state(<CancelledError> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CancelledError__set_state(CancelledError __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 0)
*/

static PyObject *__pyx_f_4cyjs_5_cyjs___pyx_unpickle_CancelledError__set_state(struct __pyx_obj_4cyjs_5_cyjs_CancelledError *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CancelledError__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_CancelledError__set_state(CancelledError __pyx_result, __pyx_state: tuple):
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 0)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 0); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_CancelledError__set_state(<CancelledError> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CancelledError__set_state(CancelledError __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 0)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_CancelledError__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_InvalidStateError(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_9__pyx_unpickle_InvalidStateError(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_9__pyx_unpickle_InvalidStateError = {"__pyx_unpickle_InvalidStateError", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_9__pyx_unpickle_InvalidStateError, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_9__pyx_unpickle_InvalidStateError(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_InvalidStateError (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_InvalidStateError", 0) < (0)) __PYX_ERR(1, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InvalidStateError", 1, 3, 3, i); __PYX_ERR(1, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InvalidStateError", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_InvalidStateError", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_8__pyx_unpickle_InvalidStateError(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_8__pyx_unpickle_InvalidStateError(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InvalidStateError", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_InvalidStateError(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')             # <<<<<<<<<<<<<<
 *     __pyx_result = InvalidStateError.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, __pyx_k__2); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
 *     __pyx_result = InvalidStateError.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InvalidStateError__set_state(<InvalidStateError> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
 *     __pyx_result = InvalidStateError.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InvalidStateError__set_state(<InvalidStateError> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = InvalidStateError.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InvalidStateError__set_state(<InvalidStateError> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_InvalidStateError__set_state(InvalidStateError __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(1, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_4cyjs_5_cyjs___pyx_unpickle_InvalidStateError__set_state(((struct __pyx_obj_4cyjs_5_cyjs_InvalidStateError *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
 *     __pyx_result = InvalidStateError.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InvalidStateError__set_state(<InvalidStateError> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InvalidStateError__set_state(<InvalidStateError> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_InvalidStateError__set_state(InvalidStateError __pyx_result, __pyx_state: tuple):
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 0)
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_InvalidStateError(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_InvalidStateError", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_InvalidStateError__set_state(<InvalidStateError> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InvalidStateError__set_state(InvalidStateError __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 0)
*/

static PyObject *__pyx_f_4cyjs_5_cyjs___pyx_unpickle_InvalidStateError__set_state(struct __pyx_obj_4cyjs_5_cyjs_InvalidStateError *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InvalidStateError__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_InvalidStateError__set_state(InvalidStateError __pyx_result, __pyx_state: tuple):
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 0)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 0); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_InvalidStateError__set_state(<InvalidStateError> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InvalidStateError__set_state(InvalidStateError __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 0)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_InvalidStateError__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_NODEFAULT(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
*/

/* Python wrapper */
static PyObject *__pyx_pw_4cyjs_5_cyjs_11__pyx_unpickle_NODEFAULT(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_4cyjs_5_cyjs_11__pyx_unpickle_NODEFAULT = {"__pyx_unpickle_NODEFAULT", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_11__pyx_unpickle_NODEFAULT, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4cyjs_5_cyjs_11__pyx_unpickle_NODEFAULT(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_NODEFAULT (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_NODEFAULT", 0) < (0)) __PYX_ERR(1, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NODEFAULT", 1, 3, 3, i); __PYX_ERR(1, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NODEFAULT", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_NODEFAULT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_4cyjs_5_cyjs_10__pyx_unpickle_NODEFAULT(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cyjs_5_cyjs_10__pyx_unpickle_NODEFAULT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_NODEFAULT", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_NODEFAULT(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')             # <<<<<<<<<<<<<<
 *     __pyx_result = NODEFAULT.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, __pyx_k__2); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
 *     __pyx_result = NODEFAULT.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NODEFAULT__set_state(<NODEFAULT> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_NODEFAULT);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
 *     __pyx_result = NODEFAULT.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_NODEFAULT__set_state(<NODEFAULT> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = NODEFAULT.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NODEFAULT__set_state(<NODEFAULT> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_NODEFAULT__set_state(NODEFAULT __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(1, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_4cyjs_5_cyjs___pyx_unpickle_NODEFAULT__set_state(((struct __pyx_obj_4cyjs_5_cyjs_NODEFAULT *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
 *     __pyx_result = NODEFAULT.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_NODEFAULT__set_state(<NODEFAULT> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NODEFAULT__set_state(<NODEFAULT> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_NODEFAULT__set_state(NODEFAULT __pyx_result, __pyx_state: tuple):
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 0)
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_NODEFAULT(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_NODEFAULT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_NODEFAULT__set_state(<NODEFAULT> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NODEFAULT__set_state(NODEFAULT __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 0)
*/

static PyObject *__pyx_f_4cyjs_5_cyjs___pyx_unpickle_NODEFAULT__set_state(struct __pyx_obj_4cyjs_5_cyjs_NODEFAULT *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_NODEFAULT__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_NODEFAULT__set_state(NODEFAULT __pyx_result, __pyx_state: tuple):
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 0)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 0); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_NODEFAULT__set_state(<NODEFAULT> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NODEFAULT__set_state(NODEFAULT __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 0)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyjs._cyjs.__pyx_unpickle_NODEFAULT__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
/* #### Code section: module_exttypes ### */
static struct __pyx_vtabstruct_4cyjs_5_cyjs_JSError __pyx_vtable_4cyjs_5_cyjs_JSError;

static PyObject *__pyx_tp_new_4cyjs_5_cyjs_JSError(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4cyjs_5_cyjs_JSError *p;
  PyObject *o = __Pyx_PyType_GetSlot((((PyTypeObject*)PyExc_Exception)), tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4cyjs_5_cyjs_JSError *)o);
  p->__pyx_vtab = __pyx_vtabptr_4cyjs_5_cyjs_JSError;
  return o;
}

static void __pyx_tp_dealloc_4cyjs_5_cyjs_JSError(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_4cyjs_5_cyjs_JSError) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  PyObject_GC_Track(o);
  __Pyx_PyType_GetSlot((((PyTypeObject*)PyExc_Exception)), tp_dealloc, destructor)(o);
}

static int __pyx_tp_traverse_4cyjs_5_cyjs_JSError(PyObject *o, visitproc v, void *a) {
  int e;
  {
    traverseproc traverse = __Pyx_PyType_GetSlot((((PyTypeObject*)PyExc_Exception)), tp_traverse, traverseproc);
    if (!traverse); else { e = traverse(o,v,a); if (e) return e; }
  }
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4cyjs_5_cyjs_JSError(PyObject *o) {
  {
    inquiry clear = __Pyx_PyType_GetSlot((((PyTypeObject*)PyExc_Exception)), tp_clear, inquiry);
    if (clear) clear(o);
  }
  return 0;
}

static PyMethodDef __pyx_methods_4cyjs_5_cyjs_JSError[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7JSError_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7JSError_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs_JSError_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs_JSError},
  {Py_tp_doc, (void *)PyDoc_STR("Represents Numerous Exceptions raised from CYJS")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4cyjs_5_cyjs_JSError},
  {Py_tp_clear, (void *)__pyx_tp_clear_4cyjs_5_cyjs_JSError},
  {Py_tp_methods, (void *)__pyx_methods_4cyjs_5_cyjs_JSError},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs_JSError},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs_JSError_spec = {
  "cyjs._cyjs.JSError",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_JSError),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_4cyjs_5_cyjs_JSError_slots,
};
#else

static PyTypeObject __pyx_type_4cyjs_5_cyjs_JSError = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""JSError", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_JSError), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs_JSError, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Represents Numerous Exceptions raised from CYJS"), /*tp_doc*/
  __pyx_tp_traverse_4cyjs_5_cyjs_JSError, /*tp_traverse*/
  __pyx_tp_clear_4cyjs_5_cyjs_JSError, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4cyjs_5_cyjs_JSError, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs_JSError, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_4cyjs_5_cyjs_MemoryUsage __pyx_vtable_4cyjs_5_cyjs_MemoryUsage;

static PyObject *__pyx_tp_new_4cyjs_5_cyjs_MemoryUsage(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 1);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *)o);
  p->__pyx_vtab = __pyx_vtabptr_4cyjs_5_cyjs_MemoryUsage;
  return o;
}

static void __pyx_tp_dealloc_4cyjs_5_cyjs_MemoryUsage(PyObject *o) {
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_malloc_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_11malloc_size_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_malloc_limit(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_12malloc_limit_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_memory_used_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_16memory_used_size_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_malloc_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_12malloc_count_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_memory_used_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_17memory_used_count_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_atom_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_10atom_count_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_atom_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_9atom_size_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_str_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_9str_count_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_str_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_8str_size_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_obj_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_9obj_count_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_obj_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_8obj_size_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_prop_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_10prop_count_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_prop_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_9prop_size_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_shape_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_11shape_count_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_shape_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_10shape_size_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_js_func_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_13js_func_count_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_js_func_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_12js_func_size_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_js_func_code_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_17js_func_code_size_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_js_func_pc2line_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_21js_func_pc2line_count_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_array_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_11array_count_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_fast_array_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_16fast_array_count_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_fast_array_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_19fast_array_elements_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_binary_object_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_19binary_object_count_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_binary_object_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_11MemoryUsage_18binary_object_size_1__get__(o);
}

static PyMethodDef __pyx_methods_4cyjs_5_cyjs_MemoryUsage[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_11MemoryUsage_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4cyjs_5_cyjs_MemoryUsage[] = {
  {"malloc_size", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_malloc_size, 0, 0, 0},
  {"malloc_limit", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_malloc_limit, 0, 0, 0},
  {"memory_used_size", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_memory_used_size, 0, 0, 0},
  {"malloc_count", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_malloc_count, 0, 0, 0},
  {"memory_used_count", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_memory_used_count, 0, 0, 0},
  {"atom_count", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_atom_count, 0, 0, 0},
  {"atom_size", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_atom_size, 0, 0, 0},
  {"str_count", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_str_count, 0, 0, 0},
  {"str_size", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_str_size, 0, 0, 0},
  {"obj_count", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_obj_count, 0, 0, 0},
  {"obj_size", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_obj_size, 0, 0, 0},
  {"prop_count", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_prop_count, 0, 0, 0},
  {"prop_size", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_prop_size, 0, 0, 0},
  {"shape_count", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_shape_count, 0, 0, 0},
  {"shape_size", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_shape_size, 0, 0, 0},
  {"js_func_count", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_js_func_count, 0, 0, 0},
  {"js_func_size", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_js_func_size, 0, 0, 0},
  {"js_func_code_size", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_js_func_code_size, 0, 0, 0},
  {"js_func_pc2line_count", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_js_func_pc2line_count, 0, 0, 0},
  {"array_count", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_array_count, 0, 0, 0},
  {"fast_array_count", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_fast_array_count, 0, 0, 0},
  {"fast_array_elements", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_fast_array_elements, 0, 0, 0},
  {"binary_object_count", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_binary_object_count, 0, 0, 0},
  {"binary_object_size", __pyx_getprop_4cyjs_5_cyjs_11MemoryUsage_binary_object_size, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs_MemoryUsage_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs_MemoryUsage},
  {Py_tp_methods, (void *)__pyx_methods_4cyjs_5_cyjs_MemoryUsage},
  {Py_tp_getset, (void *)__pyx_getsets_4cyjs_5_cyjs_MemoryUsage},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs_MemoryUsage},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs_MemoryUsage_spec = {
  "cyjs._cyjs.MemoryUsage",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER,
  __pyx_type_4cyjs_5_cyjs_MemoryUsage_slots,
};
#else

static PyTypeObject __pyx_type_4cyjs_5_cyjs_MemoryUsage = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""MemoryUsage", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs_MemoryUsage, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4cyjs_5_cyjs_MemoryUsage, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4cyjs_5_cyjs_MemoryUsage, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs_MemoryUsage, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_4cyjs_5_cyjs_PromiseHook __pyx_vtable_4cyjs_5_cyjs_PromiseHook;

static PyObject *__pyx_tp_new_4cyjs_5_cyjs_PromiseHook(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *)o);
  p->__pyx_vtab = __pyx_vtabptr_4cyjs_5_cyjs_PromiseHook;
  p->func = Py_None; Py_INCREF(Py_None);
  p->rt = ((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_4cyjs_5_cyjs_PromiseHook(PyObject *o) {
  struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *p = (struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_4cyjs_5_cyjs_PromiseHook) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->func);
  Py_CLEAR(p->rt);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_4cyjs_5_cyjs_PromiseHook(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *p = (struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->func) {
    e = (*v)(p->func, a); if (e) return e;
  }
  if (p->rt) {
    e = (*v)(((PyObject *)p->rt), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4cyjs_5_cyjs_PromiseHook(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *p = (struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *)o;
  tmp = ((PyObject*)p->func);
  p->func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->rt);
  p->rt = ((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_4cyjs_5_cyjs_PromiseHook[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_11PromiseHook_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_11PromiseHook_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs_PromiseHook_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs_PromiseHook},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4cyjs_5_cyjs_PromiseHook},
  {Py_tp_clear, (void *)__pyx_tp_clear_4cyjs_5_cyjs_PromiseHook},
  {Py_tp_methods, (void *)__pyx_methods_4cyjs_5_cyjs_PromiseHook},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs_PromiseHook},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs_PromiseHook_spec = {
  "cyjs._cyjs.PromiseHook",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_PromiseHook),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_4cyjs_5_cyjs_PromiseHook_slots,
};
#else

static PyTypeObject __pyx_type_4cyjs_5_cyjs_PromiseHook = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""PromiseHook", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_PromiseHook), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs_PromiseHook, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4cyjs_5_cyjs_PromiseHook, /*tp_traverse*/
  __pyx_tp_clear_4cyjs_5_cyjs_PromiseHook, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4cyjs_5_cyjs_PromiseHook, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs_PromiseHook, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_4cyjs_5_cyjs_Runtime __pyx_vtable_4cyjs_5_cyjs_Runtime;

static PyObject *__pyx_tp_new_4cyjs_5_cyjs_Runtime(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4cyjs_5_cyjs_Runtime *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)o);
  p->__pyx_vtab = __pyx_vtabptr_4cyjs_5_cyjs_Runtime;
  p->promise_hook = ((struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_4cyjs_5_cyjs_Runtime(PyObject *o) {
  struct __pyx_obj_4cyjs_5_cyjs_Runtime *p = (struct __pyx_obj_4cyjs_5_cyjs_Runtime *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_4cyjs_5_cyjs_Runtime) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_4cyjs_5_cyjs_7Runtime_21__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->promise_hook);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_4cyjs_5_cyjs_Runtime(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4cyjs_5_cyjs_Runtime *p = (struct __pyx_obj_4cyjs_5_cyjs_Runtime *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->promise_hook) {
    e = (*v)(((PyObject *)p->promise_hook), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4cyjs_5_cyjs_Runtime(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4cyjs_5_cyjs_Runtime *p = (struct __pyx_obj_4cyjs_5_cyjs_Runtime *)o;
  tmp = ((PyObject*)p->promise_hook);
  p->promise_hook = ((struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_4cyjs_5_cyjs_Runtime[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Runtime_23__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Runtime_25__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs_Runtime_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs_Runtime},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Represents a JavaScript runtime corresponding to an\n    object heap. Several runtimes can exist at the same time but they\n    cannot exchange objects. Inside a given runtime, no multi-threading is\n    supported.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4cyjs_5_cyjs_Runtime},
  {Py_tp_clear, (void *)__pyx_tp_clear_4cyjs_5_cyjs_Runtime},
  {Py_tp_methods, (void *)__pyx_methods_4cyjs_5_cyjs_Runtime},
  {Py_tp_init, (void *)__pyx_pw_4cyjs_5_cyjs_7Runtime_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs_Runtime},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs_Runtime_spec = {
  "cyjs._cyjs.Runtime",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_Runtime),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_4cyjs_5_cyjs_Runtime_slots,
};
#else

static PyTypeObject __pyx_type_4cyjs_5_cyjs_Runtime = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""Runtime", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_Runtime), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs_Runtime, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    Represents a JavaScript runtime corresponding to an\n    object heap. Several runtimes can exist at the same time but they\n    cannot exchange objects. Inside a given runtime, no multi-threading is\n    supported.\n    "), /*tp_doc*/
  __pyx_tp_traverse_4cyjs_5_cyjs_Runtime, /*tp_traverse*/
  __pyx_tp_clear_4cyjs_5_cyjs_Runtime, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4cyjs_5_cyjs_Runtime, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_4cyjs_5_cyjs_7Runtime_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs_Runtime, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_4cyjs_5_cyjs_Object __pyx_vtable_4cyjs_5_cyjs_Object;

static PyObject *__pyx_tp_new_4cyjs_5_cyjs_Object(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4cyjs_5_cyjs_Object *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4cyjs_5_cyjs_Object *)o);
  p->__pyx_vtab = __pyx_vtabptr_4cyjs_5_cyjs_Object;
  p->context = ((struct __pyx_obj_4cyjs_5_cyjs_Context *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_4cyjs_5_cyjs_Object(PyObject *o) {
  struct __pyx_obj_4cyjs_5_cyjs_Object *p = (struct __pyx_obj_4cyjs_5_cyjs_Object *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_4cyjs_5_cyjs_Object) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_4cyjs_5_cyjs_6Object_11__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->context);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_4cyjs_5_cyjs_Object(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4cyjs_5_cyjs_Object *p = (struct __pyx_obj_4cyjs_5_cyjs_Object *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->context) {
    e = (*v)(((PyObject *)p->context), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4cyjs_5_cyjs_Object(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4cyjs_5_cyjs_Object *p = (struct __pyx_obj_4cyjs_5_cyjs_Object *)o;
  tmp = ((PyObject*)p->context);
  p->context = ((struct __pyx_obj_4cyjs_5_cyjs_Context *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_6Object_context(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_6Object_7context_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_6Object_tag(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_6Object_3tag_1__get__(o);
}

static PyMethodDef __pyx_methods_4cyjs_5_cyjs_Object[] = {
  {"get", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_3get, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_5set, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"invoke", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_9invoke, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"items", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_13items, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"values", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_15values, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"keys", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_17keys, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_23__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6Object_25__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4cyjs_5_cyjs_Object[] = {
  {"context", __pyx_getprop_4cyjs_5_cyjs_6Object_context, 0, 0, 0},
  {"tag", __pyx_getprop_4cyjs_5_cyjs_6Object_tag, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs_Object_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs_Object},
  {Py_tp_call, (void *)__pyx_pw_4cyjs_5_cyjs_6Object_7__call__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4cyjs_5_cyjs_Object},
  {Py_tp_clear, (void *)__pyx_tp_clear_4cyjs_5_cyjs_Object},
  {Py_tp_methods, (void *)__pyx_methods_4cyjs_5_cyjs_Object},
  {Py_tp_getset, (void *)__pyx_getsets_4cyjs_5_cyjs_Object},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs_Object},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs_Object_spec = {
  "cyjs._cyjs.Object",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_Object),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_4cyjs_5_cyjs_Object_slots,
};
#else

static PyTypeObject __pyx_type_4cyjs_5_cyjs_Object = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""Object", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_Object), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs_Object, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  __pyx_pw_4cyjs_5_cyjs_6Object_7__call__, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4cyjs_5_cyjs_Object, /*tp_traverse*/
  __pyx_tp_clear_4cyjs_5_cyjs_Object, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4cyjs_5_cyjs_Object, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4cyjs_5_cyjs_Object, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs_Object, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_4cyjs_5_cyjs_JSFunction __pyx_vtable_4cyjs_5_cyjs_JSFunction;

static PyObject *__pyx_tp_new_4cyjs_5_cyjs_JSFunction(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4cyjs_5_cyjs_JSFunction *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)o);
  p->__pyx_vtab = __pyx_vtabptr_4cyjs_5_cyjs_JSFunction;
  p->context = ((struct __pyx_obj_4cyjs_5_cyjs_Context *)Py_None); Py_INCREF(Py_None);
  p->func = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_4cyjs_5_cyjs_JSFunction(PyObject *o) {
  struct __pyx_obj_4cyjs_5_cyjs_JSFunction *p = (struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_4cyjs_5_cyjs_JSFunction) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_4cyjs_5_cyjs_10JSFunction_1__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->context);
  Py_CLEAR(p->func);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_4cyjs_5_cyjs_JSFunction(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4cyjs_5_cyjs_JSFunction *p = (struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->context) {
    e = (*v)(((PyObject *)p->context), a); if (e) return e;
  }
  if (p->func) {
    e = (*v)(p->func, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4cyjs_5_cyjs_JSFunction(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4cyjs_5_cyjs_JSFunction *p = (struct __pyx_obj_4cyjs_5_cyjs_JSFunction *)o;
  tmp = ((PyObject*)p->context);
  p->context = ((struct __pyx_obj_4cyjs_5_cyjs_Context *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->func);
  p->func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_10JSFunction_object(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_10JSFunction_6object_1__get__(o);
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_10JSFunction_context(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_10JSFunction_7context_1__get__(o);
}

static PyMethodDef __pyx_methods_4cyjs_5_cyjs_JSFunction[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_10JSFunction_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_10JSFunction_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4cyjs_5_cyjs_JSFunction[] = {
  {"object", __pyx_getprop_4cyjs_5_cyjs_10JSFunction_object, 0, PyDoc_STR("Obtains the value as a Javascript Object that could in theory be manipulated\n        NOTE: that this calls upon JS_DupValue to prevent derefing the values"), 0},
  {"context", __pyx_getprop_4cyjs_5_cyjs_10JSFunction_context, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs_JSFunction_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs_JSFunction},
  {Py_tp_call, (void *)__pyx_pw_4cyjs_5_cyjs_10JSFunction_3__call__},
  {Py_tp_doc, (void *)PyDoc_STR("Used for acting as a bridge between Javascript and Python\n    it is meant to be used as an access-point for Javascript (ECMA) to be called upon\n    through python. It's not subclassed to Object however due to it's own nature\n    but if needed can be accessed from the object property")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4cyjs_5_cyjs_JSFunction},
  {Py_tp_clear, (void *)__pyx_tp_clear_4cyjs_5_cyjs_JSFunction},
  {Py_tp_methods, (void *)__pyx_methods_4cyjs_5_cyjs_JSFunction},
  {Py_tp_getset, (void *)__pyx_getsets_4cyjs_5_cyjs_JSFunction},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs_JSFunction},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs_JSFunction_spec = {
  "cyjs._cyjs.JSFunction",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_JSFunction),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_4cyjs_5_cyjs_JSFunction_slots,
};
#else

static PyTypeObject __pyx_type_4cyjs_5_cyjs_JSFunction = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""JSFunction", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_JSFunction), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs_JSFunction, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  __pyx_pw_4cyjs_5_cyjs_10JSFunction_3__call__, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Used for acting as a bridge between Javascript and Python\n    it is meant to be used as an access-point for Javascript (ECMA) to be called upon\n    through python. It's not subclassed to Object however due to it's own nature\n    but if needed can be accessed from the object property"), /*tp_doc*/
  __pyx_tp_traverse_4cyjs_5_cyjs_JSFunction, /*tp_traverse*/
  __pyx_tp_clear_4cyjs_5_cyjs_JSFunction, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4cyjs_5_cyjs_JSFunction, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4cyjs_5_cyjs_JSFunction, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs_JSFunction, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_4cyjs_5_cyjs_Context __pyx_vtable_4cyjs_5_cyjs_Context;

static PyObject *__pyx_tp_new_4cyjs_5_cyjs_Context(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4cyjs_5_cyjs_Context *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4cyjs_5_cyjs_Context *)o);
  p->__pyx_vtab = __pyx_vtabptr_4cyjs_5_cyjs_Context;
  p->runtime = ((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)Py_None); Py_INCREF(Py_None);
  p->_cb_exception = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_4cyjs_5_cyjs_Context(PyObject *o) {
  struct __pyx_obj_4cyjs_5_cyjs_Context *p = (struct __pyx_obj_4cyjs_5_cyjs_Context *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_4cyjs_5_cyjs_Context) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->runtime);
  Py_CLEAR(p->_cb_exception);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_4cyjs_5_cyjs_Context(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4cyjs_5_cyjs_Context *p = (struct __pyx_obj_4cyjs_5_cyjs_Context *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->runtime) {
    e = (*v)(((PyObject *)p->runtime), a); if (e) return e;
  }
  if (p->_cb_exception) {
    e = (*v)(p->_cb_exception, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4cyjs_5_cyjs_Context(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4cyjs_5_cyjs_Context *p = (struct __pyx_obj_4cyjs_5_cyjs_Context *)o;
  tmp = ((PyObject*)p->runtime);
  p->runtime = ((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_cb_exception);
  p->_cb_exception = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_4cyjs_5_cyjs_7Context_runtime(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4cyjs_5_cyjs_7Context_7runtime_1__get__(o);
}

static PyMethodDef __pyx_methods_4cyjs_5_cyjs_Context[] = {
  {"get_global", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_7get_global, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"json_parse", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_9json_parse, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_11get, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4cyjs_5_cyjs_7Context_10get},
  {"set", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_13set, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4cyjs_5_cyjs_7Context_12set},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_21__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Context_23__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4cyjs_5_cyjs_Context[] = {
  {"runtime", __pyx_getprop_4cyjs_5_cyjs_7Context_runtime, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs_Context_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs_Context},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4cyjs_5_cyjs_Context},
  {Py_tp_clear, (void *)__pyx_tp_clear_4cyjs_5_cyjs_Context},
  {Py_tp_methods, (void *)__pyx_methods_4cyjs_5_cyjs_Context},
  {Py_tp_getset, (void *)__pyx_getsets_4cyjs_5_cyjs_Context},
  {Py_tp_init, (void *)__pyx_pw_4cyjs_5_cyjs_7Context_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs_Context},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs_Context_spec = {
  "cyjs._cyjs.Context",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_Context),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_4cyjs_5_cyjs_Context_slots,
};
#else

static PyTypeObject __pyx_type_4cyjs_5_cyjs_Context = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""Context", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_Context), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs_Context, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4cyjs_5_cyjs_Context, /*tp_traverse*/
  __pyx_tp_clear_4cyjs_5_cyjs_Context, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4cyjs_5_cyjs_Context, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4cyjs_5_cyjs_Context, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_4cyjs_5_cyjs_7Context_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs_Context, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_4cyjs_5_cyjs__OView __pyx_vtable_4cyjs_5_cyjs__OView;

static PyObject *__pyx_tp_new_4cyjs_5_cyjs__OView(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4cyjs_5_cyjs__OView *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4cyjs_5_cyjs__OView *)o);
  p->__pyx_vtab = __pyx_vtabptr_4cyjs_5_cyjs__OView;
  p->obj = ((struct __pyx_obj_4cyjs_5_cyjs_Object *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_4cyjs_5_cyjs__OView(PyObject *o) {
  struct __pyx_obj_4cyjs_5_cyjs__OView *p = (struct __pyx_obj_4cyjs_5_cyjs__OView *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_4cyjs_5_cyjs__OView) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_4cyjs_5_cyjs_6_OView_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->obj);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_4cyjs_5_cyjs__OView(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4cyjs_5_cyjs__OView *p = (struct __pyx_obj_4cyjs_5_cyjs__OView *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->obj) {
    e = (*v)(((PyObject *)p->obj), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4cyjs_5_cyjs__OView(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4cyjs_5_cyjs__OView *p = (struct __pyx_obj_4cyjs_5_cyjs__OView *)o;
  tmp = ((PyObject*)p->obj);
  p->obj = ((struct __pyx_obj_4cyjs_5_cyjs_Object *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_4cyjs_5_cyjs__OView[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6_OView_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_6_OView_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs__OView_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs__OView},
  {Py_sq_length, (void *)__pyx_pw_4cyjs_5_cyjs_6_OView_5__len__},
  {Py_mp_length, (void *)__pyx_pw_4cyjs_5_cyjs_6_OView_5__len__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4cyjs_5_cyjs__OView},
  {Py_tp_clear, (void *)__pyx_tp_clear_4cyjs_5_cyjs__OView},
  {Py_tp_methods, (void *)__pyx_methods_4cyjs_5_cyjs__OView},
  {Py_tp_init, (void *)__pyx_pw_4cyjs_5_cyjs_6_OView_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs__OView},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs__OView_spec = {
  "cyjs._cyjs._OView",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs__OView),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_4cyjs_5_cyjs__OView_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence__OView = {
  __pyx_pw_4cyjs_5_cyjs_6_OView_5__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping__OView = {
  __pyx_pw_4cyjs_5_cyjs_6_OView_5__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_4cyjs_5_cyjs__OView = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""_OView", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs__OView), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs__OView, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence__OView, /*tp_as_sequence*/
  &__pyx_tp_as_mapping__OView, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4cyjs_5_cyjs__OView, /*tp_traverse*/
  __pyx_tp_clear_4cyjs_5_cyjs__OView, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4cyjs_5_cyjs__OView, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_4cyjs_5_cyjs_6_OView_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs__OView, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_4cyjs_5_cyjs__ObjectItemsView __pyx_vtable_4cyjs_5_cyjs__ObjectItemsView;

static PyObject *__pyx_tp_new_4cyjs_5_cyjs__ObjectItemsView(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4cyjs_5_cyjs__ObjectItemsView *p;
  PyObject *o = __pyx_tp_new_4cyjs_5_cyjs__OView(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4cyjs_5_cyjs__ObjectItemsView *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_4cyjs_5_cyjs__OView*)__pyx_vtabptr_4cyjs_5_cyjs__ObjectItemsView;
  return o;
}

static PyMethodDef __pyx_methods_4cyjs_5_cyjs__ObjectItemsView[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_16_ObjectItemsView_4__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_16_ObjectItemsView_6__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs__ObjectItemsView_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs__OView},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4cyjs_5_cyjs__OView},
  {Py_tp_clear, (void *)__pyx_tp_clear_4cyjs_5_cyjs__OView},
  {Py_tp_iter, (void *)__pyx_pw_4cyjs_5_cyjs_16_ObjectItemsView_1__iter__},
  {Py_tp_methods, (void *)__pyx_methods_4cyjs_5_cyjs__ObjectItemsView},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs__ObjectItemsView},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs__ObjectItemsView_spec = {
  "cyjs._cyjs._ObjectItemsView",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs__ObjectItemsView),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_4cyjs_5_cyjs__ObjectItemsView_slots,
};
#else

static PyTypeObject __pyx_type_4cyjs_5_cyjs__ObjectItemsView = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""_ObjectItemsView", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs__ObjectItemsView), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs__OView, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4cyjs_5_cyjs__OView, /*tp_traverse*/
  __pyx_tp_clear_4cyjs_5_cyjs__OView, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_4cyjs_5_cyjs_16_ObjectItemsView_1__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4cyjs_5_cyjs__ObjectItemsView, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_4cyjs_5_cyjs_6_OView_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs__ObjectItemsView, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_4cyjs_5_cyjs__ObjectKeysView __pyx_vtable_4cyjs_5_cyjs__ObjectKeysView;

static PyObject *__pyx_tp_new_4cyjs_5_cyjs__ObjectKeysView(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView *p;
  PyObject *o = __pyx_tp_new_4cyjs_5_cyjs__OView(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_4cyjs_5_cyjs__OView*)__pyx_vtabptr_4cyjs_5_cyjs__ObjectKeysView;
  return o;
}

static PyMethodDef __pyx_methods_4cyjs_5_cyjs__ObjectKeysView[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_15_ObjectKeysView_6__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_15_ObjectKeysView_8__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs__ObjectKeysView_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs__OView},
  {Py_sq_contains, (void *)__pyx_pw_4cyjs_5_cyjs_15_ObjectKeysView_4__contains__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4cyjs_5_cyjs__OView},
  {Py_tp_clear, (void *)__pyx_tp_clear_4cyjs_5_cyjs__OView},
  {Py_tp_iter, (void *)__pyx_pw_4cyjs_5_cyjs_15_ObjectKeysView_1__iter__},
  {Py_tp_methods, (void *)__pyx_methods_4cyjs_5_cyjs__ObjectKeysView},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs__ObjectKeysView},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs__ObjectKeysView_spec = {
  "cyjs._cyjs._ObjectKeysView",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_4cyjs_5_cyjs__ObjectKeysView_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence__ObjectKeysView = {
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_4cyjs_5_cyjs_6_OView_5__len__, /*sq_length*/
  #else
  0, /*sq_length*/
  #endif
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_4cyjs_5_cyjs_15_ObjectKeysView_4__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyTypeObject __pyx_type_4cyjs_5_cyjs__ObjectKeysView = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""_ObjectKeysView", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs__ObjectKeysView), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs__OView, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence__ObjectKeysView, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4cyjs_5_cyjs__OView, /*tp_traverse*/
  __pyx_tp_clear_4cyjs_5_cyjs__OView, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_4cyjs_5_cyjs_15_ObjectKeysView_1__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4cyjs_5_cyjs__ObjectKeysView, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_4cyjs_5_cyjs_6_OView_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs__ObjectKeysView, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_4cyjs_5_cyjs__ObjectValuesView __pyx_vtable_4cyjs_5_cyjs__ObjectValuesView;

static PyObject *__pyx_tp_new_4cyjs_5_cyjs__ObjectValuesView(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView *p;
  PyObject *o = __pyx_tp_new_4cyjs_5_cyjs__OView(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_4cyjs_5_cyjs__OView*)__pyx_vtabptr_4cyjs_5_cyjs__ObjectValuesView;
  return o;
}

static PyMethodDef __pyx_methods_4cyjs_5_cyjs__ObjectValuesView[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_17_ObjectValuesView_6__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_17_ObjectValuesView_8__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs__ObjectValuesView_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs__OView},
  {Py_sq_contains, (void *)__pyx_pw_4cyjs_5_cyjs_17_ObjectValuesView_4__contains__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4cyjs_5_cyjs__OView},
  {Py_tp_clear, (void *)__pyx_tp_clear_4cyjs_5_cyjs__OView},
  {Py_tp_iter, (void *)__pyx_pw_4cyjs_5_cyjs_17_ObjectValuesView_1__iter__},
  {Py_tp_methods, (void *)__pyx_methods_4cyjs_5_cyjs__ObjectValuesView},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs__ObjectValuesView},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs__ObjectValuesView_spec = {
  "cyjs._cyjs._ObjectValuesView",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_4cyjs_5_cyjs__ObjectValuesView_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence__ObjectValuesView = {
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_4cyjs_5_cyjs_6_OView_5__len__, /*sq_length*/
  #else
  0, /*sq_length*/
  #endif
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_4cyjs_5_cyjs_17_ObjectValuesView_4__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyTypeObject __pyx_type_4cyjs_5_cyjs__ObjectValuesView = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""_ObjectValuesView", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs__ObjectValuesView), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs__OView, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence__ObjectValuesView, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4cyjs_5_cyjs__OView, /*tp_traverse*/
  __pyx_tp_clear_4cyjs_5_cyjs__OView, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_4cyjs_5_cyjs_17_ObjectValuesView_1__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4cyjs_5_cyjs__ObjectValuesView, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_4cyjs_5_cyjs_6_OView_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs__ObjectValuesView, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_4cyjs_5_cyjs_CancelledError(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = __Pyx_PyType_GetSlot((((PyTypeObject*)PyExc_Exception)), tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_4cyjs_5_cyjs_CancelledError(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_4cyjs_5_cyjs_CancelledError) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  PyObject_GC_Track(o);
  __Pyx_PyType_GetSlot((((PyTypeObject*)PyExc_Exception)), tp_dealloc, destructor)(o);
}

static int __pyx_tp_traverse_4cyjs_5_cyjs_CancelledError(PyObject *o, visitproc v, void *a) {
  int e;
  {
    traverseproc traverse = __Pyx_PyType_GetSlot((((PyTypeObject*)PyExc_Exception)), tp_traverse, traverseproc);
    if (!traverse); else { e = traverse(o,v,a); if (e) return e; }
  }
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4cyjs_5_cyjs_CancelledError(PyObject *o) {
  {
    inquiry clear = __Pyx_PyType_GetSlot((((PyTypeObject*)PyExc_Exception)), tp_clear, inquiry);
    if (clear) clear(o);
  }
  return 0;
}

static PyMethodDef __pyx_methods_4cyjs_5_cyjs_CancelledError[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_14CancelledError_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_14CancelledError_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs_CancelledError_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs_CancelledError},
  {Py_tp_doc, (void *)PyDoc_STR("Promise was rejected")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4cyjs_5_cyjs_CancelledError},
  {Py_tp_clear, (void *)__pyx_tp_clear_4cyjs_5_cyjs_CancelledError},
  {Py_tp_methods, (void *)__pyx_methods_4cyjs_5_cyjs_CancelledError},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs_CancelledError},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs_CancelledError_spec = {
  "cyjs._cyjs.CancelledError",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_CancelledError),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_4cyjs_5_cyjs_CancelledError_slots,
};
#else

static PyTypeObject __pyx_type_4cyjs_5_cyjs_CancelledError = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""CancelledError", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_CancelledError), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs_CancelledError, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Promise was rejected"), /*tp_doc*/
  __pyx_tp_traverse_4cyjs_5_cyjs_CancelledError, /*tp_traverse*/
  __pyx_tp_clear_4cyjs_5_cyjs_CancelledError, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4cyjs_5_cyjs_CancelledError, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs_CancelledError, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_4cyjs_5_cyjs_InvalidStateError(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = __Pyx_PyType_GetSlot((((PyTypeObject*)PyExc_Exception)), tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_4cyjs_5_cyjs_InvalidStateError(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_4cyjs_5_cyjs_InvalidStateError) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  PyObject_GC_Track(o);
  __Pyx_PyType_GetSlot((((PyTypeObject*)PyExc_Exception)), tp_dealloc, destructor)(o);
}

static int __pyx_tp_traverse_4cyjs_5_cyjs_InvalidStateError(PyObject *o, visitproc v, void *a) {
  int e;
  {
    traverseproc traverse = __Pyx_PyType_GetSlot((((PyTypeObject*)PyExc_Exception)), tp_traverse, traverseproc);
    if (!traverse); else { e = traverse(o,v,a); if (e) return e; }
  }
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4cyjs_5_cyjs_InvalidStateError(PyObject *o) {
  {
    inquiry clear = __Pyx_PyType_GetSlot((((PyTypeObject*)PyExc_Exception)), tp_clear, inquiry);
    if (clear) clear(o);
  }
  return 0;
}

static PyMethodDef __pyx_methods_4cyjs_5_cyjs_InvalidStateError[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_17InvalidStateError_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_17InvalidStateError_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs_InvalidStateError_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs_InvalidStateError},
  {Py_tp_doc, (void *)PyDoc_STR("Promise on inavlid state")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4cyjs_5_cyjs_InvalidStateError},
  {Py_tp_clear, (void *)__pyx_tp_clear_4cyjs_5_cyjs_InvalidStateError},
  {Py_tp_methods, (void *)__pyx_methods_4cyjs_5_cyjs_InvalidStateError},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs_InvalidStateError},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs_InvalidStateError_spec = {
  "cyjs._cyjs.InvalidStateError",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_InvalidStateError),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_4cyjs_5_cyjs_InvalidStateError_slots,
};
#else

static PyTypeObject __pyx_type_4cyjs_5_cyjs_InvalidStateError = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""InvalidStateError", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_InvalidStateError), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs_InvalidStateError, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Promise on inavlid state"), /*tp_doc*/
  __pyx_tp_traverse_4cyjs_5_cyjs_InvalidStateError, /*tp_traverse*/
  __pyx_tp_clear_4cyjs_5_cyjs_InvalidStateError, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4cyjs_5_cyjs_InvalidStateError, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs_InvalidStateError, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_4cyjs_5_cyjs_Promise __pyx_vtable_4cyjs_5_cyjs_Promise;

static PyObject *__pyx_tp_new_4cyjs_5_cyjs_Promise(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4cyjs_5_cyjs_Promise *p;
  PyObject *o = __pyx_tp_new_4cyjs_5_cyjs_Object(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4cyjs_5_cyjs_Promise *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_4cyjs_5_cyjs_Object*)__pyx_vtabptr_4cyjs_5_cyjs_Promise;
  p->_callbacks = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_exception = Py_None; Py_INCREF(Py_None);
  p->_result = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_4cyjs_5_cyjs_Promise(PyObject *o) {
  struct __pyx_obj_4cyjs_5_cyjs_Promise *p = (struct __pyx_obj_4cyjs_5_cyjs_Promise *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_4cyjs_5_cyjs_Promise) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_callbacks);
  Py_CLEAR(p->_exception);
  Py_CLEAR(p->_result);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_4cyjs_5_cyjs_Object(o);
}

static int __pyx_tp_traverse_4cyjs_5_cyjs_Promise(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4cyjs_5_cyjs_Promise *p = (struct __pyx_obj_4cyjs_5_cyjs_Promise *)o;
  e = __pyx_tp_traverse_4cyjs_5_cyjs_Object(o, v, a); if (e) return e;
  if (p->_callbacks) {
    e = (*v)(p->_callbacks, a); if (e) return e;
  }
  if (p->_exception) {
    e = (*v)(p->_exception, a); if (e) return e;
  }
  if (p->_result) {
    e = (*v)(p->_result, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4cyjs_5_cyjs_Promise(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4cyjs_5_cyjs_Promise *p = (struct __pyx_obj_4cyjs_5_cyjs_Promise *)o;
  __pyx_tp_clear_4cyjs_5_cyjs_Object(o);
  tmp = ((PyObject*)p->_callbacks);
  p->_callbacks = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_exception);
  p->_exception = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_result);
  p->_result = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_4cyjs_5_cyjs_Promise[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Promise_13__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_7Promise_15__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs_Promise_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs_Promise},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4cyjs_5_cyjs_Promise},
  {Py_tp_clear, (void *)__pyx_tp_clear_4cyjs_5_cyjs_Promise},
  {Py_tp_methods, (void *)__pyx_methods_4cyjs_5_cyjs_Promise},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs_Promise},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs_Promise_spec = {
  "cyjs._cyjs.Promise",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_Promise),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_4cyjs_5_cyjs_Promise_slots,
};
#else

static PyTypeObject __pyx_type_4cyjs_5_cyjs_Promise = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""Promise", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_Promise), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs_Promise, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_4cyjs_5_cyjs_6Object_7__call__, /*tp_call*/
  #else
  0, /*tp_call*/
  #endif
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4cyjs_5_cyjs_Promise, /*tp_traverse*/
  __pyx_tp_clear_4cyjs_5_cyjs_Promise, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4cyjs_5_cyjs_Promise, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs_Promise, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_4cyjs_5_cyjs_NODEFAULT(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_4cyjs_5_cyjs_NODEFAULT(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_4cyjs_5_cyjs_NODEFAULT) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static PyMethodDef __pyx_methods_4cyjs_5_cyjs_NODEFAULT[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_9NODEFAULT_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4cyjs_5_cyjs_9NODEFAULT_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs_NODEFAULT_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs_NODEFAULT},
  {Py_tp_methods, (void *)__pyx_methods_4cyjs_5_cyjs_NODEFAULT},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs_NODEFAULT},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs_NODEFAULT_spec = {
  "cyjs._cyjs.NODEFAULT",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_NODEFAULT),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_4cyjs_5_cyjs_NODEFAULT_slots,
};
#else

static PyTypeObject __pyx_type_4cyjs_5_cyjs_NODEFAULT = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""NODEFAULT", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs_NODEFAULT), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs_NODEFAULT, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4cyjs_5_cyjs_NODEFAULT, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs_NODEFAULT, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_4cyjs_5_cyjs___pyx_scope_struct____iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_4cyjs_5_cyjs___pyx_scope_struct____iter__ > 0) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct____iter__, sizeof(struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__))))
  {
    o = (PyObject*)__pyx_mstate_global->__pyx_freelist_4cyjs_5_cyjs___pyx_scope_struct____iter__[--__pyx_mstate_global->__pyx_freecount_4cyjs_5_cyjs___pyx_scope_struct____iter__];
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(Py_TYPE(o));
    #endif
    memset(o, 0, sizeof(struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__));
    #if CYTHON_COMPILING_IN_LIMITED_API
    (void) PyObject_Init(o, t);
    #else
    (void) PyObject_INIT(o, t);
    #endif
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = __Pyx_AllocateExtensionType(t, 1);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4cyjs_5_cyjs___pyx_scope_struct____iter__(PyObject *o) {
  struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__ *p = (struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__ *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_4cyjs_5_cyjs___pyx_scope_struct____iter__) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_4cyjs_5_cyjs___pyx_scope_struct____iter__ < 8) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(Py_TYPE(o), __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct____iter__, sizeof(struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__))))
  {
    __pyx_mstate_global->__pyx_freelist_4cyjs_5_cyjs___pyx_scope_struct____iter__[__pyx_mstate_global->__pyx_freecount_4cyjs_5_cyjs___pyx_scope_struct____iter__++] = ((struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__ *)o);
  } else
  #endif
  {
    PyTypeObject *tp = Py_TYPE(o);
    #if CYTHON_USE_TYPE_SLOTS
    (*tp->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(tp);
    #endif
  }
}

static int __pyx_tp_traverse_4cyjs_5_cyjs___pyx_scope_struct____iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__ *p = (struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__ *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs___pyx_scope_struct____iter___slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs___pyx_scope_struct____iter__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4cyjs_5_cyjs___pyx_scope_struct____iter__},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs___pyx_scope_struct____iter__},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs___pyx_scope_struct____iter___spec = {
  "cyjs._cyjs.__pyx_scope_struct____iter__",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_4cyjs_5_cyjs___pyx_scope_struct____iter___slots,
};
#else

static PyTypeObject __pyx_type_4cyjs_5_cyjs___pyx_scope_struct____iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""__pyx_scope_struct____iter__", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct____iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs___pyx_scope_struct____iter__, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4cyjs_5_cyjs___pyx_scope_struct____iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs___pyx_scope_struct____iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_4cyjs_5_cyjs___pyx_scope_struct_1___iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ > 0) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_1___iter__, sizeof(struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__))))
  {
    o = (PyObject*)__pyx_mstate_global->__pyx_freelist_4cyjs_5_cyjs___pyx_scope_struct_1___iter__[--__pyx_mstate_global->__pyx_freecount_4cyjs_5_cyjs___pyx_scope_struct_1___iter__];
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(Py_TYPE(o));
    #endif
    memset(o, 0, sizeof(struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__));
    #if CYTHON_COMPILING_IN_LIMITED_API
    (void) PyObject_Init(o, t);
    #else
    (void) PyObject_INIT(o, t);
    #endif
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = __Pyx_AllocateExtensionType(t, 1);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4cyjs_5_cyjs___pyx_scope_struct_1___iter__(PyObject *o) {
  struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ *p = (struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_4cyjs_5_cyjs___pyx_scope_struct_1___iter__) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ < 8) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(Py_TYPE(o), __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_1___iter__, sizeof(struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__))))
  {
    __pyx_mstate_global->__pyx_freelist_4cyjs_5_cyjs___pyx_scope_struct_1___iter__[__pyx_mstate_global->__pyx_freecount_4cyjs_5_cyjs___pyx_scope_struct_1___iter__++] = ((struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ *)o);
  } else
  #endif
  {
    PyTypeObject *tp = Py_TYPE(o);
    #if CYTHON_USE_TYPE_SLOTS
    (*tp->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(tp);
    #endif
  }
}

static int __pyx_tp_traverse_4cyjs_5_cyjs___pyx_scope_struct_1___iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ *p = (struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs___pyx_scope_struct_1___iter___slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs___pyx_scope_struct_1___iter__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4cyjs_5_cyjs___pyx_scope_struct_1___iter__},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs___pyx_scope_struct_1___iter__},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs___pyx_scope_struct_1___iter___spec = {
  "cyjs._cyjs.__pyx_scope_struct_1___iter__",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_4cyjs_5_cyjs___pyx_scope_struct_1___iter___slots,
};
#else

static PyTypeObject __pyx_type_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""__pyx_scope_struct_1___iter__", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_1___iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs___pyx_scope_struct_1___iter__, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4cyjs_5_cyjs___pyx_scope_struct_1___iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs___pyx_scope_struct_1___iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_4cyjs_5_cyjs___pyx_scope_struct_2___iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ > 0) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_2___iter__, sizeof(struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__))))
  {
    o = (PyObject*)__pyx_mstate_global->__pyx_freelist_4cyjs_5_cyjs___pyx_scope_struct_2___iter__[--__pyx_mstate_global->__pyx_freecount_4cyjs_5_cyjs___pyx_scope_struct_2___iter__];
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(Py_TYPE(o));
    #endif
    memset(o, 0, sizeof(struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__));
    #if CYTHON_COMPILING_IN_LIMITED_API
    (void) PyObject_Init(o, t);
    #else
    (void) PyObject_INIT(o, t);
    #endif
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = __Pyx_AllocateExtensionType(t, 1);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4cyjs_5_cyjs___pyx_scope_struct_2___iter__(PyObject *o) {
  struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ *p = (struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_4cyjs_5_cyjs___pyx_scope_struct_2___iter__) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ < 8) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(Py_TYPE(o), __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_2___iter__, sizeof(struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__))))
  {
    __pyx_mstate_global->__pyx_freelist_4cyjs_5_cyjs___pyx_scope_struct_2___iter__[__pyx_mstate_global->__pyx_freecount_4cyjs_5_cyjs___pyx_scope_struct_2___iter__++] = ((struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ *)o);
  } else
  #endif
  {
    PyTypeObject *tp = Py_TYPE(o);
    #if CYTHON_USE_TYPE_SLOTS
    (*tp->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(tp);
    #endif
  }
}

static int __pyx_tp_traverse_4cyjs_5_cyjs___pyx_scope_struct_2___iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ *p = (struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4cyjs_5_cyjs___pyx_scope_struct_2___iter___slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4cyjs_5_cyjs___pyx_scope_struct_2___iter__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4cyjs_5_cyjs___pyx_scope_struct_2___iter__},
  {Py_tp_new, (void *)__pyx_tp_new_4cyjs_5_cyjs___pyx_scope_struct_2___iter__},
  {0, 0},
};
static PyType_Spec __pyx_type_4cyjs_5_cyjs___pyx_scope_struct_2___iter___spec = {
  "cyjs._cyjs.__pyx_scope_struct_2___iter__",
  sizeof(struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_4cyjs_5_cyjs___pyx_scope_struct_2___iter___slots,
};
#else

static PyTypeObject __pyx_type_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyjs._cyjs.""__pyx_scope_struct_2___iter__", /*tp_name*/
  sizeof(struct __pyx_obj_4cyjs_5_cyjs___pyx_scope_struct_2___iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4cyjs_5_cyjs___pyx_scope_struct_2___iter__, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4cyjs_5_cyjs___pyx_scope_struct_2___iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4cyjs_5_cyjs___pyx_scope_struct_2___iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_EnumBase = Py_None; Py_INCREF(Py_None);
  __Pyx_FlagBase = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_4cyjs_5_cyjs_JSError = &__pyx_vtable_4cyjs_5_cyjs_JSError;
  __pyx_vtable_4cyjs_5_cyjs_JSError.__pyx_new = (struct __pyx_obj_4cyjs_5_cyjs_JSError *(*)(JSContext *, JSValue))__pyx_f_4cyjs_5_cyjs_7JSError_new;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)(((PyTypeObject*)PyExc_Exception))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSError = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs_JSError_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSError)) __PYX_ERR(0, 33, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs_JSError_spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSError) < (0)) __PYX_ERR(0, 33, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSError = &__pyx_type_4cyjs_5_cyjs_JSError;
  #endif
  if (sizeof(struct __pyx_obj_4cyjs_5_cyjs_JSError) != sizeof(PyBaseExceptionObject)) {
    if (__Pyx_validate_extern_base((((PyTypeObject*)PyExc_Exception))) < (0)) __PYX_ERR(0, 33, __pyx_L1_error)
  }
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_JSError->tp_base = (((PyTypeObject*)PyExc_Exception));
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSError) < (0)) __PYX_ERR(0, 33, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSError);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSError->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSError->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSError->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSError, __pyx_vtabptr_4cyjs_5_cyjs_JSError) < (0)) __PYX_ERR(0, 33, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSError) < (0)) __PYX_ERR(0, 33, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_JSError, (PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSError) < (0)) __PYX_ERR(0, 33, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSError) < (0)) __PYX_ERR(0, 33, __pyx_L1_error)
  __pyx_vtabptr_4cyjs_5_cyjs_MemoryUsage = &__pyx_vtable_4cyjs_5_cyjs_MemoryUsage;
  __pyx_vtable_4cyjs_5_cyjs_MemoryUsage.__pyx_new = (struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *(*)(JSMemoryUsage const *))__pyx_f_4cyjs_5_cyjs_11MemoryUsage_new;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs_MemoryUsage_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage)) __PYX_ERR(0, 63, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs_MemoryUsage_spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage) < (0)) __PYX_ERR(0, 63, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage = &__pyx_type_4cyjs_5_cyjs_MemoryUsage;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage) < (0)) __PYX_ERR(0, 63, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage, __pyx_vtabptr_4cyjs_5_cyjs_MemoryUsage) < (0)) __PYX_ERR(0, 63, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage) < (0)) __PYX_ERR(0, 63, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_MemoryUsage, (PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage) < (0)) __PYX_ERR(0, 63, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage) < (0)) __PYX_ERR(0, 63, __pyx_L1_error)
  __pyx_vtabptr_4cyjs_5_cyjs_PromiseHook = &__pyx_vtable_4cyjs_5_cyjs_PromiseHook;
  __pyx_vtable_4cyjs_5_cyjs_PromiseHook.__pyx_new = (struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *(*)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, PyObject *))__pyx_f_4cyjs_5_cyjs_11PromiseHook_new;
  __pyx_vtable_4cyjs_5_cyjs_PromiseHook.hook = (void (*)(struct __pyx_obj_4cyjs_5_cyjs_PromiseHook *, JSContext *, JSPromiseHookType, JSValue, JSValue))__pyx_f_4cyjs_5_cyjs_11PromiseHook_hook;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_PromiseHook = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs_PromiseHook_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_PromiseHook)) __PYX_ERR(0, 110, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs_PromiseHook_spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_PromiseHook) < (0)) __PYX_ERR(0, 110, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_PromiseHook = &__pyx_type_4cyjs_5_cyjs_PromiseHook;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_PromiseHook) < (0)) __PYX_ERR(0, 110, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_PromiseHook);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_PromiseHook->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_PromiseHook->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_PromiseHook->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_PromiseHook, __pyx_vtabptr_4cyjs_5_cyjs_PromiseHook) < (0)) __PYX_ERR(0, 110, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_PromiseHook) < (0)) __PYX_ERR(0, 110, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_PromiseHook) < (0)) __PYX_ERR(0, 110, __pyx_L1_error)
  __pyx_vtabptr_4cyjs_5_cyjs_Runtime = &__pyx_vtable_4cyjs_5_cyjs_Runtime;
  __pyx_vtable_4cyjs_5_cyjs_Runtime.compute_memory_usage = (struct __pyx_obj_4cyjs_5_cyjs_MemoryUsage *(*)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, int __pyx_skip_dispatch))__pyx_f_4cyjs_5_cyjs_7Runtime_compute_memory_usage;
  __pyx_vtable_4cyjs_5_cyjs_Runtime.dump_memory_usage = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, PyObject *, int __pyx_skip_dispatch))__pyx_f_4cyjs_5_cyjs_7Runtime_dump_memory_usage;
  __pyx_vtable_4cyjs_5_cyjs_Runtime.execute_pending_job = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, int __pyx_skip_dispatch))__pyx_f_4cyjs_5_cyjs_7Runtime_execute_pending_job;
  __pyx_vtable_4cyjs_5_cyjs_Runtime.is_job_pending = (int (*)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, int __pyx_skip_dispatch))__pyx_f_4cyjs_5_cyjs_7Runtime_is_job_pending;
  __pyx_vtable_4cyjs_5_cyjs_Runtime.new_context = (JSContext *(*)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *))__pyx_f_4cyjs_5_cyjs_7Runtime_new_context;
  __pyx_vtable_4cyjs_5_cyjs_Runtime.run_gc = (void (*)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, int __pyx_skip_dispatch))__pyx_f_4cyjs_5_cyjs_7Runtime_run_gc;
  __pyx_vtable_4cyjs_5_cyjs_Runtime.set_memory_limit = (void (*)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, size_t, int __pyx_skip_dispatch))__pyx_f_4cyjs_5_cyjs_7Runtime_set_memory_limit;
  __pyx_vtable_4cyjs_5_cyjs_Runtime.set_max_stack_size = (void (*)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, size_t, int __pyx_skip_dispatch))__pyx_f_4cyjs_5_cyjs_7Runtime_set_max_stack_size;
  __pyx_vtable_4cyjs_5_cyjs_Runtime.update_statck_top = (void (*)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, int __pyx_skip_dispatch))__pyx_f_4cyjs_5_cyjs_7Runtime_update_statck_top;
  __pyx_vtable_4cyjs_5_cyjs_Runtime.set_promise_hook = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Runtime *, PyObject *, int __pyx_skip_dispatch))__pyx_f_4cyjs_5_cyjs_7Runtime_set_promise_hook;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Runtime = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs_Runtime_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Runtime)) __PYX_ERR(0, 137, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs_Runtime_spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Runtime) < (0)) __PYX_ERR(0, 137, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Runtime = &__pyx_type_4cyjs_5_cyjs_Runtime;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Runtime) < (0)) __PYX_ERR(0, 137, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Runtime);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Runtime->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Runtime->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Runtime->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Runtime, __pyx_vtabptr_4cyjs_5_cyjs_Runtime) < (0)) __PYX_ERR(0, 137, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Runtime) < (0)) __PYX_ERR(0, 137, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_Runtime, (PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Runtime) < (0)) __PYX_ERR(0, 137, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Runtime) < (0)) __PYX_ERR(0, 137, __pyx_L1_error)
  __pyx_vtabptr_4cyjs_5_cyjs_Object = &__pyx_vtable_4cyjs_5_cyjs_Object;
  __pyx_vtable_4cyjs_5_cyjs_Object.init = (void (*)(struct __pyx_obj_4cyjs_5_cyjs_Object *, struct __pyx_obj_4cyjs_5_cyjs_Context *, JSValue))__pyx_f_4cyjs_5_cyjs_6Object_init;
  __pyx_vtable_4cyjs_5_cyjs_Object.to_json = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Object *, int __pyx_skip_dispatch))__pyx_f_4cyjs_5_cyjs_6Object_to_json;
  __pyx_vtable_4cyjs_5_cyjs_Object.eval = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Object *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_6Object_eval *__pyx_optional_args))__pyx_f_4cyjs_5_cyjs_6Object_eval;
  __pyx_vtable_4cyjs_5_cyjs_Object.eval_module = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Object *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_6Object_eval_module *__pyx_optional_args))__pyx_f_4cyjs_5_cyjs_6Object_eval_module;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Object = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs_Object_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Object)) __PYX_ERR(0, 543, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs_Object_spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Object) < (0)) __PYX_ERR(0, 543, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Object = &__pyx_type_4cyjs_5_cyjs_Object;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Object) < (0)) __PYX_ERR(0, 543, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Object);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Object->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Object->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Object->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Object, __pyx_vtabptr_4cyjs_5_cyjs_Object) < (0)) __PYX_ERR(0, 543, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Object) < (0)) __PYX_ERR(0, 543, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_Object, (PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Object) < (0)) __PYX_ERR(0, 543, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Object) < (0)) __PYX_ERR(0, 543, __pyx_L1_error)
  __pyx_vtabptr_4cyjs_5_cyjs_JSFunction = &__pyx_vtable_4cyjs_5_cyjs_JSFunction;
  __pyx_vtable_4cyjs_5_cyjs_JSFunction.__pyx_new = (struct __pyx_obj_4cyjs_5_cyjs_JSFunction *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *))__pyx_f_4cyjs_5_cyjs_10JSFunction_new;
  __pyx_vtable_4cyjs_5_cyjs_JSFunction.call_js = (JSValue (*)(struct __pyx_obj_4cyjs_5_cyjs_JSFunction *, JSContext *, JSValue, int, JSValue *, int))__pyx_f_4cyjs_5_cyjs_10JSFunction_call_js;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSFunction = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs_JSFunction_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSFunction)) __PYX_ERR(0, 706, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs_JSFunction_spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSFunction) < (0)) __PYX_ERR(0, 706, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSFunction = &__pyx_type_4cyjs_5_cyjs_JSFunction;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSFunction) < (0)) __PYX_ERR(0, 706, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSFunction);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSFunction->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSFunction->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSFunction->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSFunction, __pyx_vtabptr_4cyjs_5_cyjs_JSFunction) < (0)) __PYX_ERR(0, 706, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSFunction) < (0)) __PYX_ERR(0, 706, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_JSFunction, (PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSFunction) < (0)) __PYX_ERR(0, 706, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_JSFunction) < (0)) __PYX_ERR(0, 706, __pyx_L1_error)
  __pyx_vtabptr_4cyjs_5_cyjs_Context = &__pyx_vtable_4cyjs_5_cyjs_Context;
  __pyx_vtable_4cyjs_5_cyjs_Context.has_exception = (int (*)(struct __pyx_obj_4cyjs_5_cyjs_Context *))__pyx_f_4cyjs_5_cyjs_7Context_has_exception;
  __pyx_vtable_4cyjs_5_cyjs_Context.get_exception = (JSValue (*)(struct __pyx_obj_4cyjs_5_cyjs_Context *))__pyx_f_4cyjs_5_cyjs_7Context_get_exception;
  __pyx_vtable_4cyjs_5_cyjs_Context.raise_exception = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *))__pyx_f_4cyjs_5_cyjs_7Context_raise_exception;
  __pyx_vtable_4cyjs_5_cyjs_Context.ceval = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_ceval *__pyx_optional_args))__pyx_f_4cyjs_5_cyjs_7Context_ceval;
  __pyx_vtable_4cyjs_5_cyjs_Context.eval = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval *__pyx_optional_args))__pyx_f_4cyjs_5_cyjs_7Context_eval;
  __pyx_vtable_4cyjs_5_cyjs_Context.eval_module = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_module *__pyx_optional_args))__pyx_f_4cyjs_5_cyjs_7Context_eval_module;
  __pyx_vtable_4cyjs_5_cyjs_Context.add_function = (struct __pyx_obj_4cyjs_5_cyjs_JSFunction *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_add_function *__pyx_optional_args))__pyx_f_4cyjs_5_cyjs_7Context_add_function;
  __pyx_vtable_4cyjs_5_cyjs_Context.eval_this = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_this *__pyx_optional_args))__pyx_f_4cyjs_5_cyjs_7Context_eval_this;
  __pyx_vtable_4cyjs_5_cyjs_Context.eval_this_with_module = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4cyjs_5_cyjs_7Context_eval_this_with_module *__pyx_optional_args))__pyx_f_4cyjs_5_cyjs_7Context_eval_this_with_module;
  __pyx_vtable_4cyjs_5_cyjs_Context.__pyx_fuse_0ceval_this = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, struct __pyx_obj_4cyjs_5_cyjs_JSFunction *, struct __pyx_fuse_0__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args))__pyx_fuse_0__pyx_f_4cyjs_5_cyjs_7Context_ceval_this;
  __pyx_vtable_4cyjs_5_cyjs_Context.__pyx_fuse_1ceval_this = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, struct __pyx_obj_4cyjs_5_cyjs_Object *, struct __pyx_fuse_1__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args))__pyx_fuse_1__pyx_f_4cyjs_5_cyjs_7Context_ceval_this;
  __pyx_vtable_4cyjs_5_cyjs_Context.__pyx_fuse_2ceval_this = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, PyObject *, struct __pyx_fuse_2__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args))__pyx_fuse_2__pyx_f_4cyjs_5_cyjs_7Context_ceval_this;
  __pyx_vtable_4cyjs_5_cyjs_Context.__pyx_fuse_3ceval_this = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, PyObject *, struct __pyx_fuse_3__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args))__pyx_fuse_3__pyx_f_4cyjs_5_cyjs_7Context_ceval_this;
  __pyx_vtable_4cyjs_5_cyjs_Context.__pyx_fuse_4ceval_this = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, PyObject *, struct __pyx_fuse_4__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args))__pyx_fuse_4__pyx_f_4cyjs_5_cyjs_7Context_ceval_this;
  __pyx_vtable_4cyjs_5_cyjs_Context.__pyx_fuse_5ceval_this = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, PyObject *, struct __pyx_fuse_5__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args))__pyx_fuse_5__pyx_f_4cyjs_5_cyjs_7Context_ceval_this;
  __pyx_vtable_4cyjs_5_cyjs_Context.__pyx_fuse_6ceval_this = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, int, struct __pyx_fuse_6__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args))__pyx_fuse_6__pyx_f_4cyjs_5_cyjs_7Context_ceval_this;
  __pyx_vtable_4cyjs_5_cyjs_Context.__pyx_fuse_7ceval_this = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, float, struct __pyx_fuse_7__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args))__pyx_fuse_7__pyx_f_4cyjs_5_cyjs_7Context_ceval_this;
  __pyx_vtable_4cyjs_5_cyjs_Context.__pyx_fuse_8ceval_this = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, PyObject *, struct __pyx_fuse_8__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args))__pyx_fuse_8__pyx_f_4cyjs_5_cyjs_7Context_ceval_this;
  __pyx_vtable_4cyjs_5_cyjs_Context.__pyx_fuse_9ceval_this = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Context *, PyObject *, int, struct __pyx_fuse_9__pyx_opt_args_4cyjs_5_cyjs_7Context_ceval_this *__pyx_optional_args))__pyx_fuse_9__pyx_f_4cyjs_5_cyjs_7Context_ceval_this;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Context = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs_Context_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Context)) __PYX_ERR(0, 776, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs_Context_spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Context) < (0)) __PYX_ERR(0, 776, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Context = &__pyx_type_4cyjs_5_cyjs_Context;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Context) < (0)) __PYX_ERR(0, 776, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Context);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Context->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Context->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Context->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Context, __pyx_vtabptr_4cyjs_5_cyjs_Context) < (0)) __PYX_ERR(0, 776, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Context) < (0)) __PYX_ERR(0, 776, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_Context, (PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Context) < (0)) __PYX_ERR(0, 776, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Context) < (0)) __PYX_ERR(0, 776, __pyx_L1_error)
  __pyx_vtabptr_4cyjs_5_cyjs__OView = &__pyx_vtable_4cyjs_5_cyjs__OView;
  __pyx_vtable_4cyjs_5_cyjs__OView.cnext = (int (*)(struct __pyx_obj_4cyjs_5_cyjs__OView *, size_t *, JSAtom *, JSValue *))__pyx_f_4cyjs_5_cyjs_6_OView_cnext;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__OView = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs__OView_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__OView)) __PYX_ERR(0, 450, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs__OView_spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__OView) < (0)) __PYX_ERR(0, 450, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__OView = &__pyx_type_4cyjs_5_cyjs__OView;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__OView) < (0)) __PYX_ERR(0, 450, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__OView);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__OView->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__OView->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__OView->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__OView, __pyx_vtabptr_4cyjs_5_cyjs__OView) < (0)) __PYX_ERR(0, 450, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__OView) < (0)) __PYX_ERR(0, 450, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_OView, (PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__OView) < (0)) __PYX_ERR(0, 450, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__OView) < (0)) __PYX_ERR(0, 450, __pyx_L1_error)
  __pyx_vtabptr_4cyjs_5_cyjs__ObjectItemsView = &__pyx_vtable_4cyjs_5_cyjs__ObjectItemsView;
  __pyx_vtable_4cyjs_5_cyjs__ObjectItemsView.__pyx_base = *__pyx_vtabptr_4cyjs_5_cyjs__OView;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs__OView); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs__ObjectItemsView_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView)) __PYX_ERR(0, 495, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs__ObjectItemsView_spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView) < (0)) __PYX_ERR(0, 495, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView = &__pyx_type_4cyjs_5_cyjs__ObjectItemsView;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView->tp_base = __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs__OView;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView) < (0)) __PYX_ERR(0, 495, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView, __pyx_vtabptr_4cyjs_5_cyjs__ObjectItemsView) < (0)) __PYX_ERR(0, 495, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView) < (0)) __PYX_ERR(0, 495, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_ObjectItemsView, (PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView) < (0)) __PYX_ERR(0, 495, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectItemsView) < (0)) __PYX_ERR(0, 495, __pyx_L1_error)
  __pyx_vtabptr_4cyjs_5_cyjs__ObjectKeysView = &__pyx_vtable_4cyjs_5_cyjs__ObjectKeysView;
  __pyx_vtable_4cyjs_5_cyjs__ObjectKeysView.__pyx_base = *__pyx_vtabptr_4cyjs_5_cyjs__OView;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs__OView); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs__ObjectKeysView_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView)) __PYX_ERR(0, 505, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs__ObjectKeysView_spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView) < (0)) __PYX_ERR(0, 505, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView = &__pyx_type_4cyjs_5_cyjs__ObjectKeysView;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView->tp_base = __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs__OView;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView) < (0)) __PYX_ERR(0, 505, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView, __pyx_vtabptr_4cyjs_5_cyjs__ObjectKeysView) < (0)) __PYX_ERR(0, 505, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView) < (0)) __PYX_ERR(0, 505, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_ObjectKeysView, (PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView) < (0)) __PYX_ERR(0, 505, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectKeysView) < (0)) __PYX_ERR(0, 505, __pyx_L1_error)
  __pyx_vtabptr_4cyjs_5_cyjs__ObjectValuesView = &__pyx_vtable_4cyjs_5_cyjs__ObjectValuesView;
  __pyx_vtable_4cyjs_5_cyjs__ObjectValuesView.__pyx_base = *__pyx_vtabptr_4cyjs_5_cyjs__OView;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs__OView); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs__ObjectValuesView_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView)) __PYX_ERR(0, 522, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs__ObjectValuesView_spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView) < (0)) __PYX_ERR(0, 522, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView = &__pyx_type_4cyjs_5_cyjs__ObjectValuesView;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView->tp_base = __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs__OView;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView) < (0)) __PYX_ERR(0, 522, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView, __pyx_vtabptr_4cyjs_5_cyjs__ObjectValuesView) < (0)) __PYX_ERR(0, 522, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView) < (0)) __PYX_ERR(0, 522, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_ObjectValuesView, (PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView) < (0)) __PYX_ERR(0, 522, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs__ObjectValuesView) < (0)) __PYX_ERR(0, 522, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)(((PyTypeObject*)PyExc_Exception))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_CancelledError = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs_CancelledError_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_CancelledError)) __PYX_ERR(0, 1137, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs_CancelledError_spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_CancelledError) < (0)) __PYX_ERR(0, 1137, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_CancelledError = &__pyx_type_4cyjs_5_cyjs_CancelledError;
  #endif
  if (sizeof(struct __pyx_obj_4cyjs_5_cyjs_CancelledError) != sizeof(PyBaseExceptionObject)) {
    if (__Pyx_validate_extern_base((((PyTypeObject*)PyExc_Exception))) < (0)) __PYX_ERR(0, 1137, __pyx_L1_error)
  }
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_CancelledError->tp_base = (((PyTypeObject*)PyExc_Exception));
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_CancelledError) < (0)) __PYX_ERR(0, 1137, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_CancelledError);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_CancelledError->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_CancelledError->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_CancelledError->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_CancelledError, (PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_CancelledError) < (0)) __PYX_ERR(0, 1137, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_CancelledError) < (0)) __PYX_ERR(0, 1137, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)(((PyTypeObject*)PyExc_Exception))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs_InvalidStateError_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError)) __PYX_ERR(0, 1140, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs_InvalidStateError_spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError) < (0)) __PYX_ERR(0, 1140, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError = &__pyx_type_4cyjs_5_cyjs_InvalidStateError;
  #endif
  if (sizeof(struct __pyx_obj_4cyjs_5_cyjs_InvalidStateError) != sizeof(PyBaseExceptionObject)) {
    if (__Pyx_validate_extern_base((((PyTypeObject*)PyExc_Exception))) < (0)) __PYX_ERR(0, 1140, __pyx_L1_error)
  }
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError->tp_base = (((PyTypeObject*)PyExc_Exception));
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError) < (0)) __PYX_ERR(0, 1140, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_InvalidStateError, (PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError) < (0)) __PYX_ERR(0, 1140, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError) < (0)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __pyx_vtabptr_4cyjs_5_cyjs_Promise = &__pyx_vtable_4cyjs_5_cyjs_Promise;
  __pyx_vtable_4cyjs_5_cyjs_Promise.__pyx_base = *__pyx_vtabptr_4cyjs_5_cyjs_Object;
  __pyx_vtable_4cyjs_5_cyjs_Promise.add_done_callback = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Promise *, PyObject *, int __pyx_skip_dispatch))__pyx_f_4cyjs_5_cyjs_7Promise_add_done_callback;
  __pyx_vtable_4cyjs_5_cyjs_Promise.exception = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Promise *, int __pyx_skip_dispatch))__pyx_f_4cyjs_5_cyjs_7Promise_exception;
  __pyx_vtable_4cyjs_5_cyjs_Promise.done = (int (*)(struct __pyx_obj_4cyjs_5_cyjs_Promise *, int __pyx_skip_dispatch))__pyx_f_4cyjs_5_cyjs_7Promise_done;
  __pyx_vtable_4cyjs_5_cyjs_Promise.remove_done_callback = (Py_ssize_t (*)(struct __pyx_obj_4cyjs_5_cyjs_Promise *, PyObject *, int __pyx_skip_dispatch))__pyx_f_4cyjs_5_cyjs_7Promise_remove_done_callback;
  __pyx_vtable_4cyjs_5_cyjs_Promise.result = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Promise *, int __pyx_skip_dispatch))__pyx_f_4cyjs_5_cyjs_7Promise_result;
  __pyx_vtable_4cyjs_5_cyjs_Promise.complete = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Promise *))__pyx_f_4cyjs_5_cyjs_7Promise_complete;
  __pyx_vtable_4cyjs_5_cyjs_Promise.poll = (PyObject *(*)(struct __pyx_obj_4cyjs_5_cyjs_Promise *, int __pyx_skip_dispatch))__pyx_f_4cyjs_5_cyjs_7Promise_poll;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Promise = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs_Promise_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Promise)) __PYX_ERR(0, 1144, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs_Promise_spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Promise) < (0)) __PYX_ERR(0, 1144, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Promise = &__pyx_type_4cyjs_5_cyjs_Promise;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Promise->tp_base = __pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Promise) < (0)) __PYX_ERR(0, 1144, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Promise);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Promise->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Promise->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Promise->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Promise, __pyx_vtabptr_4cyjs_5_cyjs_Promise) < (0)) __PYX_ERR(0, 1144, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Promise) < (0)) __PYX_ERR(0, 1144, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_Promise, (PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Promise) < (0)) __PYX_ERR(0, 1144, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_Promise) < (0)) __PYX_ERR(0, 1144, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_NODEFAULT = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs_NODEFAULT_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_NODEFAULT)) __PYX_ERR(0, 1243, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs_NODEFAULT_spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_NODEFAULT) < (0)) __PYX_ERR(0, 1243, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_NODEFAULT = &__pyx_type_4cyjs_5_cyjs_NODEFAULT;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_NODEFAULT) < (0)) __PYX_ERR(0, 1243, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_NODEFAULT);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_NODEFAULT->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_NODEFAULT->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_NODEFAULT->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs_NODEFAULT) < (0)) __PYX_ERR(0, 1243, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct____iter__ = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs___pyx_scope_struct____iter___spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct____iter__)) __PYX_ERR(0, 496, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs___pyx_scope_struct____iter___spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct____iter__) < (0)) __PYX_ERR(0, 496, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct____iter__ = &__pyx_type_4cyjs_5_cyjs___pyx_scope_struct____iter__;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct____iter__) < (0)) __PYX_ERR(0, 496, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct____iter__);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct____iter__->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct____iter__->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct____iter__->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs___pyx_scope_struct_1___iter___spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_1___iter__)) __PYX_ERR(0, 506, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs___pyx_scope_struct_1___iter___spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_1___iter__) < (0)) __PYX_ERR(0, 506, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_1___iter__ = &__pyx_type_4cyjs_5_cyjs___pyx_scope_struct_1___iter__;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_1___iter__) < (0)) __PYX_ERR(0, 506, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_1___iter__);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_1___iter__->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_1___iter__->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_1___iter__->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4cyjs_5_cyjs___pyx_scope_struct_2___iter___spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_2___iter__)) __PYX_ERR(0, 523, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4cyjs_5_cyjs___pyx_scope_struct_2___iter___spec, __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_2___iter__) < (0)) __PYX_ERR(0, 523, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_2___iter__ = &__pyx_type_4cyjs_5_cyjs___pyx_scope_struct_2___iter__;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_2___iter__) < (0)) __PYX_ERR(0, 523, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_2___iter__);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_2___iter__->tp_dictoffset && __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_2___iter__->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_4cyjs_5_cyjs___pyx_scope_struct_2___iter__->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType_3_2_4(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyLongObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyLongObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyLongObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyLongObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(3, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_2_4(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_4type_type) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__cyjs(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__cyjs},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "_cyjs",
      0, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit__cyjs(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__cyjs(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__cyjs(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  size_t __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_cyjs' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "_cyjs" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_NOT_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit__cyjs", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_cyjs___cyjs) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "cyjs._cyjs")) {
      if (unlikely((PyDict_SetItemString(modules, "cyjs._cyjs", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  (void)__Pyx_modinit_function_export_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_type_init_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  (void)__Pyx_modinit_function_import_code(__pyx_mstate);
  /*--- Execution code ---*/

  /* "EnumBase":10
 * 
 * cdef object __Pyx_EnumBase
 * from enum import IntEnum as __Pyx_EnumBase             # <<<<<<<<<<<<<<
 * 
 * cdef object __Pyx_FlagBase
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntEnum};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_enum, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 10, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntEnum};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 10, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      switch (__pyx_t_3) {
        case 0:
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_XGOTREF(__Pyx_EnumBase);
        __Pyx_DECREF_SET(__Pyx_EnumBase, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        break;
        default:;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumBase":13
 * 
 * cdef object __Pyx_FlagBase
 * from enum import IntFlag as __Pyx_FlagBase             # <<<<<<<<<<<<<<
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntFlag};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_enum, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntFlag};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      switch (__pyx_t_3) {
        case 0:
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_XGOTREF(__Pyx_FlagBase);
        __Pyx_DECREF_SET(__Pyx_FlagBase, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        break;
        default:;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":23
 * 
 * 
 * PromiseHookType = __Pyx_FlagBase('PromiseHookType', [             # <<<<<<<<<<<<<<
 *     ('INIT', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(INIT)),
 *     ('BEFORE', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(BEFORE)),
*/
  __pyx_t_4 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_5 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * PromiseHookType = __Pyx_FlagBase('PromiseHookType', [
 *     ('INIT', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(INIT)),             # <<<<<<<<<<<<<<
 *     ('BEFORE', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(BEFORE)),
 *     ('AFTER', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(AFTER)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(__pyx_e_4cyjs_5_cyjs_INIT); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_INIT, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":25
 * PromiseHookType = __Pyx_FlagBase('PromiseHookType', [
 *     ('INIT', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(INIT)),
 *     ('BEFORE', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(BEFORE)),             # <<<<<<<<<<<<<<
 *     ('AFTER', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(AFTER)),
 *     ('RESOLVE', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(RESOLVE)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(__pyx_e_4cyjs_5_cyjs_BEFORE); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_BEFORE, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":26
 *     ('INIT', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(INIT)),
 *     ('BEFORE', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(BEFORE)),
 *     ('AFTER', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(AFTER)),             # <<<<<<<<<<<<<<
 *     ('RESOLVE', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(RESOLVE)),
 * 
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(__pyx_e_4cyjs_5_cyjs_AFTER); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_AFTER, __pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":27
 *     ('BEFORE', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(BEFORE)),
 *     ('AFTER', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(AFTER)),
 *     ('RESOLVE', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(RESOLVE)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'cyjs._cyjs'))
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(__pyx_e_4cyjs_5_cyjs_RESOLVE); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_RESOLVE, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":23
 * 
 * 
 * PromiseHookType = __Pyx_FlagBase('PromiseHookType', [             # <<<<<<<<<<<<<<
 *     ('INIT', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(INIT)),
 *     ('BEFORE', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(BEFORE)),
*/
  __pyx_t_6 = __Pyx_PyList_Pack(4, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":29
 *     ('RESOLVE', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(RESOLVE)),
 * 
 * ], module=globals().get("__module__", 'cyjs._cyjs'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_10 = __Pyx_Globals(); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_11 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PromiseHookType, __pyx_t_6};
    __pyx_t_9 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module_2, __pyx_t_10, __pyx_t_9, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_11, (3-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_PromiseHookType, __pyx_t_2) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":31
 * ], module=globals().get("__module__", 'cyjs._cyjs'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_12 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_12) {

    /* "EnumType":36
 * 
 * 
 *     PromiseHookType._member_names_ = list(PromiseHookType.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PromiseHookType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PromiseHookType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_2) < (0)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "EnumType":31
 * ], module=globals().get("__module__", 'cyjs._cyjs'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "cyjs/_cyjs.pyx":30
 * # TODO: I'm writing a new cython utils library and a yyjson writer would be a good use-case here...
 * 
 * import json             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_json, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_json, __pyx_t_5) < (0)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7JSError_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_JSError___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[3])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_JSError, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_JSError, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_JSError__set_state(self, __pyx_state)
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7JSError_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_JSError___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_JSError, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_11MemoryUsage_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_MemoryUsage___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MemoryUsage, (type(self), 0xe4999e7, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MemoryUsage__set_state(self, __pyx_state)
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_11MemoryUsage_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_MemoryUsage___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[6])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_MemoryUsage, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_11PromiseHook_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PromiseHook___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[7])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_PromiseHook, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromiseHook, (type(self), 0x84cfe3b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromiseHook__set_state(self, __pyx_state)
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_11PromiseHook_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PromiseHook___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[8])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_PromiseHook, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":143
 *         self.has_promise_hook = False
 * 
 *     cpdef MemoryUsage compute_memory_usage(self):             # <<<<<<<<<<<<<<
 *         cdef JSMemoryUsage mu
 *         JS_ComputeMemoryUsage(self.rt, &mu)
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Runtime_3compute_memory_usage, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Runtime_compute_memory_usage, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[9])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime, __pyx_mstate_global->__pyx_n_u_compute_memory_usage, __pyx_t_5) < (0)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":148
 *         return MemoryUsage.new(&mu)
 * 
 *     cpdef object dump_memory_usage(self, object file):             # <<<<<<<<<<<<<<
 *         if CYJS_DumpMemoryUsage(self.rt, file) < 0:
 *             raise
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Runtime_5dump_memory_usage, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Runtime_dump_memory_usage, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[10])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime, __pyx_mstate_global->__pyx_n_u_dump_memory_usage, __pyx_t_5) < (0)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":152
 *             raise
 * 
 *     cpdef object execute_pending_job(self):             # <<<<<<<<<<<<<<
 *         cdef JSContext* ctx = NULL
 *         cdef Context py_context
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Runtime_7execute_pending_job, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Runtime_execute_pending_job, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[11])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime, __pyx_mstate_global->__pyx_n_u_execute_pending_job, __pyx_t_5) < (0)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":166
 * 
 * 
 *     cpdef bint is_job_pending(self):             # <<<<<<<<<<<<<<
 *         return JS_IsJobPending(self.rt)
 * 
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Runtime_9is_job_pending, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Runtime_is_job_pending, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[12])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime, __pyx_mstate_global->__pyx_n_u_is_job_pending, __pyx_t_5) < (0)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":176
 *         return ctx
 * 
 *     cpdef void run_gc(self):             # <<<<<<<<<<<<<<
 *         """
 *         Runs QuickJS-NG's internal Garbage Collector
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Runtime_11run_gc, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Runtime_run_gc, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[13])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime, __pyx_mstate_global->__pyx_n_u_run_gc, __pyx_t_5) < (0)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":185
 * 
 *     # While I would go ahead start to inline these, python also needs a way to call them...
 *     cpdef void set_memory_limit(self, size_t limit):             # <<<<<<<<<<<<<<
 *         JS_SetMemoryLimit(self.rt, limit)
 * 
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Runtime_13set_memory_limit, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Runtime_set_memory_limit, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[14])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime, __pyx_mstate_global->__pyx_n_u_set_memory_limit, __pyx_t_5) < (0)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":188
 *         JS_SetMemoryLimit(self.rt, limit)
 * 
 *     cpdef void set_max_stack_size(self, size_t max_stack_size):             # <<<<<<<<<<<<<<
 *         JS_SetMaxStackSize(self.rt, max_stack_size)
 * 
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Runtime_15set_max_stack_size, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Runtime_set_max_stack_size, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[15])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime, __pyx_mstate_global->__pyx_n_u_set_max_stack_size, __pyx_t_5) < (0)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":191
 *         JS_SetMaxStackSize(self.rt, max_stack_size)
 * 
 *     cpdef void update_statck_top(self):             # <<<<<<<<<<<<<<
 *         JS_UpdateStackTop(self.rt)
 * 
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Runtime_17update_statck_top, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Runtime_update_statck_top, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[16])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime, __pyx_mstate_global->__pyx_n_u_update_statck_top, __pyx_t_5) < (0)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":194
 *         JS_UpdateStackTop(self.rt)
 * 
 *     cpdef object set_promise_hook(self, object func):             # <<<<<<<<<<<<<<
 *         if not callable(func):
 *             raise TypeError("Promise hook must be callable.")
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Runtime_19set_promise_hook, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Runtime_set_promise_hook, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[17])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime, __pyx_mstate_global->__pyx_n_u_set_promise_hook, __pyx_t_5) < (0)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.rt cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Runtime_23__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Runtime___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[18])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.rt cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.rt cannot be converted to a Python object for pickling"
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Runtime_25__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Runtime___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[19])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_6_OView_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_OView___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[20])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_6_OView_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_OView___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[21])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_16_ObjectItemsView_4__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ObjectItemsView___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[22])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_16_ObjectItemsView_6__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ObjectItemsView___setstate_cyth, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[23])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_15_ObjectKeysView_6__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ObjectKeysView___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[24])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_15_ObjectKeysView_8__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ObjectKeysView___setstate_cytho, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[25])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_17_ObjectValuesView_6__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ObjectValuesView___reduce_cytho, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[26])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.tab,self.value cannot be converted to a Python object for pickling"
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_17_ObjectValuesView_8__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ObjectValuesView___setstate_cyt, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[27])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":552
 *         self.tag = JS_VALUE_GET_TAG(self.value)
 * 
 *     cpdef bytes to_json(self):             # <<<<<<<<<<<<<<
 *         """
 *         Useful when debugging or handling unknown js to py conversions
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_6Object_1to_json, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Object_to_json, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[28])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object, __pyx_mstate_global->__pyx_n_u_to_json, __pyx_t_5) < (0)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":567
 *     # serves for debugging type tags
 * 
 *     def get(self, key):             # <<<<<<<<<<<<<<
 * 
 *         cdef JSValue v
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_6Object_3get, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Object_get, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[29])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object, __pyx_mstate_global->__pyx_n_u_get, __pyx_t_5) < (0)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":582
 *         return to_python(ctx, JS_DupValue(ctx, v))
 * 
 *     def set(self, object key, object value):             # <<<<<<<<<<<<<<
 *         cdef JSContext* ctx = self.ctx
 *         cdef JSValue js_value
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_6Object_5set, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Object_set, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[30])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object, __pyx_mstate_global->__pyx_n_u_set, __pyx_t_5) < (0)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":628
 *             arg_parser_finish(&p)
 * 
 *     def invoke(self, object func, *args):             # <<<<<<<<<<<<<<
 *         cdef JSAtom atom
 *         cdef JSContext* ctx = self.ctx
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_6Object_9invoke, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Object_invoke, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[31])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object, __pyx_mstate_global->__pyx_n_u_invoke, __pyx_t_5) < (0)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":652
 * 
 *     # Mapping attributes can help with debugging attributes
 *     def items(self):             # <<<<<<<<<<<<<<
 *         return _ObjectItemsView(self)
 * 
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_6Object_13items, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Object_items, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[32])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object, __pyx_mstate_global->__pyx_n_u_items, __pyx_t_5) < (0)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":655
 *         return _ObjectItemsView(self)
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         return _ObjectValuesView(self)
 * 
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_6Object_15values, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Object_values, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[33])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object, __pyx_mstate_global->__pyx_n_u_values, __pyx_t_5) < (0)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":658
 *         return _ObjectValuesView(self)
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         return _ObjectKeysView(self)
 * 
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_6Object_17keys, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Object_keys, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[34])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object, __pyx_mstate_global->__pyx_n_u_keys, __pyx_t_5) < (0)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":663
 *     # eval_this functions (proxied for the sake of easier access)
 * 
 *     cpdef object eval(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_6Object_19eval, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Object_eval, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[35])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_mstate_global->__pyx_tuple[1]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object, __pyx_mstate_global->__pyx_n_u_eval, __pyx_t_5) < (0)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":673
 *         return self.context.ceval_this(code, self, filename, False, strict, backtrace_barrier, promise)
 * 
 *     cpdef object eval_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_6Object_21eval_module, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Object_eval_module, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[36])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_mstate_global->__pyx_tuple[1]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Object, __pyx_mstate_global->__pyx_n_u_eval_module, __pyx_t_5) < (0)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_6Object_23__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Object___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[37])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_6Object_25__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Object___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[38])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_10JSFunction_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_JSFunction___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[39])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_10JSFunction_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_JSFunction___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[40])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":778
 * cdef class Context: # type: ignore
 *     def __init__(self,
 *         Runtime runtime = Runtime(),             # <<<<<<<<<<<<<<
 *         bint base_objects = True,
 *         bint date = True,
*/
  __pyx_t_2 = NULL;
  __pyx_t_11 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Runtime, __pyx_callargs+__pyx_t_11, (1-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_5);
  }
  __pyx_mstate_global->__pyx_k_ = ((struct __pyx_obj_4cyjs_5_cyjs_Runtime *)__pyx_t_5);
  __Pyx_GIVEREF((PyObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":893
 *         return to_python(self.ctx, val)
 * 
 *     cpdef object eval(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Context_3eval, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Context_eval, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[41])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_mstate_global->__pyx_tuple[1]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Context, __pyx_mstate_global->__pyx_n_u_eval, __pyx_t_5) < (0)) __PYX_ERR(0, 893, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":903
 *         return self.ceval(code, filename, module=False, strict=strict, backtrace_barrier=backtrace_barrier, promise=promise)
 * 
 *     cpdef object eval_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Context_5eval_module, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Context_eval_module, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[42])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_mstate_global->__pyx_tuple[1]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Context, __pyx_mstate_global->__pyx_n_u_eval_module, __pyx_t_5) < (0)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":916
 *     # Otherwise Run JS_GetGlobalObject immediately for duping
 *     # and free on __dealloc__ instead.
 *     def get_global(self):             # <<<<<<<<<<<<<<
 *         return to_python(self.ctx, JS_GetGlobalObject(self.ctx))
 * 
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Context_7get_global, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Context_get_global, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[43])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Context, __pyx_mstate_global->__pyx_n_u_get_global, __pyx_t_5) < (0)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":920
 * 
 *     # also inspired from seeing rquickjs (rust) and retired python-quickjs
 *     def json_parse(self, object json):             # <<<<<<<<<<<<<<
 *         cdef JSValue ret
 *         cdef Py_buffer view
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Context_9json_parse, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Context_json_parse, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[44])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Context, __pyx_mstate_global->__pyx_n_u_json_parse, __pyx_t_5) < (0)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":936
 *     # pretty much supporting
 *     # def get(self, name: bytes | bytearray | array.array | memoryview | str)
 *     def get(self, object name):             # <<<<<<<<<<<<<<
 *         """Implements a Shortcut for converting a global object to a python object and setting a value to utilize
 *         off of."""
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Context_11get, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Context_get, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[45])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Context, __pyx_mstate_global->__pyx_n_u_get, __pyx_t_5) < (0)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":948
 *         return to_python(self.ctx, val)
 * 
 *     def set(self, object name, object item):             # <<<<<<<<<<<<<<
 *         """Sets an item to the current globalThis object"""
 *         cdef JSValue value
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Context_13set, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Context_set, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[46])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Context, __pyx_mstate_global->__pyx_n_u_set, __pyx_t_5) < (0)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":966
 * 
 * 
 *     cpdef JSFunction add_function(             # <<<<<<<<<<<<<<
 *         self,
 *         object func,
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Context_15add_function, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Context_add_function, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[47])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_mstate_global->__pyx_tuple[2]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Context, __pyx_mstate_global->__pyx_n_u_add_function, __pyx_t_5) < (0)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":1075
 *         return to_python(self.ctx, val)
 * 
 *     cpdef object eval_this(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Context_17eval_this, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Context_eval_this, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[48])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_mstate_global->__pyx_tuple[1]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Context, __pyx_mstate_global->__pyx_n_u_eval_this, __pyx_t_5) < (0)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":1086
 *         return self.ceval_this(code, this, filename, False, strict, backtrace_barrier, promise)
 * 
 *     cpdef object eval_this_with_module(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Context_19eval_this_with_module, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Context_eval_this_with_module, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[49])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_mstate_global->__pyx_tuple[1]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Context, __pyx_mstate_global->__pyx_n_u_eval_this_with_module, __pyx_t_5) < (0)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.rt cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Context_21__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Context___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[50])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.rt cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.rt cannot be converted to a Python object for pickling"
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Context_23__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Context___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[51])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_14CancelledError_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CancelledError___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[52])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_CancelledError, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CancelledError, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CancelledError__set_state(self, __pyx_state)
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_14CancelledError_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CancelledError___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[53])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_CancelledError, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_17InvalidStateError_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_InvalidStateError___reduce_cytho, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[54])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InvalidStateError, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InvalidStateError__set_state(self, __pyx_state)
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_17InvalidStateError_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_InvalidStateError___setstate_cyt, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[55])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_InvalidStateError, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":1151
 *         bint completed
 * 
 *     cpdef object add_done_callback(self, object fn):             # <<<<<<<<<<<<<<
 *         """Attaches a callable callback when promise finishes or raises an exception"""
 *         if not callable(fn):
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Promise_1add_done_callback, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Promise_add_done_callback, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[56])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Promise, __pyx_mstate_global->__pyx_n_u_add_done_callback, __pyx_t_5) < (0)) __PYX_ERR(0, 1151, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":1161
 *             fn(self)
 * 
 *     cpdef object exception(self):             # <<<<<<<<<<<<<<
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)
 *         if state == JS_PROMISE_REJECTED:
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Promise_3exception, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Promise_exception, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[57])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Promise, __pyx_mstate_global->__pyx_n_u_exception, __pyx_t_5) < (0)) __PYX_ERR(0, 1161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":1169
 *         return self._exception
 * 
 *     cpdef bint done(self):             # <<<<<<<<<<<<<<
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)
 *         return state == JS_PROMISE_FULFILLED or state == JS_PROMISE_REJECTED
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Promise_5done, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Promise_done, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[58])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Promise, __pyx_mstate_global->__pyx_n_u_done, __pyx_t_5) < (0)) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":1176
 * 
 * 
 *     cpdef Py_ssize_t remove_done_callback(self, object fn):             # <<<<<<<<<<<<<<
 *         """Remove all instances of a callback from the "call when done" list.
 * 
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Promise_7remove_done_callback, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Promise_remove_done_callback, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[59])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Promise, __pyx_mstate_global->__pyx_n_u_remove_done_callback, __pyx_t_5) < (0)) __PYX_ERR(0, 1176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":1190
 *         return removed_count
 * 
 *     cpdef object result(self):             # <<<<<<<<<<<<<<
 *         cdef JSPromiseStateEnum state = JS_PromiseState(self.ctx, self.value)
 *         if state == JS_PROMISE_REJECTED:
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Promise_9result, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Promise_result, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[60])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Promise, __pyx_mstate_global->__pyx_n_u_result, __pyx_t_5) < (0)) __PYX_ERR(0, 1190, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":1218
 *     # TODO: might be able to utilize __await__ and __iter__ for polling
 * 
 *     cpdef object poll(self):             # <<<<<<<<<<<<<<
 *         """Polls QuickJS Eventloop a single cycle while attempting
 *         to wait for this Promise to complete"""
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Promise_11poll, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Promise_poll, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[61])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_Promise, __pyx_mstate_global->__pyx_n_u_poll, __pyx_t_5) < (0)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Promise_13__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Promise___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[62])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ctx,self.value cannot be converted to a Python object for pickling"
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7Promise_15__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Promise___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[63])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_9NODEFAULT_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NODEFAULT___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[64])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_NODEFAULT, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NODEFAULT, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NODEFAULT__set_state(self, __pyx_state)
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_9NODEFAULT_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NODEFAULT___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[65])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_4cyjs_5_cyjs_NODEFAULT, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_5) < (0)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_JSError(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_1__pyx_unpickle_JSError, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_JSError, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[66])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_JSError, __pyx_t_5) < (0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * cdef extern from *:             # <<<<<<<<<<<<<<
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_3__pyx_unpickle_MemoryUsage, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_MemoryUsage, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[67])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_MemoryUsage, __pyx_t_5) < (0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_PromiseHook(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x84cfe3b, 0x3581cd1, 0x33c4952, b'func, rt')
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_5__pyx_unpickle_PromiseHook, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_PromiseHook, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[68])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_PromiseHook, __pyx_t_5) < (0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * cdef extern from *:             # <<<<<<<<<<<<<<
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_7__pyx_unpickle_CancelledError, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_CancelledError, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[69])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_CancelledError, __pyx_t_5) < (0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_InvalidStateError(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe3b0c44, 0xda39a3e, 0xd41d8cd, b'')
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_9__pyx_unpickle_InvalidStateError, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_InvalidStateError, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[70])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_InvalidStateError, __pyx_t_5) < (0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * cdef extern from *:             # <<<<<<<<<<<<<<
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4cyjs_5_cyjs_11__pyx_unpickle_NODEFAULT, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_NODEFAULT, NULL, __pyx_mstate_global->__pyx_n_u_cyjs__cyjs, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[71])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_NODEFAULT, __pyx_t_5) < (0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cyjs/_cyjs.pyx":1
 * # cython: freethreading_compatible = True             # <<<<<<<<<<<<<<
 * from cpython.bool cimport PyBool_FromLong
 * from cpython.buffer cimport PyObject_CheckBuffer
*/
  __pyx_t_5 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_5) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init cyjs._cyjs", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init cyjs._cyjs");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);

  /* Cached unbound methods */
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.method_name = &__pyx_mstate->__pyx_n_u_items;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.method_name = &__pyx_mstate->__pyx_n_u_values;
  return 0;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "EnumType":29
 *     ('RESOLVE', __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(RESOLVE)),
 * 
 * ], module=globals().get("__module__", 'cyjs._cyjs'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_mstate_global->__pyx_tuple[0] = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_module, __pyx_mstate_global->__pyx_kp_u_cyjs__cyjs); if (unlikely(!__pyx_mstate_global->__pyx_tuple[0])) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[0]);

  /* "cyjs/_cyjs.pyx":663
 *     # eval_this functions (proxied for the sake of easier access)
 * 
 *     cpdef object eval(             # <<<<<<<<<<<<<<
 *         self,
 *         object code,
*/
  __pyx_mstate_global->__pyx_tuple[1] = PyTuple_Pack(4, Py_None, Py_False, Py_False, Py_False); if (unlikely(!__pyx_mstate_global->__pyx_tuple[1])) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[1]);

  /* "cyjs/_cyjs.pyx":966
 * 
 * 
 *     cpdef JSFunction add_function(             # <<<<<<<<<<<<<<
 *         self,
 *         object func,
*/
  __pyx_mstate_global->__pyx_tuple[2] = PyTuple_Pack(3, Py_None, __pyx_mstate_global->__pyx_int_neg_1, __pyx_mstate_global->__pyx_int_0); if (unlikely(!__pyx_mstate_global->__pyx_tuple[2])) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[2]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[2]);
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_tuple;
    for (Py_ssize_t i=0; i<3; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 9; } index[] = {{14},{28},{40},{179},{30},{20},{1},{1},{8},{10},{14},{7},{31},{6},{37},{26},{2},{9},{53},{68},{80},{71},{59},{14},{10},{15},{5},{6},{14},{32},{34},{7},{25},{27},{20},{12},{19},{17},{29},{11},{18},{18},{11},{4},{7},{7},{17},{35},{37},{7},{25},{27},{10},{28},{30},{11},{29},{31},{27},{29},{6},{24},{26},{6},{16},{25},{34},{36},{15},{24},{33},{35},{17},{26},{35},{37},{24},{26},{11},{18},{10},{13},{12},{11},{10},{14},{13},{7},{15},{29},{31},{25},{27},{25},{12},{17},{12},{28},{14},{20},{7},{7},{25},{27},{28},{25},{27},{22},{14},{26},{24},{24},{25},{17},{12},{15},{4},{18},{4},{17},{12},{6},{8},{18},{5},{4},{20},{5},{3},{10},{4},{8},{5},{13},{4},{17},{5},{4},{4},{11},{9},{21},{9},{19},{4},{8},{2},{4},{8},{3},{10},{12},{4},{1},{14},{6},{13},{14},{4},{5},{8},{7},{8},{4},{10},{3},{4},{5},{5},{8},{7},{14},{14},{11},{10},{6},{8},{4},{7},{4},{3},{1},{11},{4},{3},{3},{7},{5},{14},{12},{11},{10},{29},{32},{22},{26},{24},{26},{14},{12},{15},{10},{17},{13},{6},{15},{20},{6},{3},{6},{7},{4},{4},{3},{18},{16},{12},{16},{10},{12},{19},{5},{6},{8},{4},{5},{7},{12},{6},{17},{12},{1},{3},{5},{6},{4},{8},{7},{2},{11},{207},{12},{13},{59},{62},{69},{90},{89},{89},{89},{13},{108},{26},{11},{80},{79},{11},{13},{69},{122},{118},{104},{13},{71},{12},{23},{57},{77},{35},{73},{41},{41},{43},{43},{41},{41},{22},{9},{11},{111},{372},{11},{57},{55},{55},{58},{55},{11},{11}};
    #if (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (3076 bytes) */
const char* const cstring = "BZh91AY&SY|\215\030\207\000\003\003\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\300@@@@@@@@@@@@\000@\000`\013\375\333\270\371\227\261\222\366\007\227\013fZV\2425\265\204\210Os\324\000\316\241(D\250d\364\233P\323\320\323\0212\036\247\352xE2\000\014\203jh\006\200\003\324\006\206\200l\220\032hhz\215\014\202I\t\246\215\010\001\240\204\3232h\232OT\364\365O\322L\203i\244\r\000\000\321\372\240=@d\310i\220\r\0002\000h\004P\211\240\320\014\232\0004h\310\014#@\311\220h\321\352\001\246\000\206\2004\3654db\003M4\002MIG\251\244\247\264\223#\023\3216\243#A\240\301\017@\206\206\200\000\006A\220\0324\304\311\241\3514\001\243i\006D\032b`A\200\t\246\232i\200\010\300F&L\002`\000#\010\300\000#\000&L\230L@\221 @\223\023\024\323)\275O\"\247\251\355S\362S&\217S\324\365=\250\2152\r\006@\000\r\000\365\000\001\240h\006\233S\324\016\270\020\312M\336\205gfElXm%\027\222O\211A\003\245\325\325:\232\241!\017\320`\020\364\201~\324\231\234\225\034\206\243\206PZ]j~\036\376\007\262\245\021\250\322\211,\265v\332\035N\376_\312k\336\367\301\246 \032\210*\032\377\345\177/\237\341\237J\2015/J\t=\037F\342d\311\223,\262\336!\177\257W\326\375\376\337\271\267)\030\3061\247\245#\324\342\365\261{UU\356\353W-f\231\327]u\327\\\343\273S\245\342@;o\006\367\275\365\322\033P\333\n0`\250\t\025\221U\021U@H\221@X\210\030h\233)\"ne\346\177]\205\2001\033\010d\013\\0\223(\212\020U\"\263\005\340I\224\227T\303\273\340\016x\021\003\241\217'w\021\004\304Q\203M\275\273T)\014\026(3\021!E\256\261|\361JAz4AH\221\347QDdb,R\272\204\020`\370-\205\206Y\025Q\256\030\344W9HD\352\272\202\017\364q\013%\325\322\235MJE\"\246\220T\265q.\t\343\370\211\\\255\225%\237\025*\010*\032\346+\313p\237\023\344\n\322+\240\205\236W\036\377\320\330\216\376)x\227\257\3523\327\361\177P\350\230xl\007{\000\201\2230%\203\367\243\000D*\213$\225\226\232M\210B\211\351\200\241\315\322P\036\204U\262\212\222JC\332\336l\332\315Z\305\237~\276$\251=>E\n\016M\273\241\335\024jDF!\022\004X\256]\242q\364K\013+\016\203\351TZ\255V(\362\321C\323""\223\250y\253\267f\311g\2177\3250\227\254\332\003\353U\n\014\037\371\376\201\376\032\301S\333 \361\323n\035\334\213\320\\\331\230\333\017\034\035\336h\204\252\2539\315\322Ed\\J\237\266;\235\n\t\211\246\r\247FHr\032\210\210#\256\267\373:\353\324x5nEm\310\216hx\017\231\220.A@m9\357*\023p\242]\307\222\213\035\313\216\032\251(\211w\261\352\341\260Q\265y*r\362>\205A\004V`\262T\n\203\031\326\362\232 \365$\244Hz\221\024\257\337\n\265]\252e\035\220x{\343\265\273u|u\351}\035\\\006\\*\024\203\020\225\0078B\242\324\311\314\251I\rCA\r\037\323\325s\207\223\236\267JG\245w\267X\034\320cl\265\364/!\354\245\311V\352q8.\313i\335$\317\334\342h\301;\036u\373\234\3479\316s\262\213\006\017c7\312\340\016\335;\232\331\251\213>s:\236\336Q@1\006,GEut[\363\260_\365\367rw\232\220\307\367\031\245\032\026\3660\307\311\313Pe>\312\335\r0\321Z\253K}\377\345\000\311\256W\323#B\331\325P\272m\306\335\361\331\350B\276}<\217\314wb\274\371\325c\337\006k$\346\006\237\202\264C\272\203\2009u\232\007;\275\001;\227LZ\036\224r\361u\333~\214\271\006=\035\336\343\343|\361\257\272\213\366z`Q\324\032t\310u>W\3648\371\235n\342\235ee]\316\245\330\024U\251e\016E\300\300\3143*\025k\020q\361\303\371\365\340\322\336\362\337{\224S'?\r\021\373W}\007\374k<\273al\326\313\206\177}|.|\377\017{\014Ot\202\212\n:>\205\240\245UE\002\336\225\001\274\204F 0\306\225\004\337\252)\332\334\247\035\254\326\366\021\351\036\376\245\031Q*b\025\367|\311\300\222o;{\252\261%u0g\244\240\343\035\205\3354\311\205\321\343\247\274\300\250\001B\320\336\226g\234\314\2102\035\001\207\270\300q\351\034$\025+\020Ff\365\370\335\034AMpel\216\364\016\245\001-oS1`\025\232\232\201\310\304\034\313\246\330\262\014\360\317S\311\267Pn\315\341\310\213\261\223\322h(\2328\264\275\017M\327\201\326\367'\2021+\376\361m\241r \372wI\035\260kBZ\274\001<\325\007\330\321\006$\326e\031\261\230\347\231-\214 c#\234\312\031I\252:\272\371\367ze\244\357\341\244\\\242]EZ\305\215\304\342\361\274\315\274\326\315$\315c_\241\244l\200\247+\256\246\227|\004\333\304\n\360\234\373\213""\314\250\2304e\225\315\000\352\216\332+\027M\035q\227\2443C\354\026\355N;Q\231\022\224\257hJ\346\303\001@AV\345:q0\200\323\014\262p*\201\266\"\236\025\205\031l\301aX:\330WtB/f\020R\224\034KXj\026|]^>\346\205e\343\203&\0368GE\r{\2669\320\333m\311\302h\366R%p0\244\274\214x\351uO\371d\257\275AB-Y\024\300\023\030\212\250\253\025\025EW<\254\263\026vu\314H\246\337*\331\251\267/\010\222\263[K\353\242\001\234\2144\262\302\014\367\271)uF/\206f\272\306\274\204[@f\270~\300@\254HZm\237,\341\"\366\321Tf9iYl!l\203\241\014\375c\257s\215n\262_\rr\252\210\013d\244\201\217H32Z/\337}\264$\2639\362\337\005\321LK\002\3702\344;\300\371C\227\225\357\363{^v\3000\006-n\034\032!vM\014s\334m\273\304\336l\203{)ig\201k\325\3600\024j\246\013D&\370;\227\013\231\202F6 \342\257\007F\032\366\354\354\363\010\352\347\263v\007\355\322]\275\206u\324U^$*\223Q\215\357/h\276\300eg\367l:uQ}HR\274\347\266\233^Q\314%\201\206\251\320$\342\311:\240\357\247OB\333\257^\333\304\226&+\006\335\341\007\324\341\326\361\322\371\327\310\334\234\3130\313;\222q*\347\004\211\261C\034\233O<\334\212\326\301\200k\275b\021\315\2615\260Q\006\3450\222\220\344\255\240C\r\207Z\373\302y\332\n\204GaS\031b\0302\001[\273\326\363\351\203}\216\016fC\225\306\000\260D\300p\355D\004\212:+\332\220*\024l\201oV\3371\260\235\320\004\024\270\224\330\310fi\302\346\020\321\242V0\310\244\301\275\031,\355\354\nV\260\271\002\303-\252\202\265\237pr\220&\205\021\350c@\222&j\337\355u\010\203\035k\030\315\200WCF\324\2411t\r\320G\235\265 $\271\355d28Z\244\221\032(\274\215:a+\0341\352\2128a\226\266a`\215\315\002\030c\250\017\003p[M\266L\254\307'\235\2208rf)\340V\247\206\n\037X\242\022\177?\007\r\334w\346\304\321l\032)\224\214\252\221\217/3\347\335\320\367J$\367\242\005\344\224:Uf\212(e%\351\212\252h3\206\206\201\203Oq\301\315/!\362p\363T\003H\370\334\365\334\345\333\307\r\252\n\2637\242%\202s\005#t\353g\316\n \304\201;\371n\021\326Z\372\270s\273\345wf[\333\nvy\247\021\021\020D7\034/|\353T\332\370\213\214f;\324\316$yX\n\330\014\262\306\370SR\366\302\2727""\336\354F\302\365SS\261\315\235\270X\003\205\251\252\234\360Wh\314\207c\312\262\251R\214\0030\0035o\237_\211\303\303\016h\236\036/}\272T\330\263\206\\eC\223M\247R)\322\333cm\246\330\333j\221-\203\230\014\225\322\025\316\221'+\352\203\257\345I\037\030\032\021\211\230\302\327N`,\201\221\230##\313\007V\034SW\202\276f\304Q\032\304\202YU\013\r\246\211\354`fk\030gJ\313\004c\360e\364\245\204\322\327aXgi\007!\016\322\301\226\230\334\351y@t\035+~\365\356\353\374\3226\2101$|\277\014`\225\01336\0273\326#z\363\006f\341\310*\274@\243\301\246tD8\273\351\356Z\207\025\351W\324#\0104\315Ss\363m\243\216d\221\201\212\245\345\212\313\003b:}\374@9d[b\204\270\216\303\263\004\317\201\261t^\315\311\240\204J A\216\225o6\274[\373\375\220\351b\332,s\353\364p\t\213GV\001\276\360\002 \027N\220\361\357Q\253k\342\334\210\363\337\025E\307\365:t\374\317-\207\312?\177\020\321\216\375\236\325\224C*I\227\3070\224Y\017k\311\340+\177)\255e'\252\354\220\002\246P#\346`\243\020N\304\0315\017\010\214\313\361\310{\307\327U\031I+\260\264\235I\261\334\034\270NWi\250J\205\037|*\206\"\274\307f#\242M\2111t\337\356$;,\013%\245\244\210\251\235\275M\313!|\242OV\\\035\271\354Q\221\334\346~\037\255\026\221\232w\3608\304a\265\311\364\376o\223\307\243\253\016\363\210\335\225\022\336\2715?\240\342i\202#\270\336\255\254\020\3644\261\334\3125Fr\375\264\325e\022!\312\301\370\324\215\230\377\n\357\235\241F  \265\031\252^\207-\234&#~\255\367\003\205\313\250\226\337\343J\271\230\366\200e\202\002\303\345/\314@Z\256*\275\261]\374\210\223\322\251(\252-\224\212=\000\327\025\231\254@C\221f\245G\323\024f\\\030\271\313\307U$[=RB\300\226\345o&\343\346\355]\306t\265M\233\231\301\221\014z\242!GVz\275\351[\207\232\246X\351F\224\251)\035s\3720S\312\216U\342\261\353\213\006\023\214\316P\210\312\373G\022{\236\367\367\240\030\226\234b\361\252H\367\264]\235\013\t#\031\022\230r\007\3678\253WMY\350O\212H\302G\220\351\345c1\250\260\222\033\305\307\306y\024r\\3r\234W\037\213\311.+\233je6V\342\024\323\010\302\300L#\204\032>\313\3724\212[\372\035o""\000\324\230\353!\371\277\240\224y4\n\005\351\365\331\3703\035>\362}\252\323\327\273\223|\226\247\005\345\327\007'\233\231M!\273x\233Z\323I\2752\265\275\317\230\215<\312\"s\014]\201zpp\362\346\323\300B\251\277\203yG\212\375\357z\277>d\306n\335\032(n\346_.\323\233\241j\316\211\233\3608\376\253\341:\000\266=\025\347&\321\020\353\"\230\2662\277!\"\rC\005\\\330)\254\315GP:\275$ \350\n\221\024H\024@\223\242+)\003 \363\222\2340W\213\277P\232\316 1\314\247|\213H\212\326\316\316q[7i\373\002\370B\276[Er+\"\354\313\tG\246,ij\001DEP$\021\002lt\025gWK+u\214v\372\323\255@g\200\356\210\216\346n\256\313)J\204y\223\234\242\222\303Cy\365B3\212p\r\213\351\177\361w$S\205\t\007\310\321\210p";
    PyObject *data = __Pyx_DecompressString(cstring, 3076, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (2923 bytes) */
const char* const cstring = "x\332\245\027K[\333\326\022\2031\2468\200\003\005\322G*7\344\3216%uJI\332\364&u\003\204<.` \220\246mt\205ul\204m\311\326\003\354\266\271e\351\245\226Zj\251\245\226^z\311RK/\375\023\362\023\356\314\221l\313\310$\367k\026\326\314\231s\346q\346u\306\017%\255\3003\242\2442\nQ\231UN(\020\036QU\020s\214z@\230#N\026\270\375\002Y\360\366T\211\311H\342\021\221U\246TU\017$\021N\0244\202\364'\333)U*\256K*\254\0168\225y\350\356\013\n\303\223\202\260OdN%\205*\243\250\262\220Q\211\214\207Dfse\363\353\305\273\213\014'\362\214L\016IFU\030E\333\317\0248E!\n#e\231}M(\2009\214Z-\021e\201y\234e\252\222\306\210\3045\246\004\347\374\014`\263H\357\202\306_\347D\270\032\247\n\222\310\002;\334\351:\303\0132(\021\216\250\311\253\\A!\013\233\262T\024\024\302\034HR\236)j\212\312\354\023&\303\025\n\364\342[D\321\n*^\003\335$\023\216\257.<X\340x\236\2055\311T\017\225\005\026\277\370\273E\261\205R\265\302\013\nr\363\222\350\212\332\3472y% \234\210\370\315v\\\233\003\303\245}\364\302u\205\001\263J\340\347*#rE\242d51\203\027\t\032\230\313\200\361T\020/s\200\362\014\251dH\211\036>\026 \006\032X\217,\030\323\264&d\362\207\n\210\227\216E&\313\301\325\024R\310.d\324\312M\212@`3\324mT\207\033j\3278\216\331t#\352Z\310d%\231)\201\264\002\310\355\225\241r\373.\342\346\306{\313{?1\377\354F?b\236\202\004I\2233\344>\236\336\240\20741/\242\353<\226\324\352\316\312\326\317+\253\033[+\01791C\n\020\205\025Y\226\344\336\325\002\313\312\204\3272\004\022\0045\262l`\037\222V\201\\\365\235\220D\225TT\017\364\221\320\3318\217\225fi;q\3324\002\336\364\343lQ\342\265\002\351!\251\007\202\022 \260\230LgNC\302\372P6W\220\366\273\322\017\025\260\245\304\311J\3478\030\372x\375\361\316cQ]\021\265\"\200\325\002\227{,\202\006\201\337\306\033Pg\004\010\301\273\367;\022\360\302\223m\272\345\201\240\220\356F\037\326U\317m]\254\237\000\337^@\306\277IQ\222\253\317\025.G|hPJ\357f@\314\372\306\362\312j\352\371\263\235 \247\177+\300\307n\354\n\344\330\375\006Y;\364\000\237\233\346\036x\254\222\242\342\312\351]\003\247\000M\234e\373l""\004T\005N\3641\226\236yJ\252~m\355\345Ye>z\177]\276\003\347\251\332\305\236\342W\326%\234U\327\263\323_a\317\221s\\\032d\356\320\317\343\300\332\363\241^\361y\024(8\017\023\304#)\337&\013\350g\017\317\203\037<T\351\234V%\026+\323[\321\336\252xO\240\007\326\340!\334\201\347\326\267\014\032\337\273\031\270\201\267}.\337\333x\260m\341\313\311\266_\316\366\006\022\333x\347\221k\023JR\241\320\306e\250\251#\322_\210L\037t\320V\255\300o\031\206\021v\035\232\323\026\311n\255lo<\333]\331\322DU(\022\017\004o\320\335\010\334\240\275\225\221\212%\r\350EZ\333\254\206\305\335\336\343\265b\251\357\006\251\220\0142\225\210\310\303\323\303\036J\373\355-\350\276\260j\357\264\251\262&\262\271L{\005\306\260E\256\302\202E\231<\253\010\177\220\236\035WaA(\n\252\237^r\275\302\342\360\323\246k%\036/\205W\003I\252T\n\004\304\377\260\260lw\314R`!\347\024N\251\212\031A\0027\3100}\010\"Q8\030\r\221Q\2259\360\344>'\313\002\221\3679P\354>\244\312\276\220\023D\210\nj\200/<\301\204\025`l\303\363\310\230)H\n\311H<\351\347\332\214\004\226\303\270\320\235\306\350\005X\036\203\353\002^*\262\235\224\301\273\004\302\200\004\030\244\264\"\226\232\257\334:\017`\337\227\260#\263O\364\2620\327\341\017\007\270\254\210\376bY\367\013\265\333}/\351\322\315#J\021\300\0210{(B\206E\225nu\263\240\266\343\316\336t\300\212\247U\337\356[\207\n\215\002\000Z\335X\355\335\267\030z\002\266\005\232\010E.\007J k\300\323\360-aB\367\246\020z\010Fw\226\316\240\340Jw\211\230w\377.\304#\235/\264U\370TT\210n\t\246X\230\254\2128\360`\215\226\244RIR\274\274\003P\251\262,L\314l\346\200\300\214\254\025\335U\273J\021\367|\203(\376\013p1M\244\243\03293I\235\331\014\014\ng\366\275\031\340\014\325\367$\237\331\351<\271g\350\276n\350\356\034\2518\221\243\237\312\032Wp\275B\307\363n\026v\032\013\033|V\332\004RA4G*%\367\313b\372CJ\311\375:\234\3533\231\250nc\220\335r\306!\030\362\232\307f\025h\021g[\003\355i^\020\317\266\007X\363\204\376e\360u>\266O\027\244+\367\177\036\313\"\306\322\262Q\017d\351\330{}0\216<d\251\314U\025\267\335\004\232""\216\006j\333\242\217 \2124\227\335\347\352\010\036\332c\302\345\3015\331\037\005\021\372\301\375\223\324I\250\031\276i\245\254];y\222z\023\031\230\235\203\317\314lk\372\3037\337\014\214\306j\213\265\212^6B\315\330E\375\013#i\244Z\321I}X\3376\302F\352\315\007\003\243\2235M\177\010\007\302\316'\267\355k\365\301\372\215F\002\004\000\343\003#a\3341\023\346\222u\321J6c\027\232\321\321V\2141\303f\312\334E\302\214.\002\272c\3158\267V\032\277;[\273\316\356o\316oU\247\372\367\233\201\201\237\006\037\r\002x4\270\216`}pc\260\031\233\250\375\327H\033\\sr\332\231\2367S\315\311x\013\204pz\271\031\233\322\2777R\010\222\240\373\002c\224\315\220\031?I5\243\023\316\004\323\306?\002\203\026\r\315L\235\300%\306\372\\l\214\332\274h\034\201Us\366\224\r\022'\3651z\211d+6^\343je\224\023dmFcN\354\272\251ZIk\325N\320\025hmF/\031qc\336 \316\225E\373Q}\261^\246\016\366\\\263d\306\315k\326\240\225\240\256\031\257\025<\325\277\333\363v\266\276\334\2104\270f\364c\224n\244[\321\211Z\325\010\031\323\306\256\231<\t\275\211\016\014_\250\245\232\341X\355\221\236\324W\215\233&\327\014\217\234\034\325\366\364\0258\224nF/\352`\306\254^n\341\007\367\312\324\027\267A8g\037\327\363\247\263\316&\212\365\223\342\357#\371\206\025\206$\332\263\237\326\217\033\\\253\227R~\037\301_Z\363\026g\035\333\371\306\354i\250\325Ky/\213oY\252\235\204\270\334k\344N]Wt)\034D*60\032\327C\372\264\236v\027\261\332\017zZ\347\215kf\304\344\314#k\313*\323\320\215u\350!s\326\212X\373v\210\322A`\002b<\017\366DG\333\331\313C\314#V\026\224\334\256\247\233\2233\272\200\271\332\212\315u\242\013y\370)=x\325,[!\314\332y\203\303C\230znBq\210^687\213\377\206\222\210\201\262+f\322\\\266\206\255\264E\354\333v\272[&x\365\330%\343\222\363\371wv\332\306\303\265?\2148*?\304$\235\300\254\246\245\3543x\354\234\313\216\327DH\310\035s\006h\340\332i=\245\357\0303\006z\324\227\241Xxh\227Q\356\242*\322!Z\243\237\301\335\274\026\201:\210\361\0358n\306\224\2558hhEc\265u\250\204\357\240O\254Y\264\\&\251\340\017\215\274\025\267\256Ze\367\356(\"z\236\2101""\217\2324\236\2309\327\364I=\352\314\335\260\300\323t\377>\3701\321\227z\025\014\275k-\331\037\321\000A\201/\230!'\221\264\343v\002C\343j\376\240\253\331_\3064\301\300\326)c\031\334&`6y\241\247\2760\207\274^8\031ow\216\025s\332|\201\255\327^\255'\352I\257\221\031\317\315D+\206Y@=\330qZ\0232c\307\234\263\342n\3405\2356\250Q\210%v\276v\377J\372\262!d}H\363qrZ_s>\2016\337\237\023?\335\364\"4R\220\361\0379\227\276\000\247\317C\276.\327\243\215\361S\3108\306\234\260\312v\230\332\273\324\230n@\022\\r.\335\266\023\366R=\336\214b?V\321)^^\3212\370\254\235\276c\316\030\342as\305\232\261d\033\243\205\344\033\326\204]\256\207\353\253\215\371\006\315\275%}\n-\244\316o^\230\322\277\325!k/:\027\361R4\241T\327\336\2673v\327\310\360\243q\014\211\rI\317c\344\247\364E\254\212\266\314>\302&jG\372\2663\367\265\265mG\354L=\356\272\344\006$\331\224\225j^\276b~k\036X<t\373\313\t\363\",\200e\316\270`n\203\317\247\255\337\353\237\327\037\326\313\rjD\322}F\272N\201\254\376\370\223\346\314l3\361ysv\256E\357\263\004\256\277\3460I\373S\347\376\013\347\205\342(\352;\016^v\036\374\342\374\242:\252\326\377\340]\343\036-\335]\373\207\372n\343\256\363\364\225\363\212}\367\321\275\306\232\363\214u\330\377\320\304\013\034\275\0079y\004\017\370\017\366n\375\256\363\020\036\355\275w\035\334\253\2579\313{\316\336\013\267\031xm\302\271\202\335\210\257'p\000\371\240\366\025\264Y@\256\232i3k\245\334\276>^\333\321gq\320\370\177\273;f\336,\264\300mH\320\345\372H\335\327\355oZ\313v\310\236\255\2175\356\234&Z\275\224dW\335e(\377\177\001\367\263F\330Y~\351\274\004\237A(\216a\006\251\204\236\342(\362\024F\021\000\033\203\004\001\031\024\020\010\203\257\021\274\036L\r\001H\r\245\021\244\207\236#x>TDP\034*#(\017\375\024\306\361&\274\214`9\274\215`;\274\207`/\234E\220\r\037\"8\014W\020T\302\177!\370+\2749\014`sx\007\301\316p\016An8\217 ?|\214\340x\370O\004\177\016\257F\000\254F\236 x\022\331C\260\027\371\025\301\257\221\034\202\\$\217 \037\251\"\250F^#x\035y2\202\014#\353\010\326G^\"x9\362\n\301\253\021v\344\037""\274\260\357\360y3\214\335\345\232\035:\t\265\302\320\027\034\346\033{\014\036\337\274\223\336j\206\247\364\373\330\352\240\261\202\274\277\364;P\265\321o\354\220\223\374\271\261\000\263K3\014\205u\016\343\035\343\006\216l>FT\364\245\275\336H\274\215\355{c\r\006\322\264\217\355+\030\350\236\325\307\241\035\234\317\346L\315\343Ke\371\371\276\265\223\316\342\332\3518\024\307\333\024\3363\016\240\354\312>\306\257\341\365\026\353\353\247m;\233\341[\320m#6\207\350u\030\010\"\026\367?\3235:\234";
    PyObject *data = __Pyx_DecompressString(cstring, 2923, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (6781 bytes) */
const char* const bytes = "Could not set Failed setting the variable.Failed to convert python value to JSAtomNote that Cython is deliberately stricter than PEP-484 and rejects subclasses of builtin types. If you need to pass subclasses then set the 'annotation_typing' directive to False.Promise hook must be callable.Result is not ready.?.add_notecyjs._cyjscyjs/_cyjs.pyxdisabledone callbacks must be callableenablefailed to get object's Property namesfunction must be callable.gcisenabledraised exception without it being Quickjs's own faultself.ctx,self.rt cannot be converted to a Python object for picklingself.ctx,self.tab,self.value cannot be converted to a Python object for picklingself.ctx,self.value cannot be converted to a Python object for picklingself.rt cannot be converted to a Python object for pickling<stringsource> to Objectunknown object AFTERBEFORECancelledErrorCancelledError.__reduce_cython__CancelledError.__setstate_cython__ContextContext.__reduce_cython__Context.__setstate_cython__Context.add_functionContext.evalContext.eval_moduleContext.eval_thisContext.eval_this_with_moduleContext.getContext.get_globalContext.json_parseContext.setINITIntEnumIntFlagInvalidStateErrorInvalidStateError.__reduce_cython__InvalidStateError.__setstate_cython__JSErrorJSError.__reduce_cython__JSError.__setstate_cython__JSFunctionJSFunction.__reduce_cython__JSFunction.__setstate_cython__MemoryUsageMemoryUsage.__reduce_cython__MemoryUsage.__setstate_cython__NODEFAULT.__reduce_cython__NODEFAULT.__setstate_cython___OView_OView.__reduce_cython___OView.__setstate_cython__Object_ObjectItemsView_ObjectItemsView.__iter___ObjectItemsView.__reduce_cython___ObjectItemsView.__setstate_cython___ObjectKeysView_ObjectKeysView.__iter___ObjectKeysView.__reduce_cython___ObjectKeysView.__setstate_cython___ObjectValuesView_ObjectValuesView.__iter___ObjectValuesView.__reduce_cython___ObjectValuesView.__setstate_cython__Object.__reduce_cython__Object.__setstate_cython__Object.evalObject.eval_moduleObject.getObject.invokeObje""ct.itemsObject.keysObject.setObject.to_jsonObject.valuesPromisePromiseHookTypePromiseHook.__reduce_cython__PromiseHook.__setstate_cython__Promise.__reduce_cython__Promise.__setstate_cython__Promise.add_done_callbackPromise.donePromise.exceptionPromise.pollPromise.remove_done_callbackPromise.result__Pyx_PyDict_NextRefRESOLVERuntimeRuntime.__reduce_cython__Runtime.__setstate_cython__Runtime.compute_memory_usageRuntime.dump_memory_usageRuntime.execute_pending_jobRuntime.is_job_pendingRuntime.run_gcRuntime.set_max_stack_sizeRuntime.set_memory_limitRuntime.set_promise_hookRuntime.update_statck_topadd_done_callbackadd_function__annotations__argsasyncio.coroutinesatombacktrace_barrierbase_objectsbigint__call__cline_in_tracebackclosecodecompute_memory_usagecountctxcyjs._cyjsdate__dict___dictdom_exceptiondonedump_memory_usagedumpsenumevaleval_moduleeval_thiseval_this_with_moduleexceptionexecute_pending_jobfilefilenamefnfunc__func__getget_global__getstate__globiintrinsic_evalinvoke_is_coroutineis_job_pendingitemitems__iter__js_argsjs_valuejsonjson_parsekeykeyslimitmagic__main__map_setmax_stack_size_member_names___members____module__module__name__name__new__nextobjpperformancepollpoppospromiseproxy__pyx_checksum__pyx_result__pyx_state__pyx_type__pyx_unpickle_CancelledError__pyx_unpickle_InvalidStateError__pyx_unpickle_JSError__pyx_unpickle_MemoryUsage__pyx_unpickle_NODEFAULT__pyx_unpickle_PromiseHook__pyx_vtable____qualname__raise_exception__reduce____reduce_cython____reduce_ex__regexpregexp_compilerremove_done_callbackresultretrun_gcruntimeselfsendsetset_max_stack_sizeset_memory_limit__set_name__set_promise_hooksetdefault__setstate____setstate_cython__statestrict__test__thisthrowto_jsontyped_arraysupdateupdate_statck_topuse_setstatevvalvaluevaluesviewweak_ref<input>\200A\200\001\330\004,\250A\250V\2601\200A\360\006\000\027\030\360\006\000\026\027\340\024\025\3600\000\t\014\2104\210x\220q\230\001\330\014\022\220)\2301\230A\340\010\020\220\005\220S\230\004\230A\360\n\000\t\020""\210u\220C\220q\230\004\320\0342\260&\270\002\270(\300!\360\006\000\t\014\210?\230!\2307\240!\2406\250\022\2501\330\014\r\330\010\t\340\014 \240\004\240A\240V\2501\330\014\026\220n\240A\240T\250\026\320/E\300]\320RV\320V\\\320\\y\320y\177\360\000\000@\002G\002\360\000\000G\002N\002\360\000\000N\002O\002\330\014\017\210~\230Q\230a\330\020\024\320\024$\240A\330\020\021\340\014\026\220a\220q\330\014\023\2209\230A\330\014\023\2201\360\006\000\r \230q\240\001\240\021\200A\330\010\017\320\017 \240\001\240\021\200A\330\010\032\230!\2304\230u\240A\200A\340\010\013\2104\210x\220q\230\001\330\014\022\220)\2301\230A\340\010\013\210?\230!\2304\230v\240T\250\030\260\023\260A\330\014\020\220\013\2307\240!\2401\340\014\016\210a\210q\200A\330\010\013\2104\210x\220q\230\001\330\014\022\220)\2301\230A\330\010\014\320\014'\240t\2501\250F\260!\330\010\014\320\014 \240\001\330\010\031\230\021\230$\230e\320#4\260G\2704\270q\200A\360\006\000\t\014\210?\230!\2306\240\021\240&\250\002\250!\330\014\r\330\010\016\210l\230!\2304\230v\240]\260$\260f\270D\300\006\300a\330\010\033\2301\230A\230Q\340\010\017\210y\230\001\230\024\230V\2401\200\001\360\010\000\005\r\210A\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220q\330\004\007\200q\330\010\017\320\0172\260$\260a\260w\270k\310\027\320PQ\340\010\017\320\0172\260$\260a\260w\270k\310\021\200\001\360\010\000\005\r\210A\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220q\330\004\007\200q\330\010\017\320\017(\250\004\250A\250W\260K\270w\300a\340\010\017\320\017(\250\004\250A\250W\260K\270q\200\001\360\010\000\005\r\210A\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220q\330\004\007\200q\330\010\017\320\017*\250$\250a\250w\260k\300\027\310\001\340\010\017\320\017*\250$\250a\250w\260k\300\021\200\001\360\010\000\005\r\210A""\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220q\330\004\007\200q\330\010\017\320\017/\250t\2601\260G\270;\300g\310Q\340\010\017\320\017/\250t\2601\260G\270;\300a\200A\360\014\000\t\021\220\001\220\024\220Q\200A\360\014\000\t\014\210:\220Q\220d\230&\240\006\240a\240v\250R\250q\330\014\r\330\010\013\210:\220Q\220d\230&\240\001\240\027\250\006\250b\260\001\330\014\r\330\010\017\320\017!\240\021\240$\240a\330\010\t\330\014\017\210~\230Q\230d\240&\250\006\250f\260G\2702\270Q\330\020\026\220i\230q\240\001\340\014\030\230\001\230\024\230V\2401\200A\340\010\035\230Q\230d\240%\240q\250\001\330\010\032\230$\230a\230q\240\001\200A\330\010\031\230\021\230$\230a\200A\330\010\036\230a\340\010\013\2104\210\177\230a\330\014\017\320\017#\2401\240D\250\005\250Q\250e\2602\260Q\330\020\024\320\024$\240A\330\020\027\220q\340\014\031\230\031\320\"5\260Q\260a\330\014\017\210z\230\021\330\020\026\220j\240\001\330\010\017\210q\200A\360\n\000\t\020\320\017!\240\021\240$\240a\330\010\013\210:\220Q\220d\230&\240\006\240a\240v\250R\250q\330\014\r\330\010\016\210n\230A\230T\240\026\240v\250Q\330\010\024\220A\220T\230\026\230q\330\010\017\210y\230\001\230\024\230V\2401\200A\330\010\017\210\177\230a\230q\200A\330\010\017\210\177\230a\230t\2401\200A\360\010\000\t\037\230d\240!\360\006\000\t\014\210:\220Q\220e\2305\240\001\240\026\240r\250\021\330\014\r\340\010\014\210N\230!\2305\240\004\240H\250A\360\006\000\t\020\210y\230\001\230\025\230k\250\021\250%\250q\200A\330\010\036\230d\240!\360\010\000\t\014\210:\220Q\220e\2305\240\001\240\026\240r\250\021\330\014\r\340\010\013\210:\220Q\220e\2301\230J\240g\250R\250q\330\014\020\220\010\320\030(\250\001\340\014\r\340\010\013\210>\230\021\230!\330\014\020\220\010\320\030(\250\001\340\014\r\340\010\013\210>\230\021\230%\230t\2408\2506\260\032\2702\270Q\330\014\022\220.\240\001\320!1\260\021\260!\200A\340\010\036\230d\240!\360\n\000\t\014\210:\220Q\220e\2306\240\021\240&\250""\002\250!\330\014\r\340\010\027\220q\230\001\230\023\230D\240\006\240i\250t\2601\330\010\t\330\014\017\210\177\230a\230q\240\003\2406\250\022\2501\330\020\021\330\014\022\220)\2301\230E\240\024\240X\250V\2601\260F\270!\2701\330\014\023\2209\230A\230U\240!\340\014\035\230Q\230a\230q\200A\360\010\000\t\037\230d\240!\330\010\035\230T\240\030\250\021\340\010\013\2104\210u\220A\330\014\020\220\t\230\021\330\014\023\2201\340\010\013\210?\230!\2301\330\014\017\320\017#\2401\240D\250\001\250\025\250b\260\001\330\020\024\220H\320\034,\250A\340\010\013\2104\210u\220A\330\014\020\220\t\230\021\330\014\023\2201\340\014\023\2201\200A\330\010\031\230\021\230$\230e\2401\200A\360\014\000\t\032\320\031)\250\021\250$\250f\260D\270\010\300\016\310a\340\010 \240\017\250q\260\004\260F\270!\2706\300\024\300Q\330\010\031\320\0312\260!\2606\270\021\330\010\026\220a\220t\2306\240\021\330\010\017\210q\200A\330\010\017\320\017\037\230q\240\001\200A\330\010\013\320\013\037\230q\240\004\240E\250\026\250r\260\021\330\014\r\200A\330\010(\250\017\260q\270\004\270F\300$\300a\330\010\013\2106\220\023\220A\330\014\022\220.\240\001\330\r\023\2203\220a\330\014\022\320\022#\2401\240A\330\010\017\210t\2201\200A\330\010(\250\017\260q\270\004\270F\300$\300a\330\010\013\2106\220\023\220A\330\014\022\220.\240\001\330\010\013\2106\220\023\220A\330\014\017\210t\220<\230w\240a\330\020\026\220d\230!\330\014\023\2204\220q\340\014\022\320\022#\2401\240A\200A\330\010(\250\017\260q\270\004\270F\300$\300a\330\010\017\210v\220S\320\030-\250S\260\006\260c\270\021\200A\360\014\000\t(\240r\250\023\250A\330\036#\2403\240h\250d\260!\330\036!\240\022\2403\240a\330\010\030\230\r\240S\250\001\250\024\250]\270\"\270C\270q\300\001\330\010\013\2101\330\014\020\220\013\2306\240\021\330\010\017\210q\200A\360\006\000\033\034\330\026\027\330!\"\330\027\030\340\010\017\210t\2206\230\021\230&\320 1\260\035\320>X\320Xs\320st\200A\360\006\000\033\034\330\026\027\330!\"\330\027\030\340\010\017\210t\2206\230\021\230&\320 1\260\036\320?Y""\320Yt\320tu\200A\360\006\000\033\034\330\026\027\330!\"\330\027\030\340\010\017\210t\2208\230;\240a\240v\250V\260:\270V\3008\320K^\320^_\200A\360\006\000\033\034\330\026\027\330!\"\330\027\030\340\010\017\210t\2208\230;\240a\240v\250V\260:\270W\300H\320L_\320_`\200A\360\010\000\033\034\330\026\027\330!\"\330\027\030\340\010\017\210t\220;\230a\230v\240V\250:\260V\2708\320CV\320VW\200A\360\010\000\033\034\330\026\027\330!\"\330\027\030\340\010\017\210t\220;\230a\230v\240V\250:\260W\270H\320DW\320WX\200A\330\010\017\210y\230\001\230\024\230V\320#5\260Q\260d\270!\200\001\330\004\n\210+\220Q\200\001\330\004%\240Q\240f\250A\200\001\360\010\000\005\016\210T\220\027\230\004\230A\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220t\2306\240\027\250\005\250S\260\004\260D\270\007\270q\330\004\007\200q\330\010\017\320\017,\250D\260\001\260\027\270\013\3007\310!\340\010\017\320\017,\250D\260\001\260\027\270\013\3001\200\001\360\010\000\005\016\210T\220\036\230t\240=\260\004\260L\300\004\320DZ\320Z^\320^s\320sw\360\000\000x\001K\002\360\000\000K\002O\002\360\000\000O\002e\002\360\000\000e\002i\002\360\000\000i\002}\002\360\000\000}\002A\003\360\000\000A\003Q\003\360\000\000Q\003U\003\360\000\000U\003m\003\360\000\000m\003q\003\360\000\000q\003@\004\360\000\000@\004D\004\360\000\000D\004S\004\360\000\000S\004W\004\360\000\000W\004f\004\360\000\000f\004j\004\360\000\000j\004x\004\360\000\000x\004|\004\360\000\000|\004P\005\360\000\000P\005T\005\360\000\000T\005g\005\360\000\000g\005k\005\360\000\000k\005w\005\360\000\000w\005{\005\360\000\000{\005F\006\360\000\000F\006J\006\360\000\000J\006W\006\360\000\000W\006[\006\360\000\000[\006g\006\360\000\000g\006k\006\360\000\000k\006y\006\360\000\000y\006}\006\360\000\000}\006J\007\360\000\000J\007N\007\360\000\000N\007Z\007\360\000\000Z\007^\007\360\000\000^\007_\007\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!""\330\010\027\220q\340\010\027\220q\330\004\007\200q\330\010\017\320\017,\250D\260\001\260\027\270\013\3007\310!\340\010\017\320\017,\250D\260\001\260\027\270\013\3001\200\001\330\004)\250\021\250&\260\001\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\220>\240\030\250\021\250!\330\004\007\200|\2207\230!\330\0100\260\001\3201B\300.\320PQ\330\004\013\2101\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\2207\230(\240!\2401\330\004\007\200|\2207\230!\330\010)\250\021\250*\260N\300!\330\004\013\2101\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\2209\230H\240A\240Q\330\004\007\200|\2207\230!\330\010+\2501\250L\270\016\300a\330\004\013\2101\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\320\023$\240H\250A\250Q\330\004\007\200|\2207\230!\330\0103\2601\3204H\310\016\320VW\330\004\013\2101\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\220;\230h\240a\240q\330\004\007\200|\2207\230!\330\010-\250Q\250n\270N\310!\330\004\013\2101\200\001\330\004/\250q\260\006\260a\200\001\330\004'\240q\250\006\250a";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 237; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 26) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    for (int i = 237; i < 289; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyBytes_FromStringAndSize(bytes + pos, bytes_length);
      stringtab[i] = string;
      pos += bytes_length;
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 289; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    #if CYTHON_IMMORTAL_CONSTANTS
    {
      PyObject **table = stringtab + 237;
      for (Py_ssize_t i=0; i<52; ++i) {
        #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
        #if PY_VERSION_HEX < 0x030E0000
        if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
        #else
        if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
        #endif
        {
          Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
        }
        #else
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
        #endif
      }
    }
    #endif
  }
  {
    PyObject **numbertab = __pyx_mstate->__pyx_number_tab + 0;
    int8_t const cint_constants_1[] = {0,-1};
    int32_t const cint_constants_4[] = {139263547L,238750788L,239704551L};
    for (int i = 0; i < 5; i++) {
      numbertab[i] = PyLong_FromLong((i < 2 ? cint_constants_1[i - 0] : cint_constants_4[i - 2]));
      if (unlikely(!numbertab[i])) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_number_tab;
    for (Py_ssize_t i=0; i<5; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
typedef struct {
    unsigned int argcount : 3;
    unsigned int num_posonly_args : 1;
    unsigned int num_kwonly_args : 1;
    unsigned int nlocals : 4;
    unsigned int flags : 10;
    unsigned int first_line : 11;
} __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR), 496};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_val, __pyx_mstate->__pyx_n_u_pos};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_iter, __pyx_mstate->__pyx_kp_b_iso88591_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR), 506};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_i};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_iter, __pyx_mstate->__pyx_kp_b_iso88591_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR), 523};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_val, __pyx_mstate->__pyx_n_u_pos};
    __pyx_mstate_global->__pyx_codeobj_tab[2] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_iter, __pyx_mstate->__pyx_kp_b_iso88591_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[2])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[3] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_A_G1F_a_vWE_Q_q_q_q_AWKwa_AWKq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[3])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[4] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_QfA, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[4])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[5] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_t_L_DZZ_ssw_x_K_K_O_O_e_e_i_i, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[5])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[6] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591__5, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[6])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[7] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_A_G1F_a_vWE_Q_q_t6_S_D_q_q_D, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[7])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[8] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591__5, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[8])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 143};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[9] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_compute_memory_usage, __pyx_mstate->__pyx_kp_b_iso88591_A_Qd_q_aq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[9])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 148};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_file};
    __pyx_mstate_global->__pyx_codeobj_tab[10] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_dump_memory_usage, __pyx_mstate->__pyx_kp_b_iso88591_A_q_E_r, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[10])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 152};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[11] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_execute_pending_job, __pyx_mstate->__pyx_kp_b_iso88591_A_a_4_a_1D_Qe2Q_A_q_5Qa_z_j_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[11])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 166};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[12] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_is_job_pending, __pyx_mstate->__pyx_kp_b_iso88591_A_at1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[12])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 176};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[13] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_run_gc, __pyx_mstate->__pyx_kp_b_iso88591_A_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[13])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 185};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_limit};
    __pyx_mstate_global->__pyx_codeobj_tab[14] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_set_memory_limit, __pyx_mstate->__pyx_kp_b_iso88591_A_e1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[14])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 188};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_max_stack_size};
    __pyx_mstate_global->__pyx_codeobj_tab[15] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_set_max_stack_size, __pyx_mstate->__pyx_kp_b_iso88591_A_4uA, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[15])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 191};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[16] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_update_statck_top, __pyx_mstate->__pyx_kp_b_iso88591_A_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[16])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 194};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_func};
    __pyx_mstate_global->__pyx_codeobj_tab[17] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_set_promise_hook, __pyx_mstate->__pyx_kp_b_iso88591_A_4xq_1A_t1F_e_4G4q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[17])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[18] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[18])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[19] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[19])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[20] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[20])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[21] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[21])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[22] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[22])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[23] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[23])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[24] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[24])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[25] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[25])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[26] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[26])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[27] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[27])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 552};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[28] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_to_json, __pyx_mstate->__pyx_kp_b_iso88591_A_fD_a_q_F_6_Q_2_6_at6_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[28])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 567};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_v, __pyx_mstate->__pyx_n_u_atom, __pyx_mstate->__pyx_n_u_ctx};
    __pyx_mstate_global->__pyx_codeobj_tab[29] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_get, __pyx_mstate->__pyx_kp_b_iso88591_A_d_Qe5_r_N_5_HA_y_k_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[29])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 582};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_value, __pyx_mstate->__pyx_n_u_ctx, __pyx_mstate->__pyx_n_u_js_value, __pyx_mstate->__pyx_n_u_atom};
    __pyx_mstate_global->__pyx_codeobj_tab[30] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_set, __pyx_mstate->__pyx_kp_b_iso88591_A_d_Qe5_r_Qe1JgRq_t86_2Q_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[30])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 8, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS), 628};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_func, __pyx_mstate->__pyx_n_u_args, __pyx_mstate->__pyx_n_u_atom, __pyx_mstate->__pyx_n_u_ctx, __pyx_mstate->__pyx_n_u_ret, __pyx_mstate->__pyx_n_u_js_args, __pyx_mstate->__pyx_n_u_p};
    __pyx_mstate_global->__pyx_codeobj_tab[31] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_invoke, __pyx_mstate->__pyx_kp_b_iso88591_A_d_Qe6_q_D_it1_aq_6_1_1E_XV1F_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[31])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 652};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[32] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_items, __pyx_mstate->__pyx_kp_b_iso88591_A_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[32])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 655};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[33] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_values, __pyx_mstate->__pyx_kp_b_iso88591_A_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[33])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 658};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[34] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_keys, __pyx_mstate->__pyx_kp_b_iso88591_A_aq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[34])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 663};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_code, __pyx_mstate->__pyx_n_u_filename, __pyx_mstate->__pyx_n_u_strict, __pyx_mstate->__pyx_n_u_backtrace_barrier, __pyx_mstate->__pyx_n_u_promise};
    __pyx_mstate_global->__pyx_codeobj_tab[35] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_eval, __pyx_mstate->__pyx_kp_b_iso88591_A_t8_avV_WHL, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[35])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 673};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_code, __pyx_mstate->__pyx_n_u_filename, __pyx_mstate->__pyx_n_u_strict, __pyx_mstate->__pyx_n_u_backtrace_barrier, __pyx_mstate->__pyx_n_u_promise};
    __pyx_mstate_global->__pyx_codeobj_tab[36] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_eval_module, __pyx_mstate->__pyx_kp_b_iso88591_A_t8_avV_V8K, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[36])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[37] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[37])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[38] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[38])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[39] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[39])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[40] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[40])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 893};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_code, __pyx_mstate->__pyx_n_u_filename, __pyx_mstate->__pyx_n_u_strict, __pyx_mstate->__pyx_n_u_backtrace_barrier, __pyx_mstate->__pyx_n_u_promise};
    __pyx_mstate_global->__pyx_codeobj_tab[41] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_eval, __pyx_mstate->__pyx_kp_b_iso88591_A_t6_1_YYttu, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[41])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 903};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_code, __pyx_mstate->__pyx_n_u_filename, __pyx_mstate->__pyx_n_u_strict, __pyx_mstate->__pyx_n_u_backtrace_barrier, __pyx_mstate->__pyx_n_u_promise};
    __pyx_mstate_global->__pyx_codeobj_tab[42] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_eval_module, __pyx_mstate->__pyx_kp_b_iso88591_A_t6_1_XXsst, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[42])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 916};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[43] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_get_global, __pyx_mstate->__pyx_kp_b_iso88591_A_y_V_5Qd, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[43])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 920};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_json, __pyx_mstate->__pyx_n_u_ret, __pyx_mstate->__pyx_n_u_view};
    __pyx_mstate_global->__pyx_codeobj_tab[44] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_json_parse, __pyx_mstate->__pyx_kp_b_iso88591_A_6_l_4v_fD_a_1AQ_y_V1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[44])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 936};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_name_2, __pyx_mstate->__pyx_n_u_glob, __pyx_mstate->__pyx_n_u_val, __pyx_mstate->__pyx_n_u_atom};
    __pyx_mstate_global->__pyx_codeobj_tab[45] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_get, __pyx_mstate->__pyx_kp_b_iso88591_A_a_Qd_avRq_nAT_vQ_AT_q_y_V1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[45])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 948};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_name_2, __pyx_mstate->__pyx_n_u_item, __pyx_mstate->__pyx_n_u_value, __pyx_mstate->__pyx_n_u_atom, __pyx_mstate->__pyx_n_u_glob};
    __pyx_mstate_global->__pyx_codeobj_tab[46] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_set, __pyx_mstate->__pyx_kp_b_iso88591_A_Qd_avRq_Qd_b_a_Qd_fG2Q_iq_V1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[46])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 966};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_func, __pyx_mstate->__pyx_n_u_name_2, __pyx_mstate->__pyx_n_u_count, __pyx_mstate->__pyx_n_u_magic};
    __pyx_mstate_global->__pyx_codeobj_tab[47] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_add_function, __pyx_mstate->__pyx_kp_b_iso88591_A_0_4xq_1A_S_A_uCq_2_7_6_1_AV1_n, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[47])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {7, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1075};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_code, __pyx_mstate->__pyx_n_u_this, __pyx_mstate->__pyx_n_u_filename, __pyx_mstate->__pyx_n_u_strict, __pyx_mstate->__pyx_n_u_backtrace_barrier, __pyx_mstate->__pyx_n_u_promise};
    __pyx_mstate_global->__pyx_codeobj_tab[48] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_eval_this, __pyx_mstate->__pyx_kp_b_iso88591_A_t_avV_WHDWWX, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[48])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {7, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1086};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_code, __pyx_mstate->__pyx_n_u_this, __pyx_mstate->__pyx_n_u_filename, __pyx_mstate->__pyx_n_u_strict, __pyx_mstate->__pyx_n_u_backtrace_barrier, __pyx_mstate->__pyx_n_u_promise};
    __pyx_mstate_global->__pyx_codeobj_tab[49] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_eval_this_with_module, __pyx_mstate->__pyx_kp_b_iso88591_A_t_avV_V8CVVW, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[49])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[50] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[50])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[51] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[51])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[52] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_A_G1F_a_vWE_Q_q_q_q_t1G_gQ_t1G, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[52])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[53] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_AV1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[53])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[54] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_A_G1F_a_vWE_Q_q_q_q_2_awk_PQ_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[54])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[55] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_q_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[55])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1151};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_fn};
    __pyx_mstate_global->__pyx_codeobj_tab[56] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_add_done_callback, __pyx_mstate->__pyx_kp_b_iso88591_A_4xq_1A_4vT_A_7_1_aq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[56])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1161};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[57] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_exception, __pyx_mstate->__pyx_kp_b_iso88591_A_q_F_a_6_A_3a_1A_t1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[57])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1169};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[58] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_done, __pyx_mstate->__pyx_kp_b_iso88591_A_q_F_a_vS_S_c, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[58])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1176};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_fn};
    __pyx_mstate_global->__pyx_codeobj_tab[59] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_remove_done_callback, __pyx_mstate->__pyx_kp_b_iso88591_A_r_A_3hd_3a_S_Cq_1_6_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[59])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1190};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[60] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_result, __pyx_mstate->__pyx_kp_b_iso88591_A_q_F_a_6_A_6_A_t_wa_d_4q_1A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[60])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1218};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[61] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyjs__cyjs_pyx, __pyx_mstate->__pyx_n_u_poll, __pyx_mstate->__pyx_kp_b_iso88591_A_d_T_4uA_1_1_1D_b_H_A_4uA_1_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[61])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[62] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[62])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[63] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[63])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[64] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_A_G1F_a_vWE_Q_q_q_q_awk_awk, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[64])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[65] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_q_a_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[65])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[66] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_pyx_unpickle_JSError, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_7_1_7_N_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[66])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[67] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_pyx_unpickle_MemoryUsage, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_haq_7_QnN_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[67])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[68] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_pyx_unpickle_PromiseHook, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_haq_7_QnN_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[68])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[69] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_pyx_unpickle_CancelledError, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_7_0_1B_PQ_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[69])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[70] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_pyx_unpickle_InvalidStateError, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_HAQ_7_314H_VW_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[70])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[71] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_pyx_unpickle_NODEFAULT, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_9HAQ_7_1L_a_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[71])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CommonTypesMetaclass.init */
  if (likely(__pyx_CommonTypesMetaclass_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
  {
      PyObject *typesModule=NULL;
      typesModule = PyImport_ImportModule("types");
      if (typesModule) {
          __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
          Py_DECREF(typesModule);
      }
  } // error handling follows
  #endif
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CythonFunctionShared.init */
  if (likely(__pyx_CyFunction_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* Generator.init */
  if (likely(__pyx_Generator_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError (used by GetBuiltinName) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName (used by GetModuleGlobalName) */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* PyDictVersioning (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* TupleAndListFromArray (used by fastcall) */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < (0))) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyObjectCallOneArg (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* UnpackUnboundCMethod (used by CallUnboundCMethod0) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod0 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        if (likely(cfunc->flag == METH_NOARGS))
            return __Pyx_CallCFunction(cfunc, self, NULL);
        if (likely(cfunc->flag == METH_FASTCALL))
            return __Pyx_CallCFunctionFast(cfunc, self, NULL, 0);
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, NULL, 0, NULL);
        if (likely(cfunc->flag == (METH_VARARGS | METH_KEYWORDS)))
            return __Pyx_CallCFunctionWithKeywords(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple, NULL);
        if (cfunc->flag == METH_VARARGS)
            return __Pyx_CallCFunction(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple);
        return __Pyx__CallUnboundCMethod0(cfunc, self);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod0(&tmp_cfunc, self);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod0(cfunc, self);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *result;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
    result = __Pyx_PyObject_CallOneArg(cfunc->method, self);
    return result;
}

/* py_dict_items (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_items, d);
}

/* py_dict_values (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_values, d);
}

/* OwnedDictNext (used by RejectKeywords) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue) {
    PyObject *next = NULL;
    if (!*ppos) {
        if (pvalue) {
            PyObject *dictview = pkey ? __Pyx_PyDict_Items(p) : __Pyx_PyDict_Values(p);
            if (unlikely(!dictview)) goto bad;
            *ppos = PyObject_GetIter(dictview);
            Py_DECREF(dictview);
        } else {
            *ppos = PyObject_GetIter(p);
        }
        if (unlikely(!*ppos)) goto bad;
    }
    next = PyIter_Next(*ppos);
    if (!next) {
        if (PyErr_Occurred()) goto bad;
        return 0;
    }
    if (pkey && pvalue) {
        *pkey = __Pyx_PySequence_ITEM(next, 0);
        if (unlikely(*pkey)) goto bad;
        *pvalue = __Pyx_PySequence_ITEM(next, 1);
        if (unlikely(*pvalue)) goto bad;
        Py_DECREF(next);
    } else if (pkey) {
        *pkey = next;
    } else {
        assert(pvalue);
        *pvalue = next;
    }
    return 1;
  bad:
    Py_XDECREF(next);
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
    PyErr_FormatUnraisable("Exception ignored in __Pyx_PyDict_NextRef");
#else
    PyErr_WriteUnraisable(__pyx_mstate_global->__pyx_n_u_Pyx_PyDict_NextRef);
#endif
    if (pkey) *pkey = NULL;
    if (pvalue) *pvalue = NULL;
    return 0;
}
#else // !CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue) {
    int result = PyDict_Next(p, ppos, pkey, pvalue);
    if (likely(result == 1)) {
        if (pkey) Py_INCREF(*pkey);
        if (pvalue) Py_INCREF(*pvalue);
    }
    return result;
}
#endif

/* RejectKeywords */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds) {
    PyObject *key = NULL;
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds))) {
        key = __Pyx_PySequence_ITEM(kwds, 0);
    } else {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *pos = NULL;
#else
        Py_ssize_t pos = 0;
#endif
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
        if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return;
#endif
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL);
#if CYTHON_AVOID_BORROWED_REFS
        Py_XDECREF(pos);
#endif
    }
    if (likely(key)) {
        PyErr_Format(PyExc_TypeError,
            "%s() got an unexpected keyword argument '%U'",
            function_name, key);
        Py_DECREF(key);
    }
}

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyUnicode_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* RaiseDoubleKeywords (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* ParseKeywordsImpl (used by ParseKeywords) */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos = NULL;
    #else
    Py_ssize_t pos = 0;
    #endif
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (
        #if CYTHON_AVOID_BORROWED_REFS
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL)
        #else
        PyDict_Next(kwds, &pos, &key, NULL)
        #endif
    ) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        #endif
    }
    __Pyx_END_CRITICAL_SECTION();
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(pos);
    #endif
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}

/* ParseKeywords */
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* GetTopmostException (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C0000
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#elif __PYX_LIMITED_VERSION_HEX > 0x030C0000
    local_value = PyErr_GetRaisedException();
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
#if __PYX_LIMITED_VERSION_HEX > 0x030C0000
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
#else
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
#endif // __PYX_LIMITED_VERSION_HEX > 0x030C0000
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#elif __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    PyErr_SetHandledException(local_value);
    Py_XDECREF(local_value);
    Py_XDECREF(local_type);
    Py_XDECREF(local_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
#if __PYX_LIMITED_VERSION_HEX <= 0x030C0000
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
#endif
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
    CYTHON_UNUSED_VAR(clineno);
    CYTHON_UNUSED_VAR(lineno);
    CYTHON_UNUSED_VAR(filename);
    CYTHON_MAYBE_UNUSED_VAR(nogil);
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(0);
    }
    ctx = PyUnicode_FromString(name);
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
    if (nogil)
        PyGILState_Release(state);
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* ReRaiseException */
static CYTHON_INLINE void __Pyx_ReraiseException(void) {
    PyObject *type = NULL, *value = NULL, *tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = PyThreadState_GET();
  #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    value = exc_info->exc_value;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (unlikely(value == Py_None)) {
        value = NULL;
    } else if (value) {
        Py_INCREF(value);
        type = (PyObject*) Py_TYPE(value);
        Py_INCREF(type);
        tb = PyException_GetTraceback(value);
    }
    #else
    type = exc_info->exc_type;
    tb = exc_info->exc_traceback;
    Py_XINCREF(type);
    Py_XINCREF(value);
    Py_XINCREF(tb);
    #endif
  #else
    type = tstate->exc_type;
    value = tstate->exc_value;
    tb = tstate->exc_traceback;
    Py_XINCREF(type);
    Py_XINCREF(value);
    Py_XINCREF(tb);
  #endif
#else
    PyErr_GetExcInfo(&type, &value, &tb);
#endif
    if (unlikely(!type || type == Py_None)) {
        Py_XDECREF(type);
        Py_XDECREF(value);
        Py_XDECREF(tb);
        PyErr_SetString(PyExc_RuntimeError,
            "No active exception to reraise");
    } else {
        PyErr_Restore(type, value, tb);
    }
}

/* PyObjectFastCallMethod */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf) {
    PyObject *result;
    PyObject *attr = PyObject_GetAttr(args[0], name);
    if (unlikely(!attr))
        return NULL;
    result = __Pyx_PyObject_FastCall(attr, args+1, nargsf - 1);
    Py_DECREF(attr);
    return result;
}
#endif

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectFormatAndDecref */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result;
    if (unlikely(!s)) return NULL;
    result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* ArgTypeTestFunc (used by ArgTypeTest) */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    PyObject *extra_info = __pyx_mstate_global->__pyx_empty_unicode;
    int from_annotation_subclass = 0;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (!exact) {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    } else if (exact == 2) {
        if (__Pyx_TypeCheck(obj, type)) {
            from_annotation_subclass = 1;
            extra_info = __pyx_mstate_global->__pyx_kp_u_Note_that_Cython_is_deliberately;
        }
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")"
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        "%s%U"
#endif
        , name, type_name, obj_type_name
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        , (from_annotation_subclass ? ". " : ""), extra_info
#endif
        );
#if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    if (exact == 2 && from_annotation_subclass) {
        PyObject *res;
        PyObject *vargs[2];
        vargs[0] = PyErr_GetRaisedException();
        vargs[1] = extra_info;
        res = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_kp_u_add_note, vargs, 2, NULL);
        Py_XDECREF(res);
        PyErr_SetRaisedException(vargs[0]);
    }
#endif
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* pep479 */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc, *new_exc;
    __Pyx_PyThreadState_declare
    int is_async_stopiteration = 0;
    CYTHON_MAYBE_UNUSED_VAR(in_async_gen);
    __Pyx_PyThreadState_assign
    cur_exc = __Pyx_PyErr_CurrentExceptionType();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else {
            return;
        }
    }
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(tb);
    new_exc = PyObject_CallFunction(PyExc_RuntimeError, "s",
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        "generator raised StopIteration");
    if (!new_exc) {
        Py_XDECREF(val);
        return;
    }
    PyException_SetCause(new_exc, val); // steals ref to val
    PyErr_SetObject(PyExc_RuntimeError, new_exc);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    if (max_char > 1114111) max_char = 1114111;
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - result_ulength < 0))
        goto overflow;
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = values[i];
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (__Pyx_PyUnicode_READY(uval) == (-1))
            goto bad;
        #endif
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(ulength < 0)) goto bad;
        #endif
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    Py_ssize_t i;
    PyObject *result = NULL;
    PyObject *value_tuple = PyTuple_New(value_count);
    if (unlikely(!value_tuple)) return NULL;
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    for (i=0; i<value_count; i++) {
        if (__Pyx_PyTuple_SET_ITEM(value_tuple, i, values[i]) != (0)) goto bad;
        Py_INCREF(values[i]);
    }
    result = PyUnicode_Join(__pyx_mstate_global->__pyx_empty_unicode, value_tuple);
bad:
    Py_DECREF(value_tuple);
    return result;
#endif
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    const char* function_name,
    PyObject *kw)
{
#if CYTHON_COMPILING_IN_PYPY && !defined(PyArg_ValidateKeywordArguments)
    CYTHON_UNUSED_VAR(function_name);
    CYTHON_UNUSED_VAR(kw);
    return 0;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
#if PY_VERSION_HEX >= 0x03090000
        CYTHON_UNUSED_VAR(function_name);
#else
        Py_ssize_t kwsize;
        #if CYTHON_ASSUME_SAFE_SIZE
        kwsize = PyTuple_GET_SIZE(kw);
        #else
        kwsize = PyTuple_Size(kw);
        if (unlikely(kwsize < 0)) return -1;
        #endif
        for (Py_ssize_t pos = 0; pos < kwsize; pos++) {
            PyObject* key = NULL;
            #if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, pos);
            #else
            key = PyTuple_GetItem(kw, pos);
            if (unlikely(!key)) return -1;
            #endif
            if (unlikely(!PyUnicode_Check(key))) {
                PyErr_Format(PyExc_TypeError,
                    "%.200s() keywords must be strings", function_name);
                return -1;
            }
        }
#endif
    } else {
        if (unlikely(!PyArg_ValidateKeywordArguments(kw))) return -1;
    }
    return 0;
#endif
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* SliceObject */
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(PyObject* obj, PyObject* value,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
    __Pyx_TypeName obj_type_name;
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_ass_subscript))
#endif
    {
        int result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyLong_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyLong_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_ass_subscript(obj, py_slice, value);
#else
        result = value ? PyObject_SetItem(obj, py_slice, value) : PyObject_DelItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object does not support slice %.10s",
        obj_type_name, value ? "assignment" : "deletion");
    __Pyx_DECREF_TypeName(obj_type_name);
bad:
    return -1;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS)) {
        return __Pyx_PyList_GetItemRefFast(o, wrapped_i, unsafe_shared);
    } else
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyList_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyTuple_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)) {
            return __Pyx_PyList_GetItemRefFast(o, n, unsafe_shared);
        } else if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            return __Pyx_NewRef(PyList_GET_ITEM(o, n));
        }
    } else
    #if !CYTHON_AVOID_BORROWED_REFS
    if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            return __Pyx_NewRef(PyTuple_GET_ITEM(o, n));
        }
    } else
    #endif
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_subscript) {
            PyObject *r, *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
}

/* CallTypeTraverse */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
    if (__Pyx_get_runtime_version() < 0x03090000) return 0;
    #endif
    if (!always_call) {
        PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
        unsigned long flags = PyType_GetFlags(base);
        if (flags & Py_TPFLAGS_HEAPTYPE) {
            return 0;
        }
    }
    Py_VISIT((PyObject*)Py_TYPE(o));
    return 0;
}
#endif

/* AllocateExtensionType */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final) {
    if (is_final || likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
        allocfunc alloc_func = __Pyx_PyType_GetSlot(t, tp_alloc, allocfunc);
        return alloc_func(t, 0);
    } else {
        newfunc tp_new = __Pyx_PyType_TryGetSlot(&PyBaseObject_Type, tp_new, newfunc);
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (!tp_new) {
            PyObject *new_str = PyUnicode_FromString("__new__");
            if (likely(new_str)) {
                PyObject *o = PyObject_CallMethodObjArgs((PyObject *)&PyBaseObject_Type, new_str, t, NULL);
                Py_DECREF(new_str);
                return o;
            } else
                return NULL;
        } else
    #endif
        return tp_new(t, __pyx_mstate_global->__pyx_empty_tuple, 0);
    }
}

/* LimitedApiGetTypeDict (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
    PyObject *tp_dictoffset_o;
    Py_ssize_t tp_dictoffset;
    tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
    if (unlikely(!tp_dictoffset_o)) return -1;
    tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
    Py_DECREF(tp_dictoffset_o);
    if (unlikely(tp_dictoffset == 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' doesn't have a dictoffset");
        return -1;
    } else if (unlikely(tp_dictoffset < 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' has an unexpected negative dictoffset. "
            "Please report this as Cython bug");
        return -1;
    }
    return tp_dictoffset;
}
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
    static Py_ssize_t tp_dictoffset = 0;
    if (unlikely(tp_dictoffset == 0)) {
        tp_dictoffset = __Pyx_GetTypeDictOffset();
        if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
            tp_dictoffset = 0; // try again next time?
            return NULL;
        }
    }
    return *(PyObject**)((char*)tp + tp_dictoffset);
}
#endif

/* SetItemOnTypeDict (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_SetItem(tp_dict, k, v);
    if (likely(!result)) {
        PyType_Modified(tp);
        if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
            PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
            if (!setNameResult) return -1;
            Py_DECREF(setNameResult);
        }
    }
    return result;
}

/* FixUpExtensionType */
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if __PYX_LIMITED_VERSION_HEX > 0x030900B1
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#else
    const PyType_Slot *slot = spec->slots;
    int changed = 0;
#if !CYTHON_COMPILING_IN_LIMITED_API
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
#if !CYTHON_COMPILING_IN_CPYTHON
        const
#endif  // !CYTHON_COMPILING_IN_CPYTHON)
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_vectorcall_offset = memb->offset;
                    changed = 1;
                }
#endif  // CYTHON_METH_FASTCALL
#if !CYTHON_COMPILING_IN_PYPY
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                    Py_DECREF(descr);
                    if (unlikely(set_item_result < 0)) {
                        return -1;
                    }
                    changed = 1;
                }
#endif  // !CYTHON_COMPILING_IN_PYPY
            }
            memb++;
        }
    }
#endif  // !CYTHON_COMPILING_IN_LIMITED_API
#if !CYTHON_COMPILING_IN_PYPY
    slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_getset)
        slot++;
    if (slot && slot->slot == Py_tp_getset) {
        PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
        while (getset && getset->name) {
            if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                PyObject *descr = PyDescr_NewGetSet(type, getset);
                if (unlikely(!descr))
                    return -1;
                #if CYTHON_COMPILING_IN_LIMITED_API
                PyObject *pyname = PyUnicode_FromString(getset->name);
                if (unlikely(!pyname)) {
                    Py_DECREF(descr);
                    return -1;
                }
                int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                Py_DECREF(pyname);
                #else
                CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                #endif
                Py_DECREF(descr);
                if (unlikely(set_item_result < 0)) {
                    return -1;
                }
                changed = 1;
            }
            ++getset;
        }
    }
#else
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#endif  // !CYTHON_COMPILING_IN_PYPY
    if (changed)
        PyType_Modified(type);
#endif  // PY_VERSION_HEX > 0x030900B1
    return 0;
}

/* FormatTypeName (used by ValidateExternBase) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__3);
    }
    goto done;
}
#endif

/* ValidateExternBase */
static int __Pyx_validate_extern_base(PyTypeObject *base) {
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_itemsize;
#endif
#if !CYTHON_COMPILING_IN_LIMITED_API
    itemsize = ((PyTypeObject *)base)->tp_itemsize;
#else
    py_itemsize = PyObject_GetAttrString((PyObject*)base, "__itemsize__");
    if (!py_itemsize)
        return -1;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        return -1;
#endif
    if (itemsize) {
        __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(base);
        PyErr_Format(PyExc_TypeError,
                "inheritance from PyVarObject types like '" __Pyx_FMT_TYPENAME "' not currently supported", b_name);
        __Pyx_DECREF_TypeName(b_name);
        return -1;
    }
    return 0;
}

/* PyObjectCallNoArg (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(tp);
    PyErr_Format(PyExc_AttributeError,
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}
#endif

/* PyObjectCallMethod0 (used by PyType_Ready) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[1] = {obj};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_CallNoArg;
    return PyObject_VectorcallMethod(method_name, args, 1 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
#endif
}

/* ValidateBasesTuple (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_SIZE
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (unlikely(n < 0)) return -1;
#endif
    for (i = 1; i < n; i++)
    {
        PyTypeObject *b;
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !CYTHON_USE_TYPE_SLOTS
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
CYTHON_UNUSED static int __Pyx_PyType_HasMultipleInheritance(PyTypeObject *t) {
    while (t) {
        PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
        if (bases) {
            return 1;
        }
        t = __Pyx_PyType_GetSlot(t, tp_base, PyTypeObject*);
    }
    return 0;
}
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !CYTHON_COMPILING_IN_CPYTHON || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    if (!__Pyx_PyType_HasMultipleInheritance(t)) {
        return PyType_Ready(t);
    }
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400) &&\
                !CYTHON_COMPILING_IN_GRAAL
        gc = PyImport_GetModule(__pyx_mstate_global->__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_mstate_global->__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable (used by MergeVTables) */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i=0;
    Py_ssize_t size;
    void** base_vtables;
    __Pyx_TypeName tp_base_name = NULL;
    __Pyx_TypeName base_name = NULL;
    void* unknown = (void*)-1;
    PyObject* bases = __Pyx_PyType_GetSlot(type, tp_bases, PyObject*);
    int base_depth = 0;
    {
        PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        while (base) {
            base_depth += 1;
            base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
        }
    }
    base_vtables = (void**) PyMem_Malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
#if CYTHON_COMPILING_IN_LIMITED_API
    size = PyTuple_Size(bases);
    if (size < 0) goto other_failure;
#else
    size = PyTuple_GET_SIZE(bases);
#endif
    for (i = 1; i < size; i++) {
        PyObject *basei;
        void* base_vtable;
#if CYTHON_AVOID_BORROWED_REFS
        basei = PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#else
        basei = PyTuple_GET_ITEM(bases, i);
#endif
        base_vtable = __Pyx_GetVtable((PyTypeObject*)basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
            }
        }
    }
    PyErr_Clear();
    PyMem_Free(base_vtables);
    return 0;
bad:
    {
        PyTypeObject* basei = NULL;
        PyTypeObject* tp_base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        tp_base_name = __Pyx_PyType_GetFullyQualifiedName(tp_base);
#if CYTHON_AVOID_BORROWED_REFS
        basei = (PyTypeObject*)PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = (PyTypeObject*)PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#else
        basei = (PyTypeObject*)PyTuple_GET_ITEM(bases, i);
#endif
        base_name = __Pyx_PyType_GetFullyQualifiedName(basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
    }
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
really_bad: // bad has failed!
#endif
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
other_failure:
#endif
    PyMem_Free(base_vtables);
    return -1;
}

/* DelItemOnTypeDict (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_DelItem(tp_dict, k);
    if (likely(!result)) PyType_Modified(tp);
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_mstate_global->__pyx_n_u_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_mstate_global->__pyx_n_u_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython);
            if (likely(reduce_cython)) {
                ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_mstate_global->__pyx_n_u_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetFullyQualifiedName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_2_4
#define __PYX_HAVE_RT_ImportType_3_2_4
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size)
{
    PyObject *result = 0;
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API)
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !( defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API) )
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    if (size == 0) {
        return (PyTypeObject *)result;
    }
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_2_4 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_2_4 && (size_t)basicsize > size) {
        if (PyErr_WarnFormat(NULL, 0,
                "%.200s.%.200s size changed, may indicate binary incompatibility. "
                "Expected %zd from C header, got %zd from PyObject",
                module_name, class_name, size, basicsize) < 0) {
            goto bad;
        }
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* HasAttr (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!PyUnicode_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_PyObject_GetAttrStrNoError(o, n);
    if (!r) {
        return (unlikely(PyErr_Occurred())) ? -1 : 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* ImportImpl (used by Import) */
static int __Pyx__Import_GetModule(PyObject *qualname, PyObject **module) {
    PyObject *imported_module = PyImport_GetModule(qualname);
    if (unlikely(!imported_module)) {
        *module = NULL;
        if (PyErr_Occurred()) {
            return -1;
        }
        return 0;
    }
    *module = imported_module;
    return 1;
}
static int __Pyx__Import_Lookup(PyObject *qualname, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject **module) {
    PyObject *imported_module;
    PyObject *top_level_package_name;
    Py_ssize_t i;
    int status, module_found;
    Py_ssize_t dot_index;
    module_found = __Pyx__Import_GetModule(qualname, &imported_module);
    if (unlikely(!module_found || module_found == -1)) {
        *module = NULL;
        return module_found;
    }
    if (imported_names) {
        for (i = 0; i < len_imported_names; i++) {
            PyObject *imported_name = imported_names[i];
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
            int has_imported_attribute = PyObject_HasAttr(imported_module, imported_name);
#else
            int has_imported_attribute = PyObject_HasAttrWithError(imported_module, imported_name);
            if (unlikely(has_imported_attribute == -1)) goto error;
#endif
            if (!has_imported_attribute) {
                goto not_found;
            }
        }
        *module = imported_module;
        return 1;
    }
    dot_index = PyUnicode_FindChar(qualname, '.', 0, PY_SSIZE_T_MAX, 1);
    if (dot_index == -1) {
        *module = imported_module;
        return 1;
    }
    if (unlikely(dot_index == -2)) goto error;
    top_level_package_name = PyUnicode_Substring(qualname, 0, dot_index);
    if (unlikely(!top_level_package_name)) goto error;
    Py_DECREF(imported_module);
    status = __Pyx__Import_GetModule(top_level_package_name, module);
    Py_DECREF(top_level_package_name);
    return status;
error:
    Py_DECREF(imported_module);
    *module = NULL;
    return -1;
not_found:
    Py_DECREF(imported_module);
    *module = NULL;
    return 0;
}
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *from_list = 0;
    int module_found;
    if (!qualname) {
        qualname = name;
    }
    module_found = __Pyx__Import_Lookup(qualname, imported_names, len_imported_names, &module);
    if (likely(module_found == 1)) {
        return module;
    } else if (unlikely(module_found == -1)) {
        return NULL;
    }
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (imported_names) {
#if CYTHON_COMPILING_IN_CPYTHON
        from_list = __Pyx_PyList_FromArray(imported_names, len_imported_names);
        if (unlikely(!from_list))
            goto bad;
#else
        from_list = PyList_New(len_imported_names);
        if (unlikely(!from_list)) goto bad;
        for (Py_ssize_t i=0; i<len_imported_names; ++i) {
            if (PyList_SetItem(from_list, i, __Pyx_NewRef(imported_names[i])) < 0) goto bad;
        }
#endif
    }
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, moddict, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, moddict, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(from_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level) {
    return __Pyx__Import(name, imported_names, len_imported_names, qualname, __pyx_mstate_global->__pyx_d, level);
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_mstate_global->__pyx_kp_u__4);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
                CYTHON_COMPILING_IN_GRAAL
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
    }
    return value;
}

/* ListPack */
static PyObject *__Pyx_PyList_Pack(Py_ssize_t n, ...) {
    va_list va;
    PyObject *l = PyList_New(n);
    va_start(va, n);
    if (unlikely(!l)) goto end;
    for (Py_ssize_t i=0; i<n; ++i) {
        PyObject *arg = va_arg(va, PyObject*);
        Py_INCREF(arg);
        if (__Pyx_PyList_SET_ITEM(l, i, arg) != (0)) {
            Py_CLEAR(l);
            goto end;
        }
    }
    end:
    va_end(va);
    return l;
}

/* GetAttr (used by Globals) */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
    if (likely(PyUnicode_Check(n)))
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* Globals */
static PyObject* __Pyx_Globals(void) {
    return __Pyx_NewRef(__pyx_mstate_global->__pyx_d);
}

/* PyObjectVectorCallKwBuilder */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* dict_setdefault (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
    PyObject* value;
#if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
    PyDict_SetDefaultRef(d, key, default_value, &value);
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    PyObject *args[] = {d, key, default_value};
    value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* AddModuleRef (used by FetchSharedCythonModule) */
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
      PyObject *module_dict = PyImport_GetModuleDict();
      PyObject *m;
      if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
          return NULL;
      }
      if (m != NULL && PyModule_Check(m)) {
          return m;
      }
      Py_XDECREF(m);
      m = PyModule_NewObject(name);
      if (m == NULL)
          return NULL;
      if (PyDict_CheckExact(module_dict)) {
          PyObject *new_m;
          (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
          Py_DECREF(m);
          return new_m;
      } else {
           if (PyObject_SetItem(module_dict, name, m) != 0) {
                Py_DECREF(m);
                return NULL;
            }
            return m;
      }
  }
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *py_name = PyUnicode_FromString(name);
      if (!py_name) return NULL;
      PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
      Py_DECREF(py_name);
      return module;
  }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif

/* FetchSharedCythonModule (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
}

/* FetchCommonType (used by CommonTypesMetaclass) */
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
    if (result && metaclass) {
        PyObject *old_tp = (PyObject*)Py_TYPE(result);
    Py_INCREF((PyObject*)metaclass);
#if __PYX_LIMITED_VERSION_HEX >= 0x03090000
        Py_SET_TYPE(result, metaclass);
#else
        result->ob_type = metaclass;
#endif
        Py_DECREF(old_tp);
    }
    return result;
}
#else
#define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
#endif
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t expected_basicsize) {
    Py_ssize_t basicsize;
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (expected_basicsize == 0) {
        return 0; // size is inherited, nothing useful to check
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
    if (unlikely(!py_basicsize)) return -1;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = NULL;
    if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
#else
    basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
#endif
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
    int get_item_ref_result;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    py_object_name = PyUnicode_FromString(object_name);
    if (!py_object_name) return NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) goto done;
    abi_module_dict = PyModule_GetDict(abi_module);
    if (!abi_module_dict) goto done;
    get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
    if (get_item_ref_result == 1) {
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else if (unlikely(get_item_ref_result == -1)) {
        goto bad;
    }
    cached_type = __Pyx_PyType_FromMetaclass(
        metaclass,
        CYTHON_USE_MODULE_STATE ? module : abi_module,
        spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type);
    if (unlikely(new_cached_type != cached_type)) {
        if (unlikely(!new_cached_type)) goto bad;
        Py_DECREF(cached_type);
        cached_type = new_cached_type;
        if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else {
        Py_DECREF(new_cached_type);
    }
done:
    Py_XDECREF(abi_module);
    Py_DECREF(py_object_name);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CommonTypesMetaclass (used by CythonFunctionShared) */
static PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
    return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
}
#if __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject* __pyx_CommonTypesMetaclass_call(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *args, CYTHON_UNUSED PyObject *kwds) {
    PyErr_SetString(PyExc_TypeError, "Cannot instantiate Cython internal types");
    return NULL;
}
static int __pyx_CommonTypesMetaclass_setattr(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *attr, CYTHON_UNUSED PyObject *value) {
    PyErr_SetString(PyExc_TypeError, "Cython internal types are immutable");
    return -1;
}
#endif
static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
    {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
    {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
    #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {Py_tp_call, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_new, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_setattro, (void*)__pyx_CommonTypesMetaclass_setattr},
    #endif
    {0, 0}
};
static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
    __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
    0,
    0,
    Py_TPFLAGS_IMMUTABLETYPE |
    Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT,
    __pyx_CommonTypesMetaclass_slots
};
static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    PyObject *bases = PyTuple_Pack(1, &PyType_Type);
    if (unlikely(!bases)) {
        return -1;
    }
    mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
    Py_DECREF(bases);
    if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
        return -1;
    }
    return 0;
}

/* PyMethodNew (used by CythonFunctionShared) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *result;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    {
        PyObject *args[] = {func, self};
        result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
    }
    #else
    result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
    #endif
    return result;
}
#else
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#endif

/* PyVectorcallFastCallDict (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i;
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos;
    #else
    Py_ssize_t pos;
    #endif
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    #if !CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t nkw = PyDict_Size(kw);
    if (unlikely(nkw == -1)) return NULL;
    #else
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    #endif
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = 0;
    i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (__Pyx_PyDict_NextRef(kw, &pos, &key, &value)) {
        keys_are_strings &=
        #if CYTHON_COMPILING_IN_LIMITED_API
            PyType_GetFlags(Py_TYPE(key));
        #else
            Py_TYPE(key)->tp_flags;
        #endif
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
        #else
        PyTuple_SET_ITEM(kwnames, i, key);
        #endif
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(pos);
    #endif
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    Py_ssize_t kw_size =
        likely(kw == NULL) ?
        0 :
#if !CYTHON_ASSUME_SAFE_SIZE
        PyDict_Size(kw);
#else
        PyDict_GET_SIZE(kw);
#endif
    if (kw_size == 0) {
        return vc(func, args, nargs, NULL);
    }
#if !CYTHON_ASSUME_SAFE_SIZE
    else if (unlikely(kw_size == -1)) {
        return NULL;
    }
#endif
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared (used by CythonFunction) */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
        int result;
        PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func_2);
        if (unlikely(!newFunc)) {
            PyErr_Clear(); // It's only an optimization, so don't throw an error
            return 0;
        }
        result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
        Py_DECREF(newFunc);
        return result;
    }
    return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if (PyMethod_Check(func)) {
        func = PyMethod_GET_FUNCTION(func);
    }
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
    PyObject *result;
    CYTHON_UNUSED_VAR(closure);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_doc_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#else
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_name_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    Py_INCREF(op->func_qualname);
    result = op->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
#endif
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_defaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_kwdefaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_annotations_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
    int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
    if (is_coroutine) {
        PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(is_coroutine_value)) {
            return is_coroutine_value;
        }
ignore:
        PyErr_Clear();
    }
    return __Pyx_PyBool_FromLong(is_coroutine);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    result = __Pyx_CyFunction_get_is_coroutine_value(op);
    if (unlikely(!result))
        return NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    if (op->func_is_coroutine) {
        Py_DECREF(result);
        result = __Pyx_NewRef(op->func_is_coroutine);
    } else {
        op->func_is_coroutine = __Pyx_NewRef(result);
    }
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        py_name, message, size);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        name, message, size);
#endif
}
static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s",
        py_name, message);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s",
        name, message);
#endif
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
#else
    {"func_dict", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
#endif
    {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#endif
#if CYTHON_METH_FASTCALL
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(args);
    __Pyx_BEGIN_CRITICAL_SECTION(m);
    Py_INCREF(m->func_qualname);
    result = m->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_dict = NULL;
#endif
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func_dict);
#elif PY_VERSION_HEX < 0x030d0000
    _PyObject_ClearManagedDict((PyObject*)m);
#else
    PyObject_ClearManagedDict((PyObject*)m);
#endif
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    Py_CLEAR(m->defaults);
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    {
        int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func_dict);
#else
    {
        int e =
#if PY_VERSION_HEX < 0x030d0000
            _PyObject_VisitManagedDict
#else
            PyObject_VisitManagedDict
#endif
                ((PyObject*)m, visit, arg);
        if (e != 0) return e;
    }
#endif
    __Pyx_VISIT_CONST(m->func_name);
    __Pyx_VISIT_CONST(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    __Pyx_VISIT_CONST(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    Py_VISIT(m->defaults);
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
    PyObject *repr;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
    __Pyx_END_CRITICAL_SECTION();
    return repr;
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes no arguments", size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes exactly one argument", size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    __Pyx_CyFunction_raise_type_error(
        (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_SIZE
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(argc < 0)) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            __Pyx_CyFunction_raise_type_error(
                cyfunc, "needs an argument");
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
        __Pyx_CyFunction_raise_type_error(
            cyfunc, "takes no keyword arguments");
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes no arguments", nargs);
        return NULL;
    }
    return meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes exactly one argument", nargs);
        return NULL;
    }
    return meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    #if PY_VERSION_HEX < 0x030e00A6
    size_t nargs_value = (size_t) nargs;
    #else
    Py_ssize_t nargs_value = nargs;
    #endif
    return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, nargs_value, kwnames);
}
#endif
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if CYTHON_METH_FASTCALL
#if defined(Py_TPFLAGS_HAVE_VECTORCALL)
    Py_TPFLAGS_HAVE_VECTORCALL |
#elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
#endif // CYTHON_METH_FASTCALL
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_DICT |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
static int __pyx_CyFunction_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
    if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
    if (unlikely(!m->defaults))
        return NULL;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
#define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
#elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
#else
PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
    PyObject **dict_ptr = _PyObject_GetDictPtr(o);
    return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
}
#endif
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline = NULL;
    PyObject *ptype, *pvalue, *ptraceback;
    PyObject *cython_runtime_dict;
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
    cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, cython_runtime_dict,
            __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
    }
    if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    Py_XDECREF(cython_runtime_dict);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache (used by AddTraceback) */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        if (c_line) {
            py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        } else {
            py_funcname = PyUnicode_FromString(funcname);
        }
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CheckUnpickleChecksum */
static void __Pyx_RaiseUnpickleChecksumError(long checksum, long checksum1, long checksum2, long checksum3, const char *members) {
    PyObject *pickle_module = PyImport_ImportModule("pickle");
    if (unlikely(!pickle_module)) return;
    PyObject *pickle_error = PyObject_GetAttrString(pickle_module, "PickleError");
    Py_DECREF(pickle_module);
    if (unlikely(!pickle_error)) return;
    if (checksum2 == checksum1) {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x) = (%s))",
            checksum, checksum1, members);
    } else if (checksum3 == checksum2) {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x, 0x%x) = (%s))",
            checksum, checksum1, checksum2, members);
    } else {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x, 0x%x, 0x%x) = (%s))",
            checksum, checksum1, checksum2, checksum3, members);
    }
    Py_DECREF(pickle_error);
}
static int __Pyx_CheckUnpickleChecksum(long checksum, long checksum1, long checksum2, long checksum3, const char *members) {
    int found = 0;
    found |= checksum1 == checksum;
    found |= checksum2 == checksum;
    found |= checksum3 == checksum;
    if (likely(found))
        return 0;
    __Pyx_RaiseUnpickleChecksumError(checksum, checksum1, checksum2, checksum3, members);
    return -1;
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int64_t __Pyx_PyLong_As_int64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyLong_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 2 * PyLong_SHIFT)) {
                            return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 3 * PyLong_SHIFT)) {
                            return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 4 * PyLong_SHIFT)) {
                            return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
        } else if ((sizeof(int64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int64_t) 1) << (sizeof(int64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyLong_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyLong_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 2 * PyLong_SHIFT)) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 3 * PyLong_SHIFT)) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 4 * PyLong_SHIFT)) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (size_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(size_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(size_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
        } else if ((sizeof(size_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        size_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (size_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (size_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (size_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (size_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(size_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((size_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(size_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((size_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((size_t) 1) << (sizeof(size_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (size_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint32_t __Pyx_PyLong_As_uint32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        uint32_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (uint32_t) -1;
        val = __Pyx_PyLong_As_uint32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 2 * PyLong_SHIFT)) {
                            return (uint32_t) (((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 3 * PyLong_SHIFT)) {
                            return (uint32_t) (((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 4 * PyLong_SHIFT)) {
                            return (uint32_t) (((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint32_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint32_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint32_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint32_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, long, PyLong_AsLong(x))
        } else if ((sizeof(uint32_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        uint32_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint32_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint32_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint32_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint32_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint32_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint32_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint32_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint32_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint32_t) 1) << (sizeof(uint32_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint32_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint32_t");
    return (uint32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint32_t");
    return (uint32_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint32_t(uint32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint32_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint32_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint32_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint32_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int32_t(int32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int32_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int32_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int32_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int32_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum____pyx_t_4cyjs_5_cyjs_PromiseHookType(enum __pyx_t_4cyjs_5_cyjs_PromiseHookType value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_4cyjs_5_cyjs_PromiseHookType neg_one = (enum __pyx_t_4cyjs_5_cyjs_PromiseHookType) -1, const_zero = (enum __pyx_t_4cyjs_5_cyjs_PromiseHookType) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_4cyjs_5_cyjs_PromiseHookType) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_4cyjs_5_cyjs_PromiseHookType) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum __pyx_t_4cyjs_5_cyjs_PromiseHookType) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_4cyjs_5_cyjs_PromiseHookType) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_4cyjs_5_cyjs_PromiseHookType) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_4cyjs_5_cyjs_PromiseHookType),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum __pyx_t_4cyjs_5_cyjs_PromiseHookType));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* UpdateUnpickledDict */
static int __Pyx__UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index) {
    PyObject *state_dict = __Pyx_PySequence_ITEM(state, index);
    if (unlikely(!state_dict)) {
        return -1;
    }
    int non_empty = PyObject_IsTrue(state_dict);
    if (non_empty == 0) {
        Py_DECREF(state_dict);
        return 0;
    } else if (unlikely(non_empty == -1)) {
        return -1;
    }
    PyObject *dict;
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    dict = PyObject_GetAttrString(obj, "__dict__");
    #else
    dict = PyObject_GenericGetDict(obj, NULL);
    #endif
    if (unlikely(!dict)) {
        Py_DECREF(state_dict);
        return -1;
    }
    int result;
    if (likely(PyDict_CheckExact(dict))) {
        result = PyDict_Update(dict, state_dict);
    } else {
        PyObject *obj_result = __Pyx_PyObject_CallMethod1(dict, __pyx_mstate_global->__pyx_n_u_update, state_dict);
        if (likely(obj_result)) {
            Py_DECREF(obj_result);
            result = 0;
        } else {
            result = -1;
        }
    }
    Py_DECREF(state_dict);
    Py_DECREF(dict);
    return result;
}
static int __Pyx_UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index) {
    Py_ssize_t state_size = __Pyx_PyTuple_GET_SIZE(state);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(state_size == -1)) return -1;
    #endif
    if (state_size <= index) {
        return 0;
    }
    return __Pyx__UpdateUnpickledDict(obj, state, index);
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
void __Pyx_init_runtime_version(void) {
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
}
#endif
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    return __Pyx_cached_runtime_version;
#endif
}

/* IterNextPlain (used by CoroutineBase) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *__Pyx_GetBuiltinNext_LimitedAPI(void) {
    if (unlikely(!__pyx_mstate_global->__Pyx_GetBuiltinNext_LimitedAPI_cache))
        __pyx_mstate_global->__Pyx_GetBuiltinNext_LimitedAPI_cache = __Pyx_GetBuiltinName(__pyx_mstate_global->__pyx_n_u_next);
    return __pyx_mstate_global->__Pyx_GetBuiltinNext_LimitedAPI_cache;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next_Plain(PyObject *iterator) {
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    PyObject *result;
    PyObject *next = __Pyx_GetBuiltinNext_LimitedAPI();
    if (unlikely(!next)) return NULL;
    result = PyObject_CallFunctionObjArgs(next, iterator, NULL);
    return result;
#else
    (void)__Pyx_GetBuiltinName; // only for early limited API
    iternextfunc iternext = __Pyx_PyObject_GetIterNextFunc(iterator);
    assert(iternext);
    return iternext(iterator);
#endif
}

/* ReturnWithStopIteration (used by CoroutineBase) */
static void __Pyx__ReturnWithStopIteration(PyObject* value, int async);
static CYTHON_INLINE void __Pyx_ReturnWithStopIteration(PyObject* value, int async, int iternext) {
    if (value == Py_None) {
        if (async || !iternext)
            PyErr_SetNone(async ? PyExc_StopAsyncIteration : PyExc_StopIteration);
        return;
    }
    __Pyx__ReturnWithStopIteration(value, async);
}
static void __Pyx__ReturnWithStopIteration(PyObject* value, int async) {
#if CYTHON_COMPILING_IN_CPYTHON
    __Pyx_PyThreadState_declare
#endif
    PyObject *exc;
    PyObject *exc_type = async ? PyExc_StopAsyncIteration : PyExc_StopIteration;
#if CYTHON_COMPILING_IN_CPYTHON
    if ((PY_VERSION_HEX >= (0x030C00A6)) || unlikely(PyTuple_Check(value) || PyExceptionInstance_Check(value))) {
        if (PY_VERSION_HEX >= (0x030e00A1)) {
            exc = __Pyx_PyObject_CallOneArg(exc_type, value);
        } else {
            PyObject *args_tuple = PyTuple_New(1);
            if (unlikely(!args_tuple)) return;
            Py_INCREF(value);
            PyTuple_SET_ITEM(args_tuple, 0, value);
            exc = PyObject_Call(exc_type, args_tuple, NULL);
            Py_DECREF(args_tuple);
        }
        if (unlikely(!exc)) return;
    } else {
        Py_INCREF(value);
        exc = value;
    }
    #if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    #if CYTHON_USE_EXC_INFO_STACK
    if (!__pyx_tstate->exc_info->exc_value)
    #else
    if (!__pyx_tstate->exc_type)
    #endif
    {
        Py_INCREF(exc_type);
        __Pyx_ErrRestore(exc_type, exc, NULL);
        return;
    }
    #endif
#else
    exc = __Pyx_PyObject_CallOneArg(exc_type, value);
    if (unlikely(!exc)) return;
#endif
    PyErr_SetObject(exc_type, exc);
    Py_DECREF(exc);
}

/* CoroutineBase (used by Generator) */
#if !CYTHON_COMPILING_IN_LIMITED_API
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6 && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#endif // CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void
__Pyx_Coroutine_Undelegate(__pyx_CoroutineObject *gen) {
#if CYTHON_USE_AM_SEND
    gen->yieldfrom_am_send = NULL;
#endif
    Py_CLEAR(gen->yieldfrom);
}
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            #if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
            value = PyObject_GetAttr(ev, __pyx_mstate_global->__pyx_n_u_value);
            if (unlikely(!value)) goto limited_api_failure;
            #else
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            #endif
            Py_DECREF(ev);
        }
        else if (unlikely(PyTuple_Check(ev))) {
            Py_ssize_t tuple_size = __Pyx_PyTuple_GET_SIZE(ev);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely(tuple_size < 0)) {
                Py_XDECREF(tb);
                Py_DECREF(ev);
                Py_DECREF(et);
                return -1;
            }
            #endif
            if (tuple_size >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#elif CYTHON_ASSUME_SAFE_MACROS
                value = PySequence_ITEM(ev, 0);
#else
                value = PySequence_GetItem(ev, 0);
                if (!value) goto limited_api_failure;
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_GetAttr(ev, __pyx_mstate_global->__pyx_n_u_value);
#else
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
#endif
    Py_DECREF(ev);
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!value)) return -1;
#endif
    *pvalue = value;
    return 0;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL || !CYTHON_ASSUME_SAFE_MACROS
  limited_api_failure:
    Py_XDECREF(et);
    Py_XDECREF(tb);
    Py_XDECREF(ev);
    return -1;
#endif
}
static CYTHON_INLINE
__Pyx_PySendResult __Pyx_Coroutine_status_from_result(PyObject **retval) {
    if (*retval) {
        return PYGEN_NEXT;
    } else if (likely(__Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, retval) == 0)) {
        return PYGEN_RETURN;
    } else {
        return PYGEN_ERROR;
    }
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
static void __Pyx_Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
__Pyx_PySendResult __Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, PyObject **result, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(__Pyx_Coroutine_get_is_running(self));  // Callers should ensure is_running
    if (unlikely(self->resume_label == -1)) {
        __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
        return PYGEN_ERROR;
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #elif PY_VERSION_HEX >= 0x030B00a4
        exc_tb = ((PyBaseExceptionObject*) exc_state->exc_value)->traceback;
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    retval = self->body(self, tstate, value);
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    *result = retval;
    if (self->resume_label == -1) {
        return likely(retval) ? PYGEN_RETURN : PYGEN_ERROR;
    }
    return PYGEN_NEXT;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(exc_state);
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
#define __Pyx_Coroutine_MethodReturnFromResult(gen, result, retval, iternext)\
    ((result) == PYGEN_NEXT ? (retval) : __Pyx__Coroutine_MethodReturnFromResult(gen, result, retval, iternext))
static PyObject *
__Pyx__Coroutine_MethodReturnFromResult(PyObject* gen, __Pyx_PySendResult result, PyObject *retval, int iternext) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (likely(result == PYGEN_RETURN)) {
        int is_async = 0;
        #ifdef __Pyx_AsyncGen_USED
        is_async = __Pyx_AsyncGen_CheckExact(gen);
        #endif
        __Pyx_ReturnWithStopIteration(retval, is_async, iternext);
        Py_XDECREF(retval);
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
#if PY_VERSION_HEX < 0x030d00A1
            _PyGen_SetStopIterationValue(result);
#else
            if (!PyTuple_Check(result) && !PyExceptionInstance_Check(result)) {
                PyErr_SetObject(PyExc_StopIteration, result);
            } else {
                PyObject *exc = __Pyx_PyObject_CallOneArg(PyExc_StopIteration, result);
                if (likely(exc != NULL)) {
                    PyErr_SetObject(PyExc_StopIteration, exc);
                    Py_DECREF(exc);
                }
            }
#endif
        }
        Py_DECREF(result);
        result = NULL;
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE __Pyx_PySendResult
__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen, PyObject** retval) {
    __Pyx_PySendResult result;
    PyObject *val = NULL;
    assert(__Pyx_Coroutine_get_is_running(gen));
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    result = __Pyx_Coroutine_SendEx(gen, val, retval, 0);
    Py_XDECREF(val);
    return result;
}
#if CYTHON_USE_AM_SEND
static __Pyx_PySendResult
__Pyx_Coroutine_SendToDelegate(__pyx_CoroutineObject *gen, __Pyx_pyiter_sendfunc gen_am_send, PyObject *value, PyObject **retval) {
    PyObject *ret = NULL;
    __Pyx_PySendResult delegate_result, result;
    assert(__Pyx_Coroutine_get_is_running(gen));
    delegate_result = gen_am_send(gen->yieldfrom, value, &ret);
    if (delegate_result == PYGEN_NEXT) {
        assert (ret != NULL);
        *retval = ret;
        return PYGEN_NEXT;
    }
    assert (delegate_result != PYGEN_ERROR || ret == NULL);
    __Pyx_Coroutine_Undelegate(gen);
    result = __Pyx_Coroutine_SendEx(gen, ret, retval, 0);
    Py_XDECREF(ret);
    return result;
}
#endif
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval = NULL;
    __Pyx_PySendResult result = __Pyx_Coroutine_AmSend(self, value, &retval);
    return __Pyx_Coroutine_MethodReturnFromResult(self, result, retval, 0);
}
static __Pyx_PySendResult
__Pyx_Coroutine_AmSend(PyObject *self, PyObject *value, PyObject **retval) {
    __Pyx_PySendResult result;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen))) {
        *retval = __Pyx_Coroutine_AlreadyRunningError(gen);
        return PYGEN_ERROR;
    }
    #if CYTHON_USE_AM_SEND
    if (gen->yieldfrom_am_send) {
        result = __Pyx_Coroutine_SendToDelegate(gen, gen->yieldfrom_am_send, value, retval);
    } else
    #endif
    if (gen->yieldfrom) {
        PyObject *yf = gen->yieldfrom;
        PyObject *ret;
      #if !CYTHON_USE_AM_SEND
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
      #endif
        {
            #if !CYTHON_COMPILING_IN_LIMITED_API || __PYX_LIMITED_VERSION_HEX >= 0x03080000
            if (value == Py_None && PyIter_Check(yf))
                ret = __Pyx_PyIter_Next_Plain(yf);
            else
            #endif
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_mstate_global->__pyx_n_u_send, value);
        }
        if (likely(ret)) {
            __Pyx_Coroutine_unset_is_running(gen);
            *retval = ret;
            return PYGEN_NEXT;
        }
        result = __Pyx_Coroutine_FinishDelegation(gen, retval);
    } else {
        result = __Pyx_Coroutine_SendEx(gen, value, retval, 0);
    }
    __Pyx_Coroutine_unset_is_running(gen);
    return result;
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    __Pyx_PySendResult result;
    PyObject *retval = NULL;
    CYTHON_UNUSED_VAR(gen);
    assert(__Pyx_Coroutine_get_is_running(gen));
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        result = __Pyx_Coroutine_Close(yf, &retval);
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        result = __Pyx_Coroutine_Close(yf, &retval);
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        result = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf);
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
        result = PYGEN_RETURN;
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
        result = PYGEN_RETURN;
    } else
    #endif
    {
        PyObject *meth;
        result = PYGEN_RETURN;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_mstate_global->__pyx_n_u_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval)) {
                result = PYGEN_ERROR;
            }
        }
    }
    Py_XDECREF(retval);
    return result == PYGEN_ERROR ? -1 : 0;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __Pyx_PySendResult result;
    PyObject *retval = NULL;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen))) {
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    }
    #if CYTHON_USE_AM_SEND
    if (gen->yieldfrom_am_send) {
        result = __Pyx_Coroutine_SendToDelegate(gen, gen->yieldfrom_am_send, Py_None, &retval);
    } else
    #endif
    if (gen->yieldfrom) {
        PyObject *yf = gen->yieldfrom;
        PyObject *ret;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && (PY_VERSION_HEX < 0x030A00A3 || !CYTHON_USE_AM_SEND)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
            ret = __Pyx_PyIter_Next_Plain(yf);
        if (likely(ret)) {
            __Pyx_Coroutine_unset_is_running(gen);
            return ret;
        }
        result = __Pyx_Coroutine_FinishDelegation(gen, &retval);
    } else {
        result = __Pyx_Coroutine_SendEx(gen, Py_None, &retval, 0);
    }
    __Pyx_Coroutine_unset_is_running(gen);
    return __Pyx_Coroutine_MethodReturnFromResult(self, result, retval, 1);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    PyObject *retval = NULL;
    __Pyx_PySendResult result;
    CYTHON_UNUSED_VAR(arg);
    result = __Pyx_Coroutine_Close(self, &retval);
    if (unlikely(result == PYGEN_ERROR))
        return NULL;
    Py_XDECREF(retval);
    Py_RETURN_NONE;
}
static __Pyx_PySendResult
__Pyx_Coroutine_Close(PyObject *self, PyObject **retval) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PySendResult result;
    PyObject *yf;
    int err = 0;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen))) {
        *retval = __Pyx_Coroutine_AlreadyRunningError(gen);
        return PYGEN_ERROR;
    }
    yf = gen->yieldfrom;
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    result = __Pyx_Coroutine_SendEx(gen, NULL, retval, 1);
    if (result == PYGEN_ERROR) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_Coroutine_unset_is_running(gen);
        if (!__Pyx_PyErr_Occurred()) {
            return PYGEN_RETURN;
        } else if (likely(__Pyx_PyErr_ExceptionMatches2(PyExc_GeneratorExit, PyExc_StopIteration))) {
            __Pyx_PyErr_Clear();
            return PYGEN_RETURN;
        }
        return PYGEN_ERROR;
    } else if (likely(result == PYGEN_RETURN && *retval == Py_None)) {
        __Pyx_Coroutine_unset_is_running(gen);
        return PYGEN_RETURN;
    } else {
        const char *msg;
        Py_DECREF(*retval);
        *retval = NULL;
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
            msg = "async generator ignored GeneratorExit";
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        __Pyx_Coroutine_unset_is_running(gen);
        return PYGEN_ERROR;
    }
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen)))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    yf = gen->yieldfrom;
    if (yf) {
        __Pyx_PySendResult result;
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                goto propagate_exception;
            goto throw_here;
        }
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_mstate_global->__pyx_n_u_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    __Pyx_Coroutine_unset_is_running(gen);
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        Py_DECREF(yf);
        if (ret) {
            __Pyx_Coroutine_unset_is_running(gen);
            return ret;
        }
        result = __Pyx_Coroutine_FinishDelegation(gen, &ret);
        __Pyx_Coroutine_unset_is_running(gen);
        return __Pyx_Coroutine_MethodReturnFromResult(self, result, ret, 0);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
propagate_exception:
    {
        PyObject *retval = NULL;
        __Pyx_PySendResult result = __Pyx_Coroutine_SendEx(gen, NULL, &retval, 0);
        __Pyx_Coroutine_unset_is_running(gen);
        return __Pyx_Coroutine_MethodReturnFromResult(self, result, retval, 0);
    }
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, "throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    {
        int e = __Pyx_call_type_traverse((PyObject*)gen, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    #if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    if (gen->gi_weakreflist != NULL)
    #endif
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        {
            destructor del = __Pyx_PyObject_GetSlot(gen, tp_del, destructor);
            if (del) del(self);
        }
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
#if CYTHON_USE_TP_FINALIZE
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *retval = NULL;
        __Pyx_PySendResult result = __Pyx_Coroutine_Close(self, &retval);
        if (result == PYGEN_ERROR) {
            PyErr_WriteUnraisable(self);
        } else {
            Py_XDECREF(retval);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
}
#endif
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx__Coroutine_get_frame(__pyx_CoroutineObject *self)
{
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyObject *frame;
    #if PY_VERSION_HEX >= 0x030d0000
    Py_BEGIN_CRITICAL_SECTION(self);
    #endif
    frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        PyObject *globals = PyDict_New();
        if (unlikely(!globals)) return NULL;
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            globals,                        /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        Py_DECREF(globals);
        if (unlikely(!frame))
            return NULL;
        if (unlikely(self->gi_frame)) {
            Py_DECREF(frame);
            frame = self->gi_frame;
        } else {
            self->gi_frame = frame;
        }
    }
    Py_INCREF(frame);
    #if PY_VERSION_HEX >= 0x030d0000
    Py_END_CRITICAL_SECTION();
    #endif
    return frame;
#else
    CYTHON_UNUSED_VAR(self);
    Py_RETURN_NONE;
#endif
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context) {
    CYTHON_UNUSED_VAR(context);
    PyObject *frame = self->gi_frame;
    if (frame)
        return __Pyx_NewRef(frame);
    return __Pyx__Coroutine_get_frame(self);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->yieldfrom_am_send = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_LIMITED_API
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    gen->gi_weakreflist = NULL;
#endif
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}
static char __Pyx_Coroutine_test_and_set_is_running(__pyx_CoroutineObject *gen) {
    char result;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_BEGIN_CRITICAL_SECTION(gen);
    #endif
    result = gen->is_running;
    gen->is_running = 1;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_END_CRITICAL_SECTION();
    #endif
    return result;
}
static void __Pyx_Coroutine_unset_is_running(__pyx_CoroutineObject *gen) {
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_BEGIN_CRITICAL_SECTION(gen);
    #endif
    assert(gen->is_running);
    gen->is_running = 0;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_END_CRITICAL_SECTION();
    #endif
}
static char __Pyx_Coroutine_get_is_running(__pyx_CoroutineObject *gen) {
    char result;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_BEGIN_CRITICAL_SECTION(gen);
    #endif
    result = gen->is_running;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_END_CRITICAL_SECTION();
    #endif
    return result;
}
static PyObject *__Pyx_Coroutine_get_is_running_getter(PyObject *gen, void *closure) {
    CYTHON_UNUSED_VAR(closure);
    char result = __Pyx_Coroutine_get_is_running((__pyx_CoroutineObject*)gen);
    if (result) Py_RETURN_TRUE;
    else Py_RETURN_FALSE;
}
#if __PYX_HAS_PY_AM_SEND == 2
static void __Pyx_SetBackportTypeAmSend(PyTypeObject *type, __Pyx_PyAsyncMethodsStruct *static_amsend_methods, __Pyx_pyiter_sendfunc am_send) {
    Py_ssize_t ptr_offset = (char*)(type->tp_as_async) - (char*)type;
    if (ptr_offset < 0 || ptr_offset > type->tp_basicsize) {
        return;
    }
    memcpy((void*)static_amsend_methods, (void*)(type->tp_as_async), sizeof(*type->tp_as_async));
    static_amsend_methods->am_send = am_send;
    type->tp_as_async = __Pyx_SlotTpAsAsync(static_amsend_methods);
}
#endif
static PyObject *__Pyx_Coroutine_fail_reduce_ex(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    __Pyx_TypeName self_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE((PyObject*)self));
    PyErr_Format(PyExc_TypeError, "cannot pickle '" __Pyx_FMT_TYPENAME "' object",
                         self_type_name);
    __Pyx_DECREF_TypeName(self_type_name);
    return NULL;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {"__reduce_ex__", (PyCFunction) __Pyx_Coroutine_fail_reduce_ex, METH_O, 0},
    {"__reduce__", (PyCFunction) __Pyx_Coroutine_fail_reduce_ex, METH_NOARGS, 0},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {"gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     PyDoc_STR("object being iterated by 'yield from', or None")},
    {"gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {"__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {"__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     PyDoc_STR("name of the generator"), 0},
    {"__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     PyDoc_STR("qualified name of the generator"), 0},
    {"gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     PyDoc_STR("Frame of the generator"), 0},
    {"gi_running", __Pyx_Coroutine_get_is_running_getter, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_GeneratorType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)PyObject_SelfIter},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Generator_methods},
    {Py_tp_members, (void *)__pyx_Generator_memberlist},
    {Py_tp_getset, (void *)__pyx_Generator_getsets},
    {Py_tp_getattro, (void *) PyObject_GenericGetAttr},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    {Py_am_send, (void *)__Pyx_Coroutine_AmSend},
#endif
    {0, 0},
};
static PyType_Spec __pyx_GeneratorType_spec = {
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_WEAKREF |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | __Pyx_TPFLAGS_HAVE_AM_SEND,
    __pyx_GeneratorType_slots
};
#if __PYX_HAS_PY_AM_SEND == 2
static __Pyx_PyAsyncMethodsStruct __pyx_Generator_as_async;
#endif
static int __pyx_Generator_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_GeneratorType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_GeneratorType_spec, NULL);
    if (unlikely(!mstate->__pyx_GeneratorType)) {
        return -1;
    }
#if __PYX_HAS_PY_AM_SEND == 2
    __Pyx_SetBackportTypeAmSend(mstate->__pyx_GeneratorType, &__pyx_Generator_as_async, &__Pyx_Coroutine_AmSend);
#endif
    return 0;
}
static PyObject *__Pyx_Generator_GetInlinedResult(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *retval = NULL;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen))) {
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    }
    __Pyx_PySendResult result = __Pyx_Coroutine_SendEx(gen, Py_None, &retval, 0);
    __Pyx_Coroutine_unset_is_running(gen);
    (void) result;
    assert (result == PYGEN_RETURN || result == PYGEN_ERROR);
    assert ((result == PYGEN_RETURN && retval != NULL) || (result == PYGEN_ERROR && retval == NULL));
    return retval;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* NewCodeObj */
#if CYTHON_COMPILING_IN_LIMITED_API
    static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                       PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                       PyObject *fv, PyObject *cell, PyObject* fn,
                                       PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030B0000
  static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030c00A1
    if (likely(result))
        result->_co_firsttraceable = 0;
    #endif
    return result;
  }
#elif !CYTHON_COMPILING_IN_PYPY
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
) {
    PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL;
    Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
    PyObject *varnames_tuple = PyTuple_New(var_count);
    if (unlikely(!varnames_tuple)) return NULL;
    for (Py_ssize_t i=0; i < var_count; i++) {
        Py_INCREF(varnames[i]);
        if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
    }
    #if CYTHON_COMPILING_IN_LIMITED_API
    varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
    if (!varnames_tuple_dedup) {
        if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
        varnames_tuple_dedup = varnames_tuple;
    }
    #else
    varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
    if (unlikely(!varnames_tuple_dedup)) goto done;
    #endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(varnames_tuple_dedup);
    #endif
    if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL && !CYTHON_COMPILING_IN_GRAAL) {
        Py_ssize_t line_table_length = __Pyx_PyBytes_GET_SIZE(line_table);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(line_table_length == -1)) goto done;
        #endif
        Py_ssize_t code_len = (line_table_length * 2 + 4) & ~3LL;
        code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
        if (unlikely(!code_bytes)) goto done;
        char* c_code_bytes = PyBytes_AsString(code_bytes);
        if (unlikely(!c_code_bytes)) goto done;
        memset(c_code_bytes, 0, (size_t) code_len);
    }
    code_obj = (PyObject*) __Pyx__PyCode_New(
        (int) descr.argcount,
        (int) descr.num_posonly_args,
        (int) descr.num_kwonly_args,
        (int) descr.nlocals,
        0,
        (int) descr.flags,
        code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        varnames_tuple_dedup,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        filename,
        funcname,
        (int) descr.first_line,
        (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table) ? line_table : __pyx_mstate_global->__pyx_empty_bytes
    );
done:
    Py_XDECREF(code_bytes);
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(varnames_tuple_dedup);
    #endif
    Py_DECREF(varnames_tuple);
    return code_obj;
}

/* DecompressString */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
    PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
    const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
    PyObject *methodname = PyUnicode_FromString("decompress");
    if (unlikely(!methodname)) return NULL;
    #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
    if (algo == 3) {
        PyObject *fromlist = Py_BuildValue("[O]", methodname);
        if (unlikely(!fromlist)) goto bad;
        module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
    } else
    #endif
        module = PyImport_ImportModule(module_name);
    if (unlikely(!module)) goto import_failed;
    decompress = PyObject_GetAttr(module, methodname);
    if (unlikely(!decompress)) goto import_failed;
    {
        #ifdef __cplusplus
            char *memview_bytes = const_cast<char*>(s);
        #else
            #if defined(__clang__)
              #pragma clang diagnostic push
              #pragma clang diagnostic ignored "-Wcast-qual"
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic push
              #pragma GCC diagnostic ignored "-Wcast-qual"
            #endif
            char *memview_bytes = (char*) s;
            #if defined(__clang__)
              #pragma clang diagnostic pop
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic pop
            #endif
        #endif
        #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
        int memview_flags = 0x100;
        #else
        int memview_flags = PyBUF_READ;
        #endif
        compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
    }
    if (unlikely(!compressed_bytes)) {
        Py_DECREF(decompress);
        goto bad;
    }
    decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
    Py_DECREF(compressed_bytes);
    Py_DECREF(decompress);
    Py_DECREF(module);
    Py_DECREF(methodname);
    return decompressed;
import_failed:
    PyErr_Format(PyExc_ImportError,
        "Failed to import '%.20s.decompress' - cannot initialise module strings. "
        "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
        module_name, algo);
bad:
    Py_XDECREF(module);
    Py_DECREF(methodname);
    return NULL;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
